****************************************************************************
*   This file contains useful functions and subroutines for building a 
*   phasespace generator.
c***************************************************************************
c   LIST OF ALL FUNCTIONS AND SUBROUTINES IN THIS FILE:
c
c     FUNCTION Resonance(xm2, xmg, rm2min, rm2max, R, weight, qsq)
c     FUNCTION NonResonance(rm2min, rm2max, R, weight, qsq)
c     FUNCTION TwoBodyDecay(R1, R2, P_in, M2_in, M2_out1, M2_out2, P_out1, P_out2, weight)
c     FUNCTION TwoToJetsPlusX(N, R, RN, ECM, PTmin, XQ2, K1, K2, X1, X2, Q, P, weight)
c     SUBROUTINE TwoBodyDecay0(R1, R2, P_in, M2_in, P_out1, P_out2, weight)
c     SUBROUTINE ThreeBodyDecay0(R, P_in, M2_in, P_1, P_2, P_3, weight)
c     SUBROUTINE TwoToTwo(R,RN,ECM,Qsqr1,Qsqr2,K1,K2,X1,X2,Q1,Q2,weight)
c
c***************************************************************************



****************************************************************************
      LOGICAL FUNCTION Resonance(xm2, xmg, rm2min, rm2max, R, weight, qsq)
****************************************************************************
*     Calculates the Q-square for a resonance
*
*    INPUT
*       xm2       : mass**2 of the resonance
*       xmg       : mass*width of the resonance
*       rm2min    : lowerBound**2 of the resonance
*       rm2max    : upperBound**2 of the resonance
*       R         : a random number between 0 and 1
*       weight    : phasespace weight
*
*    OUTPUT
*       weight    : new phasespace weight (multiplied by Jacobi factor)
*       qsq       : the Q**2 
*       Resonance : .true.,  if everything is ok 
*                   .false., if kinematics not ok -> weight = 0
****************************************************************************
      IMPLICIT NONE
      real*8 xm2, xmg, rm2min, rm2max, R, weight, qsq, tanx
      real*8 twopi
      parameter (twopi = 2d0*3.141592653589793d0)
      real*8 xmin, xmax, deltax, x 

      IF (rm2max.LT.rm2min) THEN
         weight = 0d0 
         Resonance = .false.
         RETURN
      ENDIF
      
      xmin = ATAN( (rm2min - xm2)/xmg )
      xmax = ATAN( (rm2max - xm2)/xmg )
      deltax = xmax - xmin
      x = xmin + deltax * R
      tanx = tan(x)
      qsq = xm2 + tanx * xmg
      
      IF (qsq.LE.0 .or. deltax.le.0) THEN
         weight = 0d0 
         Resonance = .false.
         RETURN 
      ENDIF

      weight = weight * deltax / twopi
      weight = weight * xmg * (1 + tanx**2 )

      Resonance = .true.
      RETURN 
      END
      

****************************************************************************
      LOGICAL FUNCTION NonResonance(rm2min, rm2max, R, weight, qsq)
****************************************************************************
*     Calculates the Q-square for a non-resonance (a flat distribution)
*
*    INPUT
*       rm2min       : lowerBound**2 of the resonance
*       rm2max       : upperBound**2 of the resonance
*       R            : a random number between 0 and 1
*       weight       : phasespace weight
*
*    OUTPUT
*       weight       : new phasespace weight (multiplied by Jacobi factor)
*       qsq          : the Q**2 
*       NonResonance : .true.,  if everything is ok 
*                      .false., if kinematics not ok -> weight = 0
****************************************************************************
      IMPLICIT NONE
      real*8 rm2min, rm2max, R, weight, qsq
      real*8 pi
      parameter (pi = 3.141592653589793d0)
      real*8 minMass, maxMass, deltaMass, mass

      IF (rm2max.LT.rm2min) THEN
         weight = 0d0 
         NonResonance = .false.
         RETURN 
      ENDIF
      
      minMass = sqrt(rm2min)
      maxMass = sqrt(rm2max)
      deltaMass = maxMass - minMass
      mass = minMass + deltaMass * R
      qsq = mass**2

      weight = weight * deltaMass * mass / pi

      NonResonance = .true.
      RETURN 
      END

****************************************************************************
      LOGICAL FUNCTION TwoBodyDecay(R1, R2, P_in, M2_in, M2_out1, M2_out2, 
     &                              P_out1, P_out2, weight)
****************************************************************************
*     Calculates the two-body decay P_in -> P_out1 + P_out2
*
*    INPUT
*       R1, R2         : two random numbers between 0 and 1
*       P_in           : 4-momentum of mother particle
*       M_in           : mass**2 of mother particle
*       M_out1, M_out2 : mass**2 of daughter particles
*       weight         : phasespace weight
*
*    OUTPUT
*       weight         : new phasespace weight (multiplied by Jacobi factor)
*       P_out1, P_out2 : 4-momenta of daughter particles
*       TwoBodyDecay   : .true.,  if everything is ok 
*                        .false., if kinematics not ok -> weight = 0
****************************************************************************
      IMPLICIT NONE
      real*8 R1, R2, P_in(0:3), M2_in, M2_out1, M2_out2
      real*8 P_out1(0:3), P_out2(0:3), weight
      real*8 twopi
      parameter (twopi = 2d0*3.141592653589793d0)
      integer*4 mu
      real*8 Q(0:3), CTH, STH, PHID, M_in, M_out1, M_out2, E1, P1

      M_in = sqrt(M2_in)
      M_out1 = sqrt(M2_out1)
      M_out2 = sqrt(M2_out2)

      IF (M_in .LT. (M_out1 + M_out2) ) THEN
         weight = 0d0 
         TwoBodyDecay = .false.
         RETURN 
      ENDIF

c generate outgoing momenta in CMS and boost back
      CTH = -1d0 + 2d0 * R1
      STH = sqrt(1d0 - CTH**2 )
      PHID = twopi * R2 - twopi/2.0d0
      
      E1 = (M2_in - M2_out2 + M2_out1) / (2d0*M_in)
      P1 = sqrt(E1**2 - M2_out1)
      Q(0) = E1
      Q(1) = P1 * STH*COS(PHID)
      Q(2) = P1 * STH*SIN(PHID)
      Q(3) = P1 * CTH

      call boostn(Q,P_in,P_out1)
      do mu = 0,3
         P_out2(mu) = P_in(mu) - P_out1(mu)
      enddo

      weight = weight * P1/(M_in * 2d0 * twopi) 
      
      TwoBodyDecay = .true.
      RETURN
      END

****************************************************************************
      SUBROUTINE TwoBodyDecay0(R1, R2, P_in, M2_in, P_out1, P_out2, weight)
****************************************************************************
*     Calculates the two-body decay P_in -> P_out1 + P_out2 
*     for MASSLESS decay particles 
*
*    INPUT
*       R1, R2         : two random numbers between 0 and 1
*       P_in           : 4-momentum of mother particle
*       M_in           : mass**2 of mother particle
*       weight         : phasespace weight
*
*    OUTPUT
*       weight         : new phasespace weight (multiplied by Jacobi factor)
*       P_out1, P_out2 : 4-momenta of daughter particles
****************************************************************************
      IMPLICIT NONE
      real*8 R1, R2, P_in(0:3), M2_in
      real*8 P_out1(0:3), P_out2(0:3), weight
      real*8 twopi
      parameter (twopi = 2d0*3.141592653589793d0)
      integer*4 mu
      real*8 Q(0:3), CTH, STH, PHID, M_in, E1

      M_in = sqrt(M2_in)

c generate outgoing momenta in CMS and boost back
      CTH = -1d0 + 2d0 * R1
      STH = sqrt(1d0 - CTH**2 )
      PHID = twopi * R2 - twopi/2d0
      
      E1 = 0.5d0 * M_in
      Q(0) = E1
      Q(1) = E1 * STH*COS(PHID)
      Q(2) = E1 * STH*SIN(PHID)
      Q(3) = E1 * CTH

      call boostn(Q,P_in,P_out1)
      do mu = 0,3
         P_out2(mu) = P_in(mu) - P_out1(mu)
      enddo

      weight = weight / (4d0 * twopi) 
      
      RETURN
      END

****************************************************************************
      SUBROUTINE ThreeBodyDecay0(R, P_in, M2_in, P_1, P_2, P_3, weight)
****************************************************************************
*     Calculates the three-body decay P_in -> P_1 + P_2 + p_3
*     for MASSLESS decay particles via Euler angles
*
*    INPUT
*       R              : array of 5 random numbers between 0 and 1
*       P_in           : 4-momentum of mother particle
*       M_in           : mass**2 of mother particle
*       weight         : phasespace weight
*
*    OUTPUT
*       weight         : new phasespace weight (multiplied by Jacobi factor)
*       P_1, P_2, P_3  : 4-momenta of daughter particles
****************************************************************************
      IMPLICIT NONE
      REAL*8 R(*)
      real*8 P_in(0:3), M2_in
      real*8 P_1(0:3), P_2(0:3), P_3(0:3), weight
      real*8 twopi
      parameter (twopi = 2d0*3.141592653589793d0)
      integer*4 mu
      real*8 alpha, sin_alpha, cos_alpha
      real*8 gamma, sin_gamma, cos_gamma
      real*8 cos_beta, sin_beta, s1, c1
      real*8 E1, E2, M
      
      alpha = twopi*R(1)
      cos_beta = 1-2*R(2)
      gamma = twopi*R(3)
      
      sin_beta = sqrt(1-cos_beta**2)
      sin_alpha = sin(alpha)
      cos_alpha = cos(alpha)
      sin_gamma = sin(gamma)
      cos_gamma = cos(gamma)
      
      M = sqrt(M2_in)
      E1 = 0.5d0*R(4)*M
      E2 = 0.5d0*M*(1-R(4)+R(5)*R(4))
      
      P_1(0) = E1
      P_1(1) = E1*(  cos_alpha*cos_gamma - cos_beta*sin_alpha*sin_gamma )
      P_1(2) = E1*( -cos_alpha*sin_gamma - cos_beta*cos_gamma*sin_alpha )
      P_1(3) = E1*sin_alpha*sin_beta

      c1 = -(2d0*E1*E2 - 2d0*(E1 + E2)*M + M2_in)/(2d0*E1*E2)
      s1 = sqrt(1-c1**2)

      P_2(0) = E2
      P_2(1) = E2*( s1*(cos_gamma*sin_alpha + cos_alpha*cos_beta*sin_gamma) - 
     &              c1*(cos_gamma*cos_alpha - sin_alpha*cos_beta*sin_gamma) )
      P_2(2) = E2*( s1*( cos_alpha*cos_beta*cos_gamma - sin_alpha*sin_gamma) - 
     &              c1*(-sin_alpha*cos_beta*cos_gamma - cos_alpha*sin_gamma) )
      P_2(3) = E2*sin_beta*(-s1*cos_alpha - c1*sin_alpha)

      CALL BOOSTN(P_1, P_in, P_1)      
      CALL BOOSTN(P_2, P_in, P_2)      
      do mu = 0,3
         P_3(mu) = P_in(mu)-P_1(mu)-P_2(mu)
      enddo

      weight = weight*(1d0/twopi**3)*(M2_in*R(4))/(16d0)
      
      RETURN
      END


***************************************************************************
      LOGICAL FUNCTION ThreeBodyDecay(R, RN, P_in, q1_out, q2_out, q3_out,
     &                  m1sq, m2sq, m3sq, w)
****************************************************************************
*     Calculates the three-body decay P_in -> q1_out + q2_out + q3_out
*
*    INPUT
*       R(8), rn       : eight random numbers between 0 and 1
*       P_in           : 4-momentum of mother particle
*       w              : phasespace weight
*
*    OUTPUT
*       w                      : new phasespace weight
*       q1_out, q2_out, q3_out : 4-momenta of daughter particles
*       m1sq, m2sq, m3sq       : mass squared of outgoing particles
*       ThreeBodyDecay         : .true.,  if everything is ok 
*                                .false., if kinematics not ok -> weight = 0
****************************************************************************
      IMPLICIT NONE

      REAL*8 RM2, RMG, RM2MIN, RM2MAX, S, PTC
      REAL*8 RM2MIN12, RM2MAX12, RM2MIN23, RM2MAX23, M2MIN
      REAL*8 RM2MAX1, RM2MAX2
      COMMON /TRIBOSBLIPSI/ RM2(3), RMG(3), RM2MIN(3), RM2MAX(3),
     1                RM2MIN12, RM2MAX12, RM2MIN23, RM2MAX23, M2MIN,
     2                S, PTC, RM2MAX1, RM2MAX2
      DOUBLE PRECISION RMLIMIT
      COMMON /TRIBOSWWWBLIPSI/ RMLIMIT(2)

      DOUBLE PRECISION CLR, XM2, XMG, B, VE, A
      COMMON /BKOPOU/ CLR(4,5,-1:1), XM2(6), XMG(6),
     1                B(6,6,6), VE(4,5), A(4,5)

      INTEGER I, J, II, JJ, MU, ic

      DOUBLE PRECISION m1sq, m2sq, m3sq, m12sq, m23sq, mw, m2
      DOUBLE PRECISION P_in(0:4), q1_out(0:3), q2_out(0:3), q3_out(0:3)

      DOUBLE PRECISION dotrr
      DOUBLE PRECISION M, E1, E2, E3, Abs1, Abs3, C, A1, A2, D
      DOUBLE PRECISION alpha, gamma, cosa, cosb, cosg, sina, sinb, sing

      DOUBLE PRECISION R(7), RN
      DOUBLE PRECISION w
      DOUBLE PRECISION twopi
      PARAMETER (twopi = 2d0*3.141592653589793d0)

c symmetrisation of q1 and q3
      INTEGER symm
      DOUBLE PRECISION QL1(0:3), QL3(0:3), msqlocal1, msqlocal3
      data symm /1/
      save symm    

      DOUBLE PRECISION xxwidth, xxm
      PARAMETER ( xxwidth = 200d0, xxm = 170d0 )

      LOGICAL RESONANCE
      EXTERNAL RESONANCE

      M = sqrt(p_in(4))

c generate two invariant masses (qi + qj)^2  = mij^2 first **********

c m_12^2 = (q1 + q2)^2 out of R(1)
      if ( r(1) .le. rmlimit(1) ) then
         if (.not. Resonance(xxm**2, xxm*xxwidth, rm2min12, rm2max1,
     1                              r(1)/rmlimit(1), W, m12sq)) return
         W = W/rmlimit(1)
      else if ( r(1) .le. rmlimit(2) ) then
         if (.not. Resonance(xm2(6), xmg(6), rm2max1, rm2max2,
     1        (r(1)-rmlimit(1))/(rmlimit(2)-rmlimit(1)), W, m12sq)) return
         W = W/(rmlimit(2)-rmlimit(1))
      else
         if (.not. Resonance(xxm**2, xxm*xxwidth, rm2max2, rm2max12,
     1        (r(1)-rmlimit(2))/(1d0-rmlimit(2)), W, m12sq)) return
         W = W/(1d0-rmlimit(2))
      endif

c or without the Higgs resonance
c      if (.not. Resonance(xxm**2, xxm*xxwidth, rm2min12, rm2max12,
c     1                              r(1), W, m12sq)) return

      if (m12sq.ge.p_in(4)) then
         W = 0.0d0
         return
      endif

c m_23^2 = (q2 + q3)^2 out of r(2)
      if ( r(2) .le. rmlimit(1) ) then
         if (.not. Resonance(xxm**2, xxm*xxwidth, rm2min23, rm2max1,
     1                              r(2)/rmlimit(1), W, m23sq)) return
         W = W/rmlimit(1)
      else if ( r(2) .le. rmlimit(2) ) then
         if (.not. Resonance(xm2(6), xmg(6), rm2max1, rm2max2,
     1        (r(2)-rmlimit(1))/(rmlimit(2)-rmlimit(1)), W, m23sq)) return
         W = W/(rmlimit(2)-rmlimit(1))
      else
         if (.not. Resonance(xxm**2, xxm*xxwidth, rm2max2, rm2max23,
     1        (r(2)-rmlimit(2))/(1d0-rmlimit(2)), W, m23sq)) return
         W = W/(1d0-rmlimit(2))
      endif
c or without the Higgs resonance
c      if (.not. Resonance(xxm**2, xxm*xxwidth, rm2min23, rm2max23,
c     1                              r(2), W, m23sq)) return

      if (m23sq.ge.p_in(4)) then
         W = 0.0d0
         return
      endif

c now generate m1^2 and m3^2 of the massive vector bosons **********

c m_1^2 is generated via tan mapping out of r(3)
      rm2max(1) = min(p_in(4)-m23sq,p_in(4)+m23sq-2.0d0*sqrt(p_in(4)*m23sq))
      if (rm2max(1).le.0) then
         W = 0.0d0
         return
      endif
      if(rm2max(1).ge.rm2(1)-16d0*rmg(1)) then
         if (r(3).lt.0.1d0) then
            W=W/0.1d0
            m2=max(sqrt(xm2(6))-sqrt(rm2(1)),0.0d0)
            mw=rmg(1)*10.0d0*m2
            m2 = m2**2
            if (.not.Resonance(m2, mw, rm2min(1), rm2(1)-16d0*rmg(1),
     1                              r(3)/0.1d0, W, m1sq)) RETURN 
         else
            W=W/0.9d0
            if (.not.Resonance(rm2(1), rmg(1), rm2(1)-16d0*rmg(1),
     1              rm2max(1), (r(3)-0.1d0)/0.9d0, W, m1sq)) RETURN 
         endif
      else
         if (.not. Resonance(rm2(1), rmg(1), rm2min(1), rm2max(1),
     1                              r(3), W, m1sq)) return
      endif

c m_3^2 is generated via tan mapping out of r(4)
      rm2max(3) = min(p_in(4)-m12sq, p_in(4)+m12sq-2.0d0*sqrt(p_in(4)*m12sq), 
     1                m12sq+m23sq-m1sq)
      if (rm2max(3).le.0) then
         W = 0.0d0
         return
      endif
      if (.not. Resonance(rm2(3), rmg(3), rm2min(3), rm2max(3),
     1                              r(4), W, m3sq)) return

c caculate E1, E2, E3, Abs1, Abs2 from that 
      E1 = (p_in(4) + m1sq - m23sq)/2/M
      E3 = (p_in(4) + m3sq - m12sq)/2/M
      E2 = M - E1 - E3
      Abs1 = sqrt(E1**2 - m1sq)
      Abs3 = sqrt(E3**2 - m3sq)

c m_2^2 is generated via tan mapping out of r(5)
      rm2min(2) = max(-Abs1**2-Abs3**2-2.d0*sqrt(Abs1**2*Abs3**2)+E2**2,
     1            1.0d-5)
      rm2max(2) = min(-Abs1**2-Abs3**2+2.d0*sqrt(Abs1**2*Abs3**2)+E2**2,
     1          E2**2, P_in(4))
      if (rm2max(2).le.0) then
         W = 0.0d0
         return
      endif
      if ((rm2max(2).gt.rm2(2)-16d0*rmg(2)) .and. 
     1    (rm2min(2).lt.rm2(2)-16d0*rmg(2)) ) then
         if (r(5).lt.0.1d0) then
            W=W/0.1d0
            m2=max(sqrt(xm2(6))-sqrt(rm2(2)),0.0d0)
            mw=rmg(2)*10.0d0*m2
            m2 = m2**2
            if (.not.Resonance(m2, mw, rm2min(2), rm2(2)-16d0*rmg(2),
     1                                 r(5)/0.1d0, W, m2sq)) RETURN 
         else
            W=W/0.9d0
            if (.not.Resonance(rm2(2), rmg(2), rm2(2)-16d0*rmg(2),
     1              rm2max(2), (r(5)-0.1d0)/0.9d0, W, m2sq)) RETURN 
         endif
      else
         if (.not. Resonance(rm2(2), rmg(2), rm2min(2), rm2max(2),
     1                                 r(5), W, m2sq)) return
      endif

c now generate Euler angles
      alpha = R(6) * twopi
      W = W*twopi

      cosb = -1.0d0 + R(7)*2.0d0
      W = W*2.0d0

      gamma = RN * twopi
      W = W*twopi

      cosa = cos(alpha)
      cosg = cos(gamma)
      sina = sin(alpha)
      sinb = sqrt(1 - cosb**2)
      sing = sin(gamma)

c calculate now C and D
      D = -(m1sq-m2sq+m3sq+p_in(4)-2.0d0*M*(E1+E3)+2.0d0*E1*E3)/2/Abs1/Abs3
      if (D**2.lt.1) then
         C = sqrt(1 - D**2)
      else
         W = 0.0d0
         print*,'D**2 >1'
         return
      endif

c Momenta of V1
      ql1(0) = E1
      ql1(1) = Abs1 * (cosa * cosg - cosb * sina * sing)
      ql1(2) = Abs1 * (-cosb * cosg * sina - cosa * sing)
      ql1(3) = Abs1 * (sina * sinb)

c Momenta of V3
      ql3(0) = E3
      ql3(1) = Abs3 * C * (cosg * sina + cosa * cosb * sing) 
     1      - Abs3 * D * (cosa * cosg - cosb * sina * sing)
      ql3(2) = Abs3 * C * (cosa * cosb * cosg - sina * sing)
     1      - Abs3 * D * (-cosb * cosg * sina - cosa * sing)
      ql3(3) = Abs3 * C * (- cosa * sinb) - Abs3 * D * (sina * sinb)

      call boostn( ql1(0), P_in(0), ql1(0) )
      call boostn( ql3(0), P_in(0), ql3(0) )

c Momenta of V2
      q2_out(0) = P_in(0) - ql1(0) - ql3(0)
      q2_out(1) = P_in(1) - ql1(1) - ql3(1)
      q2_out(2) = P_in(2) - ql1(2) - ql3(2)
      q2_out(3) = P_in(3) - ql1(3) - ql3(3)

c symmetrisation of q1 and q3
      msqlocal1 = m1sq
      msqlocal3 = m3sq
      symm = mod(symm,2) + 1
      if (symm.eq.1) then !interchange q1 <-> q3
         do mu = 0,3
            q1_out(mu) = ql1(mu)
            q3_out(mu) = ql3(mu)
         enddo
      else if (symm.eq.2) then
         do mu = 0,3
            q1_out(mu) = ql3(mu)
            q3_out(mu) = ql1(mu)
         enddo
         m1sq = msqlocal3
         m3sq = msqlocal1
      endif

c Additional weight factors ******************************************
      W = W/twopi**3
      W = W/32.0d0/P_in(4)

      ThreeBodyDecay = .true.

      RETURN
      END


****************************************************************************
      LOGICAL FUNCTION TwoToJetsPlusX(N,R,RN,ECM,PTmin,XQ2,K1,K2,X1,X2,Q,P,weight)
****************************************************************************
*     Calculates the phasespace for the 2 -> N massless jets + X
*     X is a pseudo-particle with arbitrary mass>0
*
*    INPUT
*       R      : array with 3*N random numbers between 0 and 1
*       RN     : one additional random number 
*       ECM    : center-of-mass energy square
*       PTmin  : minimum PT of jets
*       XQ2    : Q**2 of X-particle
*       weight : phasespace weight
*
*    OUTPUT
*       K1,K2  : 4-vectors for 2 incoming partons
*       X1,X2  : corresponding values of Feynman x
*       Q      : 4-vector of X-particle
*       P      : 4-vectors for the N massless jets
*       weight : new phasespace weight (multiplied by Jacobi factor)
****************************************************************************
      IMPLICIT NONE
      INTEGER*4 N, N1
      PARAMETER (N1 = 5)
      REAL*8 TPI,PI
      PARAMETER (PI=3.141592653589793d0)
      PARAMETER (TPI=2D0*PI )
      REAL*8 R(*),RN,ECM,PTmin,XQ2,
     1       K1(0:3),K2(0:3),Q(0:3),P(0:3,N1),
     2       X1,X2,weight
      
      INTEGER*4 I,J
      REAL*8 S,FAC,DELY,XMIN,XMAX,X,DELX,YCM,YMIN,YMAX,YBAR
      REAL*8 PCM(0:3), PNJ(0:4), Y(N1), PT(N1), PHI(N1)
      REAL*8 SINHY, COSHY, M2, M2min, QV, XMJ, PJT2, SHA, ACM
      REAL*8 SUMPST, Q0ST, RSHAT
      PARAMETER (m2min = 0.1d0**2)

      double precision dotrr
      external dotrr

      S = ECM**2
      FAC = 1d0/(16d0*PI**3)
      weight = weight*TPI

C  Reset local variables and incoming momenta      
      DO I=0,3
         PCM(I) = 0d0
         PNJ(I) = 0d0
         K1(I) = 0d0
         K2(I) = 0d0
      ENDDO
      PNJ(4) = 0d0

C  Get 4-momenta of the N massless jets
      DO I = 1,N
         weight = weight*FAC
         J = 3*(I-1)
         
         if (R(J+2).gt.0.2) then
            XMIN = 2D0/ECM
            XMAX = 1/PTmin
            DELX = XMAX-XMIN
            X = XMIN + (R(J+2)-0.2)/0.8d0*DELX
            PT(I) = 1/X
            weight = weight/0.8d0*DELX*PT(I)**3
         else
            xmin = 0
            xmax = PTmin
            DELX = XMAX-XMIN
            PT(i) = XMIN + R(J+2)/0.2d0*DELX
            weight = weight/0.2d0*DELX*PT(I)
         endif
         YMAX = ECM/(2d0*PT(I))
         YMAX = LOG(YMAX+SQRT(YMAX**2-1d0))

         DELY = 2*YMAX

         IF (I.LE.3) THEN
            Y(I) = -YMAX+R(J+1)*DELY
         ELSE
C  fix random number to 0 or 1 for delta-y = 0 with previous jet: R=0 or 1 
C  corresponds to lowest dijet invariant mass
            IF (Y(I-1).LT.0D0) THEN
               IF (Y(I-1).LT.-YMAX) THEN
                  Y(I) = -YMAX + R(J+1)*DELY
               ELSEIF (Y(I-1).LT.0d0) THEN
                  Y(I) = Y(I-1) + R(J+1)*DELY
                  IF (Y(I).GT.YMAX) Y(I) = -2*YMAX + Y(I)
               ELSE
                  Y(I) =  - R(J+1)*DELY
                  IF (Y(I).LT.-YMAX) Y(I) = 2*YMAX + Y(I)
               ENDIF
            ELSE
               IF (Y(I-1).GT.YMAX) THEN
                  Y(I) = YMAX - R(J+1)*DELY
               ELSEIF (Y(I-1).GT.0D0) THEN
                  Y(I) = Y(I-1) - R(J+1)*DELY
                  IF (Y(I).LT.-YMAX) Y(I) = 2*YMAX + Y(I)
               ELSE
                  Y(I) = R(J+1)*DELY
                  IF (Y(I).GT.YMAX) Y(I) = -2*YMAX + Y(I)
               ENDIF
            ENDIF
         ENDIF

         SINHY = SINH(Y(I))
         COSHY = SQRT(1+SINHY**2)
         weight = weight*DELY

         P(0,I) = PT(I)*COSHY

         IF (I.EQ.1) THEN
            PHI(I) = PI*(2*RN-1)
         ELSE
            PHI(I) = PHI(I-1) + TPI*R(J+3)
            if (phi(i).gt.pi) phi(i) = phi(i)-tpi
         ENDIF

         weight = weight*TPI

         P(1,I) = PT(I)*COS(PHI(I))
         P(2,I) = PT(I)*SIN(PHI(I))
         P(3,I) = PT(I)*SINHY

         DO J = 0,3
            PNJ(J) = PNJ(J)+P(J,I)
         ENDDO
      ENDDO

C  determine N-jet invariant mass
      IF (N.GT.1) THEN
         PNJ(4) = PNJ(0)**2-PNJ(1)**2-PNJ(2)**2-PNJ(3)**2
      ELSE
         PNJ(4) = 1D-18
      ENDIF
      IF (PNJ(4).GE.S) THEN
         weight=0d0
         TwoToJetsPlusX = .false.
         RETURN
      ENDIF

c  check invariant masses of all parton pairs: eliminate events which 
c  may cause numerical problems 
      do I = 1,N-1
         do J = I+1,N
            m2 = 2*dotrr(p(0,I),p(0,J))
            if (m2.lt.m2min) then
               weight = 0d0
               TwoToJetsPlusX = .false.
               RETURN
            endif
         enddo
      enddo
      
C  Next get the center of mass rapidity YCM
      QV = SQRT(XQ2)
      XMJ = SQRT(PNJ(4))
      YBAR = 0.5D0*LOG( (PNJ(0)+PNJ(3))/(PNJ(0)-PNJ(3)) )
      PJT2 = PNJ(1)**2+PNJ(2)**2
      SHA = (S-(QV+XMJ)**2-4*PJT2) / (4*(PNJ(4)+PJT2)) 
      IF (SHA.LE.0D0) THEN
        weight=0
        TwoToJetsPlusX = .false.
        RETURN
      ENDIF
      SHA = SQRT(SHA)
      ACM = LOG( SHA+SQRT(SHA**2+1) ) 
      YMIN = YBAR - ACM
      YMAX = YBAR + ACM

      DELY = YMAX-YMIN
      YCM = YMIN + R(3)*DELY
      SINHY = SINH(YCM)
      COSHY = SQRT(1+SINHY**2)
      weight = weight*DELY

C  now get Q and SHAT etc.
      SUMPST = PNJ(1)**2+PNJ(2)**2+(PNJ(3)*COSHY-PNJ(0)*SINHY)**2
      Q0ST = SQRT(XQ2+SUMPST)
      RSHAT = Q0ST + SQRT(PNJ(4)+SUMPST)
      PCM(0) = RSHAT*COSHY
      PCM(3) = RSHAT*SINHY

      DO J = 0,3
        Q(J) = PCM(J)-PNJ(J)
      ENDDO

      X1 = (PCM(0)+PCM(3))/ECM
      X2 = (PCM(0)-PCM(3))/ECM
      IF (X1.GE.1.0D0 .OR. X2.GE.1.0D0) THEN
         weight=0d0
         TwoToJetsPlusX = .false.
         RETURN
      ENDIF
      K1(0) =  X1*ECM/2
      K1(3) =  K1(0)
      K2(0) =  X2*ECM/2
      K2(3) = -K2(0)

c  check initial-final momentum transfer for all partons: eliminate events 
c  which may cause numerical problems 
      do I = 1,N         
         m2 = 2*min( p(0,I)*k1(0)-p(3,I)*k1(3),
     1               p(0,I)*k2(0)-p(3,I)*k2(3) )
         if (m2.lt.m2min) then
            weight = 0d0
            TwoToJetsPlusX = .false.
            RETURN
         endif
      enddo

C  insert Jacobian for dtau * delta(PCM(0)-Ql(0,0)-PNJ(0))/(2 Ql(0,0))
      weight = weight * RSHAT/(S*Q0ST)
      
C  finally put in flux factor for the parton collisions and conversion to fb
      weight = weight / (2*RSHAT**2) * 3.89371D+11

      TwoToJetsPlusX = .true.
      RETURN
      END


****************************************************************************
      SUBROUTINE TwoToTwo(R,RN,ECM,Qsqr1,Qsqr2,K1,K2,X1,X2,Q1,Q2,weight)
****************************************************************************
*     Calculates the phasespace for 2 -> 2 massive
*
*    INPUT
*       R            : array with 3 random numbers between 0 and 1
*       RN           : one additional random number 
*       ECM          : center-of-mass energy square
*       Qsqr1, Qsqr2 : Q**2 of the two outgoing particles
*       weight       : phasespace weight
*
*    OUTPUT
*       K1,K2  : 4-vectors for 2 incoming partons
*       X1,X2  : corresponding values of Feynman x
*       Q1,Q2  : 4-vectors of outgoing particles
*       weight : new phasespace weight (multiplied by Jacobi factor)
****************************************************************************

      IMPLICIT NONE

      REAL*8 R(*), RN, ECM, Qsqr1, Qsqr2
      REAL*8 K1(0:3), K2(0:3), Q1(0:3), Q2(0:3)
      REAL*8 X1, X2, weight
      REAL*8 TPI, PI
      PARAMETER (PI=3.141592653589793d0)
      PARAMETER (TPI=2D0*PI )

c local variables
      INTEGER*4 mu
      REAL*8 k(0:3)
      REAL*8 s, ptmax, y2_border, temp, pt, ET, y_border, y, y2
      REAL*8 phi, sinhy2, coshy2, abs_p, pL, tau
      REAL*8 A, ptc, ptmin

c test variables
      COMMON /ptdistr/ pt

      S = ecm**2

c generate y, p_T, y*, phi from plain distributed R(1),R(2),R(3), rn

      phi = rn * tpi - tpi/2d0

      ptc = 20.0d0
      pTmax = ecm/2.0d0
c      ptmax = 0.5d0*sqrt((s-(Qsqr1+Qsqr2))**2 - 4d0*Qsqr1*Qsqr2)/ecm
      A = pTmax/20.0d0*1.0d0/(2*pTmax-ptc)
      if (r(1) .le. A * ptc) then
         pT = r(1)*1.0d0/A
         Weight = Weight/A
      else
         pT = pTmax*ptc*(1-A*ptc)/(pTmax-A*ptc**2-r(1)
     &               *(pTmax-ptc))
         Weight = Weight/A/ptc**2*pT**2
      endif
c pT-distribution not continuous------------------------------
c      if (r(1).le.0.2) then
c         pTmax = ptc
c         pTmin = 0.0d0       
c         pT = r(1)*(pTmax-ptmin)/0.2d0 + ptmin
c         Weight= Weight/0.2d0
c      else
c         pTmin = 2.0d0/ecm
c         pTmax = 1/ptc
c         pT = pTmin + (r(1)-0.2)/0.8d0*(pTmax-pTmin)
c         pT = 1/pT
c         Weight = Weight/0.8d0*pT**2
c      endif
c      Weight = Weight*(ptmax-ptmin)


      ET = sqrt(pT**2+Qsqr1)
      temp = (s+Qsqr1-Qsqr2)/(2*ecm*ET)

c calculate y*
      y2_border = log( temp+sqrt(temp**2-1) )
      y2 = y2_border*2d0*( R(2) - 0.5d0 )          ! y*
      weight = weight * y2_border * 2d0      !(prop density)^-1 of y2

      sinhy2 = sinh(y2)
      coshy2 = sqrt(sinhy2**2+1)
      abs_p = sqrt(pT**2*coshy2**2+Qsqr1*sinhy2**2)
      pL = sinhy2*ET
      tau=((coshy2*ET+sqrt(coshy2**2*ET**2-Qsqr1+Qsqr2))**2)/s
      weight = weight * pT/(sqrt(tau*s))
      weight = weight/(sqrt(coshy2**2*ET**2-Qsqr1+Qsqr2))
     


      y_border = log(1/tau)
      y = y_border*( R(3) - 0.5d0 )                !y
      weight = weight * y_border !(prop density)^-1 of y


c------------------------------------------------------------------------
c generate incoming momenta in lab frame + x1 x2

      x1 = sqrt(tau) * exp(y)
      x2 = sqrt(tau) * exp(-y)

      k1(0) = 0.5d0 * x1 * ecm
      k1(1) = 0d0
      k1(2) = 0d0
      k1(3) = 0.5d0 * x1 * ecm

      k2(0) = 0.5d0 * x2 * ecm
      k2(1) = 0d0
      k2(2) = 0d0
      k2(3) = - 0.5d0 * x2 * ecm

      do mu = 0, 3
         K(mu) = k1(mu)+k2(mu)
      enddo
      
c Momenta of V1

      q1(0) = sqrt( pT**2 + pL**2 + Qsqr1 )
      q1(1) = pT * cos(phi)
      q1(2) = pT * sin(phi)
      q1(3) = pL
      

c Momenta of V2

      q2(0) = sqrt( pT**2 + pL**2 + Qsqr2 )
      q2(1) = - pT * cos(phi)
      q2(2) = - pT * sin(phi)
      q2(3) = - pL

c boost in lab frame

      call boostn( q1(0), K, q1(0) )
      call boostn( q2(0), K, q2(0) )


      weight = weight * 1d0/tpi/4d0/s
      weight = weight * 3.89379323d11         !conversion into fb

      
      RETURN
      END


****************************************************************************
      SUBROUTINE TwoToTwo2(R,RN,ECM,Qsqr1,Qsqr2,K1,K2,X1,X2,Q1,Q2,weight)
****************************************************************************
*     Calculates the phasespace for 2 -> 2 massive
*
*    INPUT
*       R            : array with 3 random numbers between 0 and 1
*       RN           : one additional random number 
*       ECM          : center-of-mass energy square
*       Qsqr1, Qsqr2 : Q**2 of the two outgoing particles
*       weight       : phasespace weight
*
*    OUTPUT
*       K1,K2  : 4-vectors for 2 incoming partons
*       X1,X2  : corresponding values of Feynman x
*       Q1,Q2  : 4-vectors of outgoing particles
*       weight : new phasespace weight (multiplied by Jacobi factor)
****************************************************************************
      IMPLICIT NONE
      REAL*8 R(*),RN,ECM,Qsqr1,Qsqr2,
     1       K1(0:3),K2(0:3),Q1(0:3),Q2(0:3),
     2       X1,X2,weight
      REAL*8 TPI,PI, ptc
      PARAMETER (PI=3.141592653589793d0)
      PARAMETER (TPI=2D0*PI )

      integer*4 mu
      real*8 k(0:3)
      real*8 s,ptmax,y2_border,temp,pt,ET,y_border,y,y2
      real*8 phi,sinhy2,coshy2,abs_p,pL,tau
      real*8 A

c test variables
      common /ptdistr/ pt


      S = ecm**2

c generate y, p_T, y*, phi from plain distributed R(1),R(2),R(3), rn

      phi = rn * tpi - tpi/2d0

      pTmax = ecm/2.0d0
c      ptmax = 0.5d0*sqrt((s-(Qsqr1+Qsqr2))**2 - 4d0*Qsqr1*Qsqr2)/ecm
      pTC = 15.0d0
c pT distribution ----constant - 1/pt^2 decrease
c      A = pTmax/pTc*1.0d0/(2*pTmax-pTc)
c      if (r(1) .le. A * ptc) then
c         pT = r(1)*1.0d0/A
c         Weight = Weight/A
c      else
c         pT = pTmax*pTc*(1-A*pTc)/(pTmax-A*pTc**2-r(1)        !1
c     &               *(pTmax-pTc))
c         pT = pTmax*pTc/(2.0d0*pTmax-r(1)*(2.0d0*pTmax-pTc))  !2
c         Weight = Weight/A/pTc**2*pT**2
c      endif
c pT distribution ----linear increase - 1/pT^2 decrease
c      A = ptmax/(3/2.d0*ptmax-ptc)/ptc**2
c      if ( r(1) .le. A*ptc**2/2.d0 ) then
c         pT = sqrt(r(1)*2.d0/A)
c         Weight = Weight/A/pT
c      else
c         pT = pTmax*pTc/(3/2.d0*pTmax-r(1)*(3/2.d0*pTmax-pTc))
c         Weight = Weight/A/ptc**3*pT**2
c      endif
c pT distribution ----linear increase - 1/pT decrease
c      A = 1.d0/ptc**2/(0.5d0+log(pTmax/pTc))
c      if ( r(1) .le. ptc**2*A/2.d0 ) then
c         pT = sqrt(r(1)*2.d0/A)
c         Weight = Weight/A/pT
c      else
c         pT = pTc*(ptmax/ptc)**r(1)*exp(0.5d0*(r(1)-1))
c         Weight = Weight/A/ptc**2*pT
c      endif
c pT distribution ----constant - 1/pt decrease
      A = 1.d0/ptc/(1+log(pTmax/ptc))
      if ( r(1) .le. A*ptc ) then
         pT = r(1)/A
         Weight = Weight/A
      else
         pT = pTc*(ptmax/ptc)**r(1)*exp(r(1)-1)
         Weight = Weight/A/ptc*pT
      endif

      ET = sqrt(pT**2+Qsqr1)
      temp = (s+Qsqr1-Qsqr2)/(2*ecm*ET)

c calculate y*
      y2_border = log( temp+sqrt(temp**2-1) )
      y2 = y2_border*2d0*( R(2) - 0.5d0 )          ! y*
      weight = weight * y2_border * 2d0      !(prop density)^-1 of y2

      sinhy2 = sinh(y2)
      coshy2 = sqrt(sinhy2**2+1)
      abs_p = sqrt(pT**2*coshy2**2+Qsqr1*sinhy2**2)
      pL = sinhy2*ET
      tau=((coshy2*ET+sqrt(coshy2**2*ET**2-Qsqr1+Qsqr2))**2)/s
      weight = weight * pT/(sqrt(tau*s))
      weight = weight/(sqrt(coshy2**2*ET**2-Qsqr1+Qsqr2))
     


      y_border = log(1/tau)
      y = y_border*( R(3) - 0.5d0 )                !y
      weight = weight * y_border !(prop density)^-1 of y


c------------------------------------------------------------------------
c generate incoming momenta in lab frame + x1 x2

      x1 = sqrt(tau) * exp(y)
      x2 = sqrt(tau) * exp(-y)

      k1(0) = 0.5d0 * x1 * ecm
      k1(1) = 0d0
      k1(2) = 0d0
      k1(3) = 0.5d0 * x1 * ecm

      k2(0) = 0.5d0 * x2 * ecm
      k2(1) = 0d0
      k2(2) = 0d0
      k2(3) = - 0.5d0 * x2 * ecm

      do mu = 0, 3
         K(mu) = k1(mu)+k2(mu)
      enddo
      
c Momenta of V1

      q1(0) = sqrt( pT**2 + pL**2 + Qsqr1 )
      q1(1) = pT * cos(phi)
      q1(2) = pT * sin(phi)
      q1(3) = pL
      

c Momenta of V2

      q2(0) = sqrt( pT**2 + pL**2 + Qsqr2 )
      q2(1) = - pT * cos(phi)
      q2(2) = - pT * sin(phi)
      q2(3) = - pL

c boost in lab frame

      call boostn( q1(0), K, q1(0) )
      call boostn( q2(0), K, q2(0) )


      weight = weight * 1d0/tpi/4d0/s
      weight = weight * 3.89379323d11         !conversion into fb

      
      RETURN
      END


****************************************************************************
      Subroutine TwoToJetPlusX(R,RN,ECM,PTC,XQ2,K1,K2,X1,X2,Q,P,weight)
****************************************************************************
*     Calculates the phasespace for the 2 -> 1 massless Jet + X
*     X is a pseudo-particle with arbitrary mass>0
*
*    INPUT
*       R      : array with 3*N random numbers between 0 and 1
*       RN     : one additional random number 
*       ECM    : center-of-mass energy square
*       PTC  : minimum PT of jets
*       XQ2    : Q**2 of X-particle
*       weight : phasespace weight
*
*    OUTPUT
*       K1,K2  : 4-vectors for 2 incoming partons
*       X1,X2  : corresponding values of Feynman x
*       Q      : 4-vector of X-particle
*       P      : 4-vectors for the N massless jets
*       weight : new phasespace weight (multiplied by Jacobi factor)
****************************************************************************

      IMPLICIT NONE
      REAL*8 R(*),RN,ECM,XQ2,
     1       K1(0:3),K2(0:3),Q(0:3),P(0:3),
     2       X1,X2,weight, ptc
      REAL*8 TPI,PI
      PARAMETER (PI=3.141592653589793d0)
      PARAMETER (TPI=2D0*PI )

c test common block for plots
      double precision ptdistr
      common /ptdistribution/ ptdistr

c local variables
      integer*4 mu
      real*8 k(0:3), m2min
      real*8 s,ptmax,y2_border,temp,pt,ET,y_border,y,y2
      real*8 phi,sinhy2,coshy2,abs_p,pL,tau

      real*8 A,B,zmax,mass1,zc,z,lambda
      parameter(lambda = 1.7d0)
      real*8 gamma, kappa
      parameter(gamma = 2.5d0)
      kappa = 0.5d0*gamma/atan(gamma)


c cutoff for numerical purposes
      m2min = 0.01d0**2

      S = ecm**2

c generate y, p_T, y*, phi from plain distributed R(1),R(2),R(3), rn

      phi = rn * tpi - tpi/2d0

      pTmax = ecm/2.0d0
c      ptmax = 0.5d0*sqrt((s-(xq2))**2 - 4d0*xq2)/ecm
c      pTC = 40.0d0
c old pt distribution in my PS generator
      z = pTmax/pTc*1.0d0/(2*pTmax-pTc)
      if (r(1) .le. z * 20.0d0) then !constant pT
         pT = r(1)*1.0d0/z
         Weight = Weight/z
      else
         pT = pTmax*pTc*(1-z*pTc)/(pTmax-z*pTc**2-r(1)
     &               *(pTmax-pTc))
         pT = pTmax*pTc/(2.0d0*pTmax-r(1)*(2.0d0*pTmax-pTc))
         Weight = Weight/z/pTc**2*pT**2
      endif

c      z = ptmax/(3/2.d0*ptmax-ptc)/ptc**2
c      if ( r(1) .le. z*ptc**2/2.d0 ) then !linear increase
c         pT = sqrt(r(1)*2.d0/z)
c         Weight = Weight/z/pT
c      else                                ! 1/pT**2
c         pT = pTmax*pTc/(3/2.d0*pTmax-r(1)*(3/2.d0*pTmax-pTc))
c         Weight = Weight/z/ptc**3*pT**2
c      endif

      ptdistr = pt

      ET = sqrt(pT**2)
      temp = (s-xq2)/(2*ecm*ET)


c calculate y*
      y2_border = log( temp+sqrt(temp**2-1) )
      y2 = y2_border*2d0*( R(2) - 0.5d0 )          ! y*
      weight = weight * y2_border * 2d0      !(prop density)^-1 of y2

      sinhy2 = sinh(y2)
      coshy2 = sqrt(sinhy2**2+1)
      abs_p = sqrt(pT**2*coshy2**2+sinhy2**2)
      pL = sinhy2*ET
      tau=((coshy2*ET+sqrt(coshy2**2*ET**2+xq2))**2)/s
      weight = weight * pT/(sqrt(tau*s))
      weight = weight/(sqrt(coshy2**2*ET**2+xq2))
     


      y_border = log(1/tau)
      y = y_border*( R(3) - 0.5d0 )                !y
      weight = weight * y_border !(prop density)^-1 of y


c------------------------------------------------------------------------
c generate incoming momenta in lab frame + x1 x2

      x1 = sqrt(tau) * exp(y)
      x2 = sqrt(tau) * exp(-y)

      k1(0) = 0.5d0 * x1 * ecm
      k1(1) = 0d0
      k1(2) = 0d0
      k1(3) = 0.5d0 * x1 * ecm

      k2(0) = 0.5d0 * x2 * ecm
      k2(1) = 0d0
      k2(2) = 0d0
      k2(3) = - 0.5d0 * x2 * ecm

      do mu = 0, 3
         K(mu) = k1(mu)+k2(mu)
      enddo
      
c Momenta of V1

      p(0) = sqrt( pT**2 + pL**2 )
      p(1) = pT * cos(phi)
      p(2) = pT * sin(phi)
      p(3) = pL

c Momenta of V2

      q(0) = sqrt( pT**2 + pL**2 + xq2 )
      q(1) = - pT * cos(phi)
      q(2) = - pT * sin(phi)
      q(3) = - pL

c boost in lab frame

      call boostn( p(0), K, p(0) )
      call boostn( q(0), K, q(0) )

      if (2.0d0*min( p(0)*k1(0)-p(3)*k1(3),
     &            p(0)*k2(0)-p(3)*k2(3) ).lt.m2min) then
         Weight = 0
         return
      endif

      weight = weight * 1d0/tpi/4d0/s
      weight = weight * 3.89379323d11         !conversion into fb

      
      RETURN
      END
