c
c---------------  tens3 = 3point tensors ------------------------------
c
      subroutine tens3( mt, p1sq, p2sq, psq, B0, C0, 
     &                  Cij)
      implicit none
      double precision mt, p1sq, p2sq, psq
      complex*16 B0(3), C0, Cij(2,4)
c
c  determine the Passarino-Veltman tensor decomposition for the 
c  three-point tensor integrals
c
c                   1                 1;  k_mu;   k_mu k_nu
c C0;C_mu;C_munu =------Int d^4k --------------------------------------
c                 i pi^2        [-k^2+mt^2][-(k+p1)^2+mt^2][-(k+p1+p2)^2+mt^2] 
c
c with
c
c   C_mu = p1_mu C11  +  p_2_mu C12
c
c  C_munu = p1_mu p1_nu C21 + p2_mu p2_nu C22 + 
c           (p1_mu p2_nu + p1_nu p2_mu) C23  -  g_munu C24
c
c  for notation see Passarino&Veltman, NP B160 (1979) 151 and my notes
c  on gg-->Hgg, pages 4.1--4.8
c
C INPUT:  mt, p1sq, p2sq, psq=(p1+p2)^2     mass and 4-momenta**2
C         B0, C0                   4 scalar integrals; the 3 B0 are, 
c                                  in PV notation:
c         B0(1) = B0(1,2)          B_0 function with subtraction of 
c         B0(2) = B0(2,3)          divergent term
c         B0(3) = B0(1,3)
c
c OUTPUT: Cij(n,m) = C_nm          form factors in the tensor integrals
c         n=1,2; m=1,2,3,4         a la PV
c
c	Dieter Zeppenfeld, <dieter@pheno.physics.wisc.edu>
c	Initial version:  2000 November
c	Last modified:    

      real*8 f1, f2, Xi(2,2), p1p2, deti
      complex*16 b13, b12, b23, r1,r2,r3,r4,r5,r6
      complex*16 C11, C12, C21, C22, C23, C24, C23p
      integer i,j, iscount
      logical ldebug
      parameter (ldebug=.false.)
      data iscount /0/
      save iscount

      real*8 srt
      external srt

      p1p2 = (psq-p1sq-p2sq)/2d0
      f1 = p1sq
      f2 = psq - p1sq
      if (psq.eq.0d0) then
         deti = 0.25d0*(p1sq-p2sq)**2
      elseif (p1sq.eq.0d0) then
         deti = 0.25d0*(psq-p2sq)**2
      elseif (p2sq.eq.0d0) then
         deti = 0.25d0*(psq-p1sq)**2
      else
         deti = p1p2**2-p1sq*p2sq
         if (abs(deti).lt.1d-30) then
            do i=1,2
               do j = 1,4
                  Cij(i,j) = 0
               enddo
            enddo
            iscount = iscount+1
            if (iscount.le.1000) then
               print*," Warning: singular point in TENS3 "
               return
            else
               print*," stop: too many singular points in TENS3 "
               stop
            endif
         endif
c         if (abs(deti).lt.1d-6*max(abs(p1p2**2),abs(p1sq*p2sq))) then
c            print*," singular Cij. deti suppressed by ",
c     1           abs(deti)/max(abs(p1p2**2),abs(p1sq*p2sq))
c            print*," p1 =  ",srt(p1sq)
c            print*," p2 =  ",srt(p2sq)
c            print*," p12 = ",srt(psq)
c         endif         
      endif
      deti = 1d0/deti
      Xi(1,1) = p2sq*deti
      Xi(1,2) = -p1p2*deti
      Xi(2,1) = -p1p2*deti
      Xi(2,2) = p1sq*deti
      b12 = B0(1)
      b23 = B0(2)
      b13 = B0(3)
c
      R1 = (f1*C0+b13-b23)/2d0
      R2 = (f2*C0+b12-b13)/2d0
      C11 = Xi(1,1)*R1 + Xi(1,2)*R2
      C12 = Xi(2,1)*R1 + Xi(2,2)*R2
      Cij(1,1) = C11
      Cij(1,2) = C12
      Cij(1,3) = 0
      Cij(1,4) = 0
c
      C24 = -0.5d0*mt**2*C0 + 0.25d0*(1d0+b23-f1*C11-f2*C12)
      R3 = -b13/4d0 + (b23+f1*C11)/2d0 - C24
      R4 = -(b13-b23)/4d0 + f1*C12/2d0
      R5 = -(b12-b13)/4d0 + f2*C11/2d0
      R6 = b13/4d0 + f2*C12/2d0 - C24

      Cij(2,1) = Xi(1,1)*R3 + Xi(1,2)*R5
      Cij(2,3) = Xi(2,1)*R3 + Xi(2,2)*R5
      Cij(2,2) = Xi(2,1)*R4 + Xi(2,2)*R6
      Cij(2,4) = C24

      if (ldebug) then
         c21 = cij(2,1)
         c22 = cij(2,2)
         c23 = cij(2,3)
         C23p= Xi(1,1)*R4 + Xi(1,2)*R6
         print*," C23 comparison ",C23, C23p,C23/(C23p+1d-30)
         print*,"r1:",p1sq*C11+p1p2*c12+0.5d0*(p1sq*c0+b13-b23)
         print*,"r2:",p1p2*C11+p2sq*C12+0.5d0*((p2sq+2*p1p2)*c0+
     &                b12-b13)
c         print*," c24 = ",c24
c         r1 = p1sq*C21+p2sq*C22+2*p1p2*C23-4*c24+b23-mt**2*C0
c         print*,r1,r1/(c24+1d-60)
         print*,"r4f",p1p2*c22+p1sq*c23+(b13-b23)/4d0+p1sq*c12/2d0
         print*,"r4:",p1p2*c22+p1sq*c23-(b13-b23)/4d0+p1sq*c12/2d0
         print*,"r5f",p1p2*c21+p2sq*c23+(b12-b13)/4d0+(2*p1p2+p2sq)*c11/2d0
         print*,"r5:",p1p2*c21+p2sq*c23-(b12-b13)/4d0+(2*p1p2+p2sq)*c11/2d0
      endif

      return
      end

c
c---------------  tens4 = 4point tensors -----------------------------
c
      subroutine tens4( mt, p1sq, p2sq, p3sq, p1p2, p1p3, p2p3, 
     &                  C0123,  C0124,  C0134,  C0234,
     &                  Cij123, Cij124, Cij134, Cij234, 
     &                  D0, Dij)
      implicit none
      double precision mt, msq, p1sq, p2sq, p3sq, p1p2, p1p3, p2p3
      complex*16 C0123,  C0124,  C0134,  C0234
      complex*16 Cij123(2,4), Cij124(2,4), Cij134(2,4), Cij234(2,4)
      complex*16 D0, Dij(3,13)
c
c  determine the Passarino-Veltman tensor decomposition for the 
c  four-point tensor integrals
c
c                                       1
c   D0; D_mu; D_mu,nu; D_mu,nu,rho = ------- Int d^4k
c                                     i pi^2 
c
c           1;  k_mu;   k_mu k_nu; k_mu k_nu k_rho
c   -------------------------------------------------------------------
c   [-k^2+mt^2][-(k+p1)^2+mt^2][-(k+p1+p2)^2+mt^2][-(k+p1+p2+p3)^2+m^2]
c
c  with
c
c   D_mu = p1_mu D11  +  p2_mu D12  +  p3_mu D13
c
c   D_munu = p1_mu p1_nu D21 + p2_mu p2_nu D22 + ...
c
c  for notation see Passarino&Veltman, NP B160 (1979) 151 and my notes
c  on gg-->Hgg, pages 4.9--4.19. Note: since I use Bjorken Drell g_munu
c  one needs to replace pi.pj --> -pi.pj, delta_munu = -g_munu in PV
c  to get the results below
c
C INPUT:  mt, p1sq, p1p2,...                mass and 4-momenta**2
C         C0123 = C0(1,2,3) = C0(p1,p2)     scalar three point 
C         C0124 = C0(1,2,4) = C0(p1,p2+p3)  functions in PV notation
C         C0134 = C0(1,3,4) = C0(p1+p2,p3)
C         C0234 = C0(2,3,4) = C0(p2,p3)
C         Cij123(n,m) = C_nm(1,2,3) ....    higher C_nm form factors
C                                           as in tens3
c         D0 = D0(p1,p2,p3)                 scalar four point function
c
c OUTPUT: Dij(n,m) = D_nm                   form factors in the tensor 
c                                           integrals a la PV
c         nm = 11, 12, 13                   ff"s for D_mu
c         nm = 21, 22, 23, 24, 25, 26, 27   ff"s for D_munu
c         nm = 31, 32, 33, ..., 39, 310, 311, 312  ff"s for D_mu,nu,rho
c
c	Dieter Zeppenfeld, <dieter@pheno.physics.wisc.edu>
c	Initial version:  2000 December 
c	Last modified:    2002 October 29
c	Last modified: 19.09.07, Michael Kubocz   
c       Additional D00ij and D0000 coefficients for Denner-Dittmaier-tensor 
c       decomposition (used in SUBROUTINE DDtens51m in ggf_amp_aux.F) 

      real*8 f1, f2, f3, Xi(3,3), X(3,3), deti
      complex*16 r(20:58), Dijp(3,13), t(9), d310pp
      integer i,j, iscount
      logical ldebug, ldb0
      parameter (ldebug=.false.)
      common /bdebug/ ldb0
      real*8 srt !, dtexm
      external srt
      data iscount /0/
      save iscount

      f1 = p1sq
      f2 = p2sq+2*p1p2
      f3 = p3sq+2*(p1p3+p2p3)
      deti = -p1sq*p2sq*p3sq-2*p1p2*p2p3*p1p3
     &       +p1sq*p2p3**2+p2sq*p1p3**2+p3sq*p1p2**2
      do i=1,3
         do j = 1,13
            Dij(i,j) = 0
         enddo
      enddo
      if (abs(deti).lt.1d-30) then
         iscount = iscount+1
         if (iscount.le.1000) then
            print*," Warning: singular point in TENS4 "
            return
         else
            print*," stop: too many singular points in TENS4 "
            stop
         endif
      endif
      deti = 1/deti
c
      Xi(1,1) = (p2sq*p3sq-p2p3**2)*deti
      Xi(1,2) = (p1p3*p2p3-p3sq*p1p2)*deti
      Xi(1,3) = (p1p2*p2p3-p2sq*p1p3)*deti

      Xi(2,1) = Xi(1,2)
      Xi(2,2) = (p1sq*p3sq-p1p3**2)*deti
      Xi(2,3) = (p1p2*p1p3-p1sq*p2p3)*deti

      Xi(3,1) = Xi(1,3)
      Xi(3,2) = Xi(2,3)
      Xi(3,3) = (p1sq*p2sq-p1p2**2)*deti

      if (ldb0) then
         print("(g15.4)")," det(-X) = ",1/deti
         print*," 1,2,3 ",srt(p1sq),srt(p2sq),srt(p3sq)
c         if (p1sq.ne.0) then
c            dtexm = (p1p2+p2p3)**2*(p1p3+p2p3)**2/
c     1              (p1sq+p2sq+p3sq+2*(p1p2+p1p3+p2p3))
c         elseif (p2sq.ne.0) then
c            dtexm = (p1p2+p1p3)**2*(p1p3+p2p3)**2/
c     1              (p1sq+p2sq+p3sq+2*(p1p2+p1p3+p2p3))
c         else
c            dtexm = (p1p2+p1p3)**2*(p1p2+p2p3)**2/
c     1              (p1sq+p2sq+p3sq+2*(p1p2+p1p3+p2p3))
c         endif
c         print*," detx ratio ",1/deti/dtexm
         print*," 12 ",srt(p1sq+2*p1p2+p2sq)," 13 ",
     1          srt(p1sq+2*p1p3+p3sq)," 23 ",srt(p2sq+2*p2p3+p3sq)
         print*," check inversion of X matrix for n = 3 "
         X(1,1) = -p1sq
         X(1,2) = -p1p2
         X(1,3) = -p1p3
         X(2,1) = -p1p2
         X(2,2) = -p2sq
         X(2,3) = -p2p3
         X(3,1) = -p1p3
         X(3,2) = -p2p3
         X(3,3) = -p3sq
         do i = 1,3
            do j = 1,3
               print*,"X*X^-1(",i,j,") = ", 
     &                X(i,1)*Xi(1,j)+X(i,2)*Xi(2,j)+X(i,3)*Xi(3,j)
            enddo
         enddo
         print*," Xi = "
         do i = 1,3
            print 13,(Xi(i,j),j=1,3)
         enddo
         print*," X = "
         do i = 1,3
            print 13,(X(i,j),j=1,3)
         enddo
 13      format (3g14.4)
      endif
c
      if (ldb0) then
         if (.false.) then
            C0123 = 0
            c0124 = 0
            c0234 = 0
            c0134 = 0
            do i=1,2
               do j = 1,4
                  cij123(i,j) = 0
                  cij124(i,j) = 0
                  cij234(i,j) = 0
                  cij134(i,j) = 0
               enddo
            enddo
         endif
         print*," input for D1j "
         print*," D0 = ",d0
         print*," C0123 = ",C0123
         print*," C0124 = ",C0124
         print*," C0234 = ",C0234
         print*," C0134 = ",C0134
      endif
      R(20) = (f1*D0+C0134-C0234)/2d0
      R(21) = (f2*D0+C0124-C0134)/2d0
      R(22) = (f3*D0+C0123-C0124)/2d0
      Dij(1,1) = Xi(1,1)*R(20) + Xi(1,2)*R(21) + Xi(1,3)*R(22)
      Dij(1,2) = Xi(2,1)*R(20) + Xi(2,2)*R(21) + Xi(2,3)*R(22)
      Dij(1,3) = Xi(3,1)*R(20) + Xi(3,2)*R(21) + Xi(3,3)*R(22)
c      do j = 4,13
c         Dij(1,j) = 0
c      enddo
ccccccc
      if (ldebug) then
         print*," r20 = ",R(20)
         print*," r21 = ",R(21)
         print*," r22 = ",R(22)
         print*,"r20c ",
     &           -(p1sq*dij(1,1)+p1p2*dij(1,2)+p1p3*dij(1,3))/R(20)
         print*,"r21c ",
     &           -(p1p2*dij(1,1)+p2sq*dij(1,2)+p2p3*dij(1,3))/R(21)
         print*,"r22c ",
     &           -(p1p3*dij(1,1)+p2p3*dij(1,2)+p3sq*dij(1,3))/R(22)
      endif
ccccccc
      Dij(2,7) = -mt**2*D0 
     &           - (f1*dij(1,1)+f2*dij(1,2)+f3*dij(1,3)-C0234)/2d0
c
      R(30) = (f1*Dij(1,1)+Cij134(1,1)+C0234)/2d0 - Dij(2,7)
      R(33) = (f1*Dij(1,2)+Cij134(1,1)-Cij234(1,1))/2d0
      R(36) = (f1*Dij(1,3)+Cij134(1,2)-Cij234(1,2))/2d0

      R(31) = (f2*Dij(1,1)+Cij124(1,1)-Cij134(1,1))/2d0
      R(34) = (f2*Dij(1,2)+Cij124(1,2)-Cij134(1,1))/2d0 - Dij(2,7)
      R(37) = (f2*Dij(1,3)+Cij124(1,2)-Cij134(1,2))/2d0

      R(32) = (f3*Dij(1,1)+Cij123(1,1)-Cij124(1,1))/2
      R(35) = (f3*Dij(1,2)+Cij123(1,2)-Cij124(1,2))/2
      R(38) = (f3*Dij(1,3)-Cij124(1,2))/2d0 - Dij(2,7)

      Dij(2,1) = Xi(1,1)*R(30) + Xi(1,2)*R(31) + Xi(1,3)*R(32)
      Dij(2,4) = Xi(2,1)*R(30) + Xi(2,2)*R(31) + Xi(2,3)*R(32)
      Dij(2,5) = Xi(3,1)*R(30) + Xi(3,2)*R(31) + Xi(3,3)*R(32)

      Dij(2,2) = Xi(2,1)*R(33) + Xi(2,2)*R(34) + Xi(2,3)*R(35)
      Dij(2,6) = Xi(3,1)*R(33) + Xi(3,2)*R(34) + Xi(3,3)*R(35)

      Dij(2,3) = Xi(3,1)*R(36) + Xi(3,2)*R(37) + Xi(3,3)*R(38)
cccccccccc
      if (ldebug) then
         Dijp(2,4) =Xi(1,1)*R(33) + Xi(1,2)*R(34) + Xi(1,3)*R(35)
         Dijp(2,5) = Xi(1,1)*R(36) + Xi(1,2)*R(37) + Xi(1,3)*R(38)
         Dijp(2,6) = Xi(2,1)*R(36) + Xi(2,2)*R(37) + Xi(2,3)*R(38)
         print*," ratios of 2 versions for D24, D25, D26 "
         print*," D24 ",dij(2,4),dij(2,4)/dijp(2,4)
         print*," D25 ",dij(2,5),dij(2,5)/dijp(2,5)
         print*," D26 ",dij(2,6),dij(2,6)/dijp(2,6)
      endif
cccccccccc
c      do j = 8,13
c         Dij(2,j) = 0
c      enddo
c
      Dij(3,11) = -mt**2/2d0*dij(1,1) 
     &            - (f1*Dij(2,1)+f2*dij(2,4)+f3*dij(2,5)+C0234)/4d0
      Dij(3,12) = -mt**2/2d0*dij(1,2) 
     &            - (f1*Dij(2,4)+f2*dij(2,2)+f3*dij(2,6)-Cij234(1,1))/4d0
      Dij(3,13) = -mt**2/2d0*dij(1,3) 
     &            - (f1*Dij(2,5)+f2*dij(2,6)+f3*dij(2,3)-Cij234(1,2))/4d0
c
      R(41) = (f1*Dij(2,1)+Cij134(2,1)-C0234)/2d0 - 2*Dij(3,11)
      R(42) = (f2*Dij(2,1)+Cij124(2,1)-Cij134(2,1))/2d0
      R(43) = (f3*Dij(2,1)+Cij123(2,1)-Cij124(2,1))/2d0

      R(44) = (f1*Dij(2,4)+Cij134(2,1)+Cij234(1,1))/2d0 - Dij(3,12)
      R(50) = (f1*Dij(2,2)+Cij134(2,1)-Cij234(2,1))/2d0 
      R(56) = (f1*Dij(2,3)+Cij134(2,2)-Cij234(2,2))/2d0

      R(45) = (f2*Dij(2,4)+Cij124(2,3)-Cij134(2,1))/2d0 - Dij(3,11)
      R(51) = (f2*Dij(2,2)+Cij124(2,2)-Cij134(2,1))/2d0 - 2*Dij(3,12)
      R(57) = (f2*Dij(2,3)+Cij124(2,2)-Cij134(2,2))/2d0 

      R(46) = (f3*Dij(2,4)+Cij123(2,3)-Cij124(2,3))/2d0
      R(52) = (f3*Dij(2,2)+Cij123(2,2)-Cij124(2,2))/2d0
      R(58) = (f3*Dij(2,3)            -Cij124(2,2))/2d0 - 2*Dij(3,13) 

      Dij(3,1) = Xi(1,1)*R(41) + Xi(1,2)*R(42) + Xi(1,3)*R(43)
      Dij(3,4) = Xi(2,1)*R(41) + Xi(2,2)*R(42) + Xi(2,3)*R(43)
      Dij(3,5) = Xi(3,1)*R(41) + Xi(3,2)*R(42) + Xi(3,3)*R(43)

      Dij(3,6) = Xi(1,1)*R(50) + Xi(1,2)*R(51) + Xi(1,3)*R(52)
      Dij(3,2) = Xi(2,1)*R(50) + Xi(2,2)*R(51) + Xi(2,3)*R(52)
      Dij(3,8) = Xi(3,1)*R(50) + Xi(3,2)*R(51) + Xi(3,3)*R(52)

      Dij(3,7) = Xi(1,1)*R(56) + Xi(1,2)*R(57) + Xi(1,3)*R(58)
      Dij(3,9) = Xi(2,1)*R(56) + Xi(2,2)*R(57) + Xi(2,3)*R(58)
      Dij(3,3) = Xi(3,1)*R(56) + Xi(3,2)*R(57) + Xi(3,3)*R(58)

      Dij(3,10) = Xi(3,1)*R(44) + Xi(3,2)*R(45) + Xi(3,3)*R(46)
*----------------------------------------------------------------------
c WARNING: Previously unused Dij(1,7-13) variables are used
c to define D00ij and D0000 functions for DD-E-coefficients in 
c amplitudes/ggf/ggf_amp_aux.F --> subroutine DDtens51m
c In PV notation we have:
c Dij(1,7)=D416, Dij(1,8)=D417, Dij(1,9)=D418
c Dij(1,10)=D419, Dij(1,11)=D420, Dij(1,12)=D421, D(1,13)=D422
      msq=mt*mt
      Dij(1,7)=(-C0234+f1*Dij(3,1)+f2*Dij(3,4)+f3*Dij(3,5)+2d0
     &        *msq*Dij(2,1))/6d0
      Dij(1,8)=(-Cij234(2,1)+f1*Dij(3,6)+f2*Dij(3,2)+f3*Dij(3,8)+2d0
     &        *msq*Dij(2,2))/6d0
      Dij(1,9)=(-Cij234(2,2)+f1*Dij(3,7)+f2*Dij(3,9)+f3*Dij(3,3)+2d0
     &        *msq*Dij(2,3))/6d0
      Dij(1,10)=(Cij234(1,1)+f1*Dij(3,4)+f2*Dij(3,6)+f3*Dij(3,10)+2d0
     &         *msq*Dij(2,4))/6d0
      Dij(1,11)=(Cij234(1,2)+f1*Dij(3,5)+f2*Dij(3,10)+f3*Dij(3,7)+2d0
     &         *msq*Dij(2,5))/6d0
      Dij(1,12)=(-Cij234(2,3)+f1*Dij(3,10)+f2*Dij(3,8)+f3*Dij(3,9)+2d0
     &         *msq*Dij(2,6))/6d0
      Dij(1,13)=(Cij234(2,4)+1d0/12d0-p1sq*Dij(1,7)-p2sq*Dij(1,8)-p3sq
     &         *Dij(1,9)-2d0*p1p2*Dij(1,10)-2d0*p1p3*Dij(1,11)-2d0*p2p3
     &         *Dij(1,12)-msq*Dij(2,7))/6d0 
*----------------------------------------------------------------------
      if (ldb0) then
         print*," R_20/R_22 = ",r(20)/r(22)
         print*," R_30/R_32 = ",r(30)/r(32)
         print*," R_33/R_35 = ",r(33)/r(35)
         print*," R_36/R_38 = ",r(36)/r(38)

         print*," R_41/R_43 = ",r(41)/r(43)
         print*," R_50/R_52 = ",r(50)/r(52)
         print*," R_56/R_58 = ",r(56)/r(58)
         print*," R_44/R_46 = ",r(44)/r(46)
      endif
      if (ldebug) then
         Dijp(3,4) = Xi(1,1)*R(44) + Xi(1,2)*R(45) + Xi(1,3)*R(46)
         Dijp(3,6) = Xi(2,1)*R(44) + Xi(2,2)*R(45) + Xi(2,3)*R(46)
         print*," ratios of 2 versions for D3j "
         print*," D34 ",dij(3,4),dij(3,4)/dijp(3,4)
         print*," D36 ",dij(3,6),dij(3,6)/dijp(3,6)
         t(1) = (cij134(2,3)+f1*Dij(2,5)+cij234(1,2))/2-dij(3,13)
         t(2) = (cij134(2,3)+f1*Dij(2,6)-cij234(2,3))/2
         t(3) = (cij134(2,4)+f1*Dij(2,7)-cij234(2,4))/2

         t(4) = (cij124(2,3)-cij134(2,3)+f2*Dij(2,5))/2
         t(5) = (cij124(2,2)-cij134(2,3)+f2*Dij(2,6))/2-dij(3,13)
         t(6) = (cij124(2,4)-cij134(2,4)+f2*Dij(2,7))/2

         t(7) = (           -cij124(2,3)+f3*Dij(2,5))/2-dij(3,11)
         t(8) = (           -cij124(2,2)+f3*Dij(2,6))/2-dij(3,12)
         t(9) = (cij123(2,4)-cij124(2,4)+f3*Dij(2,7))/2

         Dijp(3,5) = Xi(1,1)*t(1) + Xi(1,2)*t(4) + Xi(1,3)*t(7)
         Dijp(3,10)= Xi(2,1)*t(1) + Xi(2,2)*t(4) + Xi(2,3)*t(7)
         Dijp(3,7) = Xi(3,1)*t(1) + Xi(3,2)*t(4) + Xi(3,3)*t(7)

         D310pp    = Xi(1,1)*t(2) + Xi(1,2)*t(5) + Xi(1,3)*t(8)
         Dijp(3,8) = Xi(2,1)*t(2) + Xi(2,2)*t(5) + Xi(2,3)*t(8)
         Dijp(3,9) = Xi(3,1)*t(2) + Xi(3,2)*t(5) + Xi(3,3)*t(8)
         
         Dijp(3,11)= Xi(1,1)*t(3) + Xi(1,2)*t(6) + Xi(1,3)*t(9)
         Dijp(3,12)= Xi(2,1)*t(3) + Xi(2,2)*t(6) + Xi(2,3)*t(9)
         Dijp(3,13)= Xi(3,1)*t(3) + Xi(3,2)*t(6) + Xi(3,3)*t(9)

         print*," D35 ",dij(3,5),dij(3,5)/dijp(3,5)
         print*," D37 ",dij(3,7),dij(3,7)/dijp(3,7)
         print*," D38 ",dij(3,8),dij(3,8)/dijp(3,8)
         print*," D39 ",dij(3,9),dij(3,9)/dijp(3,9)
         print*," D310 ",dij(3,10),dij(3,10)/dijp(3,10)
         print*," D310pp ",d310pp,d310pp/dij(3,10)
         print*," D311 ",dij(3,11),dij(3,11)/dijp(3,11)
         print*," D312 ",dij(3,12),dij(3,12)/dijp(3,12)
         print*," D313 ",dij(3,13),dij(3,13)/dijp(3,13)
      endif
cccccccccccc
      return
      end


c
c-------------  spence function  li2(z) ------------------------
c
      complex*16 function li2(zin)
      implicit none
      complex*16 zin, z, u, u2, unpo, ans, zext
      double precision r, r2, r2n, fac
c
c determine the value of the dilogarithm 
c
c    li2(z) = - int_0^1  log(1-zt)/t dt  with cut along the positive 
c                                        real axis, z>1
c
c	Dieter Zeppenfeld, <dieter@pheno.physics.wisc.edu>
c	Initial version:  2000 November 6
c	Last modified:    2000 November 12
c  
      integer i
      double precision c0,c1,c2,c4,c6,c8,c10,c12,c14,c16,c18,c20,c22
      double precision b0,b1,b2,b4,b6,b8,b10,b12,b14,b16,b18,b20,b22
      double precision d0,d1,d2,d4,d6,d8,d10,d12,d14,d16,d18,d20,d22
      parameter (b0=1d0,            d0 =1d0,      c0= b0/d0)
      parameter (b1=-1d0/2d0,       d1 =d0*2d0,   c1= b1/d1)
      parameter (b2= 1d0/6d0,       d2 =d1*3d0,   c2= b2/d2)
      parameter (b4=-1d0/30d0,      d4 =d2*20d0,  c4= b4/d4)
      parameter (b6=1d0/42d0,       d6 =d4*42d0,  c6= b6/d6)
      parameter (b8=-1d0/30d0,      d8 =d6*72d0,  c8= b8/d8)
      parameter (b10=5d0/66d0,      d10=d8*110d0, c10=b10/d10)
      parameter (b12=-691d0/2730d0, d12=d10*156d0,c12=b12/d12)
      parameter (b14=7d0/6d0,       d14=d12*210d0,c14=b14/d14)
      parameter (b16=-3617d0/510d0, d16=d14*272d0,c16=b16/d16)
      parameter (b18=43867d0/798d0, d18=d16*342d0,c18=b18/d18)
      parameter (b20=-174611d0/330d0,d20=d18*420d0,c20=b20/d20)
      parameter (b22=854513d0/138d0,d22=d20*506d0,c22=b22/d22)
      double precision eps, epst, pi, pi2o6
      parameter (eps=1d-16, epst=1d-3)
      parameter (pi=3.14159 26535 89793238d0, pi2o6=pi**2/6d0)
c
c debug information
      logical ldebug
      parameter (ldebug=.false.)

      z = zin
c      print*," li2 call with z = ",z
      u = z**2
      r2 = dreal(z)**2+dimag(z)**2 
      if (r2.lt.eps) then
         li2 = z + u/4d0
         return
      elseif (r2.lt.epst) then
         ans = z + u/4d0
         do i = 3,11
            u = u*z
            ans = ans + u/i**2
         enddo
         li2 = ans
         return
      endif
      if (dreal(z).ge.1d0 .and. dimag(z).eq.0 ) then
         z = z + (0d0,1d0)*eps
      endif
c
c use z-->1/z and z--> 1-z mappings of the spence function to restrict 
c agument to unit circle in the complex plane with Re(z) <= 0.5
c
      zext = (0d0,0d0)
      fac = 1
      if (r2.gt.1d0) then     ! map z ---> 1/z
         fac = -fac
         zext = -pi2o6 - 0.5d0*(log(-z))**2
         z = 1d0/z
      endif
      if (dreal(z).gt.0.5d0) then     ! map new z ---> 1-z
         zext = zext + fac*(pi2o6-log(z)*log(1-z))
         fac = -fac
         z = 1-z
      endif
c
c now use t = 1 - exp(-u) mapping to write Li(z) in terms of Bernoulli 
c numbers
c
      u = - log(1-z)
      r2 = abs(u)**2
      u2 = u*u
      ans = u*(c0 + u*(c1+c2*u))
      r2n = r2*r2       !r^4

      unpo = u2*u2*u
      ans = ans + c4*unpo

      unpo = unpo*u2
      ans = ans + c6*unpo

      r = r2n*r2n       !r^8
      unpo = unpo*u2
      ans = ans + c8*unpo

      r2n = r*r2        !r^10 
      if ((r2n*c10).gt.eps) then
         unpo = unpo*u2
         ans = ans + c10*unpo
      else
         li2 = fac * ans + zext
         if (ldebug) print*," exit li2s at n=8 "
         return
      endif

      unpo = unpo*u2
      ans = ans + c12*unpo

      unpo = unpo*u2
      ans = ans + c14*unpo

      unpo = unpo*u2
      ans = ans + c16*unpo

      r2n = r2n*r
      if ((r2n*c18).gt.eps) then
         unpo = unpo*u2
         ans = ans + c18*unpo
      else
         li2 = fac * ans + zext
         if (ldebug) print*," exit li2s at n=16 "
         return
      endif

      unpo = unpo*u2
      ans = ans + c20*unpo

      unpo = unpo*u2
      ans = ans + c22*unpo

      li2 = fac * ans + zext
      end

c----  B0t(qsq): regularized 2-point function for massless loop  -----
c
      complex*16 function B0t(qsq,musq)
      implicit none
      double precision qsq, musq

c evaluate the finite piece of the massless B0 function
c  
c             d^d k 
c    B0 = Int------- [-k^2+idelta]^-1 [-(k-q)^2+idelta]^-1 
c             i pi^2
c
c       = pi^-eps Gamma(1+eps) musq^-eps [1/eps + B0t(qsq,musq)]
c
c i.e. B0t(qsq,musq) = 2 - log( (-qsq-idelta)/musq )
c
c subroutine requires positive musq
c
c	Dieter Zeppenfeld, <dieter@pheno.physics.wisc.edu>
c	Initial version:  2002 September 13
c	Last modified:    2002 September 13
c  
      complex*16 ipi
      parameter (ipi=(0d0,3.14159 26535 89793))
      logical ldebug
      parameter (ldebug=.true.)
c
      if (ldebug) then
         if (musq.le.0d0) then
            print*," Unacceptable mu^2 in B0t(q^2,mu^2): mu^2 = ",
     1             musq," ==> set B0t = 0 "
            b0t=0
            return
         endif
      endif
c
      if (qsq.lt.0d0) then
         b0t = 2d0 - log(-qsq/musq)
      else
         b0t = 2d0 - log(qsq/musq) + ipi
      endif
      end

c
c------------- C0(q1^2,q2^2,P^2)  massless 3-point fuction ------------
c
      complex*16 function C0t(q1sq,q2sq,Psq,musq)
      implicit none
      double precision q1sq, q2sq, psq, musq

c evaluate the finite part of the scalar 3-point function for zero mass
c propagators 
c  
c  C0 = 1/(i*pi^2) * Int d^dk [-k^2-i0]^-1 [-(k+q1)^2-i0]^-1 [-(k-q2)^2-i0]^-1
c
c       = pi^-eps Gamma(1+eps) musq^-eps [IR + C0t(q1sq,q2sq,psq)]
c
c where IR represents the potentially divergent terms when one or more
c of the arguments vanish. P = q1+q2. The divergent terms are
c 
c (1) IR = 0              all 3 of q1sq,q2sq,psq nonzero
c 
c (2) IR = 1/(t-qsq)*log(t/qsq)*1/eps    for 1 vanishing argument, the
c                                        other two being t and qsq
c
c (3) IR = -1/qsq*(1/eps**2 - 1/eps log(qsq/musq))  when exactly one 
c                                                   argument, qsq,
c                                                   is nonzero
c
c	Dieter Zeppenfeld, <dieter@pheno.physics.wisc.edu>
c	Initial version:  2002 September 13
c	Last modified:    2002 October 25
c  
      complex*16 ipi, z, res, li2
      parameter (ipi=(0d0,3.14159 26535 89793d0))
      external li2
      double precision pi,pi2o3,pi2o6,pi2o2
      parameter (pi=3.14159 26535 89793d0,pi2o3=pi**2/3d0)
      parameter (pi2o6=pi**2/6d0,pi2o2=pi**2/2d0)
      logical ldebug, ltaylor, first_call
      parameter (ldebug=.false.)
      double precision x, flambda, tsqlami, qsq(3), 
     1                 sum, rz, logz, ln1, ln2,
     2                 musq_t, asum, fac, resr, ser, fb(8), xn
      integer izero, inz, i, j
      data first_call /.true./
      save first_call, fb
c
      if (ldebug) then
         if (musq.le.0d0) then
            print*," Unacceptable mu^2 in C0t(q1sq,q2sq,psq,mu^2): ",
     1         "mu^2 = ",musq," ==> set C0t = 0 "
            C0t=0
            return
         endif
      endif
c
      inz = 0
      izero = 0
      do i=1,3
         qsq(i)=0
      enddo
      if (q1sq.eq.0d0) then
         izero = izero+1
      else
         inz = inz+1
         qsq(inz)=q1sq
      endif
      if (q2sq.eq.0d0) then
         izero = izero+1
      else
         inz = inz+1
         qsq(inz)=q2sq
      endif
      if (psq.eq.0d0) then
         izero = izero+1
      else
         inz = inz+1
         qsq(inz)=psq
      endif
      if (ldebug) then
         print("(a,3g12.4,a,g12.4)"), "qsq_i = ",q1sq,q2sq,psq,
     1                                " mu^2 = ",musq
         print*," number of q_i^2 = 0  ",izero
         print*," number of q_i^2 ne 0 ",inz," Sum = ",inz+izero
         if ((inz+izero).ne.3) print*," WARNING: inz+izero .ne. 3 "
      endif
c
      if (izero.eq.0) then   ! all qsq(i) nonzero, this is full C0
         flambda = (qsq(1)**2+qsq(2)**2+qsq(3)**2) - 
     1           2*(qsq(1)*qsq(2)+qsq(1)*qsq(3)+qsq(2)*qsq(3))
         sum = qsq(1)+qsq(2)+qsq(3)
         asum = abs(qsq(1))+abs(qsq(2))+abs(qsq(3))
         ltaylor = abs(abs(sum)/asum-1).lt.1d-13      ! same sign for all qsq
         asum=min(abs(sum-2*qsq(1)),abs(sum-2*qsq(2)),
     1            abs(sum-2*qsq(3)))
         ltaylor = ltaylor .and. abs(flambda).lt.(0.01*asum**2)
c
         res = (0d0,0d0)
         if (ldebug) then
            print*," lambda = ",flambda/4d0," x_taylor = ",
     1                          flambda/asum**2
            ltaylor=.false.
         endif
         if (flambda.gt.0d0 .and. .not.ltaylor) then
            tsqlami = 1d0/sqrt(flambda)
            do i=1,3
               x = (sum-2*qsq(i))*tsqlami
               if (x.gt.1d0) then
                  rz = (x-1)/(x+1)
                  z = rz
                  logz = log(rz)
                  res = res + 2*Li2(z)+0.5d0*logz**2-pi2o3 
                  if (qsq(i).gt.0d0) then
                     res = res + ipi*logz
                  else
                     res = res - ipi*logz
                  endif
               elseif (x.gt.0d0) then
                  rz = (x-1)/(x+1)
                  z = rz
                  logz = log(-rz)
                  res = res + 2*Li2(z)+0.5d0*logz**2+pi2o6
               elseif (x.gt.-1d0) then
                  rz = (x+1)/(x-1)
                  z = rz
                  logz = -log(-rz)
                  res = res - 2*Li2(z)-0.5d0*logz**2-pi2o6
               else 
                  rz = (x+1)/(x-1)
                  z = rz
                  logz = -log(rz)
                  res = res - 2*Li2(z)-0.5d0*logz**2+pi2o3 
                  if (qsq(i).gt.0d0) then
                     res = res + ipi*logz
                  else
                     res = res - ipi*logz
                  endif
               endif
               if (ldebug) print*," i,x, z = ",i,x,rz
            enddo
            C0t = res*tsqlami
         elseif (flambda.lt.0d0 .and. .not.ltaylor) then  ! the lambda < 0 case
            tsqlami = 1d0/sqrt(-flambda)         ! + imag part of 2 sqrt(lambda)
            do i=1,3
               x = (sum-2*qsq(i))*tsqlami
               z = dcmplx(x,-1d0)/dcmplx(x,1d0)
               if (ldebug) then  
                  print*," i,x, z = ",i,x,z
               endif
               res = res + dimag(Li2(z))
            enddo
            C0t = res*2*tsqlami
         endif
         if (ldebug) then
            print*," C0t via Li2 calls: ",C0t
            ltaylor = abs(flambda).lt.(0.5*asum**2)
         endif
         if (ltaylor) then
            if (first_call) then
               fb(1) = 2d0/9d0
               fb(2) = 16d0/75d0
               fb(3) = 142d0/735d0
               fb(4) = 496d0/2835d0
               fb(5) = 6086d0/38115d0
               fb(6) = 86048d0/585585d0
               fb(7) = 92054d0/675675d0
               fb(8) = 1655008d0/13018005d0 
c               fb(9) = 32976682d0/276441165d0
               first_call = .false.
            endif
            musq_t = abs(sum/3d0)
            resr = 0
            do i = 1,3
               fac = 1/(sum-2*qsq(i))
               x = flambda*fac**2
c               if (ldebug) print*," i = ",i," x = ",x
               ln1 = log(abs(qsq(i))/musq_t)
               ser = ln1
               j = 1
               xn = x
c               if (ldebug) print*," log = ",ln1
               do while (abs(xn).gt.1d-15 .and. j.lt.9)
                  ser = ser + xn*(ln1/(2*j+1)-fb(j))
c                  if (ldebug) then
c                     print*," n = ",j," a_n = ",ln1/(2*j+1)-fb(j)
c                  endif
                  xn = xn*x
                  j = j+1
               enddo
               if (ldebug) print*," C_i = ",2*ser*fac
               resr = resr + 2*ser*fac
            enddo
            if (ldebug) then
               print*," C0t via Li2 calls: ",C0t
               print*," C0t via Taylor exp.",dcmplx(resr,0d0)
            endif
            C0t = dcmplx(resr,0d0)
         endif
      elseif (izero.eq.1) then
         ln1 = log(abs(qsq(1))/musq)
         ln2 = log(abs(qsq(2))/musq)
         res = 0.5d0*(ln1**2-ln2**2)
         if (qsq(1).gt.0d0) then
            res = res - ipi*ln1 - pi2o2
         endif
         if (qsq(2).gt.0d0) then
            res = res + ipi*ln2 + pi2o2
         endif
         C0t = res/(qsq(2)-qsq(1))
      elseif (izero.eq.2) then
         ln1 = log(abs(qsq(1))/musq)
         res = 0.5d0*ln1**2-pi2o6
         if (qsq(1).gt.0d0) then
            res = res - ipi*ln1 - pi2o2
         endif
         C0t = -res/qsq(1)
      else
         print*," WARNING: C0t called with 3 zero arguments "
         print*," C0t reset to 0 "
         C0t = 0
      endif
      if (ldebug) then
         print*," output: C0t = ",C0t
      endif
      return 
      end
c

c
c------------- C(m,q1^2,q2^2,P^2)  3-point fuction ---------------------------
c
      complex*16 function I3point(m,q1sq,q2sq,Psq)
      implicit none
      double precision m, q1sq, q2sq, psq 

c evaluate scalar 3-point function for equal masses m on propagators 
c  
c  I3 = 1/(i*pi^2) * Int d^4k [k^2-m^2]^-1 [(k+q1)^2-m^2]^-1 [(k-q2)^2-m^2]^-1
c
c     = i/pi^2 C("t Hooft,Veltman) = -C0(Passarino,Veltman)
c
c in terms of 12 dilogarithms. P = q1+q2 is assumed to be space-like. q1 
c and q2 may be space-, time- or lightlike
c
c	Dieter Zeppenfeld, <dieter@pheno.physics.wisc.edu>
c	Initial version:  2000 November 12
c	Last modified:    
c  

      double precision lambda, rtlam, q1q2, q1p, q2p, fmsq
      double precision rp(3), rm(3), eps, precise, pi

      complex*16 sp(3), sm(3), fmt2, ieps, fac, z, di, li2l, li2
      external li2, li2l
      parameter (eps=1d-18, ieps = eps*(0,1), precise=1d-3 )
      parameter (pi=3.14159 26535 89793238d0, fac=-1d0/2d0)

c determine dot products of input momenta
      q1q2 = 0.5d0*(psq - q1sq - q2sq)
      q1p  = 0.5d0*(psq + q1sq - q2sq)
      q2p  = 0.5d0*(psq - q1sq + q2sq)
      lambda = q1q2**2-q1sq*q2sq
      
      
      if (lambda.le.0d0) then
         write(*,*) " singular lambda in 3-point function. Reset to 0 "
         I3point = 0
         return
      else
         rtlam = sqrt(lambda)
      endif
      
      
      
c determine factors for call of spence functions
      fmsq = 4*m**2
      fmt2 = fmsq - ieps
      di = q1sq*q2sq*psq + fmt2*lambda 
      if ( abs(dreal(di)).gt.eps )then
         di = 1d0/di
      else
         print*," singular point called in I3point "
         stop
      endif

      rp(1) = q2p + rtlam
      rm(1) = q2p - rtlam
      if (abs(rm(1)).lt.abs(rp(1))*precise ) then
         rm(1) = q2sq*psq/rp(1)
      elseif (abs(rp(1)).lt.abs(rm(1))*precise ) then
         rp(1) = q2sq*psq/rm(1)
      endif

      rp(2) = q1q2 + rtlam
      rm(2) = q1q2 - rtlam
      if (abs(rm(2)).lt.abs(rp(2))*precise ) then
         rm(2) = q1sq*q2sq/rp(2)
      elseif (abs(rp(2)).lt.abs(rm(2))*precise ) then
         rp(2) = q1sq*q2sq/rm(2)
      endif

      rp(3) = q1p + rtlam
      rm(3) = q1p - rtlam
      if (abs(rm(3)).lt.abs(rp(3))*precise ) then
         rm(3) = q1sq*psq/rp(3)
      elseif (abs(rp(3)).lt.abs(rm(3))*precise ) then
         rp(3) = q1sq*psq/rm(3)
      endif

      if (q1sq.ne.0d0) then
         z = rtlam*sqrt((q1sq-fmt2)*q1sq) 
         sp(1) = di*( q2p*q1sq + z )
         sm(1) = di*( q2p*q1sq - z )
      else
         sp(1) = 0
         sm(1) = 0
      endif

      if (psq.ne.0d0) then
         z = rtlam*sqrt((psq-fmt2)*psq) 
         sp(2) = di*( q1q2*psq + z )
         sm(2) = di*( q1q2*psq - z )
      else
         sp(2) = 0
         sm(2) = 0
      endif

      if (q2sq.ne.0d0) then
         z = rtlam*sqrt((q2sq-fmt2)*q2sq) 
         sp(3) = di*( q1p*q2sq + z )
         sm(3) = di*( q1p*q2sq - z )
      else
         sp(3) = 0
         sm(3) = 0
      endif

      z = 0
      if (q1sq.lt.0d0 .or. q1sq.gt.fmsq) then
         z = z + li2(rm(1)*sp(1)) + li2(rm(1)*sm(1)) - 
     &           li2(rp(1)*sm(1)) - li2(rp(1)*sp(1))
      elseif (q1sq.gt.0d0) then
         z = z + 2*dreal( li2(rm(1)*sp(1)) ) - 
     &           2*dreal( li2(rp(1)*sp(1)) )
      endif
      if (psq.lt.0d0 .or. psq.gt.fmsq) then
         z = z - li2(rm(2)*sp(2)) - li2(rm(2)*sm(2)) + 
     &           li2(rp(2)*sm(2)) + li2(rp(2)*sp(2))
      elseif (psq.gt.0d0) then
         z = z - 2*dreal( li2(rm(2)*sp(2)) ) + 
     &           2*dreal( li2(rp(2)*sp(2)) )
      endif
      if (q2sq.lt.0d0 .or. q2sq.gt.fmsq) then
         z = z + li2(rm(3)*sp(3)) + li2(rm(3)*sm(3)) - 
     &           li2(rp(3)*sm(3)) - li2(rp(3)*sp(3))
      elseif (q2sq.gt.0d0) then
         z = z + 2*dreal( li2(rm(3)*sp(3)) ) - 
     &           2*dreal( li2(rp(3)*sp(3)) )
      endif

      I3point = z*fac/rtlam
      return
      end

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                tens_red3 = 3-point tensors
C     Francisco Campanario
C    Email: Francam@particle.uni-karlsruhe.de
C Date: 4/04/2008
C  Tensor reduction a la Passarino-Vetlman for the tensor
C In this file, we should put as an input the 3 scalar invariants, 
C the 3 B0, and C0.
C Notation: Example CR. the "R" means the real part of C0
C                   CI, the "I" means the imaginary part of C0
C It speeds up the code. 
C OUTPUT:  is given in C0r,C0i,Cijr,CijI. "r":  real part, "I" imaginary part
C NOTE!!!! The code seems to be 3 times faster than the old one..
C The indices are pertumted in comparison with the old subroutines.
C the rank of the tensor integral is store in the second argument.
C C_old(1,2)-> C(2,1)(rank 1 coefficient 2) it is faster in this way. 
C I was not able to find this recursion relations in the literature.
C So, one has to look into my notes if want to generalize them.
C I have more version which only needs the input in complex form
C and the output is given as a complex array. The subroutine still makes the separation
C between imaginary part and real part, but it is done inside of it. Here I am saving doing that since 
C I need and use this CR and CI in my penline and boxline subroutines.
C If you need other case ask me!
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      subroutine tens_red3_new_Re_Com_1M(m,p1sq,p2sq,s12,B0_23,B0_13,B0_12, 
     #                     C0,Cij)
c$$$     # C0r,C0I,Cijr,CijI)
      implicit none
      real * 8  p1sq, p2sq, s12,m
      complex*16 B0_23, B0_13, B0_12, C0, Cij(4,2)
c
c  determine the Passarino-Veltman tensor decomposition for the three-point
c  tensor integrals
c
c                       d^4k           1;  k_mu;   k_mu k_nu
c C0;C_mu;C_munu =Int -------  -----------------------------------------
c                     (2pi)^4    [k^2-m^2][(k+p1)^2-m^2][(k+p1+p2)^2-m^2] 
c
c with
c
c    C_mu = p1_mu C11  +  p_2_mu C12
c
c  C_munu = p1_mu p1_nu C21 + p2_mu p2_nu C22 + 
c           (p1_mu p2_nu + p1_nu p2_mu) C23  -  g_munu C24
c
c  for notation see Passarino&Veltman, NP B160 (1979) 151 and my notes
c
C INPUT:  p1sq, p2sq, s12          external invariants: p1^2, p2^2, s12=(p1+p2)^2
C         B0, C0                   4 scalar integrals; the 3 B0 are, 
c                                  in PV notation:
c         B0(1) = B0(1,2) = B0(p1)  B_0 function with subtraction of 
c         B0(2) = B0(2,3) = B0(p2)  divergent term
c         B0(3) = B0(1,3) = B0(s12)
c
c OUTPUT: Cij(n,m) = C_nm          form factors in the tensor integrals
c          n=1,2,3,4; n=1,2        a la PV
c
      real*8 r1, r2r1, deter,Ideter,p1p2,IZ(2,2),msq,detAbs
      real*8 B1r_12,B1r_13,B1r_23,Rr(2),Cijr(4,2)
      real*8 B1I_12,B1I_13,B1I_23,RI(2),CijI(4,2)
      real*8 B0r_23, B0r_13, B0r_12, C0r
      real*8 B0I_23, B0I_13, B0I_12, C0I
      logical Singular 
      COMMON /GramCheck/ Singular

      p1p2 = (s12 - p1sq - p2sq)*0.5d0

      r1 = p1sq
      r2r1 = s12 - r1
      

      deter = 2.d0*(p1sq*p2sq - p1p2*p1p2)
      detAbs=  2.d0*(abs(p1sq*p2sq)+abs(p1p2*p1p2))
      If( abs(deter)/abs(detAbs).lt.10d-5) then
         Singular=.true.
      endif


      Ideter=1.d0/deter
     
      IZ(1,1) =  p2sq*Ideter
      IZ(2,1) = -p1p2*Ideter
      IZ(1,2) =  IZ(2,1)
      IZ(2,2) =  p1sq*Ideter

     
      B0r_12=Dble(B0_12)
      B0r_13=Dble(B0_13)
      B0r_23=Dble(B0_23)
      C0r=Dble(C0)

      B0I_12=DImag(B0_12)
      B0I_13=DImag(B0_13)
      B0I_23=DImag(B0_23)
      C0I=DImag(C0)

      B1r_12 = -B0r_12*0.5d0
      B1r_13 = -B0r_13*0.5d0
      B1r_23 = -B0r_23*0.5d0
      
      B1I_12 = -B0I_12*0.5d0
      B1I_13 = -B0I_13*0.5d0
      B1I_23 = -B0I_23*0.5d0
      



c 1-2
      Rr(1) = (B0r_13 - B0r_23 - C0r*r1)
      Rr(2) = (B0r_12 - B0r_13 - C0r*r2r1)
      
      RI(1) = (B0I_13 - B0I_23 - C0I*r1)
      RI(2) = (B0I_12 - B0I_13 - C0I*r2r1)
      

      Cijr(1,1) = (Rr(1)*IZ(1,1) + Rr(2)*IZ(2,1))
      Cijr(2,1) = (Rr(1)*IZ(1,2) + Rr(2)*IZ(2,2))
      
      CijI(1,1) = (RI(1)*IZ(1,1) + RI(2)*IZ(2,1))
      CijI(2,1) = (RI(1)*IZ(1,2) + RI(2)*IZ(2,2))
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
c C00
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      msq=m*m
      Cijr(4,2) = ( B0r_23 +2*msq*C0r + Cijr(1,1)*r1 +Cijr(2,1)*r2r1 +1.d0)*0.25d0
      CijI(4,2) = ( B0I_23 +2*msq*C0I + CijI(1,1)*r1 +CijI(2,1)*r2r1)*0.25d0
      
c 3-5
      Rr(1) = (B1r_13 + B0r_23 - Cijr(1,1)*r1 - Cijr(4,2)*2.d0)
      Rr(2) = (B1r_12 - B1r_13 - Cijr(1,1)*r2r1)
     
      RI(1) = (B1I_13 + B0I_23 - CijI(1,1)*r1 - CijI(4,2)*2.d0)
      RI(2) = (B1I_12 - B1I_13 - CijI(1,1)*r2r1)
      
      Cijr(1,2) =(Rr(1)*IZ(1,1)+Rr(2)*IZ(2,1))
      CijI(1,2) =(RI(1)*IZ(1,1)+RI(2)*IZ(2,1))


c 4-6
      Rr(1) = (  B1r_13 - B1r_23 - Cijr(2,1)*r1)
      Rr(2) = (- B1r_13          -Cijr(2,1)*r2r1 -Cijr(4,2)*2.d0)
      
      RI(1) = (  B1I_13 - B1I_23 - CijI(2,1)*r1)
      RI(2) = (- B1I_13          - CijI(2,1)*r2r1 -CijI(4,2)*2.d0)

      Cijr(2,2) = (Rr(1)*IZ(1,2)+Rr(2)*IZ(2,2))
      Cijr(3,2) = (Rr(1)*IZ(1,1)+Rr(2)*IZ(2,1))

      CijI(2,2) = (RI(1)*IZ(1,2)+RI(2)*IZ(2,2))
      CijI(3,2) = (RI(1)*IZ(1,1)+RI(2)*IZ(2,1))
     
      Cij(1,1)=DCMPLX(Cijr(1,1),CijI(1,1))
      Cij(2,1)=DCMPLX(Cijr(2,1),CijI(2,1))

      Cij(1,2)=DCMPLX(Cijr(1,2),CijI(1,2))
      Cij(2,2)=DCMPLX(Cijr(2,2),CijI(2,2))
      Cij(3,2)=DCMPLX(Cijr(3,2),CijI(3,2))
      Cij(4,2)=DCMPLX(Cijr(4,2),CijI(4,2))

      

      end

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                tens_red4 = 4-point tensors
C     Francisco Campanario
C    Email: Francam@particle.uni-karlsruhe.de
C Date: 4/04/2008
C  Tensor reduction a la Passarino-Vetlman for the tensor
C In this file, we should put as an input the 6 scalar invariants, 
C the 4 C0, and the Cij got it from tens_red3_new_Re_Com and D0.
C Notation: Example CR234. the "R" means the real part of C0234
C                   CI234, the "I" means the imaginary part of C0234
C This separation is done already in tens_red3_....
C It speeds up the code. 
C OUTPUT:  is given in D0r,D0i,Dijr,DijI. "r":  real part, "I" imaginary part
C NOTE!!!! The code seems to be 3 to 4 times faster than the old one..
C The indices are pertumted in comparison with the old subroutines.
C the rank of the tensor integral is store in the second argument.
C C_old(1,2)-> C(2,1)(rank 1 coefficient 2) it is faster in this way. 
C I was not able to find this recursion relations in the literature.
C One has to look into my notes if want to generalize them.
C I have more version which only needs the input in complex form
C and the output is given as a complex array. The subroutine still makes the separation
C between imaginary part and real part, but it is done inside of it. Here I am saving doing that since 
C I need and use this CR and CI in my penline and boxline subroutines.
C If you need other case ask me!
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      subroutine tens_red4_new_Re_Com_1M(m,p1sq,p2sq,p3sq,p1p2,p1p3,p2p3, 
     #                     C0r_234, C0r_134, C0r_124, C0r_123,   
     #                     Cijr_234, Cijr_134, Cijr_124, Cijr_123,             
     #                     C0I_234, C0I_134, C0I_124, C0I_123,   
     #                     CijI_234, CijI_134, CijI_124, CijI_123,   
     #                     D0,D0r,D0I,Dij)
c$$$     #   Dijr,DijI)
      implicit none
      real * 8 p1sq, p2sq, p3sq, p1p2, p1p3,p2p3,m,msq
c$$$      complex*16 C0_234, C0_134, C0_124, C0_123
c$$$      complex*16 Cij_234(4,2), Cij_134(4,2), Cij_124(4,2), Cij_123(4,2)
      complex*16 D0, Dij(13,3)
c, Dijp(13,3), Dijpp(13,3)
c      complex*16 t(9)
c
c  determine the Passarino-Veltman tensor decomposition for the four-point
c  tensor integrals
c
c                                          d^4k
c   D0; D_mu; D_mu,nu; D_mu,nu,rho =  Int ------ 
c                                         (2pi)^4
c
c              1;  k_mu;   k_mu k_nu; k_mu k_nu k_rho
c      -------------------------------------------------------------------
c         [k^2-m^2][(k+p)^2-m^2][(k+p+q)^2-m^2][(k+p+q+l)^2-m^2]
c with
c
c   D_mu = p1_mu D11  +  p2_mu D12  +  p3_mu D13
c
c   D_munu = p1_mu p2_nu D21 + p2_mu p2_nu D22 + ...
c
c  for notation see Passarino&Veltman, NP B160 (1979) 151 
c
C INPUT:  psq, pq,...                        kinematics invariants
C         C0_123 = C0(1,2,3) = C0(p1,p2)     scalar three point 
C         C0_124 = C0(1,2,4) = C0(p1,p2+p3)  functions in PV notation
C         C0_134 = C0(1,3,4) = C0(p1+p2,p3)
C         C0_234 = C0(2,3,4) = C0(p2,p3)
C         Cij_123(n,m) = C_nm(1,2,3) ....    higher C_nm form factors
C                                            as in tens_red3
c         D0 = D0(p,q,l)                  scalar four point function
c
c OUTPUT: Dij(n,m) = D_nm                    form factors in the tensor 
c                                            integrals a la PV
c         nm = 11, 21, 31                    ff's for D_mu
c         nm = 12, 22, 32, 42, 52, 62, 72    ff's for D_munu
c         nm = 13, 23, 33, ..., 93, 103, 113, 123  ff's for D_mu,nu,rho
c
      real*8 r1, r2r1, r3r2, IZ(3,3), deter,Ideter,Iv,Ivv,detAbs
      real*8 Rr(1:3), RI(1:3)
      real*8 C0r_234, C0r_134, C0r_124, C0r_123
      real*8 Cijr_234(4,2), Cijr_134(4,2), Cijr_124(4,2), Cijr_123(4,2)
      real*8 C0I_234, C0I_134, C0I_124, C0I_123
      real*8 CijI_234(4,2), CijI_134(4,2), CijI_124(4,2), CijI_123(4,2)
      real*8 D0r, Dijr(13,3)
      real*8 D0I, DijI(13,3)
c stability flag for Cijs and Dijs
      logical Singular 
      COMMON /GramCheck/ Singular
    
      r1   = p1sq
      r2r1 = p2sq+2.d0*p1p2
      r3r2 = 2.d0*(p1p3+p2p3)+p3sq

      deter =-2.d0*(-2.d0*p1p2*p1p3*p2p3 + p1p3*p1p3*p2sq + p1p2*p1p2*p3sq 
     -  + p1sq*(p2p3*p2p3 - p2sq*p3sq))
 
 
      detAbs=-2.d0*(abs(2.d0*p1p2*p1p3*p2p3) + abs(p1p3*p1p3*p2sq)
     -   +abs(p1p2*p1p2*p3sq) 
     -  + abs(p1sq*(abs(p2p3*p2p3) + abs(p2sq*p3sq))))
    
      If(abs(deter)/abs(detAbs).lt.10d-5) then
         Singular=.true.
      endif
 

      Ideter=1.d0/deter

      IZ(1,1) = (p2sq*p3sq - p2p3*p2p3)*Ideter
      IZ(2,1) = (p1p3*p2p3 - p1p2*p3sq)*Ideter
      IZ(3,1) = (p1p2*p2p3 - p1p3*p2sq)*Ideter
      IZ(1,2) = IZ(2,1)
      IZ(2,2) = (p1sq*p3sq-p1p3*p1p3)*Ideter
      IZ(3,2) = (p1p2*p1p3 - p1sq*p2p3)*Ideter
      IZ(1,3) = IZ(3,1)
      IZ(2,3) = IZ(3,2)
      IZ(3,3) = (p1sq*p2sq - p1p2*p1p2)*Ideter

      D0r=Dble(D0)
      D0I=DImag(D0)
c
      Rr(1) = (C0r_134 - C0r_234 - D0r*r1   )
      Rr(2) = (C0r_124 - C0r_134 - D0r*r2r1 )
      Rr(3) = (C0r_123 - C0r_124 - D0r*r3r2 )
c    
      Dijr(1,1) =( Rr(1)*IZ(1,1) + Rr(2)*IZ(2,1) + Rr(3)*IZ(3,1)  )
      Dijr(2,1) =( Rr(1)*IZ(1,2) + Rr(2)*IZ(2,2) + Rr(3)*IZ(3,2)  )
      Dijr(3,1) =( Rr(1)*IZ(1,3) + Rr(2)*IZ(2,3) + Rr(3)*IZ(3,3)  ) 

c
      RI(1) = (C0I_134 - C0I_234 - D0I*r1   )
      RI(2) = (C0I_124 - C0I_134 - D0I*r2r1 )
      RI(3) = (C0I_123 - C0I_124 - D0I*r3r2 )
c      
      DijI(1,1) =( RI(1)*IZ(1,1) + RI(2)*IZ(2,1) + RI(3)*IZ(3,1)  )
      DijI(2,1) =( RI(1)*IZ(1,2) + RI(2)*IZ(2,2) + RI(3)*IZ(3,2)  )
      DijI(3,1) =( RI(1)*IZ(1,3) + RI(2)*IZ(2,3) + RI(3)*IZ(3,3)  ) 
       
      
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
c D00
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      msq=m*m
c      msq=0d0
c      m=0d0
      Dijr(7,2) = (C0r_234+ 2*msq*D0r + Dijr(1,1)*r1 + Dijr(2,1)*r2r1
     #     +Dijr(3,1)*r3r2)*0.5d0

      DijI(7,2) = (C0I_234+2*msq*D0I + DijI(1,1)*r1 + DijI(2,1)*r2r1
     #     +DijI(3,1)*r3r2)*0.5d0


CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

c 30-31-32
    
c
      Rr(1) = (Cijr_134(1,1) + C0r_234 - Dijr(1,1)*r1 - Dijr(7,2)*2.d0)
      Rr(2) = (Cijr_124(1,1) - Cijr_134(1,1) - Dijr(1,1)*r2r1        )
      Rr(3) = (Cijr_123(1,1) - Cijr_124(1,1) - Dijr(1,1)*r3r2        )     
c      
      Dijr(1,2) = ( Rr(1)*IZ(1,1) + Rr(2)*IZ(2,1) + Rr(3)*IZ(3,1)  )
      Dijr(4,2) = ( Rr(1)*IZ(1,2) + Rr(2)*IZ(2,2) + Rr(3)*IZ(3,2)  )
      Dijr(5,2) = ( Rr(1)*IZ(1,3) + Rr(2)*IZ(2,3) + Rr(3)*IZ(3,3)  )
c      
      RI(1) = (CijI_134(1,1) + C0I_234  - DijI(1,1)*r1 - DijI(7,2)*2.d0)
      RI(2) = (CijI_124(1,1) - CijI_134(1,1) - DijI(1,1)*r2r1        )
      RI(3) = (CijI_123(1,1) - CijI_124(1,1) - DijI(1,1)*r3r2        )     
c
      DijI(1,2) = ( RI(1)*IZ(1,1) + RI(2)*IZ(2,1) + RI(3)*IZ(3,1)  )
      DijI(4,2) = ( RI(1)*IZ(1,2) + RI(2)*IZ(2,2) + RI(3)*IZ(3,2)  )
      DijI(5,2) = ( RI(1)*IZ(1,3) + RI(2)*IZ(2,3) + RI(3)*IZ(3,3)  )

c 33-34-35
      
      Rr(1) = (Cijr_134(1,1) -Cijr_234(1,1) - Dijr(2,1)*r1          )
      Rr(2)=(Cijr_124(2,1)-Cijr_134(1,1)-Dijr(2,1)*r2r1-Dijr(7,2)*2.d0)
      Rr(3) = (Cijr_123(2,1) -Cijr_124(2,1) - Dijr(2,1)*r3r2        )

c      Dijr(4,2) = ( Rr(1)*IZ(1,1) + Rr(2)*IZ(2,1) + Rr(3)*IZ(3,1)  )
      Dijr(2,2) = ( Rr(1)*IZ(1,2) + Rr(2)*IZ(2,2) + Rr(3)*IZ(3,2)  )
      Dijr(6,2) = ( Rr(1)*IZ(1,3) + Rr(2)*IZ(2,3) + Rr(3)*IZ(3,3)  )
      
      RI(1) = (CijI_134(1,1) -CijI_234(1,1) - DijI(2,1)*r1          )
      RI(2)=(CijI_124(2,1)-CijI_134(1,1)-DijI(2,1)*r2r1-DijI(7,2)*2.d0)
      RI(3) = (CijI_123(2,1) -CijI_124(2,1) - DijI(2,1)*r3r2        )
    
c      DijI(4,2) = ( RI(1)*IZ(1,1) + RI(2)*IZ(2,1) + RI(3)*IZ(3,1)  )
      DijI(2,2) = ( RI(1)*IZ(1,2) + RI(2)*IZ(2,2) + RI(3)*IZ(3,2)  )
      DijI(6,2) = ( RI(1)*IZ(1,3) + RI(2)*IZ(2,3) + RI(3)*IZ(3,3)  )
 
c 36-37-38
   
      Rr(1) = (Cijr_134(2,1) - Cijr_234(2,1) - Dijr(3,1)*r1      )
      Rr(2) = (Cijr_124(2,1) - Cijr_134(2,1) - Dijr(3,1)*r2r1    )
      Rr(3) = (       - Cijr_124(2,1) - Dijr(3,1)*r3r2 -Dijr(7,2)*2.d0)

c      Dijr(5,2) = ( Rr(1)*IZ(1,1) + Rr(2)*IZ(2,1) + Rr(3)*IZ(3,1)  )
c      Dijr(6,2) = ( Rr(1)*IZ(1,2) + Rr(2)*IZ(2,2) + Rr(3)*IZ(3,2)  )
      Dijr(3,2) = ( Rr(1)*IZ(1,3) + Rr(2)*IZ(2,3) + Rr(3)*IZ(3,3)  )
  
      RI(1) = (CijI_134(2,1) - CijI_234(2,1) - DijI(3,1)*r1      )
      RI(2) = (CijI_124(2,1) - CijI_134(2,1) - DijI(3,1)*r2r1    )
      RI(3) = (       - CijI_124(2,1) - DijI(3,1)*r3r2 -DijI(7,2)*2.d0)

c      DijI(5,2) = ( RI(1)*IZ(1,1) + RI(2)*IZ(2,1) + RI(3)*IZ(3,1)  )
c      DijI(6,2) = ( RI(1)*IZ(1,2) + RI(2)*IZ(2,2) + RI(3)*IZ(3,2)  )
      DijI(3,2) = ( RI(1)*IZ(1,3) + RI(2)*IZ(2,3) + RI(3)*IZ(3,3)  )



CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC



      Dijr(11,3)=0.25d0*(-C0r_234+2*msq*Dijr(1,1)+Dijr(1,2)*r1
     - +Dijr(4,2)*r2r1 +Dijr(5,2)*r3r2)     
      Dijr(12,3)=0.25d0*(Cijr_234(1,1)+2*msq*Dijr(2,1)+Dijr(4,2)*r1
     -+Dijr(2,2)*r2r1 +Dijr(6,2)*r3r2)
      Dijr(13,3)=0.25d0*(Cijr_234(2,1)+2*msq*Dijr(3,1)+Dijr(5,2)*r1 
     - + Dijr(6,2)*r2r1 + Dijr(3,2)*r3r2)

      

      DijI(11,3)=0.25d0*(-C0I_234+2*msq*DijI(1,1)+DijI(1,2)*r1
     - +DijI(4,2)*r2r1 +DijI(5,2)*r3r2)
     
      DijI(12,3)=0.25d0*(CijI_234(1,1)+2*msq*DijI(2,1)+DijI(4,2)*r1
     - +DijI(2,2)*r2r1+DijI(6,2)*r3r2)
      DijI(13,3)=0.25d0*(CijI_234(2,1)+2*msq*DijI(3,1)+DijI(5,2)*r1 
     - + DijI(6,2)*r2r1 + DijI(3,2)*r3r2)


c 41-42-43


      Rr(1) =(Cijr_134(1,2) - C0r_234 - Dijr(1,2)*r1  - Dijr(11,3)*4.d0)
      Rr(2) =(Cijr_124(1,2) - Cijr_134(1,2) - Dijr(1,2)*r2r1       )
      Rr(3) =(Cijr_123(1,2) - Cijr_124(1,2) - Dijr(1,2)*r3r2       )
  
      Dijr(1,3) = ( Rr(1)*IZ(1,1) + Rr(2)*IZ(2,1) + Rr(3)*IZ(3,1)  )
      Dijr(4,3) = ( Rr(1)*IZ(1,2) + Rr(2)*IZ(2,2) + Rr(3)*IZ(3,2)  )
      Dijr(5,3) = ( Rr(1)*IZ(1,3) + Rr(2)*IZ(2,3) + Rr(3)*IZ(3,3)  )

      RI(1) =(CijI_134(1,2) - C0I_234 - DijI(1,2)*r1  - DijI(11,3)*4.d0)
      RI(2) =(CijI_124(1,2) - CijI_134(1,2) - DijI(1,2)*r2r1       )
      RI(3) =(CijI_123(1,2) - CijI_124(1,2) - DijI(1,2)*r3r2       )

      DijI(1,3) = ( RI(1)*IZ(1,1) + RI(2)*IZ(2,1) + RI(3)*IZ(3,1)  )
      DijI(4,3) = ( RI(1)*IZ(1,2) + RI(2)*IZ(2,2) + RI(3)*IZ(3,2)  )
      DijI(5,3) = ( RI(1)*IZ(1,3) + RI(2)*IZ(2,3) + RI(3)*IZ(3,3)  )

c 50-51-52

      Rr(1) =  (Cijr_134(1,2) - Cijr_234(1,2) - Dijr(2,2)*r1)
      Rr(2) =  (Cijr_124(2,2) - Cijr_134(1,2) - Dijr(2,2)*r2r1
     #     -Dijr(12,3)*4.d0)
      Rr(3) =  (Cijr_123(2,2) - Cijr_124(2,2) - Dijr(2,2)*r3r2)

      
      Dijr(6,3) = ( Rr(1)*IZ(1,1) + Rr(2)*IZ(2,1) + Rr(3)*IZ(3,1)  )
      Dijr(2,3) = ( Rr(1)*IZ(1,2) + Rr(2)*IZ(2,2) + Rr(3)*IZ(3,2)  )
      Dijr(8,3) = ( Rr(1)*IZ(1,3) + Rr(2)*IZ(2,3) + Rr(3)*IZ(3,3)  )

      RI(1) =  (CijI_134(1,2) - CijI_234(1,2) - DijI(2,2)*r1)
      RI(2) =  (CijI_124(2,2) - CijI_134(1,2) - DijI(2,2)*r2r1
     #     -DijI(12,3)*4.d0)
      RI(3) =  (CijI_123(2,2) - CijI_124(2,2) - DijI(2,2)*r3r2)

      DijI(6,3) = ( RI(1)*IZ(1,1) + RI(2)*IZ(2,1) + RI(3)*IZ(3,1)  )
      DijI(2,3) = ( RI(1)*IZ(1,2) + RI(2)*IZ(2,2) + RI(3)*IZ(3,2)  )
      DijI(8,3) = ( RI(1)*IZ(1,3) + RI(2)*IZ(2,3) + RI(3)*IZ(3,3)  )

c
c 56-57-58
c     
c
      Rr(1) = (Cijr_134(2,2) - Cijr_234(2,2) - Dijr(3,2)*r1)
      Rr(2) = (Cijr_124(2,2) - Cijr_134(2,2) - Dijr(3,2)*r2r1)
      Rr(3) = (     -Cijr_124(2,2) - Dijr(3,2)*r3r2 - Dijr(13,3)*4.d0)
c      
      Dijr(7,3) = ( Rr(1)*IZ(1,1) + Rr(2)*IZ(2,1) + Rr(3)*IZ(3,1)  )
      Dijr(9,3) = ( Rr(1)*IZ(1,2) + Rr(2)*IZ(2,2) + Rr(3)*IZ(3,2)  )
      Dijr(3,3) = ( Rr(1)*IZ(1,3) + Rr(2)*IZ(2,3) + Rr(3)*IZ(3,3)  )
c
      RI(1) = (CijI_134(2,2) - CijI_234(2,2) - DijI(3,2)*r1)
      RI(2) = (CijI_124(2,2) - CijI_134(2,2) - DijI(3,2)*r2r1)
      RI(3) = (     -CijI_124(2,2) - DijI(3,2)*r3r2 - DijI(13,3)*4.d0)
c
      DijI(7,3) = ( RI(1)*IZ(1,1) + RI(2)*IZ(2,1) + RI(3)*IZ(3,1)  )
      DijI(9,3) = ( RI(1)*IZ(1,2) + RI(2)*IZ(2,2) + RI(3)*IZ(3,2)  )
      DijI(3,3) = ( RI(1)*IZ(1,3) + RI(2)*IZ(2,3) + RI(3)*IZ(3,3)  )
c 44-45-46
      Rr(1) = (Cijr_134(1,2) + Cijr_234(1,1) - Dijr(4,2)*r1 
     #     -Dijr(12,3)*2.d0)
      Rr(2) = (Cijr_124(3,2) - Cijr_134(1,2) - Dijr(4,2)*r2r1
     #     - Dijr(11,3)*2.d0)
      Rr(3) = (Cijr_123(3,2) - Cijr_124(3,2)  -Dijr(4,2)*r3r2)
c    
C      Dijr(4,3) = ( Rr(1)*IZ(1,1) + Rr(2)*IZ(2,1) + Rr(3)*IZ(3,1)  )
C      Dijr(6,3) = ( Rr(1)*IZ(1,2) + Rr(2)*IZ(2,2) + Rr(3)*IZ(3,2)  )
      Dijr(10,3) = ( Rr(1)*IZ(1,3) + Rr(2)*IZ(2,3) + Rr(3)*IZ(3,3)  )
c
c
      RI(1) = (CijI_134(1,2) + CijI_234(1,1) - DijI(4,2)*r1 
     #     -DijI(12,3)*2.d0)
      RI(2) = (CijI_124(3,2) - CijI_134(1,2) - DijI(4,2)*r2r1
     #     - DijI(11,3)*2.d0)
      RI(3) = (CijI_123(3,2) - CijI_124(3,2)  -DijI(4,2)*r3r2)
    
C      DijI(4,3) = ( RI(1)*IZ(1,1) + RI(2)*IZ(2,1) + RI(3)*IZ(3,1)  )
C      DijI(6,3) = ( RI(1)*IZ(1,2) + RI(2)*IZ(2,2) + RI(3)*IZ(3,2)  )
      DijI(10,3) = ( RI(1)*IZ(1,3) + RI(2)*IZ(2,3) + RI(3)*IZ(3,3)  )
c
c to define D00ij and D0000 functions!
c In PV notation we have:
c Dij(1,7)=D416, Dij(1,8)=D417, Dij(1,9)=D418
c Dij(1,10)=D419, Dij(1,11)=D420, Dij(1,12)=D421, D(1,13)=D422
c   
      Iv=1.d0/6.d0
c     
c
c      msq=0d0

      Dijr(7,1)=(C0r_234+2*msq*Dijr(1,2)+Dijr(1,3)*r1+Dijr(4,3)*r2r1+Dijr(5,3)*r3r2 )*Iv
c      
      Dijr(8,1)=(Cijr_234(1,2)+2*msq*Dijr(2,2)+Dijr(6,3)*r1+Dijr(2,3)*r2r1
     - +Dijr(8,3)*r3r2)*Iv
c
      Dijr(9,1)=(Cijr_234(2,2)+2*msq*Dijr(3,2)+Dijr(7,3)*r1+Dijr(9,3)*r2r1
     - +Dijr(3,3)*r3r2)*Iv
c
      Dijr(10,1)=(-Cijr_234(1,1)+2*msq*Dijr(4,2)+Dijr(4,3)*r1+Dijr(6,3)*r2r1
     - +Dijr(10,3)*r3r2)*Iv
c      
      Dijr(11,1)=(-Cijr_234(2,1)+2*msq*Dijr(5,2)+Dijr(5,3)*r1+Dijr(10,3)*r2r1
     - +Dijr(7,3)*r3r2)*Iv
c
      Dijr(12,1)=(Cijr_234(3,2)+2*msq*Dijr(6,2)+Dijr(10,3)*r1+Dijr(8,3)*r2r1
     - +Dijr(9,3)*r3r2)*Iv
c
      Dijr(13,1)=(Cijr_234(4,2)+2*msq*Dijr(7,2)+Dijr(11,3)*r1+Dijr(12,3)*r2r1+
     -      + Dijr(13,3)*r3r2+Iv)*Iv        
c
      DijI(7,1)=(C0I_234+2*msq*DijI(1,2)+DijI(1,3)*r1+DijI(4,3)*r2r1+DijI(5,3)*r3r2 )*Iv
c      
      DijI(8,1)=(CijI_234(1,2)+2*msq*DijI(2,2)+DijI(6,3)*r1+DijI(2,3)*r2r1
     - +DijI(8,3)*r3r2)*Iv
c
      DijI(9,1)=(CijI_234(2,2)+2*msq*DijI(3,2)+DijI(7,3)*r1+DijI(9,3)*r2r1
     -+DijI(3,3)*r3r2)*Iv
c
      DijI(10,1)=(-CijI_234(1,1)+2*msq*DijI(4,2)+DijI(4,3)*r1+DijI(6,3)*r2r1
     - +DijI(10,3)*r3r2)*Iv
c      
      DijI(11,1)=(-CijI_234(2,1)+2*msq*DijI(5,2)+DijI(5,3)*r1+DijI(10,3)*r2r1
     - +DijI(7,3)*r3r2)*Iv
c
      DijI(12,1)=(CijI_234(3,2)+2*msq*DijI(6,2)+DijI(10,3)*r1+DijI(8,3)*r2r1
     - +DijI(9,3)*r3r2)*Iv
c
      DijI(13,1)=(CijI_234(4,2)+2*msq*DijI(7,2)+DijI(11,3)*r1+DijI(12,3)*r2r1+
     -      + DijI(13,3)*r3r2)*Iv  

      Dij(1,1)=DCMPLX(Dijr(1,1),DijI(1,1))
      Dij(2,1)=DCMPLX(Dijr(2,1),DijI(2,1))
      Dij(3,1)=DCMPLX(Dijr(3,1),DijI(3,1))
c
      Dij(7,1)=DCMPLX(Dijr(7,1),DijI(7,1))
      Dij(8,1)=DCMPLX(Dijr(8,1),DijI(8,1))
      Dij(9,1)=DCMPLX(Dijr(9,1),DijI(9,1))
      Dij(10,1)=DCMPLX(Dijr(10,1),DijI(10,1))
      Dij(11,1)=DCMPLX(Dijr(11,1),DijI(11,1))
      Dij(12,1)=DCMPLX(Dijr(12,1),DijI(12,1))
      Dij(13,1)=DCMPLX(Dijr(13,1),DijI(13,1))
c      
      Dij(1,2)=DCMPLX(Dijr(1,2),DijI(1,2))
      Dij(2,2)=DCMPLX(Dijr(2,2),DijI(2,2))
      Dij(3,2)=DCMPLX(Dijr(3,2),DijI(3,2))
      Dij(4,2)=DCMPLX(Dijr(4,2),DijI(4,2))
      Dij(5,2)=DCMPLX(Dijr(5,2),DijI(5,2))
      Dij(6,2)=DCMPLX(Dijr(6,2),DijI(6,2))
      Dij(7,2)=DCMPLX(Dijr(7,2),DijI(7,2))
c      
      Dij(1,3)=DCMPLX(Dijr(1,3),DijI(1,3))
      Dij(2,3)=DCMPLX(Dijr(2,3),DijI(2,3))
      Dij(3,3)=DCMPLX(Dijr(3,3),DijI(3,3))
      Dij(4,3)=DCMPLX(Dijr(4,3),DijI(4,3))
      Dij(5,3)=DCMPLX(Dijr(5,3),DijI(5,3))
      Dij(6,3)=DCMPLX(Dijr(6,3),DijI(6,3))
      Dij(7,3)=DCMPLX(Dijr(7,3),DijI(7,3))
      Dij(8,3)=DCMPLX(Dijr(8,3),DijI(8,3))
      Dij(9,3)=DCMPLX(Dijr(9,3),DijI(9,3))
      Dij(10,3)=DCMPLX(Dijr(10,3),DijI(10,3))
      Dij(11,3)=DCMPLX(Dijr(11,3),DijI(11,3))
      Dij(12,3)=DCMPLX(Dijr(12,3),DijI(12,3))
      Dij(13,3)=DCMPLX(Dijr(13,3),DijI(13,3))
c 
      end

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                tens_red5 = 5-point tensors
C     Francisco Campanario
C Date: 4/04/2008
C  Tensor reduction for Pentagons a la Denner-Dittmeier for massless quarks and p1^2=0, p5^5=0.
C However, I am using the convention of Passarino-Vetlman for the tensor
C integrals. In this way, we avoid to convert the Dij from Passarino to Denner
C and later on to convert the Eij from Denner to Passarino Convection.
C In this file, we should put as an input the 10 scalar invariants, 
C the 5 D0, and the Dij got it from tens_red4_new_Re_Com.
C Notation: Example DR2345. the "R" means the real part of D02345
C                   DI2345, the "I" means the imaginary part of D02345
C This separation is done already in tens_red4_....
C It speeds up the code. 
C OUTPUT:  is given in E(Real part) and C(complex part) 
C NOTE!!!! The code seems to be 6 to 7 times faster than the old one and more stable.
C The indices are pertumted in comparison with the old subroutines.
C the rank of the tensor integral is store in the second argument.
C D_old(1,3)-> D(3,1)(rank 1 coefficient 3) it is faster in this way.
C The recursion relation that are implemented can not be found in the literature.
C So, one has to look into my notes if want to generalize them.
C I have more version which only needs the input in complex form
C and the output is given as a complex array. The subroutine still makes the separation
C between imaginary part and real part, but it is done inside of it. Here I am saving doing that since 
C I need and use this DR and DI in my penline subroutine.
C If you need other case ask me!
C Change:26/08/2008
C(Including general kinematics in Caley determinants, this is necessary for the hexagons)
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      subroutine tens_red5_new_Re_Com_1M(m,p1, p2, p3, p4,
     #                     p1p2, p1p3,p1p4, p2p3,p2p4,p3p4, 
     #                     DR2345,DR1345,DR1245,DR1235,DR1234,
     #                     DRR2345,DRR1345,DRR1245,DRR1235,DRR1234,
     #                     DI2345,DI1345,DI1245,DI1235,DI1234,
     #                     DII2345,DII1345,DII1245,DII1235,DII1234,
     #                     E,C)
      implicit none
      real*8 p1,p2,p3,p4,p1p2,p1p3,p1p4,p2p3,p2p4,p3p4,m,msq,eps,p5
c$$$      complex*16 D0_2345,D0_1345,D0_1245,D0_1235,D0_1234
c$$$      complex*16 Dij_2345(13,3),Dij_1345(13,3),Dij_1245(13,3),
c$$$     #     Dij_1235(13,3),Dij_1234(13,3)
      complex*16 Eij(46,4)
c                                                             d^4k
c
c   E0; E_mu; E_mu,nu; E_mu,nu,rho; E_mu,nu,rho_sig  =  Int -------
c                                                           (2pi)^4
c
c   1; k_mu; k_mu k_nu; k_mu k_nu k_rho; k_mu k_nu k_rho k_sig
c --------------------------------------------------------------------------
c   k^2  (k+p1)^2  (k+p1+p2)  (k+p1+p2+p3)^2  (k+p1+p2+p3+p4)^2
c
c  with
c
c   E_mu = p1_mu E11  +  p2_mu E12  +  p3_mu E13  +  p4_mu E14
c
c   E_munu = p1_mu p1_nu E21 + p2_mu p2_nu E22 + ...
c
c
C INPUT:  p1sq, p1p2,...                     kinematics invariants
c         D0_1345 = D00(p1+p2,p3,p4)          scalar 4-point functions
c         D0_1245 = D00(p1,p2+p3,p4)          in the PV notation
c         D0_1234 = D00(p1,p2,p3)
c         D0_1235 = D00(p1,p2,p3+p4)      
c         D0_2345 = D00(p2,p3,p4)

C         Dij_1235(n,m) = D_nm(1,2,3,5) ....    higher D_nm form factors
C                                               as in tens_red3 and tens_red4
c         E0 = E0(p,q,l,t)                      scalar five point function
c
c OUTPUT: Eij(n,m) = E_nm                    form factors in the tensor 
c                                            integrals a la PV
c         nm = 11, 12, 13, 14                ff's for E_mu
c         nm = 21, 22, 23, ..., 211    ff's for E_munu
c         nm = 31, 32, 33, ..., 323, 324     ff's for E_mu,nu,rho
c         nm = 41, 42, 43, ..., 445, 446
c
      real*8 R(4),R1(4),R2(4),R3(4),R4(4),R11(4),R12(4),R13(4)
      real*8 R14(4),R22(4),R23(4),R24(4),R33(4),R34(4),R44(4)
      real*8 R111(4),R112(4),R113(4),R114(4),R221(4),R222(4),R223(4)
      real*8 R224(4),R331(4),R332(4),R333(4),R334(4),R441(4),R442(4)
      real*8 R443(4),R444(4),R123(4),R124(4),R134(4),R234(4)

      real*8 s12,s23,s34,s45,s15,s12sq,s45sq,s1523,s15m23,s15m34,s1534
      real*8 s15mp2,p2t45,st1223mp245,aux1,aux2,aux3,s1215,s15p2,s45mp4
      real*8  p2sq,sts15m23,p5sq,d,Id,I2,I3,I4,d2,d3,d4
      real*8  Y123,Y124,Y134,Y213,Y214,Y234,Y312,Y314,Y324,Y412,Y413,Y423
      real*8  x(1:4,0:4), Y(4,4,4),ER,EI
      real*8 DR2345,DR1345,DR1245,DR1235,DR1234,DRR1234(13,3),E(46,4)
      real*8 DRR2345(13,3),DRR1345(13,3),DRR1245(13,3),DRR1235(13,3)
      real*8 DI2345,DI1345,DI1245,DI1235,DI1234,DII1234(13,3),C(46,4)
      real*8 DII2345(13,3),DII1345(13,3),DII1245(13,3),DII1235(13,3)
      Parameter(eps=1d-07)
      Integer i,j 
c with p1.p1 = 0
c p5=(-p1-p2-p3-p4) with p5.p5=0
c      if(p1.ne.0.d0) then
c         print*,' p1sq must be zero. Here p1sq = ',p1
c      endif
c      p5sq=(p1+p2+p3+p4+2.d0*(p1p2+p1p3+p1p4+p2p3+p2p4+p3p4))
c      if(p5sq.gt.1.d-07) then
c      print*, ' p5sq=(p1+p2+p3+p4)^2 must be zero. Here p5sq = ',p5sq
c      endif



c s_ij=(p_i+p_j)^2
      s12=2.d0*p1p2+p2+p1
      s23=2.d0*p2p3+p2+p3      
      s34=2.d0*p3p4+p3+p4
      s45=p1+p2+p3+2d0*(p1p2+p1p3+p2p3)
      s15=p2+p3+p4+2d0*(p2p3+p2p4+p3p4)
      p5=s15+p1+2d0*(p1p2+p1p3+p1p4)
      s12sq=s12*s12
      s45sq=s45*s45    
      msq=m*m
       If(abs(p1).lt.eps) then 
       If(abs(p2).lt.eps) then 
       If(abs(p3).lt.eps) then 
       If(abs(p4).lt.eps) then 
       If(abs(p5).lt.eps) then 
c  Configuration                 (00000)
       d= 2*(16*msq*(p1p4**2*p2p3**2+(p1p3*p2p4-p1p2*p3p4)**2-2*p1p4*p
     -   2p3*(p1p3*p2p4+p1p2*p3p4))-s12*s15*s23*s34*s45)
       x(1,1) = s12sq*(s15-s23)**2-4*msq*s23*s34*(-s15+s23+s34)+2*s12*
     -   (-s15**2+s23*s34+s15*(s23+s34))*s45+(s15-s34)**2*s45sq
       x(2,1) = s12sq*(s15-s23)**2+s12*(2*msq*(s15-s23)*s34+(-2*s15*(s
     -   15-s23)+(s15+s23)*s34)*s45)-2*msq*s34*((2*s23+s34)*(s23-s45)
     -   +s15*(-2*s23+s45))+s15*(s15-s34)*s45sq
       x(3,1) = 2*msq*(s15-s23-s34)*(s12*(-s15+s23)+s23*s34+s15*s45-s3
     -   4*s45)+s12*(s12*(s15-s23)**2+(-s15**2+s23*s34+s15*(s23+s34))
     -   *s45)
       x(4,1) = s23*(2*msq*(s12*(s15-s23-2*s34)-s23*s34-s15*s45+s34*s4
     -   5)+s12*(s12*(-s15+s23)+(s15+s34)*s45))
       x(2,2) = 4*msq*s34*(s12+s23-s45)*(s15-s23+s45)+(s12*(-s15+s23)+
     -   s15*s45)**2
       x(3,2) = s12*(s15-s23)*(s12*(s15-s23)-s15*s45)+2*msq*(s15-s23+s
     -   45)*(s23*s34+s12*(-s15+s23+2*s34)+s15*s45-s34*s45)
       x(4,2) = s12*s23*(s12*(-s15+s23)+s15*s45)-2*msq*(s12+s23-s45)*(
     -   s12*(-s15+s23)+s23*s34+s15*s45-s34*s45)
       x(3,3) = s12*(s12*(s15-s23)**2-4*msq*(s15-s23-s34)*(s15-s23+s45
     -   ))
       x(4,3) = s12*(s12*s23*(-s15+s23)-2*msq*(s12*(-s15+s23)+(2*s23+s
     -   34)*(s23-s45)+s15*(-2*s23+s45)))
       x(4,4) = s12*s23*(s12*s23-4*msq*(s12+s23-s45))
c          
       x(1,0)= -(s12sq*(s15-s23)**2)+s12*s23*(-s15+s23)*s34-2*s12*(-s1
     -   5**2+s23*s34+s15*(s23+s34))*s45-(s15-s34)*s45*(s23*s34+(s15-
     -   s34)*s45)
       x(2,0)= -(s12*(s15-s23)*(s12*(s15-s23)+s23*s34))+(2*s12*s15*(s1
     -   5-s23)-(s15*s23+s12*(s15+s23))*s34)*s45-s15*(s15-s34)*s45sq
       x(3,0)= -(s12*(s12*(s15-s23)**2+s15*s23*s34-s15**2*s45+s15*(s23
     -   +s34)*s45+s23*s34*(-s23+s45)))
       x(4,0)= s12*s23*(s12*(s15-s23)+s23*s34-(s15+s34)*s45)
       else 
c  Configuration                 (0000p5s)
       d= 2*(16*msq*(p1p4**2*p2p3**2+(p1p3*p2p4-p1p2*p3p4)**2-2*p1p4*p
     -   2p3*(p1p3*p2p4+p1p2*p3p4))+s12*s23*s34*(p5*s23-s15*s45))
       x(1,1) = s12sq*(s15-s23)**2+p5**2*s23**2-4*msq*s23*s34*(-s15+s2
     -   3+s34)-2*(p5*s23*(s15-s34)*s45-s12*(p5*s23*(s15-s23-2*s34)+(
     -   -s15**2+s23*s34+s15*(s23+s34))*s45))+(s15-s34)**2*s45sq
       x(2,1) = s12sq*(s15-s23)**2+p5**2*s23**2+p5*s23*(-2*msq*s34+(-2
     -   *s15+s34)*s45)-2*(msq*s23*s34*(-2*(s15-s23)+s34)-msq*s34*(-s
     -   15+2*s23+s34)*s45)+s12*(2*(msq*(s15-s23)*s34-p5*s23*(-s15+s2
     -   3+s34))+(-2*s15*(s15-s23)+(s15+s23)*s34)*s45)+s15*(s15-s34)*
     -   s45sq
       x(3,1) = 2*msq*(s15-s23-s34)*(-(p5*s23)+s12*(-s15+s23)+s23*s34+
     -   s15*s45-s34*s45)+s12*(s12*(s15-s23)**2+p5*s23*(s15-s23-2*s34
     -   )+(-s15**2+s23*s34+s15*(s23+s34))*s45)
       x(4,1) = s23*(2*msq*(p5*s23+s12*(s15-s23-2*s34)-s23*s34-s15*s45
     -   +s34*s45)+s12*(-(p5*s23)+s12*(-s15+s23)+(s15+s34)*s45))
       x(2,2) = -4*msq*s34*(s12+s23-s45)*(p5-s15+s23-s45)+(s12*(s15-s2
     -   3)+p5*s23-s15*s45)**2
       x(3,2) = s12*(s15-s23)*(s12*(s15-s23)+p5*s23-s15*s45)+2*msq*(p5
     -   -s15+s23-s45)*(p5*s23+s12*(s15-s23-2*s34)-s23*s34-s15*s45+s3
     -   4*s45)
       x(4,2) = s12*s23*(-(p5*s23)+s12*(-s15+s23)+s15*s45)+2*msq*(s12+
     -   s23-s45)*(s12*(s15-s23)+p5*s23-s23*s34-s15*s45+s34*s45)
       x(3,3) = s12*(s12*(s15-s23)**2+4*msq*(s15-s23-s34)*(p5-s15+s23-
     -   s45))
       x(4,3) = s12*(s12*s23*(-s15+s23)-2*msq*(p5*s23+s12*(-s15+s23)+(
     -   2*s23+s34)*(s23-s45)+s15*(-2*s23+s45)))
       x(4,4) = s12*s23*(s12*s23-4*msq*(s12+s23-s45))
c          
       x(1,0)= -(s12sq*(s15-s23)**2)+s12*s23*((-s15+s23)*s34+p5*(-2*(s
     -   15-s23)+4*s34))-2*s12*(-s15**2+s23*s34+s15*(s23+s34))*s45-(p
     -   5*s23-s23*s34-s15*s45+s34*s45)*(p5*s23+(-s15+s34)*s45)
       x(2,0)= -(s12sq*(s15-s23)**2)+s12*s23*((-s15+s23)*s34+2*p5*(-s1
     -   5+s23+s34))-s12*(-2*s15*(s15-s23)+(s15+s23)*s34)*s45-(p5*s23
     -   -s15*s45)*(p5*s23-s23*s34-s15*s45+s34*s45)
       x(3,0)= -(s12*(s12*(s15-s23)**2+p5*s23*(s15-s23-2*s34)+s15*s23*
     -   s34-s23**2*s34-s15**2*s45+s15*s23*s45+s15*s34*s45+s23*s34*s4
     -   5))
       x(4,0)= s12*s23*(s12*(s15-s23)+s23*(p5+s34)-(s15+s34)*s45)
       endif 
       else 
       If(abs(p5).lt.eps) then 
c  Configuration                 (000p4s0)
       d= 2*(16*msq*(p1p4**2*p2p3**2+(p1p3*p2p4-p1p2*p3p4)**2-2*(p1p4*
     -   p2p3*(p1p3*p2p4+p1p2*p3p4)-p1p2*p1p3*p2p3*p4))+s12*s15*s23*(
     -   p4*s12-s34*s45))
       x(1,1) = p4**2*s12sq+s12sq*(s15-s23)**2-4*msq*s23*s34*(-s15+s23
     -   +s34)-2*(s12*(s15*(s15-s23)-(s15+s23)*s34)*s45+p4*(s12sq*(s1
     -   5+s23)+2*msq*s23*(s15-s34)+s12*(-s15+s34)*s45))+(s15-s34)**2
     -   *s45sq
       x(2,1) = -(s12sq*(-(s15-s23)**2+p4*(s15+s23)))+s12*(s15*(p4-2*(
     -   s15-s23))+(s15+s23)*s34)*s45+2*msq*(p4**2*s12+s34*(s12*(s15-
     -   s23)-(2*s23+s34)*(s23-s45)+s15*(2*s23-s45))-p4*(2*s15*s23-s2
     -   3*s34+s12*(s15+s23+s34)-s15*s45+s34*s45))+s15*(s15-s34)*s45s
     -   q
       x(3,1) = s12*(s12*(s15-s23)**2-p4*s12*(s15+s23)+(-s15**2+s23*s3
     -   4+s15*(s23+s34))*s45)+2*msq*(p4*s12*(s15+s23-s34)-(s15-s23-s
     -   34)*(s12*s15-s12*s23-s23*s34-s15*s45+s34*s45))
       x(4,1) = s23*(-(s12sq*(p4+s15-s23))+s12*(s15+s34)*s45+2*msq*(p4
     -   *s12+s12*(s15-s23-2*s34)-s23*s34-s15*s45+s34*s45))
       x(2,2) = (s12*(-s15+s23)+s15*s45)**2-4*msq*(s12+s23-s45)*(p4*(s
     -   12+s15)-s34*(s15-s23+s45))
       x(3,2) = s12*(s15-s23)*(s12*(s15-s23)-s15*s45)+2*msq*(-(p4*s12*
     -   (2*s12+s15+s23-s45))+(s15-s23+s45)*(s23*s34+s12*(-s15+s23+2*
     -   s34)+s15*s45-s34*s45))
       x(4,2) = s12*s23*(s12*(-s15+s23)+s15*s45)-2*msq*(s12+s23-s45)*(
     -   p4*s12-s12*s15+s12*s23+s23*s34+s15*s45-s34*s45)
       x(3,3) = s12*(s12*(s15-s23)**2-4*msq*(p4*s12+(s15-s23-s34)*(s15
     -   -s23+s45)))
       x(4,3) = s12*(s12*s23*(-s15+s23)-2*msq*(p4*s12+s12*(-s15+s23)+s
     -   23*(-2*(s15-s23)+s34)+(s15-2*s23-s34)*s45))
       x(4,4) = s12*s23*(s12*s23-4*msq*(s12+s23-s45))
c          
       x(1,0)= -(p4**2*s12sq)-s12*(s15-s23)*(s12*(s15-s23)+s23*s34)+(s
     -   23*s34*(-s15+s34)+2*(s12*s15*(s15-s23)-s12*(s15+s23)*s34))*s
     -   45+p4*s12*(-(s23*s34)+2*(s15*s23+s12*(s15+s23)-s15*s45+s34*s
     -   45))-(s15-s34)**2*s45sq
       x(2,0)= s12*(p4*(2*s15*s23+s12*(s15+s23))-(s15-s23)*(s12*(s15-s
     -   23)+s23*s34))-(s12*s15*(p4-2*(s15-s23))+(s15*s23+s12*(s15+s2
     -   3))*s34)*s45-s15*(s15-s34)*s45sq
       x(3,0)= -(s12*(-(p4*s12*(s15+s23))+(s15-s23)*(s12*(s15-s23)+s23
     -   *s34)+(-s15**2+s23*s34+s15*(s23+s34))*s45))
       x(4,0)= s12*s23*(s12*(p4+s15-s23)+s23*s34-(s15+s34)*s45)
       else 
c  Configuration                 (000p4sp5s)
       d= 2*(16*msq*(p1p4**2*p2p3**2+(p1p3*p2p4-p1p2*p3p4)**2-2*(p1p4*
     -   p2p3*(p1p3*p2p4+p1p2*p3p4)-p1p2*p1p3*p2p3*p4))+s12*s23*(p4*s
     -   12*s15+p5*s23*s34-s15*s34*s45))
       x(1,1) = s12sq*(p4-s34)**2-4*msq*s23*(p4*(s15-s34)+s34*(-s15+s2
     -   3+s34))+s23**2*(p5-s45)**2+(-s15+s23+s34)**2*(s12-s45)**2+2*
     -   (s12*s23*(p4-s34)*(p5-s45)-s12*(p4*(s15+s23-s34)+s34*(-s15+s
     -   23+s34))*(s12-s45)+s23*(-s15+s23+s34)*(p5-s45)*(-s12+s45))
       x(2,1) = -(p4*s12sq*s15)+s12sq*s15**2+p4*p5*s12*s23-p4*s12sq*s2
     -   3+p5**2*s23**2+s12sq*s23**2+(p4*s12*s15-2*s12*s15*(s15-s23)+
     -   s12*(s15+s23)*s34+p5*s23*(-2*s15+s34))*s45-2*(-(p5*s12*s15*s
     -   23)+s12sq*s15*s23+p5*s12*s23**2+p5*s12*s23*s34+msq*(-(p4**2*
     -   s12)+p4*(-(s23*(p5-2*s15+s34))+s12*(s15+s23+s34)+(-s15+s34)*
     -   s45)+s34*(p5*s23+s12*(-s15+s23)+(2*s23+s34)*(s23-s45)+s15*(-
     -   2*s23+s45))))+s15*(s15-s34)*s45sq
       x(3,1) = s12*(s12*(s15-s23)**2+p5*s15*s23-p5*s23**2-p4*s12*(s15
     -   +s23)-2*p5*s23*s34-s15**2*s45+s15*s23*s45+s15*s34*s45+s23*s3
     -   4*s45)+2*msq*(p4*s12*(s15+s23-s34)-(s15-s23-s34)*(s12*(s15-s
     -   23)+p5*s23-s23*s34-s15*s45+s34*s45))
       x(4,1) = s23*(-(s12*(s12*(p4+s15-s23)+p5*s23))+s12*(s15+s34)*s4
     -   5+2*msq*(p4*s12+p5*s23+s12*(s15-s23-2*s34)-s23*s34-s15*s45+s
     -   34*s45))
       x(2,2) = -4*msq*(p4*(-p5+s12+s15)+s34*(p5-s15+s23-s45))*(s12+s2
     -   3-s45)+(s12*(s15-s23)+p5*s23-s15*s45)**2
       x(3,2) = s12*(s15-s23)*(s12*(s15-s23)+p5*s23-s15*s45)+2*msq*(p4
     -   *s12*(p5-2*s12-s15-s23+s45)+(p5-s15+s23-s45)*(p5*s23+s12*(s1
     -   5-s23-2*s34)-s23*s34-s15*s45+s34*s45))
       x(4,2) = s12*s23*(-(p5*s23)+s12*(-s15+s23)+s15*s45)-2*msq*(s12+
     -   s23-s45)*(p4*s12-s12*s15-p5*s23+s12*s23+s23*s34+s15*s45-s34*
     -   s45)
       x(3,3) = s12*(s12*(s15-s23)**2-4*msq*(p4*s12+(s15-s23-s34)*(-p5
     -   +s15-s23+s45)))
       x(4,3) = s12*(s12*s23*(-s15+s23)-2*msq*(p4*s12+s12*(-s15+s23)+s
     -   23*(p5-2*(s15-s23)+s34)+(s15-2*s23-s34)*s45))
       x(4,4) = s12*s23*(s12*s23-4*msq*(s12+s23-s45))
c          
       x(1,0)= -(p4**2*s12sq)-s12sq*(s15-s23)**2+s12*s23*((-s15+s23)*s
     -   34+p5*(-2*(s15-s23)+4*s34))-2*s12*(-s15**2+s23*s34+s15*(s23+
     -   s34))*s45-(p5*s23-s23*s34-s15*s45+s34*s45)*(p5*s23+(-s15+s34
     -   )*s45)+p4*s12*(-(s23*s34)+2*(-(p5*s23)+s15*s23+s12*(s15+s23)
     -   -s15*s45+s34*s45))
       x(2,0)= -(s12sq*(s15-s23)**2)+s12*s23*((-s15+s23)*s34+2*p5*(-s1
     -   5+s23+s34))-s12*(-2*s15*(s15-s23)+(s15+s23)*s34)*s45+p4*s12*
     -   (-(p5*s23)+2*s15*s23+s12*(s15+s23)-s15*s45)-(p5*s23-s15*s45)
     -   *(p5*s23-s23*s34-s15*s45+s34*s45)
       x(3,0)= -(s12*(s12*(s15-s23)**2+p5*s15*s23-p5*s23**2-p4*s12*(s1
     -   5+s23)-2*p5*s23*s34+s15*s23*s34-s23**2*s34-s15**2*s45+s15*s2
     -   3*s45+s15*s34*s45+s23*s34*s45))
       x(4,0)= s12*s23*(s12*(p4+s15-s23)+p5*s23+s23*s34-(s15+s34)*s45)
       endif 
       endif 
       else 
       If(abs(p4).lt.eps) then 
       If(abs(p5).lt.eps) then 
c  Configuration                 (00p3s00)
       d= 2*(16*msq*(p1p4**2*p2p3**2+(p1p3*p2p4-p1p2*p3p4)**2-2*p1p4*(
     -   p1p3*p2p3*p2p4-p1p2*p2p4*p3+p1p2*p2p3*p3p4))+s12*s15*(p3*s15
     -   -s23*s34)*s45)
       x(1,1) = s12sq*(s15-s23)**2-4*msq*(p3+s15-s23-s34)*(p3*s15-s23*
     -   s34)+2*s12*(-(s15*(2*p3+s15-s23))+(s15+s23)*s34)*s45+(s15-s3
     -   4)**2*s45sq
       x(2,1) = s12sq*(s15-s23)**2+s12*(-2*s15*(p3+s15-s23)+(s15+s23)*
     -   s34)*s45-2*msq*(p3**2*s15+s34*(s12*(-s15+s23)+(2*s23+s34)*(s
     -   23-s45)+s15*(-2*s23+s45))+p3*(2*s15**2+s12*(s15-s23)+s34*(-s
     -   23+s45)+s15*(-2*s23-s34+s45)))+s15*(s15-s34)*s45sq
       x(3,1) = 2*msq*(p3+s15-s23-s34)*(-(p3*s15)-s12*s15+s12*s23+s23*
     -   s34+s15*s45-s34*s45)+s12*(s12*(s15-s23)**2+(-(s15*(2*p3+s15-
     -   s23))+(s15+s23)*s34)*s45)
       x(4,1) = s12*(s12*s23*(-s15+s23)+(-2*p3*s15+s23*(s15+s34))*s45)
     -   -2*msq*(p3**2*s15-p3*(s12*(s15+s23)+s23*(s15+s34)+(s15-s34)*
     -   s45)+s23*(s23*s34+s12*(-s15+s23+2*s34)+s15*s45-s34*s45))
       x(2,2) = -4*msq*(p3*(s12+s15)-s34*(s12+s23-s45))*(s15-s23+s45)+
     -   (s12*(-s15+s23)+s15*s45)**2
       x(3,2) = s12*(s15-s23)*(s12*(s15-s23)-s15*s45)+2*msq*(s15-s23+s
     -   45)*(-(p3*(2*s12+s15))+s23*s34+s12*(-s15+s23+2*s34)+s15*s45-
     -   s34*s45)
       x(4,2) = s12*s23*(s12*(-s15+s23)+s15*s45)-2*msq*((s12+s23-s45)*
     -   (s12*(-s15+s23)+s23*s34+s15*s45-s34*s45)+p3*(s12*(s15-2*(s23
     -   -s45))+s15*(-s23+s45)))
       x(3,3) = s12*(s12*(s15-s23)**2-4*msq*(p3+s15-s23-s34)*(s15-s23+
     -   s45))
       x(4,3) = s12*(s12*s23*(-s15+s23)-2*msq*(s12*(-s15+s23)+s23*(-2*
     -   (s15-s23)+s34)+p3*(s15-2*(s23-s45))+(s15-2*s23-s34)*s45))
       x(4,4) = s12*(s12*s23**2-4*msq*(s23*(s12+s23-s45)+p3*(-s23+s45)
     -   ))
c          
       x(1,0)= s12*(s15-s23)*(p3*s15+s12*(-s15+s23)-s23*s34)+(p3*s15*(
     -   4*s12+s15-s34)+s23*s34*(-s15+s34)+2*(s12*s15*(s15-s23)-s12*(
     -   s15+s23)*s34))*s45-(s15-s34)**2*s45sq
       x(2,0)= s12*(s15-s23)*(p3*s15+s12*(-s15+s23)-s23*s34)+(p3*s15*(
     -   2*s12+s15)+2*s12*s15*(s15-s23)-(s15*s23+s12*(s15+s23))*s34)*
     -   s45-s15*(s15-s34)*s45sq
       x(3,0)= -(s12*(s15-s23)*(-(p3*s15)+s12*(s15-s23)+s23*s34))+s12*
     -   (s15*(2*p3+s15-s23)-(s15+s23)*s34)*s45
       x(4,0)= s12*s23*(-(p3*s15)+s12*(s15-s23)+s23*s34)+s12*(2*p3*s15
     -   -s23*(s15+s34))*s45
       else 
c  Configuration                 (00p3s0p5s)
       d= 2*(16*msq*(p1p4**2*p2p3**2+(p1p3*p2p4-p1p2*p3p4)**2-2*p1p4*(
     -   p1p3*p2p3*p2p4-p1p2*p2p4*p3+p1p2*p2p3*p3p4))+s12*(p3*s15-s23
     -   *s34)*(-(p5*s23)+s15*s45))
       x(1,1) = p3**2*(p5**2-4*msq*s15)+s12sq*(s15-s23)**2+s23*(p5**2*
     -   s23+4*(msq*s15*s34-msq*s34*(s23+s34)))+2*(-(p5*s23*(s15-s34)
     -   *s45)+s12*(p5*s23*(s15-s23-2*s34)+(-s15**2+s23*s34+s15*(s23+
     -   s34))*s45)+p3*(-(p5**2*s23)+p5*s12*(s15+s23)+p5*(s15-s34)*s4
     -   5+2*(msq*(-s15**2+s23*s34+s15*(s23+s34))-s12*s15*s45)))+(s15
     -   -s34)**2*s45sq
       x(2,1) = p3*p5*s12*s15+s12sq*s15**2-p3*p5**2*s23+p3*p5*s12*s23+
     -   p5**2*s23**2+s12sq*s23**2+(p3*(p5-2*s12)*s15-2*s12*s15*(s15-
     -   s23)+s12*(s15+s23)*s34+p5*s23*(-2*s15+s34))*s45-2*(-(p5*s12*
     -   s15*s23)+s12sq*s15*s23+p5*s12*s23**2+p5*s12*s23*s34-msq*(p3*
     -   *2*(p5-s15)+p3*(s12*(-s15+s23)-2*(s15**2-s15*s23)+p5*(2*s15-
     -   s23-s34)+s15*s34+s23*s34-(s15+s34)*s45)-s34*(p5*s23+s12*(-s1
     -   5+s23)+(2*s23+s34)*(s23-s45)+s15*(-2*s23+s45))))+s15*(s15-s3
     -   4)*s45sq
       x(3,1) = 2*msq*(p3+s15-s23-s34)*(p3*(p5-s15)-p5*s23+s12*(-s15+s
     -   23)+s23*s34+s15*s45-s34*s45)+s12*(s12*(s15-s23)**2+p5*s15*s2
     -   3-p5*s23**2+p3*p5*(s15+s23)-s15**2*s45+s15*s23*s45+s15*s34*s
     -   45+s23*s34*s45-2*(p5*s23*s34+p3*s15*s45))
       x(4,1) = s12*s23*(p3*p5-p5*s23+s12*(-s15+s23))+s12*(-2*p3*s15+s
     -   23*(s15+s34))*s45+2*msq*(p3**2*(p5-s15)+p3*(-2*p5*s23+s12*(s
     -   15+s23)+s23*(s15+s34)+(s15-s34)*s45)+s23*(p5*s23+s12*(s15-s2
     -   3-2*s34)-s23*s34-s15*s45+s34*s45))
       x(2,2) = -4*msq*(p3*(p5-s12-s15)+s34*(s12+s23-s45))*(p5-s15+s23
     -   -s45)+(s12*(s15-s23)+p5*s23-s15*s45)**2
       x(3,2) = s12*(s15-s23)*(s12*(s15-s23)+p5*s23-s15*s45)+2*msq*(p5
     -   -s15+s23-s45)*(p3*(-p5+2*s12+s15)+p5*s23+s12*(s15-s23-2*s34)
     -   -s23*s34-s15*s45+s34*s45)
       x(4,2) = s12*s23*(-(p5*s23)+s12*(-s15+s23)+s15*s45)+2*msq*((s12
     -   +s23-s45)*(s12*(s15-s23)+p5*s23-s23*s34-s15*s45+s34*s45)+p3*
     -   (-(s12*s15)+2*s12*s23+s15*s23-(2*s12+s15)*s45+p5*(s12-s23+s4
     -   5)))
       x(3,3) = s12*(s12*(s15-s23)**2+4*msq*(p3+s15-s23-s34)*(p5-s15+s
     -   23-s45))
       x(4,3) = s12*(s12*s23*(-s15+s23)-2*msq*(s12*(-s15+s23)+s23*(p5-
     -   2*(s15-s23)+s34)+p3*(-p5+s15-2*(s23-s45))+(s15-2*s23-s34)*s4
     -   5))
       x(4,4) = s12*(s12*s23**2-4*msq*(s23*(s12+s23-s45)+p3*(-s23+s45)
     -   ))
c          
       x(1,0)= p3**2*p5*(-p5+s15)-s12sq*(s15-s23)**2+s12*s23*((-s15+s2
     -   3)*s34+p5*(-2*(s15-s23)+4*s34))-2*s12*(-s15**2+s23*s34+s15*(
     -   s23+s34))*s45-(p5*s23-s23*s34-s15*s45+s34*s45)*(p5*s23+(-s15
     -   +s34)*s45)+p3*(s12*s15*(s15-s23)+2*p5**2*s23+s15*(4*s12+s15-
     -   s34)*s45-p5*(s23*(s15+s34)+2*(s12*(s15+s23)+(s15-s34)*s45)))
       x(2,0)= -(s12sq*(s15-s23)**2)+s12*s23*((-s15+s23)*s34+2*p5*(-s1
     -   5+s23+s34))-s12*(-2*s15*(s15-s23)+(s15+s23)*s34)*s45-(p5*s23
     -   -s15*s45)*(p5*s23-s23*s34-s15*s45+s34*s45)+p3*(p5**2*s23-p5*
     -   (s12*(s15+s23)+s15*(s23+s45))+s15*(s15*s45+s12*(s15-s23+2*s4
     -   5)))
       x(3,0)= -(s12*(s12*(s15-s23)**2+p5*s15*s23-p5*s23**2-2*p5*s23*s
     -   34+s15*s23*s34-s23**2*s34+p3*(p5*(s15+s23)+s15*(-s15+s23-2*s
     -   45))-s15**2*s45+s15*s23*s45+s15*s34*s45+s23*s34*s45))
       x(4,0)= p3*s12*(-((p5+s15)*s23)+2*s15*s45)+s12*s23*(s12*(s15-s2
     -   3)+s23*(p5+s34)-(s15+s34)*s45)
       endif 
       else 
       If(abs(p5).lt.eps) then 
c  Configuration                 (00p3sp4s0)
       d= 2*(16*msq*(p1p4**2*p2p3**2+(p1p3*p2p4-p1p2*p3p4)**2-2*p1p4*(
     -   p1p3*p2p3*p2p4-p1p2*p2p4*p3+p1p2*p2p3*p3p4)+p1p2*(2*p1p3*p2p
     -   3-p1p2*p3)*p4)+s12*s15*(p4*s12*s23+p3*s15*s45-s23*s34*s45))
       x(1,1) = p4**2*s12sq+s12sq*(s15-s23)**2-4*msq*(p3+s15-s23-s34)*
     -   (p3*s15-s23*s34)-2*(s12*(s15*(2*p3+s15-s23)-(s15+s23)*s34)*s
     -   45-p4*(2*msq*(p3-s23)*(s15-s34)-s12*(s12*(s15+s23)+(-s15+s34
     -   )*s45)))+(s15-s34)**2*s45sq
       x(2,1) = -(s12sq*(-(s15-s23)**2+p4*(s15+s23)))+s12*(-2*p3*s15+s
     -   15*(p4-2*(s15-s23))+(s15+s23)*s34)*s45-2*msq*(-(p4**2*s12)+p
     -   3**2*s15+p4*(2*s15*s23-s23*s34+s12*(s15+s23+s34)-s15*s45+s34
     -   *s45)+p3*(p4*(s12-s15)+s12*(s15-s23)+2*(s15**2-s15*s23)-s15*
     -   s34-s23*s34+(s15+s34)*s45)+s34*(s12*(-s15+s23)+(2*s23+s34)*(
     -   s23-s45)+s15*(-2*s23+s45)))+s15*(s15-s34)*s45sq
       x(3,1) = s12*(s12*(s15-s23)**2-p4*s12*(s15+s23)+(-(s15*(2*p3+s1
     -   5-s23))+(s15+s23)*s34)*s45)+2*msq*(-(p3**2*s15)+p4*s12*(s15+
     -   s23-s34)+p3*(-(p4*s12)-s12*s15-s15**2+s12*s23+s15*s23+s15*s3
     -   4+s23*s34+s15*s45-s34*s45)-(s15-s23-s34)*(s12*s15-s12*s23-s2
     -   3*s34-s15*s45+s34*s45))
       x(4,1) = -(s12sq*(p4+s15-s23)*s23)+s12*(-2*p3*s15+s23*(s15+s34)
     -   )*s45+2*msq*(-(p3**2*s15)+p3*(-(p4*s12)+s12*(s15+s23)+s23*(s
     -   15+s34)+(s15-s34)*s45)+s23*(p4*s12+s12*(s15-s23-2*s34)-s23*s
     -   34-s15*s45+s34*s45))
       x(2,2) = (s12*(-s15+s23)+s15*s45)**2-4*msq*(p4*(s12+s15)*(s12+s
     -   23-s45)+(p3*(s12+s15)-s34*(s12+s23-s45))*(s15-s23+s45))
       x(3,2) = s12*(s15-s23)*(s12*(s15-s23)-s15*s45)-2*msq*(p4*s12*(2
     -   *s12+s15+s23-s45)+(s15-s23+s45)*(p3*(2*s12+s15)+s12*(s15-s23
     -   -2*s34)-s23*s34-s15*s45+s34*s45))
       x(4,2) = s12*s23*(s12*(-s15+s23)+s15*s45)+2*msq*(-(p4*s12*(s12+
     -   s23-s45))-(s12+s23-s45)*(s12*(-s15+s23)+s23*s34+s15*s45-s34*
     -   s45)-p3*(s12*(s15-2*(s23-s45))+s15*(-s23+s45)))
       x(3,3) = s12*(s12*(s15-s23)**2-4*msq*(p4*s12+(p3+s15-s23-s34)*(
     -   s15-s23+s45)))
       x(4,3) = s12*(s12*s23*(-s15+s23)-2*msq*(p4*s12+s12*(-s15+s23)+s
     -   23*(-2*(s15-s23)+s34)+p3*(s15-2*(s23-s45))+(s15-2*s23-s34)*s
     -   45))
       x(4,4) = s12*(s12*s23**2-4*msq*(s23*(s12+s23-s45)+p3*(-s23+s45)
     -   ))
c          
       x(1,0)= s12*(-(p4**2*s12)+p4*(-(p3*s15)+2*s12*(s15+s23)+s23*(2*
     -   s15-s34))+(s15-s23)*(p3*s15+s12*(-s15+s23)-s23*s34))+(s15*(p
     -   3*(4*s12+s15)-2*s12*(p4-s15+s23))-(p3*s15+s15*s23-2*(p4*s12-
     -   s12*(s15+s23)))*s34+s23*s34**2)*s45-(s15-s34)**2*s45sq
       x(2,0)= s12*(p4*(2*s15*s23+s12*(s15+s23))+(s15-s23)*(p3*s15-s12
     -   *s15+s12*s23-s23*s34))-(-(p3*s15*(2*s12+s15))+s12*s15*(p4-2*
     -   (s15-s23))+(s15*s23+s12*(s15+s23))*s34)*s45-s15*(s15-s34)*s4
     -   5sq
       x(3,0)= -(s12*(-(p4*s12*(s15+s23))-(s15-s23)*(p3*s15-s12*s15+s1
     -   2*s23-s23*s34)+(-(s15*(2*p3+s15-s23))+(s15+s23)*s34)*s45))
       x(4,0)= s12*s23*(-(p3*s15)+s12*(p4+s15-s23)+s23*s34)-s12*(-2*p3
     -   *s15+s23*(s15+s34))*s45
       else 
c  Configuration                 (00p3sp4sp5s)
       d= 2*(16*msq*(p1p4**2*p2p3**2+(p1p3*p2p4-p1p2*p3p4)**2-2*p1p4*(
     -   p1p3*p2p3*p2p4-p1p2*p2p4*p3+p1p2*p2p3*p3p4)+p1p2*(2*p1p3*p2p
     -   3-p1p2*p3)*p4)+p3*s12*s15*(-(p5*s23)+s15*s45)+s12*s23*(p4*s1
     -   2*s15+p5*s23*s34-s15*s34*s45))
       x(1,1) = -4*msq*(s15*(p3**2+p4*s23-p3*(p4-s15+s23))-((p4+s15-s2
     -   3)*s23+p3*(-p4+s15+s23))*s34+s23*s34**2)+s12sq*(p3**2+(p4-s3
     -   4)**2-2*p3*(p4+s34))+(p3-s23)**2*(p5-s45)**2+(p3+s15-s23-s34
     -   )**2*(s12-s45)**2-2*(s12*(-p3**2+s23*(-p4+s34)+p3*(p4-2*s15+
     -   s23+s34))*(p5-s45)+s12*(p3**2+p4*(s15+s23-s34)+s34*(-s15+s23
     -   +s34)-p3*(p4-s15+s23+2*s34))*(s12-s45)-(p3-s23)*(p3+s15-s23-
     -   s34)*(p5-s45)*(-s12+s45))
       x(2,1) = p3*p5*s12*s15-p4*s12sq*s15+s12sq*s15**2-p3*p5**2*s23+p
     -   3*p5*s12*s23+p4*p5*s12*s23-p4*s12sq*s23+p5**2*s23**2+s12sq*s
     -   23**2+(s15*(p3*(p5-2*s12)+s12*(p4-2*(s15-s23))-2*p5*s23)+(p5
     -   *s23+s12*(s15+s23))*s34)*s45-2*(-(p5*s12*s15*s23)+s12sq*s15*
     -   s23+p5*s12*s23**2+p5*s12*s23*s34+msq*(-(p4**2*s12)+p3**2*(-p
     -   5+s15)+p4*(-(s23*(p5-2*s15+s34))+s12*(s15+s23+s34)+(-s15+s34
     -   )*s45)+p3*(p4*(p5+s12-s15)+s12*s15-s12*s23+2*(s15**2-s15*s23
     -   )-s15*s34-s23*s34+p5*(-2*s15+s23+s34)+(s15+s34)*s45)+s34*(p5
     -   *s23+s12*(-s15+s23)+(2*s23+s34)*(s23-s45)+s15*(-2*s23+s45)))
     -   )+s15*(s15-s34)*s45sq
       x(3,1) = s12*(s12*s15**2+p5*s15*s23-p5*s23**2+s12*s23**2+p3*p5*
     -   (s15+s23)-p4*s12*(s15+s23)-s15**2*s45+s15*s23*s45+s15*s34*s4
     -   5+s23*s34*s45-2*(s12*s15*s23+p5*s23*s34+p3*s15*s45))+2*msq*(
     -   p3**2*(p5-s15)+p4*s12*(s15+s23-s34)+p3*(-(p4*s12)-s12*s15-s1
     -   5**2+s12*s23+s15*s23+p5*(s15-2*s23-s34)+s15*s34+s23*s34+s15*
     -   s45-s34*s45)-(s15-s23-s34)*(s12*s15+p5*s23-s12*s23-s23*s34-s
     -   15*s45+s34*s45))
       x(4,1) = s12*s23*(p3*p5-s12*(p4+s15)+(-p5+s12)*s23)+s12*(-2*p3*
     -   s15+s23*(s15+s34))*s45+2*msq*(p3**2*(p5-s15)+p3*(-(p4*s12)-2
     -   *p5*s23+s12*(s15+s23)+s23*(s15+s34)+(s15-s34)*s45)+s23*(p4*s
     -   12+p5*s23+s12*(s15-s23-2*s34)-s23*s34-s15*s45+s34*s45))
       x(2,2) = -4*msq*((p4*(-p5+s12+s15)+s34*(p5-s15+s23-s45))*(s12+s
     -   23-s45)+p3*(p5-s12-s15)*(p5-s15+s23-s45))+(s12*(s15-s23)+p5*
     -   s23-s15*s45)**2
       x(3,2) = s12*(s15-s23)*(s12*(s15-s23)+p5*s23-s15*s45)-2*msq*(p3
     -   *(p5-2*s12-s15)*(p5-s15+s23-s45)+p4*s12*(-p5+2*s12+s15+s23-s
     -   45)-(p5-s15+s23-s45)*(p5*s23+s12*(s15-s23-2*s34)-s23*s34-s15
     -   *s45+s34*s45))
       x(4,2) = s12*s23*(-(p5*s23)+s12*(-s15+s23)+s15*s45)+2*msq*(-((s
     -   12+s23-s45)*(p4*s12-s12*s15-p5*s23+s12*s23+s23*s34+s15*s45-s
     -   34*s45))+p3*(-(s12*s15)+2*s12*s23+s15*s23-(2*s12+s15)*s45+p5
     -   *(s12-s23+s45)))
       x(3,3) = s12*(s12*(s15-s23)**2-4*msq*(p4*s12+(p3+s15-s23-s34)*(
     -   -p5+s15-s23+s45)))
       x(4,3) = s12*(s12*s23*(-s15+s23)-2*msq*(p4*s12+s12*(-s15+s23)+s
     -   23*(p5-2*(s15-s23)+s34)+p3*(-p5+s15-2*(s23-s45))+(s15-2*s23-
     -   s34)*s45))
       x(4,4) = s12*(s12*s23**2-4*msq*(s23*(s12+s23-s45)+p3*(-s23+s45)
     -   ))
c          
       x(1,0)= -(p4**2*s12sq)-s12sq*s15**2+p3**2*p5*(-p5+s15)-p5**2*s2
     -   3**2-s12sq*s23**2+2*(-(p5*s12*s15*s23)+s12sq*s15*s23+p5*s12*
     -   s23**2)+4*p5*s12*s23*s34-s12*s15*s23*s34+p5*s23**2*s34+s12*s
     -   23**2*s34+(s23*(2*p5-s34)*(s15-s34)+2*(s12*s15*(s15-s23)-s12
     -   *(s15+s23)*s34))*s45+p4*s12*(-(s23*s34)+2*(-(p5*s23)+s15*s23
     -   +s12*(s15+s23)-s15*s45+s34*s45))+p3*(p4*s12*(2*p5-s15)+s12*s
     -   15*(s15-s23)+2*p5**2*s23+s15*(4*s12+s15-s34)*s45-p5*(s23*(s1
     -   5+s34)+2*(s12*(s15+s23)+(s15-s34)*s45)))-(s15-s34)**2*s45sq
       x(2,0)= p4*s12sq*s15-s12sq*s15**2-p4*p5*s12*s23+p4*s12sq*s23-p5
     -   **2*s23**2-s12sq*s23**2-s12*s15*s23*s34+p5*s23**2*s34+s12*s2
     -   3**2*s34+2*(p4*s12*s15*s23-p5*s12*s15*s23+s12sq*s15*s23+p5*s
     -   12*s23**2+p5*s12*s23*s34)-(s15*(s12*(p4-2*(s15-s23))-2*p5*s2
     -   3)+(s12*s15+(p5+s12+s15)*s23)*s34)*s45+p3*(p5**2*s23-p5*(s12
     -   *(s15+s23)+s15*(s23+s45))+s15*(s15*s45+s12*(s15-s23+2*s45)))
     -   -s15*(s15-s34)*s45sq
       x(3,0)= -(s12*(s12*s15**2+p5*s15*s23-p5*s23**2+s12*s23**2-p4*s1
     -   2*(s15+s23)+s15*s23*s34-s23**2*s34-2*(s12*s15*s23+p5*s23*s34
     -   )+p3*(p5*(s15+s23)+s15*(-s15+s23-2*s45))-s15**2*s45+s15*s23*
     -   s45+s15*s34*s45+s23*s34*s45))
       x(4,0)= s12*s23*(-(p3*(p5+s15))+s12*(p4+s15-s23)+p5*s23+s23*s34
     -   )-s12*(-2*p3*s15+s23*(s15+s34))*s45
       endif 
       endif 
       endif 
       else 
       If(abs(p3).lt.eps) then 
       If(abs(p4).lt.eps) then 
       If(abs(p5).lt.eps) then 
c  Configuration                 (0p2s000)
       d= 2*(16*msq*(p1p4**2*p2p3**2+(p1p3*p2p4-p1p2*p3p4)**2-2*p1p4*(
     -   p1p3*p2p3*p2p4-p1p3*p2*p3p4+p1p2*p2p3*p3p4))+s15*s34*s45*(-(
     -   s12*s23)+p2*s45))
       x(1,1) = s12sq*(s15-s23)**2-4*msq*s34*((p2-s23)*(s15-s23)+s23*s
     -   34)+2*s12*((p2-s15)*(s15-s23)+(s15+s23)*s34)*s45+((p2-s15)**
     -   2-2*(p2+s15)*s34+s34**2)*s45sq
       x(2,1) = s12sq*(s15-s23)**2+s12*(2*msq*(s15-s23)*s34+(2*(p2-s15
     -   )*(s15-s23)+(s15+s23)*s34)*s45)-2*msq*s34*((2*s23+s34)*(s23-
     -   s45)+p2*(2*(s15-s23)+s45)+s15*(-2*s23+s45))+((p2-s15)**2-(p2
     -   +s15)*s34)*s45sq
       x(3,1) = s12sq*(s15-s23)**2+s12*((2*p2-s15)*(s15-s23)+(s15+s23)
     -   *s34)*s45+2*msq*(-2*p2*(s15-s23)*s34+s12*(s15-s23)*(-s15+s23
     -   +s34)-p2*(s15-s23+s34)*s45+(s15-s23-s34)*(s23*s34+s15*s45-s3
     -   4*s45))-p2*(-p2+s15+s34)*s45sq
       x(4,1) = (s12*s23-p2*s45)*(s12*(-s15+s23)+(-p2+s15+s34)*s45)-2*
     -   msq*(p2**2*s45+s23*(s23*s34+s12*(-s15+s23+2*s34)+s15*s45-s34
     -   *s45)+p2*(s12*(s15-s23)-s23*s34-(s15+s23+s34)*s45))
       x(2,2) = 4*msq*s34*(-p2+s12+s23-s45)*(s15-s23+s45)+(s12*(s15-s2
     -   3)+(p2-s15)*s45)**2
       x(3,2) = (s12*(s15-s23)+p2*s45)*(s12*(s15-s23)+(p2-s15)*s45)+2*
     -   msq*(s15-s23+s45)*((-2*p2+s23)*s34+s12*(-s15+s23+2*s34)-(p2-
     -   s15+s34)*s45)
       x(4,2) = (s12*s23-p2*s45)*(s12*(-s15+s23)+(-p2+s15)*s45)+2*msq*
     -   (-p2+s12+s23-s45)*(s12*(s15-s23)-s23*s34+(p2-s15+s34)*s45)
       x(3,3) = (s12*(s15-s23)+p2*s45)**2-4*msq*(s15-s23+s45)*(s12*(s1
     -   5-s23-s34)+p2*(s34+s45))
       x(4,3) = (s12*s23-p2*s45)*(s12*(-s15+s23)-p2*s45)-2*msq*(p2**2*
     -   s45+s12*(s12*(-s15+s23)+(2*s23+s34)*(s23-s45)+s15*(-2*s23+s4
     -   5))+p2*(s12*(s15-s23-s45)-s23*(s34+2*s45)+s45*(s15+s34+2*s45
     -   )))
       x(4,4) = (s12*s23-p2*s45)*(s12*s23-p2*s45+4*msq*(p2-s12-s23+s45
     -   ))
c          
       x(1,0)= -(s12*(s15-s23)*(s12*(s15-s23)+s23*s34))-(2*s12*(p2-s15
     -   )*(s15-s23)+(s15*s23+p2*(-2*s15+s23)+2*s12*(s15+s23))*s34-s2
     -   3*s34**2)*s45-(p2**2+(s15-s34)**2-2*p2*(s15+s34))*s45sq
       x(2,0)= -(s12*(s15-s23)*(s12*(s15-s23)+s23*s34))-(2*s12*(p2-s15
     -   )*(s15-s23)+(s15*s23+p2*(-2*s15+s23)+s12*(s15+s23))*s34)*s45
     -   -((p2-s15)**2-(p2+s15)*s34)*s45sq
       x(3,0)= -(s12*(s15-s23)*(s12*(s15-s23)+s23*s34))-(s12*(2*p2-s15
     -   )*(s15-s23)+(p2*(-2*s15+s23)+s12*(s15+s23))*s34)*s45+p2*(-p2
     -   +s15+s34)*s45sq
       x(4,0)= -((s12*s23-p2*s45)*(s12*(-s15+s23)-s23*s34+(-p2+s15+s34
     -   )*s45))
       else 
c  Configuration                 (0p2s00p5s)
       d= 2*(16*msq*(p1p4**2*p2p3**2+(p1p3*p2p4-p1p2*p3p4)**2-2*p1p4*(
     -   p1p3*p2p3*p2p4-p1p3*p2*p3p4+p1p2*p2p3*p3p4))+s34*(s12*s23-p2
     -   *s45)*(p5*s23-s15*s45))
       x(1,1) = s12sq*s34**2-4*msq*s34*((p2-s23)*(s15-s23)+s23*s34)+(p
     -   2-s23)**2*(p5-s45)**2+(-s15+s23+s34)**2*(s12-s45)**2+2*(s12*
     -   (p2-s23)*s34*(p5-s45)-s12*s34*(-s15+s23+s34)*(s12-s45)+(p2*(
     -   s15-s23+s34)+s23*(-s15+s23+s34))*(p5-s45)*(-s12+s45))
       x(2,1) = p2**2*p5**2+s12sq*s15**2+p5**2*s23**2+s12sq*s23**2-(2*
     -   (p2-s15)*(p2*p5-p5*s23+s12*(-s15+s23))-(p2*p5+p5*s23+s12*(s1
     -   5+s23))*s34)*s45-2*(p2*p5*s12*s15+p2*p5**2*s23-p2*p5*s12*s23
     -   -p5*s12*s15*s23+s12sq*s15*s23+p5*s12*s23**2+p5*s12*s23*s34+m
     -   sq*s34*(s12*(-s15+s23)+s23*(p5-2*(s15-s23)+s34)+(s15-2*s23-s
     -   34)*s45+p2*(-p5+2*(s15-s23)+s45)))+((p2-s15)**2-(p2+s15)*s34
     -   )*s45sq
       x(3,1) = p2**2*(p5-s45)**2+2*msq*(s15-s23-s34)*(-(p5*s23)+s12*(
     -   -s15+s23)+s23*s34+s15*s45-s34*s45)+s12*(s12*(s15-s23)**2+p5*
     -   s23*(s15-s23-2*s34)+(-s15**2+s23*s34+s15*(s23+s34))*s45)+p2*
     -   (-((p5-s45)*(2*s12*(s15-s23)+p5*s23-(s15+s34)*s45))+2*msq*(-
     -   2*(s15-s23)*s34+p5*(s15-s23+s34)-(s15-s23+s34)*s45))
       x(4,1) = (s12*s23-p2*s45)*(p2*p5-s12*s15-p5*s23+s12*s23+(-p2+s1
     -   5+s34)*s45)+2*msq*(p2**2*(p5-s45)+s23*(p5*s23+s12*(s15-s23-2
     -   *s34)-s23*s34-s15*s45+s34*s45)+p2*(-(s12*s15)-2*p5*s23+s12*s
     -   23+s23*s34+(s15+s23+s34)*s45))
       x(2,2) = 4*msq*s34*(p5-s15+s23-s45)*(p2-s12-s23+s45)+(p2*p5-s12
     -   *s15-p5*s23+s12*s23-p2*s45+s15*s45)**2
       x(3,2) = 2*msq*(p5-s15+s23-s45)*(-(p2*p5)+s12*s15+p5*s23-s12*s2
     -   3-s23*s34+2*(p2*s34-s12*s34)+(p2-s15+s34)*s45)+(s12*(s15-s23
     -   )+p2*(-p5+s45))*(s12*s15+p5*s23-s12*s23-s15*s45+p2*(-p5+s45)
     -   )
       x(4,2) = (s12*s23-p2*s45)*(-(p5*s23)+s12*(-s15+s23)+p2*(p5-s45)
     -   +s15*s45)+2*msq*(-p2+s12+s23-s45)*(s12*s15+p5*s23-s23*(s12+s
     -   34)+(-s15+s34)*s45+p2*(-p5+s45))
       x(3,3) = -4*msq*(s12*(-s15+s23+s34)+p2*(p5-s34-s45))*(p5-s15+s2
     -   3-s45)+(s12*(s15-s23)+p2*(-p5+s45))**2
       x(4,3) = (s12*(-s15+s23)+p2*(p5-s45))*(s12*s23-p2*s45)-2*msq*(p
     -   2**2*(-p5+s45)+s12*(p5*s23+s12*(-s15+s23)+(2*s23+s34)*(s23-s
     -   45)+s15*(-2*s23+s45))+p2*(p5*(s12+s23-2*s45)+s12*(s15-s23-s4
     -   5)+s15*s45-(s23-s45)*(s34+2*s45)))
       x(4,4) = (s12*s23-p2*s45)*(s12*s23-p2*s45+4*msq*(p2-s12-s23+s45
     -   ))
c          
       x(1,0)= -(s12sq*(s15-s23)**2)+s12*s23*((-s15+s23)*s34+p5*(-2*(s
     -   15-s23)+4*s34))-p2**2*(p5-s45)**2-2*s12*(-s15**2+s23*s34+s15
     -   *(s23+s34))*s45-(p5*s23-s23*s34-s15*s45+s34*s45)*(p5*s23+(-s
     -   15+s34)*s45)+p2*(2*p5**2*s23+s45*(-(s23*s34)-2*(s12*s15-s12*
     -   s23-s15*s34-(s15+s34)*s45))+p5*(-(s23*s34)+2*(s12*s15-s12*s2
     -   3-(s15+s23+s34)*s45)))
       x(2,0)= -(s12sq*(s15-s23)**2)+s12*s23*((-s15+s23)*s34+2*p5*(-s1
     -   5+s23+s34))-p2**2*(p5-s45)**2-s12*(-2*s15*(s15-s23)+(s15+s23
     -   )*s34)*s45-(p5*s23-s15*s45)*(p5*s23-s23*s34-s15*s45+s34*s45)
     -   +p2*(p5*(2*(s12*s15+p5*s23-s12*s23)-s23*s34)-(2*(s12*(s15-s2
     -   3)+p5*(s15+s23))+(p5-2*s15+s23)*s34)*s45+(2*s15+s34)*s45sq)
       x(3,0)= -(p2**2*(p5-s45)**2)-s12*(s12*(s15-s23)**2+p5*s23*(s15-
     -   s23-2*s34)+s15*s23*s34-s23**2*s34-s15**2*s45+s15*s23*s45+s15
     -   *s34*s45+s23*s34*s45)+p2*(p5*(2*s12*(s15-s23)+s23*(p5-s34))-
     -   (2*s12*(s15-s23)+(-2*s15+s23)*s34+p5*(s15+s23+s34))*s45+(s15
     -   +s34)*s45sq)
       x(4,0)= (s12*s23-p2*s45)*(s12*s15+s23*(p5-s12+s34)-(s15+s34)*s4
     -   5+p2*(-p5+s45))
       endif 
       else 
       If(abs(p5).lt.eps) then 
c  Configuration                 (0p2s0p4s0)
       d= 2*(16*msq*(p1p4**2*p2p3**2+(p1p3*p2p4-p1p2*p3p4)**2-2*p1p4*(
     -   p1p3*p2p3*p2p4-p1p3*p2*p3p4+p1p2*p2p3*p3p4)-p1p3*(p1p3*p2-2*
     -   p1p2*p2p3)*p4)+s15*(s12*s23-p2*s45)*(p4*s12-s34*s45))
       x(1,1) = s12sq*(p4**2+(s15-s23)**2-2*p4*(s15+s23))+4*msq*(-(p2*
     -   p4*(p2+p4-s15))+p4*(p2-s15)*s23+((p4+s15-s23)*s23+p2*(p4-s15
     -   +s23))*s34-s23*s34**2)+2*s12*(p4*s15-s15**2+p2*(p4+s15-s23)+
     -   s15*s23-p4*s34+s15*s34+s23*s34)*s45+(p2**2+(s15-s34)**2-2*p2
     -   *(s15+s34))*s45sq
       x(2,1) = s12sq*(s15-s23)**2+s12*(2*(p2-s15)*(s15-s23)+(s15+s23)
     -   *s34)*s45+p4*s12*(-(s12*(s15+s23))+(p2+s15)*s45)-2*msq*(2*p2
     -   **2*p4-p4**2*s12+p4*(2*s15*s23-s23*s34+s12*(s15+s23+s34)-s15
     -   *s45+s34*s45)+p2*(p4**2-p4*(2*(s12+s15+s23)+s34-s45)+s34*(2*
     -   (s15-s23)+s45))+s34*(s12*(-s15+s23)+(2*s23+s34)*(s23-s45)+s1
     -   5*(-2*s23+s45)))+((p2-s15)**2-(p2+s15)*s34)*s45sq
       x(3,1) = s12sq*(s15-s23)**2+s12*((2*p2-s15)*(s15-s23)+(s15+s23)
     -   *s34)*s45+p4*s12*(-(s12*(s15+s23))+p2*s45)+2*msq*(-2*p2**2*p
     -   4+p4*s12*(s15+s23-s34)-(s15-s23-s34)*(s12*s15-s12*s23-s23*s3
     -   4-s15*s45+s34*s45)+p2*(-2*(s15-s23)*s34+p4*(s15+s23+s34+2*(s
     -   12-s45))-(s15-s23+s34)*s45))-p2*(-p2+s15+s34)*s45sq
       x(4,1) = (s12*s23-p2*s45)*(-(s12*(p4+s15-s23))+(-p2+s15+s34)*s4
     -   5)-2*msq*(p2**2*(p4+s45)+s23*(-(p4*s12)+s23*s34+s12*(-s15+s2
     -   3+2*s34)+s15*s45-s34*s45)-p2*(s12*(-s15+s23)+s23*s34+p4*(s12
     -   +s23-2*s45)+(s15+s23+s34)*s45))
       x(2,2) = (s12*(s15-s23)+(p2-s15)*s45)**2-4*msq*(p2-s12-s23+s45)
     -   *(p2*p4-p4*(s12+s15)+s34*(s15-s23+s45))
       x(3,2) = (s12*(s15-s23)+p2*s45)*(s12*(s15-s23)+(p2-s15)*s45)-2*
     -   msq*(2*p2**2*p4+p4*s12*(2*s12+s15+s23-s45)-(s15-s23+s45)*(s2
     -   3*s34+s12*(-s15+s23+2*s34)+s15*s45-s34*s45)+p2*(-(p4*(4*s12+
     -   s15+s23-s45))+(s15-s23+s45)*(2*s34+s45)))
       x(4,2) = (s12*s23-p2*s45)*(s12*(-s15+s23)+(-p2+s15)*s45)-2*msq*
     -   (p2-s12-s23+s45)*(p2*p4-p4*s12+s12*s15-s12*s23-s23*s34+(p2-s
     -   15+s34)*s45)
       x(3,3) = (s12*(s15-s23)+p2*s45)**2-4*msq*(p2**2*p4+s12*(p4*s12+
     -   (s15-s23-s34)*(s15-s23+s45))+p2*(-2*p4*s12+(s15-s23+s45)*(s3
     -   4+s45)))
       x(4,3) = (s12*s23-p2*s45)*(s12*(-s15+s23)-p2*s45)-2*msq*(p2**2*
     -   (p4+s45)+s12*(p4*s12+s12*(-s15+s23)+s23*(-2*(s15-s23)+s34)+(
     -   s15-2*s23-s34)*s45)+p2*(-2*p4*s12+s12*(s15-s23-s45)+s15*s45-
     -   (s23-s45)*(s34+2*s45)))
       x(4,4) = (s12*s23-p2*s45)*(s12*s23-p2*s45+4*msq*(p2-s12-s23+s45
     -   ))
c          
       x(1,0)= s12*(-(p4**2*s12)+p2*p4*(p4-s15-s23)-s12*(s15-s23)**2+2
     -   *p4*(s15*s23+s12*(s15+s23))-(p4+s15-s23)*s23*s34)-(-(p2**2*p
     -   4)-2*s12*s15*(-p4+s15-s23)+(s15*s23-2*(p4*s12-s12*(s15+s23))
     -   )*s34-s23*s34**2+p2*(2*s12*(s15-s23)+(-2*s15+s23)*s34+p4*(2*
     -   s12+s15+s34)))*s45-(p2**2+(s15-s34)**2-2*p2*(s15+s34))*s45sq
       x(2,0)= s12*(-(p2*p4*(s15+s23))+p4*(2*s15*s23+s12*(s15+s23))-(s
     -   15-s23)*(s12*(s15-s23)+s23*s34))-(-(p2**2*p4)+s12*s15*(p4-2*
     -   (s15-s23))+(s15*s23+s12*(s15+s23))*s34+p2*(p4*(s12+s15)+s23*
     -   s34+2*(s12*s15-s12*s23-s15*s34)))*s45-((p2-s15)**2-(p2+s15)*
     -   s34)*s45sq
       x(3,0)= p2**2*(p4-s45)*s45+s12*(p4*s12*(s15+s23)-(s15-s23)*(s12
     -   *s15-s12*s23+s23*s34)+(s15*(s15-s23)-(s15+s23)*s34)*s45)+p2*
     -   (-(p4*s12*(s15+s23+s45))+s45*(-(s23*s34)-2*(s12*s15-s12*s23-
     -   s15*s34)+(s15+s34)*s45))
       x(4,0)= -((s12*s23-p2*s45)*(p2*p4-p4*s12-s12*s15+s12*s23-s23*s3
     -   4+(-p2+s15+s34)*s45))
       else 
c  Configuration                 (0p2s0p4sp5s)
       d= 2*(16*msq*(p1p4**2*p2p3**2+(p1p3*p2p4-p1p2*p3p4)**2-2*p1p4*(
     -   p1p3*p2p3*p2p4-p1p3*p2*p3p4+p1p2*p2p3*p3p4)-p1p3*(p1p3*p2-2*
     -   p1p2*p2p3)*p4)+(s12*s23-p2*s45)*(-(p2*p4*p5)+p4*s12*s15+p5*s
     -   23*s34-s15*s34*s45))
       x(1,1) = s12sq*(p4-s34)**2-4*msq*(p2**2*p4+s23*(p4*(s15-s34)+s3
     -   4*(-s15+s23+s34))+p2*(p4**2+(s15-s23)*s34-p4*(s15+s23+s34)))
     -   +(p2-s23)**2*(p5-s45)**2+(-4*p2*p4+(-s15+s23+s34)**2)*(s12-s
     -   45)**2-2*(s12*(p2-s23)*(p4-s34)*(p5-s45)+s12*(-2*p2*p4+p4*(s
     -   15+s23-s34)+s34*(-s15+s23+s34))*(s12-s45)-(p2*(-2*p4+s15-s23
     -   +s34)+s23*(-s15+s23+s34))*(p5-s45)*(-s12+s45))
       x(2,1) = -(p4*s12sq*s15)+s12sq*s15**2+p4*p5*s12*s23-p4*s12sq*s2
     -   3+p5**2*s23**2+s12sq*s23**2+p2**2*(p5-s45)**2+(p4*s12*s15-2*
     -   s12*s15*(s15-s23)+s12*(s15+s23)*s34+p5*s23*(-2*s15+s34))*s45
     -   +p2*(p4*(p5*(s12-2*s45)+s12*s45)-(p5-s45)*(2*(s12*(s15-s23)+
     -   p5*s23)-(2*s15+s34)*s45))-2*(-(p5*s12*s15*s23)+s12sq*s15*s23
     -   +p5*s12*s23**2+p5*s12*s23*s34+msq*(2*p2**2*p4-p4**2*s12+p4*(
     -   -(s23*(p5-2*s15+s34))+s12*(s15+s23+s34)+(-s15+s34)*s45)+s34*
     -   (p5*s23+s12*(-s15+s23)+(2*s23+s34)*(s23-s45)+s15*(-2*s23+s45
     -   ))+p2*(p4**2+s34*(-p5+2*(s15-s23)+s45)+p4*(p5-2*(s12+s15+s23
     -   )-s34+s45))))+s15*(s15-s34)*s45sq
       x(3,1) = p2**2*(p5-s45)**2+s12*(s12*(s15-s23)**2+p5*s15*s23-p5*
     -   s23**2-p4*s12*(s15+s23)-2*p5*s23*s34-s15**2*s45+s15*s23*s45+
     -   s15*s34*s45+s23*s34*s45)-2*msq*(2*p2**2*p4-p4*s12*(s15+s23-s
     -   34)-p2*(2*(-s15+s23)*s34+p5*(s15-s23+s34)+p4*(-2*(p5-s12)+s1
     -   5+s23+s34))+p2*(2*p4+s15-s23+s34)*s45+(s15-s23-s34)*(s12*(s1
     -   5-s23)+p5*s23-s23*s34-s15*s45+s34*s45))+p2*(p4*s12*(p5+s45)-
     -   (p5-s45)*(2*s12*(s15-s23)+p5*s23-(s15+s34)*s45))
       x(4,1) = (s12*s23-p2*s45)*(-(s12*(p4+s15))+(-p5+s12)*s23+p2*(p5
     -   -s45)+(s15+s34)*s45)+2*msq*(-(p2**2*(p4-p5+s45))+s23*(p4*s12
     -   +p5*s23+s12*(s15-s23-2*s34)-s23*s34-s15*s45+s34*s45)+p2*(-(s
     -   12*s15)-2*p5*s23+s12*s23+s23*s34+p4*(s12+s23-2*s45)+(s15+s23
     -   +s34)*s45))
       x(2,2) = (p2*p5-s12*s15-p5*s23+s12*s23-p2*s45+s15*s45)**2-4*msq
     -   *(p2-s12-s23+s45)*(p2*p4+p4*(p5-s12-s15)+s34*(-p5+s15-s23+s4
     -   5))
       x(3,2) = (s12*(s15-s23)+p2*(-p5+s45))*(s12*(s15-s23)+p5*s23-s15
     -   *s45+p2*(-p5+s45))-2*msq*(2*p2**2*p4+p4*s12*(-p5+2*s12+s15+s
     -   23-s45)-(p5-s15+s23-s45)*(p5*s23+s12*(s15-s23-2*s34)-s23*s34
     -   -s15*s45+s34*s45)+p2*((p5-s15+s23-s45)*(p5-2*s34-s45)+p4*(p5
     -   -4*s12-s15-s23+s45)))
       x(4,2) = (s12*s23-p2*s45)*(-(s12*s15)-p5*s23+s12*s23+p2*(p5-s45
     -   )+s15*s45)-2*msq*(p2-s12-s23+s45)*(-(p4*s12)+s12*s15+p5*s23-
     -   s12*s23-s23*s34-s15*s45+s34*s45+p2*(p4-p5+s45))
       x(3,3) = (s12*(s15-s23)+p2*(-p5+s45))**2-4*msq*(p2**2*p4+p2*(-2
     -   *p4*s12+(p5-s15+s23-s45)*(p5-s34-s45))+s12*(p4*s12+(s15-s23-
     -   s34)*(-p5+s15-s23+s45)))
       x(4,3) = (s12*(-s15+s23)+p2*(p5-s45))*(s12*s23-p2*s45)-2*msq*(p
     -   2**2*(p4-p5+s45)+s12*(p4*s12+s12*(-s15+s23)+s23*(p5-2*(s15-s
     -   23)+s34)+(s15-2*s23-s34)*s45)+p2*(s12*s15-s12*s23-s23*s34+p5
     -   *(s12+s23-2*s45)-s12*s45+s15*s45+s34*s45-2*(p4*s12+s23*s45-s
     -   45sq)))
       x(4,4) = (s12*s23-p2*s45)*(s12*s23-p2*s45+4*msq*(p2-s12-s23+s45
     -   ))
c          
       x(1,0)= -(p4**2*s12sq)-s12sq*s15**2-p5**2*s23**2-s12sq*s23**2+2
     -   *(-(p5*s12*s15*s23)+s12sq*s15*s23+p5*s12*s23**2)+4*p5*s12*s2
     -   3*s34-s12*s15*s23*s34+p5*s23**2*s34+s12*s23**2*s34+(s23*(2*p
     -   5-s34)*(s15-s34)+2*(s12*s15*(s15-s23)-s12*(s15+s23)*s34))*s4
     -   5+p2**2*(-(p5-s45)**2+p4*(p5+s45))+p4*s12*(-(s23*s34)+2*(-(p
     -   5*s23)+s15*s23+s12*(s15+s23)-s15*s45+s34*s45))-(s15-s34)**2*
     -   s45sq+p2*(p4**2*s12+p5*(2*(s12*s15+p5*s23-s12*s23)-s23*s34)-
     -   (s23*s34+2*(s12*s15-s12*s23-s15*s34+p5*(s15+s23+s34)))*s45-p
     -   4*(s12*(s15+s23)+p5*(2*s12+s23-4*s45)+(2*s12+s15+s34)*s45)+2
     -   *(s15+s34)*s45sq)
       x(2,0)= p4*s12sq*s15-s12sq*s15**2-p4*p5*s12*s23+p4*s12sq*s23-p5
     -   **2*s23**2-s12sq*s23**2-s12*s15*s23*s34+p5*s23**2*s34+s12*s2
     -   3**2*s34+2*(p4*s12*s15*s23-p5*s12*s15*s23+s12sq*s15*s23+p5*s
     -   12*s23**2+p5*s12*s23*s34)-(s15*(s12*(p4-2*(s15-s23))-2*p5*s2
     -   3)+(s12*s15+(p5+s12+s15)*s23)*s34)*s45-p2**2*((p5-s45)**2-p4
     -   *(p5+s45))-s15*(s15-s34)*s45sq-p2*(p5*(-2*(s12*s15+p5*s23-s1
     -   2*s23)+s23*s34)+(2*((p5+s12)*s15+(p5-s12)*s23)+(p5-2*s15+s23
     -   )*s34)*s45+p4*(p5*(s12+s23-2*s45)+s15*s45+s12*(s15+s23+s45))
     -   -(2*s15+s34)*s45sq)
       x(3,0)= s12*(-(s12*(s15-s23)**2)+p5*s23*(-s15+s23)+p4*s12*(s15+
     -   s23)+s23*(2*p5-s15+s23)*s34-(-s15**2+s23*s34+s15*(s23+s34))*
     -   s45)+p2**2*(-(p5-s45)**2+p4*(p5+s45))+p2*(p5*(2*s12*(s15-s23
     -   )+s23*(p5-s34))-(s23*s34+p5*(s15+s23+s34)+2*(s12*s15-s12*s23
     -   -s15*s34))*s45-p4*s12*(p5+s15+s23+s45)+(s15+s34)*s45sq)
       x(4,0)= (s12*s23-p2*s45)*(p4*s12+s12*s15+p5*s23-s12*s23+s23*s34
     -   -p2*(p4+p5-s45)-(s15+s34)*s45)
       endif 
       endif 
       else 
       If(abs(p4).lt.eps) then 
       If(abs(p5).lt.eps) then 
c  Configuration                 (0p2sp3s00)
       d= 2*(16*msq*(p1p4**2*(p2p3**2-p2*p3)+(p1p3*p2p4-p1p2*p3p4)**2-
     -   2*p1p4*(p1p3*p2p3*p2p4-p1p2*p2p4*p3-p1p3*p2*p3p4+p1p2*p2p3*p
     -   3p4))+s15*s45*(p3*s12*s15-s12*s23*s34+p2*s34*s45))
       x(1,1) = s12sq*(s15-s23)**2+4*msq*(p2*(s15-s23)*(p3-s34)-(p3+s1
     -   5-s23-s34)*(p3*s15-s23*s34))+2*s12*(p2*(s15-s23)-s15*(2*p3+s
     -   15-s23)+(s15+s23)*s34)*s45+(p2**2+(s15-s34)**2-2*p2*(s15+s34
     -   ))*s45sq
       x(2,1) = s12sq*(s15-s23)**2+s12*(2*(p2*(s15-s23)-s15*(p3+s15-s2
     -   3))+(s15+s23)*s34)*s45+2*msq*(-(p3**2*s15)+s34*(s12*(s15-s23
     -   )-(2*s23+s34)*(s23-s45)+s15*(2*s23-s45))+p2*(p3-s34)*(2*(s15
     -   -s23)+s45)+p3*(s12*(-s15+s23)-2*(s15**2-s15*s23)+s15*s34+s23
     -   *s34-(s15+s34)*s45))+((p2-s15)**2-(p2+s15)*s34)*s45sq
       x(3,1) = s12sq*(s15-s23)**2+s12*(2*p2*(s15-s23)-s15*(2*p3+s15-s
     -   23)+(s15+s23)*s34)*s45+2*msq*(2*p2*(s15-s23)*(p3-s34)+p2*(p3
     -   -s15+s23-s34)*s45-(p3+s15-s23-s34)*(p3*s15+s12*s15-s12*s23-s
     -   23*s34-s15*s45+s34*s45))-p2*(-p2+s15+s34)*s45sq
       x(4,1) = s12sq*s23*(-s15+s23)+s12*(-2*p3*s15+p2*(s15-2*s23)+s23
     -   *(s15+s34))*s45+2*msq*(-(p3**2*s15)-p2**2*s45+p3*(s12*(s15+s
     -   23)+s23*(s15+s34)+(s15-s34)*s45)-s23*(s23*s34+s12*(-s15+s23+
     -   2*s34)+s15*s45-s34*s45)+p2*(-(s12*s15)+s12*s23+s23*s34+(s15+
     -   s23+s34)*s45+p3*(s15-2*s23+s45)))+p2*(p2-s15-s34)*s45sq
       x(2,2) = -4*msq*(p3*(s12+s15)+p2*(-p3+s34)-s34*(s12+s23-s45))*(
     -   s15-s23+s45)+(s12*(s15-s23)+(p2-s15)*s45)**2
       x(3,2) = (s12*(s15-s23)+p2*s45)*(s12*(s15-s23)+(p2-s15)*s45)-2*
     -   msq*(s15-s23+s45)*(p3*s15+s12*s15-s12*s23-s23*s34-2*(p2*p3-p
     -   3*s12-p2*s34+s12*s34)+(p2-s15+s34)*s45)
       x(4,2) = (s12*s23-p2*s45)*(s12*(-s15+s23)+(-p2+s15)*s45)+2*msq*
     -   (-(p2**2*s45)-(s12+s23-s45)*(s12*(-s15+s23)+s23*s34+(s15-s34
     -   )*s45)-p3*(s12*(s15-2*(s23-s45))+s15*(-s23+s45))+p2*(p3*(s15
     -   -2*(s23-s45))+s15*s45+s12*(-s15+s23+s45)+(s23-s45)*(s34+s45)
     -   ))
       x(3,3) = (s12*(s15-s23)+p2*s45)**2-4*msq*(s15-s23+s45)*(s12*(p3
     -   +s15-s23-s34)+p2*(-p3+s34+s45))
       x(4,3) = (s12*s23-p2*s45)*(s12*(-s15+s23)-p2*s45)-2*msq*(p2**2*
     -   s45+s12*(-(s12*s15)+s12*s23-2*(s15*s23-s23**2)+s23*s34+p3*(s
     -   15-2*(s23-s45))+(s15-2*s23-s34)*s45)-p2*(p3*(s15-2*(s23-s45)
     -   )-s15*s45+s12*(-s15+s23+s45)+(s23-s45)*(s34+2*s45)))
       x(4,4) = (s12*s23-p2*s45)**2-4*msq*(s12*s23*(s12+s23-s45)+p2**2
     -   *s45+p3*s12*(-s23+s45)+p2*((p3-s12)*s23-(p3+s12+s23)*s45+s45
     -   sq))
c          
       x(1,0)= -(s12*(s15-s23)*(s12*(s15-s23)+s23*s34))-(2*s12*(p2-s15
     -   )*(s15-s23)+(s15*s23+p2*(-2*s15+s23)+2*s12*(s15+s23))*s34-s2
     -   3*s34**2)*s45+p3*s15*(-((p2-s15+s34)*s45)+s12*(s15-s23+4*s45
     -   ))-(p2**2+(s15-s34)**2-2*p2*(s15+s34))*s45sq
       x(2,0)= -(s12*(s15-s23)*(s12*(s15-s23)+s23*s34))-(2*s12*(p2-s15
     -   )*(s15-s23)+((-2*p2+s12)*s15+(p2+s12+s15)*s23)*s34)*s45+p3*s
     -   15*((-p2+s15)*s45+s12*(s15-s23+2*s45))-((p2-s15)**2-(p2+s15)
     -   *s34)*s45sq
       x(3,0)= s12*(s15-s23)*(p3*s15+s12*(-s15+s23)-s23*s34)-(-(s12*s1
     -   5*(2*p3+s15-s23))+s12*(s15+s23)*s34+p2*(p3*s15+s23*s34+2*(s1
     -   2*s15-s12*s23-s15*s34)))*s45+p2*(-p2+s15+s34)*s45sq
       x(4,0)= -(p3*s15*(s12*(s23-2*s45)+p2*s45))+(s12*s23-p2*s45)*(s1
     -   2*(s15-s23)+s23*s34+p2*s45-(s15+s34)*s45)
       else 
c  Configuration                 (0p2sp3s0p5s)
       d= 2*(16*msq*(p1p4**2*(p2p3**2-p2*p3)+(p1p3*p2p4-p1p2*p3p4)**2-
     -   2*p1p4*(p1p3*p2p3*p2p4-p1p2*p2p4*p3-p1p3*p2*p3p4+p1p2*p2p3*p
     -   3p4))+(p5*s23-s15*s45)*(p2*p3*p5-p3*s12*s15+s12*s23*s34-p2*s
     -   34*s45))
       x(1,1) = s12sq*(p3-s34)**2-4*msq*(-(p2*(s15-s23)*(p3-s34))+(p3+
     -   s15-s23-s34)*(p3*s15-s23*s34))+(p2**2+(p3-s23)**2-2*p2*(p3+s
     -   23))*(p5-s45)**2+(p3+s15-s23-s34)**2*(s12-s45)**2-2*(s12*(-p
     -   3**2+p2*(p3-s34)+s23*s34+p3*(-2*s15+s23+s34))*(p5-s45)+s12*(
     -   p3-s34)*(p3+s15-s23-s34)*(s12-s45)+(-((p3-s23)*(p3+s15-s23-s
     -   34))+p2*(p3-s15+s23-s34))*(p5-s45)*(-s12+s45))
       x(2,1) = p3*p5*s12*s15+s12sq*s15**2-p3*p5**2*s23+p3*p5*s12*s23+
     -   p5**2*s23**2+s12sq*s23**2+p2**2*(p5-s45)**2+(p3*(p5-2*s12)*s
     -   15-2*s12*s15*(s15-s23)+s12*(s15+s23)*s34+p5*s23*(-2*s15+s34)
     -   )*s45-p2*(p5-s45)*(p3*p5+2*(s12*(s15-s23)+p5*s23)-(2*s15+s34
     -   )*s45)-2*(-(p5*s12*s15*s23)+s12sq*s15*s23+p5*s12*s23**2+p5*s
     -   12*s23*s34+msq*(p3**2*(-p5+s15)+p2*(p3-s34)*(p5-2*(s15-s23)-
     -   s45)+p3*(s12*(s15-s23)+2*(s15**2-s15*s23)-s15*s34-s23*s34+p5
     -   *(-2*s15+s23+s34)+(s15+s34)*s45)+s34*(p5*s23+s12*(-s15+s23)+
     -   (2*s23+s34)*(s23-s45)+s15*(-2*s23+s45))))+s15*(s15-s34)*s45s
     -   q
       x(3,1) = p2**2*(p5-s45)**2-p2*(p5-s45)*(p3*p5+2*s12*(s15-s23)+p
     -   5*s23-(s15+s34)*s45)+s12*(s12*(s15-s23)**2+p5*s15*s23-p5*s23
     -   **2+p3*p5*(s15+s23)-s15**2*s45+s15*s23*s45+s15*s34*s45+s23*s
     -   34*s45-2*(p5*s23*s34+p3*s15*s45))-2*msq*((p3+s15-s23-s34)*(p
     -   3*(-p5+s15)+s12*(s15-s23)+p5*s23-s23*s34-s15*s45+s34*s45)+p2
     -   *(2*(s15-s23)*s34-p5*(s15-s23+s34)+p3*(p5-2*(s15-s23)-s45)+(
     -   s15-s23+s34)*s45))
       x(4,1) = s12*s23*(p2*p5+p3*p5-s12*s15-p5*s23+s12*s23)+(-(p2**2*
     -   p5)+p2*(p3*p5+s12*s15+p5*s23-2*s12*s23)+s12*(-2*p3*s15+s23*(
     -   s15+s34)))*s45+2*msq*(p3**2*(p5-s15)+p2**2*(p5-s45)+p3*(-2*p
     -   5*s23+s12*(s15+s23)+s23*(s15+s34)+(s15-s34)*s45)+s23*(p5*s23
     -   +s12*(s15-s23-2*s34)-s23*s34-s15*s45+s34*s45)+p2*(-(s12*s15)
     -   -2*p5*s23+s12*s23+s23*s34+(s15+s23+s34)*s45+p3*(s15-2*(p5+s2
     -   3)+s45)))+p2*(p2-s15-s34)*s45sq
       x(2,2) = -4*msq*(p3*(p5-s12-s15)+p2*(p3-s34)+s34*(s12+s23-s45))
     -   *(p5-s15+s23-s45)+(p2*p5-s12*s15-p5*s23+s12*s23-p2*s45+s15*s
     -   45)**2
       x(3,2) = -2*msq*(p5-s15+s23-s45)*(p3*(p5-2*s12-s15)-s12*s15-p5*
     -   s23+s12*s23+2*s12*s34+s23*s34+p2*(p5+2*(p3-s34)-s45)+s15*s45
     -   -s34*s45)+(s12*(s15-s23)+p2*(-p5+s45))*(s12*(s15-s23)+p5*s23
     -   -s15*s45+p2*(-p5+s45))
       x(4,2) = (s12*s23-p2*s45)*(-(p5*s23)+s12*(-s15+s23)+p2*(p5-s45)
     -   +s15*s45)+2*msq*(p2**2*(p5-s45)+(s12+s23-s45)*(s12*(s15-s23)
     -   +p5*s23-s23*s34-s15*s45+s34*s45)+p3*(-(s12*s15)+2*s12*s23+s1
     -   5*s23-(2*s12+s15)*s45+p5*(s12-s23+s45))+p2*(-(p5*s12)-s12*s1
     -   5-2*p5*s23+s12*s23+s23*s34+p3*(-p5+s15-2*(s23-s45))+(p5+s12+
     -   s15+s23-s34)*s45-s45sq))
       x(3,3) = -4*msq*(s12*(-p3-s15+s23+s34)+p2*(p3+p5-s34-s45))*(p5-
     -   s15+s23-s45)+(s12*(s15-s23)+p2*(-p5+s45))**2
       x(4,3) = (s12*(-s15+s23)+p2*(p5-s45))*(s12*s23-p2*s45)-2*msq*(p
     -   2**2*(-p5+s45)+s12*(s12*(-s15+s23)+s23*(p5-2*(s15-s23)+s34)+
     -   p3*(-p5+s15-2*(s23-s45))+(s15-2*s23-s34)*s45)+p2*(s12*s15-s1
     -   2*s23-s23*s34+p3*(p5-s15+2*(s23-s45))+p5*(s12+s23-2*s45)-s12
     -   *s45+s15*s45+s34*s45-2*(s23*s45-s45sq)))
       x(4,4) = (s12*s23-p2*s45)**2-4*msq*(s12*s23*(s12+s23-s45)+p2**2
     -   *s45+p3*s12*(-s23+s45)+p2*((p3-s12)*s23-(p3+s12+s23)*s45+s45
     -   sq))
c          
       x(1,0)= -(s12sq*s15**2)+p3**2*p5*(-p5+s15)-p5**2*s23**2-s12sq*s
     -   23**2+2*(-(p5*s12*s15*s23)+s12sq*s15*s23+p5*s12*s23**2)+4*p5
     -   *s12*s23*s34-s12*s15*s23*s34+p5*s23**2*s34+s12*s23**2*s34-p2
     -   **2*(p5-s45)**2+(s23*(2*p5-s34)*(s15-s34)+2*(s12*s15*(s15-s2
     -   3)-s12*(s15+s23)*s34))*s45+p3*(2*p5**2*s23+s15*(s15-s34)*s45
     -   +s12*s15*(s15-s23+4*s45)-p5*(s23*(s15+s34)+2*(s12*(s15+s23)+
     -   (s15-s34)*s45)))-(s15-s34)**2*s45sq+p2*(p5*(2*(s12*s15+p5*s2
     -   3-s12*s23)+p3*(-s15+2*(p5+s23))-s23*s34)-(p3*(2*p5+s15)+s23*
     -   s34+2*(s12*s15-s12*s23-s15*s34+p5*(s15+s23+s34)))*s45+2*(s15
     -   +s34)*s45sq)
       x(2,0)= -(p3*p5*s12*s15)+p3*s12*s15**2-s12sq*s15**2+p3*p5**2*s2
     -   3-p3*p5*s12*s23-p3*p5*s15*s23-p3*s12*s15*s23-p5**2*s23**2-s1
     -   2sq*s23**2-s12*s15*s23*s34+p5*s23**2*s34+s12*s23**2*s34+2*(-
     -   (p5*s12*s15*s23)+s12sq*s15*s23+p5*s12*s23**2+p5*s12*s23*s34)
     -   -p2**2*(p5-s45)**2-(p3*(p5-2*s12-s15)*s15-2*s15*(s12*s15+p5*
     -   s23-s12*s23)+(s12*s15+(p5+s12+s15)*s23)*s34)*s45-s15*(s15-s3
     -   4)*s45sq+p2*(p5*(p3*(p5-s15+2*s23)+2*(s12*s15+p5*s23-s12*s23
     -   )-s23*s34)-(p3*(p5+s15)+2*((p5+s12)*s15+(p5-s12)*s23)+(p5-2*
     -   s15+s23)*s34)*s45+(2*s15+s34)*s45sq)
       x(3,0)= -(p2**2*(p5-s45)**2)-s12*(s12*(s15-s23)**2+p5*s15*s23-p
     -   5*s23**2-2*p5*s23*s34+s15*s23*s34-s23**2*s34+p3*(p5*(s15+s23
     -   )+s15*(-s15+s23-2*s45))-s15**2*s45+s15*s23*s45+s15*s34*s45+s
     -   23*s34*s45)+p2*(p5*(2*s12*(s15-s23)+p3*(p5-s15+2*s23)+s23*(p
     -   5-s34))-(p3*(p5+s15)+s23*s34+p5*(s15+s23+s34)+2*(s12*s15-s12
     -   *s23-s15*s34))*s45+(s15+s34)*s45sq)
       x(4,0)= p2**2*(p5-s45)*s45+p3*s12*(-((p5+s15)*s23)+2*s15*s45)+s
     -   12*s23*(s12*(s15-s23)+s23*(p5+s34)-(s15+s34)*s45)+p2*(p5*(2*
     -   p3-s12)*s23-(p3*(p5+s15)+s12*(s15-2*s23)+s23*(p5+s34))*s45+(
     -   s15+s34)*s45sq)
       endif 
       else 
       If(abs(p5).lt.eps) then 
c  Configuration                 (0p2sp3sp4s0)
       d= 2*(16*msq*(p1p4**2*(p2p3**2-p2*p3)+(p1p3*p2p4-p1p2*p3p4)**2-
     -   2*p1p4*(p1p3*p2p3*p2p4-p1p2*p2p4*p3-p1p3*p2*p3p4+p1p2*p2p3*p
     -   3p4)-(p1p3**2*p2-2*p1p2*p1p3*p2p3+p1p2**2*p3)*p4)+s15*(p4*s1
     -   2*(s12*s23-p2*s45)+s45*(p3*s12*s15-s12*s23*s34+p2*s34*s45)))
       x(1,1) = s12sq*(p4**2+(s15-s23)**2-2*p4*(s15+s23))+4*msq*(-(p2*
     -   *2*p4)-p3*s15*(p3-p4+s15)+(p3-p4)*s15*s23+((p4+s15-s23)*s23+
     -   p3*(-p4+s15+s23))*s34-s23*s34**2+p2*(-p4**2+p3*(p4+s15-s23)+
     -   (-s15+s23)*s34+p4*(s15+s23+s34)))+2*s12*(-(s15*(2*p3-p4+s15-
     -   s23))+p2*(p4+s15-s23)+(-p4+s15+s23)*s34)*s45+(p2**2+(s15-s34
     -   )**2-2*p2*(s15+s34))*s45sq
       x(2,1) = s12sq*(s15-s23)**2+s12*(2*(p2*(s15-s23)-s15*(p3+s15-s2
     -   3))+(s15+s23)*s34)*s45+p4*s12*(-(s12*(s15+s23))+(p2+s15)*s45
     -   )-2*msq*(-(p4**2*s12)+p3**2*s15+p4*s12*s15+p4*s12*s23+p4*s12
     -   *s34-s12*s15*s34-p4*s23*s34+s12*s23*s34+s23*s34**2+2*(p2**2*
     -   p4+p4*s15*s23-s15*s23*s34+s23**2*s34)-(p4*s15-(p4+s15-2*s23)
     -   *s34+s34**2)*s45+p3*(p4*(s12-s15)+s12*(s15-s23)+2*(s15**2-s1
     -   5*s23)-s15*s34-s23*s34+(s15+s34)*s45)-p2*(-p4**2+p4*(2*(s12+
     -   s15+s23)+s34-s45)-s34*(2*(s15-s23)+s45)+p3*(p4+2*(s15-s23)+s
     -   45)))+((p2-s15)**2-(p2+s15)*s34)*s45sq
       x(3,1) = -(s12sq*(-(s15-s23)**2+p4*(s15+s23)))+s12*(p2*(p4+2*(s
     -   15-s23))-s15*(2*p3+s15-s23)+(s15+s23)*s34)*s45-2*msq*(p3*p4*
     -   s12+p3**2*s15+p3*s12*s15-p4*s12*s15+p3*s15**2+s12*s15**2-p3*
     -   s12*s23-p4*s12*s23-p3*s15*s23+s12*s23**2+2*(p2**2*p4-s12*s15
     -   *s23)+p4*s12*s34-p3*s15*s34-s12*s15*s34-p3*s23*s34+s12*s23*s
     -   34-s15*s23*s34+s23**2*s34+s23*s34**2-(s15-s34)*(p3+s15-s23-s
     -   34)*s45-p2*(-2*(s15-s23)*s34+p4*(s15+s23+s34+2*(s12-s45))-(s
     -   15-s23+s34)*s45+p3*(p4+2*(s15-s23)+s45)))-p2*(-p2+s15+s34)*s
     -   45sq
       x(4,1) = -(s12sq*(p4+s15-s23)*s23)+s12*(-2*p3*s15+p2*(p4+s15-2*
     -   s23)+s23*(s15+s34))*s45+2*msq*(-(p3**2*s15)-p2**2*(p4+s45)+p
     -   3*(-(p4*s12)+s12*(s15+s23)+s23*(s15+s34)+(s15-s34)*s45)+s23*
     -   (p4*s12+s12*(s15-s23-2*s34)-s23*s34-s15*s45+s34*s45)+p2*(-(s
     -   12*s15)+s12*s23+s23*s34+p4*(s12+s23-2*s45)+(s15+s23+s34)*s45
     -   +p3*(p4+s15-2*s23+s45)))-p2*(-p2+s15+s34)*s45sq
       x(2,2) = (s12*(s15-s23)+(p2-s15)*s45)**2-4*msq*(p2**2*p4+p4*(s1
     -   2+s15)*(s12+s23-s45)+(p3*(s12+s15)-s34*(s12+s23-s45))*(s15-s
     -   23+s45)-p2*(p4*(2*s12+s15+s23-s45)+(p3-s34)*(s15-s23+s45)))
       x(3,2) = (s12*(s15-s23)+p2*s45)*(s12*(s15-s23)+(p2-s15)*s45)-2*
     -   msq*(2*p2**2*p4+p4*s12*(2*s12+s15+s23-s45)+(s15-s23+s45)*(p3
     -   *(2*s12+s15)+s12*(s15-s23-2*s34)-s23*s34-s15*s45+s34*s45)+p2
     -   *(-(p4*(4*s12+s15+s23-s45))-(2*(p3-s34)-s45)*(s15-s23+s45)))
       x(4,2) = (s12*s23-p2*s45)*(s12*(-s15+s23)+(-p2+s15)*s45)+2*msq*
     -   (-(p3*s12*s15)+s12sq*s15+2*p3*s12*s23-s12sq*s23+p3*s15*s23+s
     -   12*s15*s23-s12*s23**2-s12*s23*s34-s23**2*s34-p4*s12*(s12+s23
     -   -s45)-(2*s12*s15+p3*(2*s12+s15)-s12*s23+s15*s23-(s12+2*s23)*
     -   s34)*s45-p2**2*(p4+s45)+p2*(p3*(s15-2*(s23-s45))+p4*(2*s12+s
     -   23-s45)+s15*s45+s12*(-s15+s23+s45)+(s23-s45)*(s34+s45))+(s15
     -   -s34)*s45sq)
       x(3,3) = (s12*(s15-s23)+p2*s45)**2-4*msq*(p2**2*p4+s12*(p4*s12+
     -   (p3+s15-s23-s34)*(s15-s23+s45))+p2*(-2*p4*s12+(s15-s23+s45)*
     -   (-p3+s34+s45)))
       x(4,3) = (s12*s23-p2*s45)*(s12*(-s15+s23)-p2*s45)-2*msq*(p2**2*
     -   (p4+s45)+s12*(p4*s12+s12*(-s15+s23)+s23*(-2*(s15-s23)+s34)+p
     -   3*(s15-2*(s23-s45))+(s15-2*s23-s34)*s45)-p2*(2*p4*s12+p3*(s1
     -   5-2*(s23-s45))-s15*s45+s12*(-s15+s23+s45)+(s23-s45)*(s34+2*s
     -   45)))
       x(4,4) = (s12*s23-p2*s45)**2-4*msq*(s12*s23*(s12+s23-s45)+p2**2
     -   *s45+p3*s12*(-s23+s45)+p2*((p3-s12)*s23-(p3+s12+s23)*s45+s45
     -   sq))
c          
       x(1,0)= s12*((p4-s15)*(p2*p4-p4*s12+(-p3+s12)*s15)+(-(p2*p4)-p3
     -   *s15+2*(s12*s15+p4*(s12+s15)))*s23-s12*s23**2-(p4+s15-s23)*s
     -   23*s34)-(-(p2**2*p4)-s15*(p3*(4*s12+s15)-2*s12*(p4-s15+s23))
     -   +(p3*s15+s15*s23-2*(p4*s12-s12*(s15+s23)))*s34-s23*s34**2+p2
     -   *(p3*s15+s23*s34+p4*(2*s12+s15+s34)+2*(s12*s15-s12*s23-s15*s
     -   34)))*s45-(p2**2+(s15-s34)**2-2*p2*(s15+s34))*s45sq
       x(2,0)= s12*(-(p2*p4*(s15+s23))+p4*(2*s15*s23+s12*(s15+s23))+(s
     -   15-s23)*(p3*s15-s12*s15+s12*s23-s23*s34))-(-(p2**2*p4)-p3*s1
     -   5*(2*s12+s15)+s12*s15*(p4-2*(s15-s23))+(s12*s15+(s12+s15)*s2
     -   3)*s34+p2*(p3*s15+p4*(s12+s15)+s23*s34+2*(s12*s15-s12*s23-s1
     -   5*s34)))*s45-((p2-s15)**2-(p2+s15)*s34)*s45sq
       x(3,0)= p2**2*(p4-s45)*s45+s12*(p4*s12*(s15+s23)+(s15-s23)*(p3*
     -   s15-s12*s15+s12*s23-s23*s34)+(s15*(2*p3+s15-s23)-(s15+s23)*s
     -   34)*s45)+p2*(-(p4*s12*(s15+s23+s45))+s45*(-(p3*s15)-s23*s34-
     -   2*(s12*s15-s12*s23-s15*s34)+(s15+s34)*s45))
       x(4,0)= s12*s23*(-(p3*s15)+s12*(p4+s15-s23)+s23*s34)-s12*(-2*p3
     -   *s15+s23*(s15+s34))*s45+p2**2*(p4-s45)*s45+p2*(-(p4*s12*s23)
     -   -(p3*s15+s12*(p4+s15-2*s23)+s23*s34)*s45+(s15+s34)*s45sq)
       else 
c  Configuration                 (0p2sp3sp4sp5s)
       d= 2*(16*msq*(p1p4**2*(p2p3**2-p2*p3)+(p1p3*p2p4-p1p2*p3p4)**2-
     -   2*p1p4*(p1p3*p2p3*p2p4-p1p2*p2p4*p3-p1p3*p2*p3p4+p1p2*p2p3*p
     -   3p4)-(p1p3**2*p2-2*p1p2*p1p3*p2p3+p1p2**2*p3)*p4)+s23*((p3*p
     -   5-p4*s12)*(p2*p5-s12*s15)+p5*s12*s23*s34)+((p2*p4-p3*s15)*(p
     -   2*p5-s12*s15)-(p2*p5+s12*s15)*s23*s34)*s45+p2*s15*s34*s45sq)
       x(1,1) = s12sq*(p3**2+(p4-s34)**2-2*p3*(p4+s34))-4*msq*(p2**2*p
     -   4+s15*(p3**2+p4*s23-p3*(p4-s15+s23))-((p4+s15-s23)*s23+p3*(-
     -   p4+s15+s23))*s34+s23*s34**2-p2*(-p4**2+p3*(p4+s15-s23)+(-s15
     -   +s23)*s34+p4*(s15+s23+s34)))+(p2**2+(p3-s23)**2-2*p2*(p3+s23
     -   ))*(p5-s45)**2+(-4*p2*p4+(p3+s15-s23-s34)**2)*(s12-s45)**2-2
     -   *(s12*(-p3**2+p2*(p3+p4-s34)+s23*(-p4+s34)+p3*(p4-2*s15+s23+
     -   s34))*(p5-s45)+s12*(p3**2-2*p2*p4+p4*(s15+s23-s34)+s34*(-s15
     -   +s23+s34)-p3*(p4-s15+s23+2*s34))*(s12-s45)+(-((p3-s23)*(p3+s
     -   15-s23-s34))+p2*(p3+2*p4-s15+s23-s34))*(p5-s45)*(-s12+s45))
       x(2,1) = p3*p5*s12*s15-p4*s12sq*s15+s12sq*s15**2-p3*p5**2*s23+p
     -   3*p5*s12*s23+p4*p5*s12*s23-p4*s12sq*s23+p5**2*s23**2+s12sq*s
     -   23**2+p2**2*(p5-s45)**2+(s15*(p3*(p5-2*s12)+s12*(p4-2*(s15-s
     -   23))-2*p5*s23)+(p5*s23+s12*(s15+s23))*s34)*s45+p2*(p3*p5*(-p
     -   5+s45)+p4*(p5*s12-2*p5*s45+s12*s45)-(p5-s45)*(2*(s12*(s15-s2
     -   3)+p5*s23)-(2*s15+s34)*s45))-2*(-(p5*s12*s15*s23)+s12sq*s15*
     -   s23+p5*s12*s23**2+p5*s12*s23*s34+msq*(-(p4**2*s12)+p4*s12*s1
     -   5+p3**2*(-p5+s15)-p4*p5*s23+p4*s12*s23+p4*s12*s34-s12*s15*s3
     -   4-p4*s23*s34+p5*s23*s34+s12*s23*s34+s23*s34**2+2*(p2**2*p4+p
     -   4*s15*s23-s15*s23*s34+s23**2*s34)-(p4*s15-(p4+s15-2*s23)*s34
     -   +s34**2)*s45+p3*(p4*(p5+s12-s15)+s12*s15-s12*s23+2*(s15**2-s
     -   15*s23)-s15*s34-s23*s34+p5*(-2*s15+s23+s34)+(s15+s34)*s45)+p
     -   2*(p4**2+s34*(-p5+2*(s15-s23)+s45)-p3*(p4-p5+2*(s15-s23)+s45
     -   )+p4*(p5-2*(s12+s15+s23)-s34+s45))))+s15*(s15-s34)*s45sq
       x(3,1) = p2**2*(p5-s45)**2+s12*(s12*s15**2+p5*s15*s23-p5*s23**2
     -   +s12*s23**2+p3*p5*(s15+s23)-p4*s12*(s15+s23)-s15**2*s45+s15*
     -   s23*s45+s15*s34*s45+s23*s34*s45-2*(s12*s15*s23+p5*s23*s34+p3
     -   *s15*s45))+p2*(p3*p5*(-p5+s45)+p4*s12*(p5+s45)-(p5-s45)*(p5*
     -   s23+2*(s12*s15-s12*s23)-(s15+s34)*s45))-2*msq*(-(p4*s12*s15)
     -   +s12*s15**2+p3**2*(-p5+s15)-p4*s12*s23+p5*s15*s23-p5*s23**2+
     -   s12*s23**2+2*(p2**2*p4-s12*s15*s23)+p4*s12*s34-s12*s15*s34-p
     -   5*s23*s34+s12*s23*s34-s15*s23*s34+s23**2*s34+s23*s34**2-(s15
     -   -s34)*(s15-s23-s34)*s45+p3*(p4*s12+s12*s15+s15**2-s12*s23-s1
     -   5*s23-s15*s34-s23*s34+p5*(-s15+2*s23+s34)-s15*s45+s34*s45)-p
     -   2*(p5*s15-p5*s23+p5*s34-2*(s15*s34-s23*s34)-(s15-s23+s34)*s4
     -   5+p3*(p4-p5+2*(s15-s23)+s45)+p4*(s15+s23+s34-2*(p5-s12+s45))
     -   ))
       x(4,1) = s12*s23*(p2*p5+p3*p5-s12*(p4+s15)+(-p5+s12)*s23)+(-(p2
     -   **2*p5)+p2*(p3*p5+s12*(p4+s15-2*s23)+p5*s23)+s12*(-2*p3*s15+
     -   s23*(s15+s34)))*s45+2*msq*(p3**2*(p5-s15)-p2**2*(p4-p5+s45)+
     -   p3*(-(p4*s12)-2*p5*s23+s12*(s15+s23)+s23*(s15+s34)+(s15-s34)
     -   *s45)+s23*(p4*s12+p5*s23+s12*(s15-s23-2*s34)-s23*s34-s15*s45
     -   +s34*s45)+p2*(-(s12*s15)-2*p5*s23+s12*s23+s23*s34+p4*(s12+s2
     -   3-2*s45)+(s15+s23+s34)*s45+p3*(p4+s15-2*(p5+s23)+s45)))+p2*(
     -   p2-s15-s34)*s45sq
       x(2,2) = (p2*p5-s12*s15-p5*s23+s12*s23-p2*s45+s15*s45)**2-4*msq
     -   *(p2**2*p4+(p4*(-p5+s12+s15)+s34*(p5-s15+s23-s45))*(s12+s23-
     -   s45)+p3*(p5-s12-s15)*(p5-s15+s23-s45)+p2*(p3*(p5-s15+s23-s45
     -   )+p4*(p5-2*s12-s15-s23+s45)+s34*(-p5+s15-s23+s45)))
       x(3,2) = (s12*(s15-s23)+p2*(-p5+s45))*(s12*(s15-s23)+p5*s23-s15
     -   *s45+p2*(-p5+s45))-2*msq*(-(p4*p5*s12)+p4*s12*s15-p5*s12*s15
     -   +s12*s15**2-p5**2*s23+p4*s12*s23+p5*s12*s23+p5*s15*s23-p5*s2
     -   3**2+s12*s23**2+p5*s23*s34-s15*s23*s34+s23**2*s34+2*(p2**2*p
     -   4+p4*s12sq-s12*s15*s23+p5*s12*s34-s12*s15*s34+s12*s23*s34)+p
     -   3*(p5-2*s12-s15)*(p5-s15+s23-s45)-(p4*s12+(-s12+s15)*(s15-s2
     -   3)-p5*(s15+s23-s34)+(-s15+2*(s12+s23))*s34)*s45+p2*(2*p3*(p5
     -   -s15+s23-s45)+(p5-s15+s23-s45)*(p5-2*s34-s45)+p4*(p5-4*s12-s
     -   15-s23+s45))-(s15-s34)*s45sq)
       x(4,2) = (s12*s23-p2*s45)*(-(p5*s23)+s12*(-s15+s23)+p2*(p5-s45)
     -   +s15*s45)-2*msq*(p2**2*(p4-p5+s45)+(s12+s23-s45)*(p4*s12-p5*
     -   s23+s12*(-s15+s23)+s23*s34+s15*s45-s34*s45)+p3*(s12*(s15-2*(
     -   s23-s45))+s15*(-s23+s45)-p5*(s12-s23+s45))+p2*(p5*s12+s12*s1
     -   5+2*p5*s23-s12*s23-s23*s34+p3*(p5-s15+2*(s23-s45))-(p5+s12+s
     -   15+s23-s34)*s45+p4*(-2*s12-s23+s45)+s45sq))
       x(3,3) = (s12*(s15-s23)+p2*(-p5+s45))**2-4*msq*(p2**2*p4+p2*(-2
     -   *p4*s12+(p5-s15+s23-s45)*(p3+p5-s34-s45))+s12*(p4*s12+(p3+s1
     -   5-s23-s34)*(-p5+s15-s23+s45)))
       x(4,3) = (s12*(-s15+s23)+p2*(p5-s45))*(s12*s23-p2*s45)-2*msq*(p
     -   2**2*(p4-p5+s45)+s12*(p4*s12+s12*(-s15+s23)+s23*(p5-2*(s15-s
     -   23)+s34)+p3*(-p5+s15-2*(s23-s45))+(s15-2*s23-s34)*s45)+p2*(s
     -   12*s15-s12*s23-s23*s34+p3*(p5-s15+2*(s23-s45))+p5*(s12+s23-2
     -   *s45)-s12*s45+s15*s45+s34*s45-2*(p4*s12+s23*s45-s45sq)))
       x(4,4) = (s12*s23-p2*s45)**2-4*msq*(s12*s23*(s12+s23-s45)+p2**2
     -   *s45+p3*s12*(-s23+s45)+p2*((p3-s12)*s23-(p3+s12+s23)*s45+s45
     -   sq))
c          
       x(1,0)= -(p3**2*p5**2)-p4**2*s12sq+p3**2*p5*s15-p3*p4*s12*s15+p
     -   3*s12*s15**2-s12sq*s15**2-p3*p5*s15*s23-p3*s12*s15*s23-p5**2
     -   *s23**2-s12sq*s23**2-p3*p5*s23*s34-p4*s12*s23*s34+4*p5*s12*s
     -   23*s34-s12*s15*s23*s34+p5*s23**2*s34+s12*s23**2*s34-p2*(-(p4
     -   **2*s12)+p4*s12*(2*p5+s15)+p4*(p5+s12)*s23+p3*p5*(p4+s15-2*(
     -   p5+s23))+p5*(-2*(s12*s15+p5*s23-s12*s23)+s23*s34))-(-(s15*(p
     -   3*(-2*p5+4*s12+s15)+2*(p5*s23-s12*(p4-s15+s23))))+(p3*(-2*p5
     -   +s15)-2*(p4*s12-s12*s15)+(2*(p5+s12)+s15)*s23)*s34-s23*s34**
     -   2)*s45-p2*(p3*(2*p5+s15)+2*((p5+s12)*s15+(p5-s12)*s23)+(2*(p
     -   5-s15)+s23)*s34+p4*(-4*p5+2*s12+s15+s34))*s45+p2**2*(-(p5-s4
     -   5)**2+p4*(p5+s45))-(s15-s34)**2*s45sq+2*(p3*p4*p5*s12-p3*p5*
     -   s12*s15+p4*s12sq*s15+p3*p5**2*s23-p3*p5*s12*s23-p4*p5*s12*s2
     -   3+p4*s12sq*s23+p4*s12*s15*s23-p5*s12*s15*s23+s12sq*s15*s23+p
     -   5*s12*s23**2+p2*(s15+s34)*s45sq)
       x(2,0)= p4*s12sq*s15-s12sq*s15**2-p4*p5*s12*s23+p4*s12sq*s23-p5
     -   **2*s23**2-s12sq*s23**2-s12*s15*s23*s34+p5*s23**2*s34+s12*s2
     -   3**2*s34+2*(p4*s12*s15*s23-p5*s12*s15*s23+s12sq*s15*s23+p5*s
     -   12*s23**2+p5*s12*s23*s34)-(s15*(s12*(p4-2*(s15-s23))-2*p5*s2
     -   3)+(s12*s15+(p5+s12+s15)*s23)*s34)*s45+p2**2*(-(p5-s45)**2+p
     -   4*(p5+s45))+p3*(p5**2*s23-p5*(s12*(s15+s23)+s15*(s23+s45))+s
     -   15*(s15*s45+s12*(s15-s23+2*s45)))-s15*(s15-s34)*s45sq+p2*(p3
     -   *p5*(p5-s15+2*s23)+p5*(2*(s12*s15+p5*s23-s12*s23)-s23*s34)-p
     -   3*(p5+s15)*s45-(2*((p5+s12)*s15+(p5-s12)*s23)+(p5-2*s15+s23)
     -   *s34)*s45-p4*(p5*(s12+s23-2*s45)+s15*s45+s12*(s15+s23+s45))+
     -   (2*s15+s34)*s45sq)
       x(3,0)= -(s12*(s12*s15**2+p5*s15*s23-p5*s23**2+s12*s23**2-p4*s1
     -   2*(s15+s23)+s15*s23*s34-s23**2*s34-2*(s12*s15*s23+p5*s23*s34
     -   )+p3*(p5*(s15+s23)+s15*(-s15+s23-2*s45))-s15**2*s45+s15*s23*
     -   s45+s15*s34*s45+s23*s34*s45))+p2**2*(-(p5-s45)**2+p4*(p5+s45
     -   ))+p2*(p3*p5*(p5-s15+2*s23)+p5*(2*s12*(s15-s23)+s23*(p5-s34)
     -   )-p3*(p5+s15)*s45-(s23*s34+p5*(s15+s23+s34)+2*(s12*s15-s12*s
     -   23-s15*s34))*s45-p4*s12*(p5+s15+s23+s45)+(s15+s34)*s45sq)
       x(4,0)= p2**2*(p4+p5-s45)*s45+p3*s12*(-((p5+s15)*s23)+2*s15*s45
     -   )+s12*s23*(p4*s12+s12*(s15-s23)+s23*(p5+s34)-(s15+s34)*s45)+
     -   p2*((2*p3*p5-(p4+p5)*s12)*s23-(p3*(p5+s15)+s12*(p4+s15-2*s23
     -   )+p5*s23+s23*s34)*s45+(s15+s34)*s45sq)
       endif 
       endif 
       endif 
       endif 
       else 
       If(abs(p2).lt.eps) then 
       If(abs(p3).lt.eps) then 
       If(abs(p4).lt.eps) then 
       If(abs(p5).lt.eps) then 
c  Configuration                 (p1s0000)
       d= 2*(16*msq*((p1p4*p2p3-p1p3*p2p4)**2-2*(p1p2*p1p4*p2p3+p1p2*p
     -   1p3*p2p4-p1*p2p3*p2p4)*p3p4+p1p2**2*p3p4**2)+s23*s34*(-(s12*
     -   s15)+p1*s34)*s45)
       x(1,1) = s12sq*(s15-s23)**2+p1**2*s34**2-4*msq*s23*s34*(-s15+s2
     -   3+s34)-2*(p1*s34*(-s15+2*s23+s34)*s45-s12*(p1*(-s15+s23)*s34
     -   +(-s15**2+s23*s34+s15*(s23+s34))*s45))+(s15-s34)**2*s45sq
       x(2,1) = s12sq*(s15-s23)**2+2*(msq-p1)*s12*(s15-s23)*s34+s34*(4
     -   *msq*(s15-s23)*s23+p1**2*s34-2*msq*(p1+s23)*s34)+s34*(-2*(ms
     -   q-p1)*(s15-2*s23)+(2*msq-p1)*s34)*s45+s12*(-2*s15*(s15-s23)+
     -   (s15+s23)*s34)*s45+s15*(s15-s34)*s45sq
       x(3,1) = (s12*(s15-s23)-p1*s34)**2-(s12*s15*(s15-s23)-s12*(s15+
     -   s23)*s34+p1*s34*(-s15+2*s23+s34))*s45+2*msq*(s15-s23-s34)*(s
     -   12*(-s15+s23)+s34*(p1+s23-s45)+s15*s45)
       x(4,1) = s23*(s12sq*(-s15+s23)+p1*s12*s34+s12*(s15+s34)*s45-2*(
     -   p1*s34*s45-msq*(s12*(s15-s23-2*s34)-s15*s45+s34*(p1-s23+s45)
     -   )))
       x(2,2) = p1**2*s34**2+(s12*(-s15+s23)+s15*s45)**2+2*p1*s34*(s12
     -   *(-s15+s23)+(s15-2*s23)*s45)-4*msq*s34*(p1*s34-(s12+s23-s45)
     -   *(s15-s23+s45))
       x(3,2) = (s12*(s15-s23)-p1*s34)**2+(s12*s15*(-s15+s23)+p1*(s15-
     -   2*s23)*s34)*s45+2*msq*(-(s12*(s15-s23-2*s34)*(s15-s23+s45))-
     -   p1*s34*(-s15+s23+2*s34+s45)+(s15-s23+s45)*(s23*s34+s15*s45-s
     -   34*s45))
       x(4,2) = p1*s23*s34*(s12-2*s45)+s12*s23*(s12*(-s15+s23)+s15*s45
     -   )+2*msq*(s12sq*(s15-s23)+p1*s34*(s23+s45)-(s23-s45)*(s23*s34
     -   +(s15-s34)*s45)+s12*((s15-s23)*s23-(p1+s23)*s34+(-2*s15+s23+
     -   s34)*s45))
       x(3,3) = (s12*(s15-s23)-p1*s34)**2+4*msq*(s15-s23-s34)*(-(s12*(
     -   s15-s23+s45))+p1*(s34+s45))
       x(4,3) = s12sq*s23*(-s15+s23)+p1*s12*s23*s34-2*msq*(-(p1**2*s34
     -   )+p1*(s23*s34+s12*(s15-s23+s34)+(-s15+2*s23+s34)*s45)+s12*(s
     -   12*(-s15+s23)+(2*s23+s34)*(s23-s45)+s15*(-2*s23+s45)))
       x(4,4) = s23*(s12sq*s23-4*msq*(s12*(s12+s23-s45)+p1*(-s12+s45))
     -   )
c          
       x(1,0)= -(s12sq*(s15-s23)**2)+s12*(2*p1-s23)*(s15-s23)*s34-p1**
     -   2*s34**2+2*s12*(s15**2-s23*s34-s15*(s23+s34))*s45-(s15-s34)*
     -   s45*(s23*s34+s15*s45-s34*s45)+p1*s34*(s23*s34+2*(-s15+2*s23+
     -   s34)*s45)
       x(2,0)= -(s12sq*(s15-s23)**2)+s12*(2*p1-s23)*(s15-s23)*s34-p1**
     -   2*s34**2+s12*(2*s15*(s15-s23)-(s15+s23)*s34)*s45-s15*s45*(s2
     -   3*s34+s15*s45-s34*s45)+p1*s34*(s23*s34+(-2*s15+4*s23+s34)*s4
     -   5)
       x(3,0)= -(s12sq*(s15-s23)**2)+s12*(2*p1-s23)*(s15-s23)*s34+s12*
     -   (s15**2-s23*s34-s15*(s23+s34))*s45+p1*s34*(-(p1*s34)+s23*s34
     -   +(-s15+2*s23+s34)*s45)
       x(4,0)= s23*(s12sq*(s15-s23)+2*p1*s34*s45-s12*(s15*s45+s34*(p1-
     -   s23+s45)))
       else 
c  Configuration                 (p1s000p5s)
       d= 2*(16*msq*((p1p4*p2p3-p1p3*p2p4)**2-2*(p1p2*p1p4*p2p3+p1p2*p
     -   1p3*p2p4-p1*p2p3*p2p4)*p3p4+p1p2**2*p3p4**2)+s23*s34*(p5*s12
     -   *s23-s12*s15*s45+p1*s34*s45))
       x(1,1) = (p1-s12)**2*s34**2-4*msq*s23*s34*(-s15+s23+s34)+s23**2
     -   *(p5-s45)**2+(-s15+s23+s34)**2*(s12-s45)**2+2*((p1-s12)*s23*
     -   s34*(p5-s45)+(p1-s12)*s34*(-s15+s23+s34)*(s12-s45)+s23*(-s15
     -   +s23+s34)*(p5-s45)*(-s12+s45))
       x(2,1) = s12sq*(s15-s23)**2+p5**2*s23**2+s34*(4*msq*(s15-s23)*s
     -   23+p1**2*s34-2*msq*(p1+s23)*s34)+s34*(-2*(msq-p1)*(s15-2*s23
     -   )+(2*msq-p1)*s34)*s45+s12*(2*((msq-p1)*(s15-s23)*s34-p5*s23*
     -   (-s15+s23+s34))+(-2*s15*(s15-s23)+(s15+s23)*s34)*s45)+p5*s23
     -   *(-2*s15*s45+s34*(-2*(msq-p1)+s45))+s15*(s15-s34)*s45sq
       x(3,1) = s12sq*(s15-s23)**2+2*msq*(s15-s23-s34)*(-(p5*s23)+s12*
     -   (-s15+s23)+s34*(p1+s23-s45)+s15*s45)+p1*s34*(p5*s23+p1*s34+(
     -   s15-2*s23-s34)*s45)+s12*(p5*s23*(s15-s23-2*s34)-2*p1*(s15-s2
     -   3)*s34+(-s15**2+s23*s34+s15*(s23+s34))*s45)
       x(4,1) = s23*(s12sq*(-s15+s23)+s12*(-(p5*s23)+p1*s34+(s15+s34)*
     -   s45)-2*(p1*s34*s45-msq*(p5*s23+s12*(s15-s23-2*s34)-s15*s45+s
     -   34*(p1-s23+s45))))
       x(2,2) = p1**2*s34**2-4*msq*s34*(p1*s34+(s12+s23-s45)*(p5-s15+s
     -   23-s45))+(s12*(s15-s23)+p5*s23-s15*s45)**2+2*p1*s34*(p5*s23+
     -   s12*(-s15+s23)+(s15-2*s23)*s45)
       x(3,2) = s12sq*(s15-s23)**2+s12*(s15-s23)*(p5*s23-2*p1*s34-s15*
     -   s45)+p1*s34*(p5*s23+p1*s34+(s15-2*s23)*s45)+2*msq*(p5**2*s23
     -   +s34*((s15-s23)*(p1+s23)-2*p1*s34)+(s15*(s15-s23)-(p1+s15-2*
     -   s23)*s34)*s45-s12*(s15-s23-2*s34)*(s15-s23+s45)+p5*(s12*(s15
     -   -s23-2*s34)+p1*s34+(s23-s34)*(s23-s45)-s15*(s23+s45))+(s15-s
     -   34)*s45sq)
       x(4,2) = p1*s23*s34*(s12-2*s45)+s12*s23*(-(p5*s23)+s12*(-s15+s2
     -   3)+s15*s45)+2*msq*(s12sq*(s15-s23)+(p1-s23)*s23*s34+p5*s23*(
     -   s23-s45)+(-(s15*s23)+p1*s34+2*s23*s34)*s45+s12*(s23*(p5+s15-
     -   s23-s34)-p1*s34+(-2*s15+s23+s34)*s45)+(s15-s34)*s45sq)
       x(3,3) = (s12*(s15-s23)-p1*s34)**2+4*msq*(s15-s23-s34)*(s12*(p5
     -   -s15+s23-s45)+p1*(-p5+s34+s45))
       x(4,3) = s12*s23*(s12*(-s15+s23)+p1*s34)+2*msq*(p1**2*s34+s12*(
     -   s12*(s15-s23)-s23*(p5-2*(s15-s23)+s34)+(-s15+2*s23+s34)*s45)
     -   -p1*(-(p5*s23)+s23*s34+s12*(s15-s23+s34)+(-s15+2*s23+s34)*s4
     -   5))
       x(4,4) = s23*(s12sq*s23-4*msq*(s12*(s12+s23-s45)+p1*(-s12+s45))
     -   )
c          
       x(1,0)= -(s12sq*(s15-s23)**2)-p5**2*s23**2-p1*(p1-s23)*s34**2+s
     -   34*(4*p1*s23-s15*s23+s23*s34-2*(p1*s15-p1*s34))*s45+s12*((2*
     -   p1-s23)*(s15-s23)*s34+2*(p5*s23*(-s15+s23+2*s34)+(s15*(s15-s
     -   23)-(s15+s23)*s34)*s45))+p5*s23*(2*s15*s45+s34*(s23-2*(p1+s4
     -   5)))-(s15-s34)**2*s45sq
       x(2,0)= -(s12sq*(s15-s23)**2)-p5**2*s23**2-p1*(p1-s23)*s34**2+s
     -   34*(-(s15*s23)+p1*(-2*s15+4*s23+s34))*s45+p5*s23*(s34*(-2*p1
     -   +s23-s45)+2*s15*s45)-s12*(s15*s23*s34-s23**2*s34+s15*s34*s45
     -   +s23*s34*s45-2*(p1*s15*s34-p1*s23*s34+p5*s23*(-s15+s23+s34)+
     -   s15**2*s45-s15*s23*s45))+s15*(-s15+s34)*s45sq
       x(3,0)= -(p1**2*s34**2)-s12*(s12*(s15-s23)**2+p5*s23*(s15-s23-2
     -   *s34)+s15*s23*s34-s23**2*s34-s15**2*s45+s15*s23*s45+s15*s34*
     -   s45+s23*s34*s45)+p1*s34*(2*s12*(s15-s23)-p5*s23+s23*s34+(-s1
     -   5+2*s23+s34)*s45)
       x(4,0)= s23*(s12*(s12*(s15-s23)+s23*(p5+s34))-p1*s34*(s12-2*s45
     -   )-s12*(s15+s34)*s45)
       endif 
       else 
       If(abs(p5).lt.eps) then 
c  Configuration                 (p1s00p4s0)
       d= 2*(16*msq*(p1p4**2*p2p3**2+p1p3**2*p2p4**2+p1p2**2*p3p4**2-2
     -   *(p1p2*p1p3*p2p4*p3p4-p1*p2p3*p2p4*p3p4+p1p4*p2p3*(p1p3*p2p4
     -   +p1p2*p3p4))+p2p3*(2*p1p2*p1p3-p1*p2p3)*p4)+s23*(s12*s15-p1*
     -   s34)*(p4*s12-s34*s45))
       x(1,1) = (p1-s12)**2*(p4-s34)**2-4*msq*s23*(p4*(s15-s34)+s34*(-
     -   s15+s23+s34))+(-s15+s23+s34)**2*(s12-s45)**2+2*((p1-s12)*(p4
     -   *(s15+s23-s34)+s34*(-s15+s23+s34))*(s12-s45)+(p1-s12)*s23*(p
     -   4-s34)*s45-s23*(-s15+s23+s34)*s45*(-s12+s45))+s23**2*s45sq
       x(2,1) = -(s12sq*(-(s15-s23)**2+p4*(s15+s23)))+p1**2*(p4-s34)**
     -   2+p1*s12*(-(p4*(p4-2*(s15+s23)))+(p4-2*(s15-s23))*s34)+s12*(
     -   s15*(p4-2*(s15-s23))+(s15+s23)*s34)*s45-p1*(2*p4*s15-(p4+2*s
     -   15-4*s23)*s34+s34**2)*s45+2*msq*(p4**2*s12-p1*(p4-s34)**2-p4
     -   *(2*s15*s23-s23*s34+s12*(s15+s23+s34)-s15*s45+s34*s45)-s34*(
     -   s12*(-s15+s23)+(2*s23+s34)*(s23-s45)+s15*(-2*s23+s45)))+s15*
     -   (s15-s34)*s45sq
       x(3,1) = p1**2*s34*(-p4+s34)+p1*s12*(2*(-s15+s23)*s34+p4*(s15+s
     -   23+s34))+p1*(s15-2*s23-s34)*s34*s45+s12*(s12*(s15-s23)**2-p4
     -   *s12*(s15+s23)+(-s15**2+s23*s34+s15*(s23+s34))*s45)+2*msq*(p
     -   4*s12*(s15+s23-s34)+p1*(-(p4*(s15+s23))+(p4+s15-s23)*s34-s34
     -   **2)-(s15-s23-s34)*(s12*s15-s12*s23-s23*s34-s15*s45+s34*s45)
     -   )
       x(4,1) = s23*(p1*s12*(p4+s34)+s12*(-(s12*(p4+s15-s23))+(s15+s34
     -   )*s45)-2*(p1*s34*s45-msq*(p4*s12+s12*s15-s12*s23-2*s12*s34-s
     -   23*s34+p1*(-p4+s34)-s15*s45+s34*s45)))
       x(2,2) = p1**2*(p4-s34)**2+(s12*(-s15+s23)+s15*s45)**2+2*p1*(p4
     -   *s12*(s15+s23)-p4*s15*s45+s34*(s12*(-s15+s23)+(s15-2*s23)*s4
     -   5))-4*msq*(p1*(p4-s34)**2+(s12+s23-s45)*(p4*(s12+s15)-s34*(s
     -   15-s23+s45)))
       x(3,2) = p1**2*s34*(-p4+s34)+s12*(s15-s23)*(s12*(s15-s23)-s15*s
     -   45)-2*msq*(p4*s12*(2*s12+s15+s23-s45)-(s15-s23+s45)*(s23*s34
     -   +s12*(-s15+s23+2*s34)+s15*s45-s34*s45)+p1*(s34*(-s15+s23+2*s
     -   34+s45)+p4*(s15+s23-2*(s12+s34)+s45)))+p1*(p4*s12*(s15+s23)+
     -   s34*(s15*s45-2*(s12*s15-s12*s23+s23*s45)))
       x(4,2) = s12*s23*(-(s12*s15)+s12*s23+s15*s45)+p1*s23*(s12*(p4+s
     -   34)-2*s34*s45)+2*msq*(p1*(p4-s34)*(s12-s23-s45)-(s12+s23-s45
     -   )*(p4*s12-s12*s15+s12*s23+s23*s34+s15*s45-s34*s45))
       x(3,3) = (s12*(s15-s23)-p1*s34)**2+4*msq*(-(p1**2*p4)-s12*(p4*s
     -   12+(s15-s23-s34)*(s15-s23+s45))+p1*(2*p4*s12+(s15-s23-s34)*(
     -   s34+s45)))
       x(4,3) = s12*s23*(s12*(-s15+s23)+p1*s34)-2*(msq*(-(p1*s12*(2*p4
     -   -s15+s23))+p1**2*(p4-s34)+p1*(s12+s23)*s34+s12*(p4*s12-(s15-
     -   s23)*(s12+2*s23)+s23*s34))-msq*(p1-s12)*(s15-2*s23-s34)*s45)
       x(4,4) = s23*(s12sq*s23-4*msq*(s12*(s12+s23-s45)+p1*(-s12+s45))
     -   )
c          
       x(1,0)= -(p4**2*s12sq)-p1**2*(p4-s34)**2-s12*(s15-s23)*(s12*(s1
     -   5-s23)+s23*s34)+(s23*s34*(-s15+s34)+2*(s12*s15*(s15-s23)-s12
     -   *(s15+s23)*s34))*s45+p4*s12*(-(s23*s34)+2*(s15*s23+s12*(s15+
     -   s23)-s15*s45+s34*s45))+p1*(2*p4**2*s12-p4*(s23*s34+2*(s12*(s
     -   15+s23+s34)-s15*s45+s34*s45))+s34*(s23*s34+2*(s12*(s15-s23)+
     -   (-s15+2*s23+s34)*s45)))-(s15-s34)**2*s45sq
       x(2,0)= -(p1**2*(p4-s34)**2)+s12*(p4*(2*s15*s23+s12*(s15+s23))-
     -   (s15-s23)*(s12*(s15-s23)+s23*s34))-(s12*s15*(p4-2*(s15-s23))
     -   +(s15*s23+s12*(s15+s23))*s34)*s45+p1*(p4**2*s12+s34*(2*s12*(
     -   s15-s23)+s23*s34+(-2*s15+4*s23+s34)*s45)-p4*(s12*(2*(s15+s23
     -   )+s34)-2*s15*s45+s34*(s23+s45)))-s15*(s15-s34)*s45sq
       x(3,0)= p1**2*(p4-s34)*s34+s12*(p4*s12*(s15+s23)-(s15-s23)*(s12
     -   *s15-s12*s23+s23*s34)+(s15*(s15-s23)-(s15+s23)*s34)*s45)-p1*
     -   (p4*s12*(s15+s23+s34)-s34*(2*s12*(s15-s23)+s23*s34+(-s15+2*s
     -   23+s34)*s45))
       x(4,0)= s23*(-(p1*s12*(p4+s34))+2*p1*s34*s45+s12*(p4*s12+s12*s1
     -   5-s12*s23+s23*s34-(s15+s34)*s45))
       else 
c  Configuration                 (p1s00p4sp5s)
       d= 2*(16*msq*(p1p4**2*p2p3**2+p1p3**2*p2p4**2+p1p2**2*p3p4**2-2
     -   *(p1p2*p1p3*p2p4*p3p4-p1*p2p3*p2p4*p3p4+p1p4*p2p3*(p1p3*p2p4
     -   +p1p2*p3p4))+p2p3*(2*p1p2*p1p3-p1*p2p3)*p4)+s23*(p4*s12*(s12
     -   *s15-p1*s34)+s34*(p5*s12*s23-s12*s15*s45+p1*s34*s45)))
       x(1,1) = (p1-s12)**2*(p4-s34)**2-4*msq*s23*(p4*(s15-s34)+s34*(-
     -   s15+s23+s34))+s23**2*(p5-s45)**2+(-s15+s23+s34)**2*(s12-s45)
     -   **2+2*((p1-s12)*s23*(-p4+s34)*(p5-s45)+(p1-s12)*(p4*(s15+s23
     -   -s34)+s34*(-s15+s23+s34))*(s12-s45)+s23*(-s15+s23+s34)*(p5-s
     -   45)*(-s12+s45))
       x(2,1) = -(p4*s12sq*s15)+s12sq*s15**2+p4*p5*s12*s23-p4*s12sq*s2
     -   3+p5**2*s23**2+s12sq*s23**2+p1**2*(p4-s34)**2+(p4*s12*s15-2*
     -   s12*s15*(s15-s23)+s12*(s15+s23)*s34+p5*s23*(-2*s15+s34))*s45
     -   +p1*(-(p4**2*s12)+s34*(-2*(s12*(s15-s23)-p5*s23)-(-2*s15+4*s
     -   23+s34)*s45)+p4*(s12*(2*(s15+s23)+s34)+s34*s45-2*(p5*s23+s15
     -   *s45)))-2*(-(p5*s12*s15*s23)+s12sq*s15*s23+p5*s12*s23**2+p5*
     -   s12*s23*s34+msq*(-(p4**2*s12)+p1*(p4-s34)**2+p4*(-(s23*(p5-2
     -   *s15+s34))+s12*(s15+s23+s34)+(-s15+s34)*s45)+s34*(p5*s23+s12
     -   *(-s15+s23)+(2*s23+s34)*(s23-s45)+s15*(-2*s23+s45))))+s15*(s
     -   15-s34)*s45sq
       x(3,1) = p1**2*s34*(-p4+s34)+s12*(s12*(s15-s23)**2+p5*s15*s23-p
     -   5*s23**2-p4*s12*(s15+s23)-2*p5*s23*s34-s15**2*s45+s15*s23*s4
     -   5+s15*s34*s45+s23*s34*s45)+p1*(p4*s12*(s15+s23+s34)+s34*(p5*
     -   s23-2*(s12*s15-s12*s23)+(s15-2*s23-s34)*s45))+2*msq*(p4*s12*
     -   (s15+s23-s34)+p1*(-(p4*(s15+s23))+(p4+s15-s23)*s34-s34**2)-(
     -   s15-s23-s34)*(s12*s15+p5*s23-s12*s23-s23*s34-s15*s45+s34*s45
     -   ))
       x(4,1) = s23*(-(s12*(s12*(p4+s15-s23)+p5*s23))+p1*s12*(p4+s34)+
     -   s12*(s15+s34)*s45-2*(p1*s34*s45-msq*(p4*s12+s12*s15+p5*s23-s
     -   12*s23-2*s12*s34-s23*s34+p1*(-p4+s34)-s15*s45+s34*s45)))
       x(2,2) = p1**2*(p4-s34)**2-4*msq*(p1*(p4-s34)**2+(p4*(-p5+s12+s
     -   15)+s34*(p5-s15+s23-s45))*(s12+s23-s45))+(s12*(s15-s23)+p5*s
     -   23-s15*s45)**2+2*p1*(p4*(-(p5*s23)+s12*(s15+s23)-s15*s45)+s3
     -   4*(-(s12*s15)+p5*s23+s12*s23+s15*s45-2*s23*s45))
       x(3,2) = p1**2*s34*(-p4+s34)+s12*(s15-s23)*(s12*(s15-s23)+p5*s2
     -   3-s15*s45)-2*msq*(p4*s12*(-p5+2*s12+s15+s23-s45)-(p5-s15+s23
     -   -s45)*(p5*s23+s12*(s15-s23-2*s34)-s23*s34-s15*s45+s34*s45)+p
     -   1*(s34*(-p5-s15+s23+2*s34+s45)+p4*(p5+s15+s23-2*(s12+s34)+s4
     -   5)))+p1*(p4*s12*(s15+s23)+s34*(p5*s23+s15*s45-2*(s12*s15-s12
     -   *s23+s23*s45)))
       x(4,2) = s12*s23*(-(s12*s15)-p5*s23+s12*s23+s15*s45)+p1*s23*(s1
     -   2*(p4+s34)-2*s34*s45)+2*msq*(p1*(p4-s34)*(s12-s23-s45)-(s12+
     -   s23-s45)*(p4*s12-s12*s15-p5*s23+s12*s23+s23*s34+s15*s45-s34*
     -   s45))
       x(3,3) = (s12*(s15-s23)-p1*s34)**2-4*(msq*(p1**2*p4+s12*(p4*s12
     -   -(p5-s15+s23)*(s15-s23-s34))+p1*(-2*p4*s12-(p5-s34)*(-s15+s2
     -   3+s34)))-msq*(p1-s12)*(s15-s23-s34)*s45)
       x(4,3) = s12*s23*(s12*(-s15+s23)+p1*s34)-2*(msq*(p1**2*(p4-s34)
     -   +s12*(p4*s12+s12*(-s15+s23)+s23*(p5-2*(s15-s23)+s34))+p1*(-2
     -   *p4*s12+s23*(-p5+s34)+s12*(s15-s23+s34)))-msq*(p1-s12)*(s15-
     -   2*s23-s34)*s45)
       x(4,4) = s23*(s12sq*s23-4*msq*(s12*(s12+s23-s45)+p1*(-s12+s45))
     -   )
c          
       x(1,0)= -(p4**2*s12sq)-s12sq*s15**2-p5**2*s23**2-s12sq*s23**2+2
     -   *(-(p5*s12*s15*s23)+s12sq*s15*s23+p5*s12*s23**2)-p1**2*(p4-s
     -   34)**2+4*p5*s12*s23*s34-s12*s15*s23*s34+p5*s23**2*s34+s12*s2
     -   3**2*s34+(s23*(2*p5-s34)*(s15-s34)+2*(s12*s15*(s15-s23)-s12*
     -   (s15+s23)*s34))*s45+p4*s12*(-(s23*s34)+2*(-(p5*s23)+s15*s23+
     -   s12*(s15+s23)-s15*s45+s34*s45))+p1*(2*p4**2*s12-p4*(s23*s34-
     -   2*(p5*s23-s12*(s15+s23+s34)+s15*s45-s34*s45))+s34*(s23*s34+2
     -   *(s12*(s15-s23)-p5*s23+(-s15+2*s23+s34)*s45)))-(s15-s34)**2*
     -   s45sq
       x(2,0)= p4*s12sq*s15-s12sq*s15**2-p4*p5*s12*s23+p4*s12sq*s23-p5
     -   **2*s23**2-s12sq*s23**2-p1**2*(p4-s34)**2-s12*s15*s23*s34+p5
     -   *s23**2*s34+s12*s23**2*s34+2*(p4*s12*s15*s23-p5*s12*s15*s23+
     -   s12sq*s15*s23+p5*s12*s23**2+p5*s12*s23*s34)-(s15*(s12*(p4-2*
     -   (s15-s23))-2*p5*s23)+(s12*s15+(p5+s12+s15)*s23)*s34)*s45+p1*
     -   (p4**2*s12+s34*(2*(s12*(s15-s23)-p5*s23)+s23*s34+(-2*s15+4*s
     -   23+s34)*s45)-p4*(s12*(2*(s15+s23)+s34)+s34*(s23+s45)-2*(p5*s
     -   23+s15*s45)))-s15*(s15-s34)*s45sq
       x(3,0)= p1**2*(p4-s34)*s34+s12*(-(s12*(s15-s23)**2)+p5*s23*(-s1
     -   5+s23)+p4*s12*(s15+s23)+s23*(2*p5-s15+s23)*s34-(-s15**2+s23*
     -   s34+s15*(s23+s34))*s45)-p1*(p4*s12*(s15+s23+s34)+s34*(p5*s23
     -   -2*(s12*s15-s12*s23)-s23*s34+(s15-2*s23-s34)*s45))
       x(4,0)= s23*(-(p1*s12*(p4+s34))+2*p1*s34*s45+s12*(p4*s12+s12*(s
     -   15-s23)+s23*(p5+s34)-(s15+s34)*s45))
       endif 
       endif 
       else 
       If(abs(p4).lt.eps) then 
       If(abs(p5).lt.eps) then 
c  Configuration                 (p1s0p3s00)
       d= 2*(16*msq*(p1p4**2*p2p3**2+p1p3**2*p2p4**2-p1*p2p4**2*p3+p1p
     -   2**2*p3p4**2-2*(p1p2*p1p3*p2p4*p3p4-p1*p2p3*p2p4*p3p4+p1p4*(
     -   p1p3*p2p3*p2p4-p1p2*p2p4*p3+p1p2*p2p3*p3p4)))+(s12*s15-p1*s3
     -   4)*(p3*s15-s23*s34)*s45)
       x(1,1) = s12sq*(s15-s23)**2+p1**2*(p3-s34)**2-4*msq*(p3+s15-s23
     -   -s34)*(p3*s15-s23*s34)-2*(s12*(s15*(2*p3+s15-s23)-(s15+s23)*
     -   s34)*s45-p1*(s12*(s15-s23)*(p3-s34)+((s15-2*s23-s34)*s34+p3*
     -   (s15+s34))*s45))+(s15-s34)**2*s45sq
       x(2,1) = (s12*(s15-s23)+p1*(p3-s34))**2+(-2*s12*s15*(p3+s15-s23
     -   )+s12*(s15+s23)*s34+p1*((2*s15-4*s23-s34)*s34+p3*(2*s15+s34)
     -   ))*s45-2*msq*(p3**2*s15+p1*(p3-s34)**2+s34*(s12*(-s15+s23)+(
     -   2*s23+s34)*(s23-s45)+s15*(-2*s23+s45))+p3*(2*s15**2+s12*(s15
     -   -s23)+s34*(-s23+s45)+s15*(-2*s23-s34+s45)))+s15*(s15-s34)*s4
     -   5sq
       x(3,1) = (s12*(s15-s23)+p1*(p3-s34))**2+(-(s12*s15*(2*p3+s15-s2
     -   3))+s12*(s15+s23)*s34+p1*((s15-2*s23-s34)*s34+p3*(s15+s34)))
     -   *s45-2*msq*(p3+s15-s23-s34)*(p3*s15+s12*s15-s12*s23+p1*(p3-s
     -   34)-s23*s34-s15*s45+s34*s45)
       x(4,1) = (p1*p3-s12*s23)*(s12*(s15-s23)+p1*(p3-s34))+(p1*(-2*s2
     -   3*s34+p3*(s15+s34))+s12*(-2*p3*s15+s23*(s15+s34)))*s45+2*msq
     -   *(-(p3**2*s15)-p1*p3*(p3+2*s15-s23)+p1*(p3+s23)*s34+p3*(s12*
     -   (s15+s23)+s23*(s15+s34)+(s15-s34)*s45)-s23*(s23*s34+s12*(-s1
     -   5+s23+2*s34)+s15*s45-s34*s45))
       x(2,2) = (s12*(s15-s23)+p1*(p3-s34))**2+2*(s12*s15*(-s15+s23)+p
     -   1*(p3*s15+s15*s34-2*s23*s34))*s45-4*msq*(p1*(p3-s34)**2+(p3*
     -   (s12+s15)-s34*(s12+s23-s45))*(s15-s23+s45))+s15**2*s45sq
       x(3,2) = (s12*(s15-s23)+p1*(p3-s34))**2+(s12*s15*(-s15+s23)+p1*
     -   (p3*s15+(s15-2*s23)*s34))*s45+2*msq*(p1*(p3-s34)*(-s15+s23-2
     -   *(p3-s34)+s45)+(s15-s23+s45)*(-(p3*(2*s12+s15))+s23*s34+s12*
     -   (-s15+s23+2*s34)+s15*s45-s34*s45))
       x(4,2) = (p1*p3-s12*s23)*(s12*(s15-s23)+p1*(p3-s34))+(p1*p3*s15
     -   +s12*s15*s23-2*p1*s23*s34)*s45+2*msq*(-((s12+s23-s45)*(s12*(
     -   -s15+s23)+s23*s34+s15*s45-s34*s45))-p3*(s12*(s15-2*(s23-s45)
     -   )+s15*(-s23+s45))+p1*(-2*p3**2+s34*(-s12+s23+s45)+p3*(s12+s2
     -   3-2*(s15-s34)+s45)))
       x(3,3) = (s12*(s15-s23)+p1*(p3-s34))**2+4*msq*(p3+s15-s23-s34)*
     -   (-(s12*(s15-s23+s45))+p1*(-p3+s34+s45))
       x(4,3) = (p1*p3-s12*s23)*(s12*(s15-s23)+p1*(p3-s34))-2*(msq*(p1
     -   **2*(p3-s34)+s12*(p3*(s15-2*s23)-(s15-s23)*(s12+2*s23)+s23*s
     -   34)+p1*(2*p3**2+s23*s34+s12*(s15-s23+s34)-p3*(s12-s15+2*(s23
     -   +s34))))-msq*(p1-s12)*(s15+2*(p3-s23)-s34)*s45)
       x(4,4) = (p1*p3-s12*s23)*(p1*p3-4*msq*(p1+p3-s12-s23)-s12*s23)+
     -   4*msq*(p1-s12)*(p3-s23)*s45
c          
       x(1,0)= -(p1**2*(p3-s34)**2)+s12*(s15-s23)*(p3*s15+s12*(-s15+s2
     -   3)-s23*s34)+p1*(p3-s34)*(p3*s15-2*(s12*s15-s12*s23)-s23*s34)
     -   -2*p1*((s15-2*s23-s34)*s34+p3*(s15+s34))*s45+(p3*s15*(4*s12+
     -   s15-s34)+s23*s34*(-s15+s34)+2*(s12*s15*(s15-s23)-s12*(s15+s2
     -   3)*s34))*s45-(s15-s34)**2*s45sq
       x(2,0)= -(p1**2*(p3-s34)**2)+s12*(s15-s23)*(p3*s15-s12*s15+s12*
     -   s23-s23*s34)+p1*(p3-s34)*(p3*s15-2*(s12*s15-s12*s23)-s23*s34
     -   )+(p3*s15*(2*s12+s15)+2*s12*s15*(s15-s23)-(s12*s15+(s12+s15)
     -   *s23)*s34)*s45-p1*((2*s15-4*s23-s34)*s34+p3*(2*s15+s34))*s45
     -   -s15*(s15-s34)*s45sq
       x(3,0)= -(p1**2*(p3-s34)**2)+p1*(p3-s34)*(p3*s15-2*(s12*s15-s12
     -   *s23)-s23*s34)-s12*(s15-s23)*(-(p3*s15)+s12*s15-s12*s23+s23*
     -   s34)+s12*(s15*(2*p3+s15-s23)-(s15+s23)*s34)*s45-p1*((s15-2*s
     -   23-s34)*s34+p3*(s15+s34))*s45
       x(4,0)= -((p1*p3-s12*s23)*(-(p3*s15)+s12*s15-s12*s23+p1*(p3-s34
     -   )+s23*s34))-(p1*(-2*s23*s34+p3*(s15+s34))+s12*(-2*p3*s15+s23
     -   *(s15+s34)))*s45
       else 
c  Configuration                 (p1s0p3s0p5s)
       d= 2*(16*msq*(p1p4**2*p2p3**2+p1p3**2*p2p4**2-p1*p2p4**2*p3+p1p
     -   2**2*p3p4**2-2*(p1p2*p1p3*p2p4*p3p4-p1*p2p3*p2p4*p3p4+p1p4*(
     -   p1p3*p2p3*p2p4-p1p2*p2p4*p3+p1p2*p2p3*p3p4)))+(p3*s15-s23*s3
     -   4)*(p1*p3*p5-p5*s12*s23+s12*s15*s45-p1*s34*s45))
       x(1,1) = (p1-s12)**2*(p3-s34)**2-4*msq*(p3+s15-s23-s34)*(p3*s15
     -   -s23*s34)+(p3-s23)**2*(p5-s45)**2+(p3+s15-s23-s34)**2*(s12-s
     -   45)**2+2*((p1-s12)*(-p3**2+s23*s34+p3*(-2*s15+s23+s34))*(p5-
     -   s45)+(p1-s12)*(p3-s34)*(p3+s15-s23-s34)*(s12-s45)+(p3-s23)*(
     -   p3+s15-s23-s34)*(p5-s45)*(-s12+s45))
       x(2,1) = p3*p5*s12*s15+s12sq*s15**2-p3*p5**2*s23+p3*p5*s12*s23+
     -   p5**2*s23**2+s12sq*s23**2+p1**2*(p3-s34)**2+(p3*(p5-2*s12)*s
     -   15-2*s12*s15*(s15-s23)+s12*(s15+s23)*s34+p5*s23*(-2*s15+s34)
     -   )*s45+p1*(-(p3**2*p5)+s34*(-2*(s12*(s15-s23)-p5*s23)-(-2*s15
     -   +4*s23+s34)*s45)+p3*(p5*(-4*s15+2*s23+s34)+s34*s45+2*(s12*s1
     -   5-s12*s23+s15*s45)))-2*(-(p5*s12*s15*s23)+s12sq*s15*s23+p5*s
     -   12*s23**2+p5*s12*s23*s34+msq*(p3**2*(-p5+s15)+p1*(p3-s34)**2
     -   +p3*(s12*(s15-s23)+2*(s15**2-s15*s23)-s15*s34-s23*s34+p5*(-2
     -   *s15+s23+s34)+(s15+s34)*s45)+s34*(p5*s23+s12*(-s15+s23)+(2*s
     -   23+s34)*(s23-s45)+s15*(-2*s23+s45))))+s15*(s15-s34)*s45sq
       x(3,1) = p1**2*(p3-s34)**2-2*msq*(p3+s15-s23-s34)*(s12*s15+p3*(
     -   -p5+s15)+p5*s23-s12*s23+p1*(p3-s34)-s23*s34-s15*s45+s34*s45)
     -   +s12*(s12*(s15-s23)**2+p5*s15*s23-p5*s23**2+p3*p5*(s15+s23)-
     -   s15**2*s45+s15*s23*s45+s15*s34*s45+s23*s34*s45-2*(p5*s23*s34
     -   +p3*s15*s45))+p1*(-(p3**2*p5)+s34*(p5*s23-2*(s12*s15-s12*s23
     -   )+(s15-2*s23-s34)*s45)+p3*(2*s12*(s15-s23)+p5*(-2*s15+s23+s3
     -   4)+(s15+s34)*s45))
       x(4,1) = (p1*p3-s12*s23)*(-(p3*p5)+s12*s15+p5*s23-s12*s23+p1*(p
     -   3-s34))+(p1*(-2*s23*s34+p3*(s15+s34))+s12*(-2*p3*s15+s23*(s1
     -   5+s34)))*s45+2*msq*(p3**2*(p5-s15)-p1*p3*(p3+2*s15-s23)+p1*(
     -   p3+s23)*s34+p3*(-2*p5*s23+s12*(s15+s23)+s23*(s15+s34)+(s15-s
     -   34)*s45)+s23*(p5*s23+s12*(s15-s23-2*s34)-s23*s34-s15*s45+s34
     -   *s45))
       x(2,2) = p1**2*(p3-s34)**2-4*msq*(p1*(p3-s34)**2+(p3*(p5-s12-s1
     -   5)+s34*(s12+s23-s45))*(p5-s15+s23-s45))+(s12*(s15-s23)+p5*s2
     -   3-s15*s45)**2+2*p1*(p3*(-2*p5*s15+s12*s15+p5*s23-s12*s23+s15
     -   *s45)+s34*(-(s12*s15)+p5*s23+s12*s23+s15*s45-2*s23*s45))
       x(3,2) = p1**2*(p3-s34)**2+s12*(s15-s23)*(s12*(s15-s23)+p5*s23-
     -   s15*s45)-2*msq*(p1*(p3-s34)*(p5+s15-s23+2*(p3-s34)-s45)+(p5-
     -   s15+s23-s45)*(p3*(p5-2*s12-s15)-p5*s23+s23*s34+s12*(-s15+s23
     -   +2*s34)+s15*s45-s34*s45))+p1*(p3*(p5*(-2*s15+s23)+2*(s12*s15
     -   -s12*s23)+s15*s45)+s34*(p5*s23+s15*s45-2*(s12*s15-s12*s23+s2
     -   3*s45)))
       x(4,2) = (p1*p3-s12*s23)*(s12*(s15-s23)+p5*s23+p1*(p3-s34))+(p1
     -   *p3*s15+s12*s15*s23-2*p1*s23*s34)*s45+2*msq*((s12+s23-s45)*(
     -   s12*(s15-s23)+p5*s23-s23*s34-s15*s45+s34*s45)+p3*(-(s12*s15)
     -   +2*s12*s23+s15*s23-(2*s12+s15)*s45+p5*(s12-s23+s45))+p1*(-2*
     -   p3**2+s34*(-s12+s23+s45)+p3*(s12+s23-2*(p5+s15-s34)+s45)))
       x(3,3) = (s12*(s15-s23)+p1*(p3-s34))**2-4*msq*(p3+s15-s23-s34)*
     -   (p1*(p3+p5-s34-s45)+s12*(-p5+s15-s23+s45))
       x(4,3) = (p1*p3-s12*s23)*(s12*(s15-s23)+p1*(p3-s34))-2*msq*(p1*
     -   *2*(p3-s34)+s12*(s12*(-s15+s23)+s23*(p5-2*(s15-s23)+s34)+p3*
     -   (-p5+s15-2*(s23-s45))+(s15-2*s23-s34)*s45)+p1*(2*p3**2-p5*s2
     -   3+s23*s34+s12*(s15-s23+s34)+(-s15+2*s23+s34)*s45+p3*(p5-s12+
     -   s15-2*(s23+s34+s45))))
       x(4,4) = (p1*p3-s12*s23)*(p1*p3-4*msq*(p1+p3-s12-s23)-s12*s23)+
     -   4*msq*(p1-s12)*(p3-s23)*s45
c          
       x(1,0)= -(s12sq*s15**2)+p3**2*p5*(-p5+s15)-p5**2*s23**2-s12sq*s
     -   23**2+2*(-(p5*s12*s15*s23)+s12sq*s15*s23+p5*s12*s23**2)-p1**
     -   2*(p3-s34)**2+4*p5*s12*s23*s34-s12*s15*s23*s34+p5*s23**2*s34
     -   +s12*s23**2*s34+(s23*(2*p5-s34)*(s15-s34)+2*(s12*s15*(s15-s2
     -   3)-s12*(s15+s23)*s34))*s45+p3*(2*p5**2*s23+s15*(s15-s34)*s45
     -   +s12*s15*(s15-s23+4*s45)-p5*(s23*(s15+s34)+2*(s12*(s15+s23)+
     -   (s15-s34)*s45)))+p1*(p3**2*(2*p5+s15)-p3*((s15+s23)*s34+2*(s
     -   12*(s15-s23)+p5*(-2*s15+s23+s34)+(s15+s34)*s45))+s34*(s23*s3
     -   4+2*(s12*(s15-s23)-p5*s23+(-s15+2*s23+s34)*s45)))-(s15-s34)*
     -   *2*s45sq
       x(2,0)= -(p3*p5*s12*s15)+p3*s12*s15**2-s12sq*s15**2+p3*p5**2*s2
     -   3-p3*p5*s12*s23-p3*p5*s15*s23-p3*s12*s15*s23-p5**2*s23**2-s1
     -   2sq*s23**2-p1**2*(p3-s34)**2-s12*s15*s23*s34+p5*s23**2*s34+s
     -   12*s23**2*s34+2*(-(p5*s12*s15*s23)+s12sq*s15*s23+p5*s12*s23*
     -   *2+p5*s12*s23*s34)-(p3*(p5-2*s12-s15)*s15-2*s15*(s12*(s15-s2
     -   3)+p5*s23)+(s12*s15+(p5+s12+s15)*s23)*s34)*s45-p1*(-(p3**2*(
     -   p5+s15))+p3*(2*s12*(s15-s23)+(s15+s23)*s34+p5*(-4*s15+2*s23+
     -   s34)+(2*s15+s34)*s45)-s34*(2*(s12*(s15-s23)-p5*s23)+s23*s34+
     -   (-2*s15+4*s23+s34)*s45))-s15*(s15-s34)*s45sq
       x(3,0)= -(p1**2*(p3-s34)**2)-s12*(s12*(s15-s23)**2+p5*s15*s23-p
     -   5*s23**2-2*p5*s23*s34+s15*s23*s34-s23**2*s34+p3*(p5*(s15+s23
     -   )+s15*(-s15+s23-2*s45))-s15**2*s45+s15*s23*s45+s15*s34*s45+s
     -   23*s34*s45)-p1*(-(p3**2*(p5+s15))+s34*(p5*s23-2*(s12*s15-s12
     -   *s23)-s23*s34+(s15-2*s23-s34)*s45)+p3*(2*s12*(s15-s23)+(s15+
     -   s23)*s34+p5*(-2*s15+s23+s34)+(s15+s34)*s45))
       x(4,0)= -((p1*p3-s12*s23)*(s12*s15-p3*(p5+s15)+p5*s23-s12*s23+p
     -   1*(p3-s34)+s23*s34))-(p1*(-2*s23*s34+p3*(s15+s34))+s12*(-2*p
     -   3*s15+s23*(s15+s34)))*s45
       endif 
       else 
       If(abs(p5).lt.eps) then 
c  Configuration                 (p1s0p3sp4s0)
       d= 2*(16*msq*(p1p4**2*p2p3**2+p1p3**2*p2p4**2-p1*p2p4**2*p3+p1p
     -   2**2*p3p4**2-2*(p1p2*p1p3*p2p4*p3p4-p1*p2p3*p2p4*p3p4+p1p4*(
     -   p1p3*p2p3*p2p4-p1p2*p2p4*p3+p1p2*p2p3*p3p4))-(-2*p1p2*p1p3*p
     -   2p3+p1*p2p3**2+p1p2**2*p3)*p4)+(s12*s15-p1*s34)*(-(p1*p3*p4)
     -   +p4*s12*s23+p3*s15*s45-s23*s34*s45))
       x(1,1) = -4*msq*(s15*(p3**2+p4*s23-p3*(p4-s15+s23))-((p4+s15-s2
     -   3)*s23+p3*(-p4+s15+s23))*s34+s23*s34**2)+(p1-s12)**2*(p3**2+
     -   (p4-s34)**2-2*p3*(p4+s34))+(p3+s15-s23-s34)**2*(s12-s45)**2-
     -   2*(-((p1-s12)*(p3**2+p4*(s15+s23-s34)+s34*(-s15+s23+s34)-p3*
     -   (p4-s15+s23+2*s34))*(s12-s45))+(p1-s12)*(-p3**2+s23*(-p4+s34
     -   )+p3*(p4-2*s15+s23+s34))*s45+(p3-s23)*(p3+s15-s23-s34)*s45*(
     -   -s12+s45))+(p3-s23)**2*s45sq
       x(2,1) = -(s12sq*(-(s15-s23)**2+p4*(s15+s23)))+p1*s12*(p3*(p4+2
     -   *(s15-s23))-p4*(p4-2*(s15+s23))+(p4-2*(s15-s23))*s34)+p1**2*
     -   (p3**2+(p4-s34)**2-2*p3*(p4+s34))+s12*(-2*p3*s15+s15*(p4-2*(
     -   s15-s23))+(s15+s23)*s34)*s45+p1*(2*(p3-p4)*s15+(p3+p4+2*s15-
     -   4*s23)*s34-s34**2)*s45-2*msq*(-(p4**2*s12)+p3**2*s15+p4*s12*
     -   s15+p4*s12*s23+p4*s12*s34-s12*s15*s34-p4*s23*s34+s12*s23*s34
     -   +s23*s34**2+2*(p4*s15*s23-s15*s23*s34+s23**2*s34)+p1*(p3**2+
     -   (p4-s34)**2-2*p3*(p4+s34))-(p4*s15-(p4+s15-2*s23)*s34+s34**2
     -   )*s45+p3*(p4*s12-p4*s15+s12*s15-s12*s23+2*(s15**2-s15*s23)-s
     -   15*s34-s23*s34+(s15+s34)*s45))+s15*(s15-s34)*s45sq
       x(3,1) = p1*s12*(p3*(p4+2*(s15-s23))+2*(-s15+s23)*s34+p4*(s15+s
     -   23+s34))+p1**2*(p3**2+s34*(-p4+s34)-p3*(p4+2*s34))+p1*((s15-
     -   2*s23-s34)*s34+p3*(s15+s34))*s45-2*msq*(p3*p4*s12+p3**2*s15+
     -   p3*s12*s15-p4*s12*s15+p3*s15**2+s12*s15**2-p3*s12*s23-p4*s12
     -   *s23-p3*s15*s23-2*s12*s15*s23+s12*s23**2+p4*s12*s34-p3*s15*s
     -   34-s12*s15*s34-p3*s23*s34+s12*s23*s34-s15*s23*s34+s23**2*s34
     -   +s23*s34**2+p1*(p3**2+p4*(s15+s23-s34)+s34*(-s15+s23+s34)-p3
     -   *(p4-s15+s23+2*s34))-(s15-s34)*(p3+s15-s23-s34)*s45)+s12*(s1
     -   2*(s15-s23)**2-p4*s12*(s15+s23)+(-(s15*(2*p3+s15-s23))+(s15+
     -   s23)*s34)*s45)
       x(4,1) = (p1*p3-s12*s23)*(s12*(p4+s15-s23)+p1*(p3-p4-s34))+(p1*
     -   (-2*s23*s34+p3*(s15+s34))+s12*(-2*p3*s15+s23*(s15+s34)))*s45
     -   +2*msq*(-(p3**2*s15)+p1*(-p3**2+s23*(-p4+s34)+p3*(p4-2*s15+s
     -   23+s34))+p3*(-(p4*s12)+s12*(s15+s23)+s23*(s15+s34)+(s15-s34)
     -   *s45)+s23*(p4*s12+s12*(s15-s23-2*s34)-s23*s34-s15*s45+s34*s4
     -   5))
       x(2,2) = p1**2*(p3**2+(p4-s34)**2-2*p3*(p4+s34))+(s12*(-s15+s23
     -   )+s15*s45)**2+2*p1*(s12*(p3*(s15-s23)+p4*(s15+s23)+(-s15+s23
     -   )*s34)+(p3*s15-p4*s15+(s15-2*s23)*s34)*s45)-4*msq*(p1*(p3**2
     -   +(p4-s34)**2-2*p3*(p4+s34))+p4*(s12+s15)*(s12+s23-s45)+(p3*(
     -   s12+s15)-s34*(s12+s23-s45))*(s15-s23+s45))
       x(3,2) = p1**2*(p3**2+s34*(-p4+s34)-p3*(p4+2*s34))+p1*s12*(p4*(
     -   s15+s23)+2*(p3*(s15-s23)+(-s15+s23)*s34))+p1*(p3*s15+(s15-2*
     -   s23)*s34)*s45+s12*(s15-s23)*(s12*(s15-s23)-s15*s45)-2*msq*(p
     -   4*s12*(2*s12+s15+s23-s45)+(s15-s23+s45)*(p3*(2*s12+s15)+s12*
     -   (s15-s23-2*s34)-s23*s34-s15*s45+s34*s45)+p1*(2*p3**2+s34*(-s
     -   15+s23+2*s34+s45)-p3*(2*p4-s15+s23+4*s34+s45)+p4*(s15+s23-2*
     -   (s12+s34)+s45)))
       x(4,2) = (p1*p3-s12*s23)*(s12*(s15-s23)+p1*(p3-p4-s34))+(p1*p3*
     -   s15+s12*s15*s23-2*p1*s23*s34)*s45+2*msq*(-(p3*s12*s15)+s12sq
     -   *s15+2*p3*s12*s23-s12sq*s23+p3*s15*s23+s12*s15*s23-s12*s23**
     -   2-s12*s23*s34-s23**2*s34-p4*s12*(s12+s23-s45)-(p3*s15+2*(p3*
     -   s12+s12*s15)-s12*s23+s15*s23-(s12+2*s23)*s34)*s45+p1*(-2*p3*
     -   *2+(p4-s34)*(s12-s23-s45)+p3*(s12+s23+2*(p4-s15+s34)+s45))+(
     -   s15-s34)*s45sq)
       x(3,3) = (s12*(s15-s23)+p1*(p3-s34))**2-4*(msq*(p1**2*p4+s12*(p
     -   4*s12+(s15-s23)*(p3+s15-s23-s34))+p1*(-2*p4*s12+(p3-s34)*(p3
     -   +s15-s23-s34)))-msq*(p1-s12)*(p3+s15-s23-s34)*s45)
       x(4,3) = (p1*p3-s12*s23)*(s12*(s15-s23)+p1*(p3-s34))-2*(msq*(p1
     -   **2*(p3+p4-s34)+s12*(p4*s12+p3*(s15-2*s23)-(s15-s23)*(s12+2*
     -   s23)+s23*s34)+p1*(2*p3**2-s12*(2*p4-s15+s23)+(s12+s23)*s34-p
     -   3*(s12-s15+2*(s23+s34))))-msq*(p1-s12)*(s15+2*(p3-s23)-s34)*
     -   s45)
       x(4,4) = (p1*p3-s12*s23)*(p1*p3-4*msq*(p1+p3-s12-s23)-s12*s23)+
     -   4*msq*(p1-s12)*(p3-s23)*s45
c          
       x(1,0)= -(p1**2*(p3**2+(p4-s34)**2-2*p3*(p4+s34)))+s12*(-(p4**2
     -   *s12)+p4*(-(p3*s15)+2*s12*(s15+s23)+s23*(2*s15-s34))+(s15-s2
     -   3)*(p3*s15+s12*(-s15+s23)-s23*s34))+(-2*s12*s15*(p4-s15+s23)
     -   +p3*s15*(4*s12+s15-s34)+(-(s15*s23)+2*(p4*s12-s12*(s15+s23))
     -   )*s34+s23*s34**2)*s45+p1*(2*p4**2*s12+p3**2*s15-p4*(s23*s34+
     -   2*(s12*(s15+s23+s34)+(-s15+s34)*s45))-p3*(p4*(s15+2*(s12-s34
     -   ))+(s15+s23)*s34+2*(s12*(s15-s23)+(s15+s34)*s45))+s34*(s23*s
     -   34+2*(s12*(s15-s23)+(-s15+2*s23+s34)*s45)))-(s15-s34)**2*s45
     -   sq
       x(2,0)= -(p1**2*(p3**2+(p4-s34)**2-2*p3*(p4+s34)))+s12*(p4*(2*s
     -   15*s23+s12*(s15+s23))+(s15-s23)*(p3*s15-s12*s15+s12*s23-s23*
     -   s34))-(-(p3*s15*(2*s12+s15))+s12*s15*(p4-2*(s15-s23))+(s15*s
     -   23+s12*(s15+s23))*s34)*s45+p1*(p4**2*s12+p3**2*s15-p3*(p4*(s
     -   12+s15)+2*s12*(s15-s23)+(s15+s23)*s34+(2*s15+s34)*s45)+s34*(
     -   2*s12*(s15-s23)+s23*s34+(-2*s15+4*s23+s34)*s45)-p4*(s12*(2*(
     -   s15+s23)+s34)-2*s15*s45+s34*(s23+s45)))-s15*(s15-s34)*s45sq
       x(3,0)= p1**2*(-p3**2+(p4-s34)*s34+p3*(p4+2*s34))+s12*(p4*s12*(
     -   s15+s23)+(s15-s23)*(p3*s15-s12*s15+s12*s23-s23*s34)+(s15*(2*
     -   p3+s15-s23)-(s15+s23)*s34)*s45)+p1*(p3**2*s15-p4*s12*(s15+s2
     -   3+s34)-p3*(p4*s12+2*s12*(s15-s23)+(s15+s23)*s34+(s15+s34)*s4
     -   5)+s34*(2*s12*(s15-s23)+s23*s34+(-s15+2*s23+s34)*s45))
       x(4,0)= -((p1*p3-s12*s23)*(-(p3*s15)+s12*(p4+s15-s23)+p1*(p3-p4
     -   -s34)+s23*s34))-(p1*(-2*s23*s34+p3*(s15+s34))+s12*(-2*p3*s15
     -   +s23*(s15+s34)))*s45
       else 
c  Configuration                 (p1s0p3sp4sp5s)
       d= 2*(16*msq*(p1p4**2*p2p3**2+p1p3**2*p2p4**2-p1*p2p4**2*p3+p1p
     -   2**2*p3p4**2-2*(p1p2*p1p3*p2p4*p3p4-p1*p2p3*p2p4*p3p4+p1p4*(
     -   p1p3*p2p3*p2p4-p1p2*p2p4*p3+p1p2*p2p3*p3p4))-(-2*p1p2*p1p3*p
     -   2p3+p1*p2p3**2+p1p2**2*p3)*p4)+(p1*p3-s12*s23)*(p3*p5*s15-p4
     -   *s12*s15+p1*p4*s34-p5*s23*s34)+(s12*s15-p1*s34)*(p3*s15-s23*
     -   s34)*s45)
       x(1,1) = -4*msq*(s15*(p3**2+p4*s23-p3*(p4-s15+s23))-((p4+s15-s2
     -   3)*s23+p3*(-p4+s15+s23))*s34+s23*s34**2)+(p1-s12)**2*(p3**2+
     -   (p4-s34)**2-2*p3*(p4+s34))+(p3-s23)**2*(p5-s45)**2+(p3+s15-s
     -   23-s34)**2*(s12-s45)**2+2*((p1-s12)*(-p3**2+s23*(-p4+s34)+p3
     -   *(p4-2*s15+s23+s34))*(p5-s45)+(p1-s12)*(p3**2+p4*(s15+s23-s3
     -   4)+s34*(-s15+s23+s34)-p3*(p4-s15+s23+2*s34))*(s12-s45)+(p3-s
     -   23)*(p3+s15-s23-s34)*(p5-s45)*(-s12+s45))
       x(2,1) = p3*p5*s12*s15-p4*s12sq*s15+s12sq*s15**2-p3*p5**2*s23+p
     -   3*p5*s12*s23+p4*p5*s12*s23-p4*s12sq*s23+p5**2*s23**2+s12sq*s
     -   23**2+p1**2*(p3**2+(p4-s34)**2-2*p3*(p4+s34))+(s15*(p3*(p5-2
     -   *s12)+s12*(p4-2*(s15-s23))-2*p5*s23)+(p5*s23+s12*(s15+s23))*
     -   s34)*s45+p1*(-(p3**2*p5)-p4**2*s12+p4*(-2*p5*s23+s12*(2*(s15
     -   +s23)+s34)+(-2*s15+s34)*s45)+s34*(-2*(s12*(s15-s23)-p5*s23)-
     -   (-2*s15+4*s23+s34)*s45)+p3*(p4*(p5+s12)+p5*(-4*s15+2*s23+s34
     -   )+s34*s45+2*(s12*s15-s12*s23+s15*s45)))-2*(-(p5*s12*s15*s23)
     -   +s12sq*s15*s23+p5*s12*s23**2+p5*s12*s23*s34+msq*(-(p4**2*s12
     -   )+p4*s12*s15+p3**2*(-p5+s15)-p4*p5*s23+p4*s12*s23+p4*s12*s34
     -   -s12*s15*s34-p4*s23*s34+p5*s23*s34+s12*s23*s34+s23*s34**2+2*
     -   (p4*s15*s23-s15*s23*s34+s23**2*s34)+p1*(p3**2+(p4-s34)**2-2*
     -   p3*(p4+s34))-(p4*s15-(p4+s15-2*s23)*s34+s34**2)*s45+p3*(p4*(
     -   p5+s12-s15)+s12*s15-s12*s23+2*(s15**2-s15*s23)-s15*s34-s23*s
     -   34+p5*(-2*s15+s23+s34)+(s15+s34)*s45)))+s15*(s15-s34)*s45sq
       x(3,1) = p1**2*(p3**2+s34*(-p4+s34)-p3*(p4+2*s34))-2*msq*(p3*p4
     -   *s12+p3*s12*s15-p4*s12*s15+p3*s15**2+s12*s15**2+p3**2*(-p5+s
     -   15)-p3*s12*s23-p4*s12*s23-p3*s15*s23+p5*s15*s23-2*s12*s15*s2
     -   3-p5*s23**2+s12*s23**2+p4*s12*s34-p3*s15*s34-s12*s15*s34-p3*
     -   s23*s34-p5*s23*s34+s12*s23*s34-s15*s23*s34+s23**2*s34+s23*s3
     -   4**2+p3*p5*(-s15+2*s23+s34)+p1*(p3**2+p4*(s15+s23-s34)+s34*(
     -   -s15+s23+s34)-p3*(p4-s15+s23+2*s34))-(s15-s34)*(p3+s15-s23-s
     -   34)*s45)+s12*(s12*s15**2+p5*s15*s23-p5*s23**2+s12*s23**2+p3*
     -   p5*(s15+s23)-p4*s12*(s15+s23)-s15**2*s45+s15*s23*s45+s15*s34
     -   *s45+s23*s34*s45-2*(s12*s15*s23+p5*s23*s34+p3*s15*s45))+p1*(
     -   -(p3**2*p5)+p4*s12*(s15+s23+s34)+s34*(p5*s23-2*(s12*s15-s12*
     -   s23)+(s15-2*s23-s34)*s45)+p3*(s12*(p4+2*(s15-s23))+p5*(-2*s1
     -   5+s23+s34)+(s15+s34)*s45))
       x(4,1) = (p1*p3-s12*s23)*(-(p3*p5)+s12*(p4+s15-s23)+p5*s23+p1*(
     -   p3-p4-s34))+(p1*(-2*s23*s34+p3*(s15+s34))+s12*(-2*p3*s15+s23
     -   *(s15+s34)))*s45+2*msq*(p3**2*(p5-s15)+p1*(-p3**2+s23*(-p4+s
     -   34)+p3*(p4-2*s15+s23+s34))+p3*(-(p4*s12)-2*p5*s23+s12*(s15+s
     -   23)+s23*(s15+s34)+(s15-s34)*s45)+s23*(p4*s12+p5*s23+s12*(s15
     -   -s23-2*s34)-s23*s34-s15*s45+s34*s45))
       x(2,2) = p1**2*(p3**2+(p4-s34)**2-2*p3*(p4+s34))-4*msq*(p1*(p3*
     -   *2+(p4-s34)**2-2*p3*(p4+s34))+(p4*(-p5+s12+s15)+s34*(p5-s15+
     -   s23-s45))*(s12+s23-s45)+p3*(p5-s12-s15)*(p5-s15+s23-s45))+(s
     -   12*(s15-s23)+p5*s23-s15*s45)**2+2*p1*(p4*(-(p5*s23)+s12*(s15
     -   +s23)-s15*s45)+p3*(s12*s15-s12*s23+p5*(-2*s15+s23)+s15*s45)+
     -   s34*(-(s12*s15)+p5*s23+s12*s23+s15*s45-2*s23*s45))
       x(3,2) = p1**2*(p3**2+s34*(-p4+s34)-p3*(p4+2*s34))+s12*(s15-s23
     -   )*(s12*(s15-s23)+p5*s23-s15*s45)+p1*(p4*s12*(s15+s23)+p3*(p5
     -   *(-2*s15+s23)+2*(s12*s15-s12*s23)+s15*s45)+s34*(p5*s23+s15*s
     -   45-2*(s12*s15-s12*s23+s23*s45)))-2*msq*(-(p4*p5*s12)+p4*s12*
     -   s15-p5*s12*s15+s12*s15**2-p5**2*s23+p4*s12*s23+p5*s12*s23+p5
     -   *s15*s23-p5*s23**2+s12*s23**2+p5*s23*s34-s15*s23*s34+s23**2*
     -   s34+2*(p4*s12sq-s12*s15*s23+p5*s12*s34-s12*s15*s34+s12*s23*s
     -   34)+p3*(p5-2*s12-s15)*(p5-s15+s23-s45)-(p4*s12+(-s12+s15)*(s
     -   15-s23)-p5*(s15+s23-s34)+(-s15+2*(s12+s23))*s34)*s45+p1*(2*p
     -   3**2+s34*(-p5-s15+s23+2*s34+s45)-p3*(2*p4-p5-s15+s23+4*s34+s
     -   45)+p4*(p5+s15+s23-2*(s12+s34)+s45))-(s15-s34)*s45sq)
       x(4,2) = (p1*p3-s12*s23)*(s12*(s15-s23)+p5*s23+p1*(p3-p4-s34))+
     -   (p1*p3*s15+s12*s15*s23-2*p1*s23*s34)*s45-2*msq*((s12+s23-s45
     -   )*(p4*s12-p5*s23+s12*(-s15+s23)+s23*s34+s15*s45-s34*s45)+p3*
     -   (s12*(s15-2*(s23-s45))+s15*(-s23+s45)-p5*(s12-s23+s45))+p1*(
     -   2*p3**2-(p4-s34)*(s12-s23-s45)-p3*(s12+s23+2*(p4-p5-s15+s34)
     -   +s45)))
       x(3,3) = (s12*(s15-s23)+p1*(p3-s34))**2-4*msq*(p1**2*p4+p1*(-2*
     -   p4*s12+(p3+s15-s23-s34)*(p3+p5-s34-s45))+s12*(p4*s12+(p3+s15
     -   -s23-s34)*(-p5+s15-s23+s45)))
       x(4,3) = (p1*p3-s12*s23)*(s12*(s15-s23)+p1*(p3-s34))-2*msq*(p1*
     -   *2*(p3+p4-s34)+s12*(p4*s12+s12*(-s15+s23)+s23*(p5-2*(s15-s23
     -   )+s34)+p3*(-p5+s15-2*(s23-s45))+(s15-2*s23-s34)*s45)+p1*(s12
     -   *s15-p5*s23-s12*s23+s12*s34+s23*s34-s15*s45+s34*s45+2*(p3**2
     -   -p4*s12+s23*s45)+p3*(p5-s12+s15-2*(s23+s34+s45))))
       x(4,4) = (p1*p3-s12*s23)*(p1*p3-4*msq*(p1+p3-s12-s23)-s12*s23)+
     -   4*msq*(p1-s12)*(p3-s23)*s45
c          
       x(1,0)= -(p3**2*p5**2)-p4**2*s12sq+p3**2*p5*s15-p3*p4*s12*s15+p
     -   3*s12*s15**2-s12sq*s15**2-p3*p5*s15*s23-p3*s12*s15*s23-p5**2
     -   *s23**2-s12sq*s23**2+2*(p3*p4*p5*s12-p3*p5*s12*s15+p4*s12sq*
     -   s15+p3*p5**2*s23-p3*p5*s12*s23-p4*p5*s12*s23+p4*s12sq*s23+p4
     -   *s12*s15*s23-p5*s12*s15*s23+s12sq*s15*s23+p5*s12*s23**2)-p3*
     -   p5*s23*s34-p4*s12*s23*s34+4*p5*s12*s23*s34-s12*s15*s23*s34+p
     -   5*s23**2*s34+s12*s23**2*s34-p1**2*(p3**2+(p4-s34)**2-2*p3*(p
     -   4+s34))-(-(s15*(p3*(-2*p5+4*s12+s15)+2*(p5*s23-s12*(p4-s15+s
     -   23))))+(p3*(-2*p5+s15)-2*(p4*s12-s12*s15)+(2*(p5+s12)+s15)*s
     -   23)*s34-s23*s34**2)*s45+p1*(2*p4**2*s12+p3**2*(2*p5+s15)-p4*
     -   (s23*s34-2*(p5*s23-s12*(s15+s23+s34)+s15*s45-s34*s45))-p3*(p
     -   4*(s15+2*(p5+s12-s34))+s15*s34+s23*s34+2*(s12*s15-s12*s23+p5
     -   *(-2*s15+s23+s34)+(s15+s34)*s45))+s34*(s23*s34+2*(s12*(s15-s
     -   23)-p5*s23+(-s15+2*s23+s34)*s45)))-(s15-s34)**2*s45sq
       x(2,0)= p4*s12sq*s15-s12sq*s15**2-p4*p5*s12*s23+p4*s12sq*s23-p5
     -   **2*s23**2-s12sq*s23**2-s12*s15*s23*s34+p5*s23**2*s34+s12*s2
     -   3**2*s34+2*(p4*s12*s15*s23-p5*s12*s15*s23+s12sq*s15*s23+p5*s
     -   12*s23**2+p5*s12*s23*s34)-p1**2*(p3**2+(p4-s34)**2-2*p3*(p4+
     -   s34))-(s15*(s12*(p4-2*(s15-s23))-2*p5*s23)+(s12*s15+(p5+s12+
     -   s15)*s23)*s34)*s45+p3*(p5**2*s23-p5*(s12*(s15+s23)+s15*(s23+
     -   s45))+s15*(s15*s45+s12*(s15-s23+2*s45)))+p1*(p4**2*s12+p3**2
     -   *(p5+s15)+s34*(2*(s12*(s15-s23)-p5*s23)+s23*s34+(-2*s15+4*s2
     -   3+s34)*s45)-p4*(s12*(2*(s15+s23)+s34)+s34*(s23+s45)-2*(p5*s2
     -   3+s15*s45))-p3*(p4*(p5+s12+s15)+s15*s34+s23*s34+p5*(-4*s15+2
     -   *s23+s34)+s34*s45+2*(s12*s15-s12*s23+s15*s45)))-s15*(s15-s34
     -   )*s45sq
       x(3,0)= -(p1**2*(p3**2+s34*(-p4+s34)-p3*(p4+2*s34)))-s12*(s12*s
     -   15**2+p5*s15*s23-p5*s23**2+s12*s23**2-p4*s12*(s15+s23)+s15*s
     -   23*s34-s23**2*s34-2*(s12*s15*s23+p5*s23*s34)+p3*(p5*(s15+s23
     -   )+s15*(-s15+s23-2*s45))-s15**2*s45+s15*s23*s45+s15*s34*s45+s
     -   23*s34*s45)-p1*(-(p3**2*(p5+s15))+p4*s12*(s15+s23+s34)+s34*(
     -   p5*s23-2*(s12*s15-s12*s23)-s23*s34+(s15-2*s23-s34)*s45)+p3*(
     -   p4*s12+2*s12*(s15-s23)+(s15+s23)*s34+p5*(-2*s15+s23+s34)+(s1
     -   5+s34)*s45))
       x(4,0)= -((p1*p3-s12*s23)*(-(p3*(p5+s15))+s12*(p4+s15-s23)+p5*s
     -   23+p1*(p3-p4-s34)+s23*s34))-(p1*(-2*s23*s34+p3*(s15+s34))+s1
     -   2*(-2*p3*s15+s23*(s15+s34)))*s45
       endif 
       endif 
       endif 
       else 
       If(abs(p3).lt.eps) then 
       If(abs(p4).lt.eps) then 
       If(abs(p5).lt.eps) then 
c  Configuration                 (p1sp2s000)
       d= 2*(16*msq*((p1p4*p2p3-p1p3*p2p4)**2-2*(-(p1p3*p1p4*p2)+p1p2*
     -   p1p4*p2p3+p1p2*p1p3*p2p4-p1*p2p3*p2p4)*p3p4+(p1p2**2-p1*p2)*
     -   p3p4**2)+s34*s45*(-(s12*s15*s23)+p1*s23*s34+p2*s15*s45))
       x(1,1) = s12sq*(s15-s23)**2+s34*(p1**2*s34+4*(msq*(p2-s23)*(-s1
     -   5+s23)-msq*s23*s34))+2*(p1*(p2+s15-2*s23-s34)*s34*s45+s12*(p
     -   1*(-s15+s23)*s34+((p2-s15)*(s15-s23)+(s15+s23)*s34)*s45))+(p
     -   2**2+(s15-s34)**2-2*p2*(s15+s34))*s45sq
       x(2,1) = s12sq*(s15-s23)**2+2*(msq-p1)*s12*(s15-s23)*s34+s34*(4
     -   *msq*(p2-s23)*(-s15+s23)+p1**2*s34-2*msq*(p1+s23)*s34)+s34*(
     -   -2*(msq-p1)*(p2+s15-2*s23)+(2*msq-p1)*s34)*s45+s12*(2*(p2-s1
     -   5)*(s15-s23)+(s15+s23)*s34)*s45+((p2-s15)**2-(p2+s15)*s34)*s
     -   45sq
       x(3,1) = (s12*(s15-s23)-p1*s34)**2+(s12*(2*p2-s15)*(s15-s23)+(p
     -   1*(s15+2*(p2-s23))+s12*(s15+s23))*s34-p1*s34**2)*s45+2*msq*(
     -   -(s12*(s15-s23)**2)+(s15-s23)*(p1-2*p2+s12+s23)*s34-(p1+s23)
     -   *s34**2+((s15-s34)*(s15-s23-s34)-p2*(s15-s23+s34))*s45)-p2*(
     -   -p2+s15+s34)*s45sq
       x(4,1) = s12sq*s23*(-s15+s23)+p1*s12*s23*s34+s12*(p2*(s15-2*s23
     -   )+s23*(s15+s34))*s45+s45*(p1*(p2-2*s23)*s34+p2*(p2-s15-s34)*
     -   s45)-2*msq*(p2**2*s45+p2*(s12*(s15-s23)+(p1-s23)*s34-(s15+s2
     -   3+s34)*s45)+s23*(s12*(-s15+s23+2*s34)+s15*s45-s34*(p1-s23+s4
     -   5)))
       x(2,2) = p1**2*s34**2+(s12*(s15-s23)+(p2-s15)*s45)**2+2*p1*s34*
     -   (s12*(-s15+s23)+(p2+s15-2*s23)*s45)-4*msq*s34*(p1*s34+(p2-s1
     -   2-s23+s45)*(s15-s23+s45))
       x(3,2) = (s12*(s15-s23)-p1*s34)**2+(s12*(2*p2-s15)*(s15-s23)+p1
     -   *(s15+2*(p2-s23))*s34)*s45+2*msq*(-(s12*(s15-s23-2*s34)*(s15
     -   -s23+s45))-p1*s34*(-s15+s23+2*s34+s45)+(s15-s23+s45)*((-2*p2
     -   +s23)*s34-(p2-s15+s34)*s45))+p2*(p2-s15)*s45sq
       x(4,2) = s12*s23*(s12*(-s15+s23)+p1*s34)+(s12*(p2*(s15-2*s23)+s
     -   15*s23)+p1*(p2-2*s23)*s34)*s45+p2*(p2-s15)*s45sq+2*msq*(-(s1
     -   2*(-p2+s12+s23)*(-s15+s23))-(p1*(p2+s12-s23)+s23*(-p2+s12+s2
     -   3))*s34-(p2**2+2*s12*s15-s12*s23+s15*s23-p2*(s12+s15+s23-s34
     -   )-(p1+s12+2*s23)*s34)*s45-(p2-s15+s34)*s45sq)
       x(3,3) = p1**2*s34**2+(s12*(s15-s23)+p2*s45)**2+2*p1*s34*(s12*(
     -   -s15+s23)+p2*s45)-4*msq*(s12*(s15-s23-s34)*(s15-s23+s45)+(s3
     -   4+s45)*(p1*(-s15+s23+s34)+p2*(s15-s23+s45)))
       x(4,3) = (s12*s23-p2*s45)*(s12*(-s15+s23)-p2*s45)+p1*s34*(s12*s
     -   23+p2*s45)-2*msq*(-(p1**2*s34)+p2**2*s45+p1*((p2+s23)*s34+s1
     -   2*(s15-s23+s34)-(p2+s15-2*s23-s34)*s45)+s12*(s12*(-s15+s23)+
     -   (2*s23+s34)*(s23-s45)+s15*(-2*s23+s45))+p2*(s12*(s15-s23-s45
     -   )-s23*(s34+2*s45)+s45*(s15+s34+2*s45)))
       x(4,4) = (s12*s23-p2*s45)**2-4*msq*(p1*(p2-s23)*(s12-s45)+(p2-s
     -   12-s23+s45)*(-(s12*s23)+p2*s45))
c          
       x(1,0)= -(s12sq*(s15-s23)**2)+s12*(2*p1-s23)*(s15-s23)*s34-p1*(
     -   p1-s23)*s34**2+s34*(-(p2*s23)-s15*s23+2*(p2*s15-p1*(p2+s15-2
     -   *s23-s34))+s23*s34)*s45-2*s12*((p2-s15)*(s15-s23)+(s15+s23)*
     -   s34)*s45-(p2**2+(s15-s34)**2-2*p2*(s15+s34))*s45sq
       x(2,0)= -(s12sq*(s15-s23)**2)+s12*(2*p1-s23)*(s15-s23)*s34-p1*(
     -   p1-s23)*s34**2-s12*(2*(p2-s15)*(s15-s23)+(s15+s23)*s34)*s45+
     -   s34*(2*p2*s15-(p2+s15)*s23+p1*(-2*(p2+s15-2*s23)+s34))*s45+(
     -   -(p2-s15)**2+(p2+s15)*s34)*s45sq
       x(3,0)= -(s12sq*(s15-s23)**2)+s12*(2*p1-s23)*(s15-s23)*s34-p1**
     -   2*s34**2-s12*((2*p2-s15)*(s15-s23)+(s15+s23)*s34)*s45+p2*s45
     -   *((2*s15-s23)*s34+(-p2+s15+s34)*s45)+p1*s34*(s23*s34+(-s15-2
     -   *(p2-s23)+s34)*s45)
       x(4,0)= s12*s23*(s12*(s15-s23)+(-p1+s23)*s34)-(p2*s12*(s15-2*s2
     -   3)+p2*(p1+s23)*s34+s23*(-2*p1*s34+s12*(s15+s34)))*s45+p2*(-p
     -   2+s15+s34)*s45sq
       else 
c  Configuration                 (p1sp2s00p5s)
       d= 2*(16*msq*((p1p4*p2p3-p1p3*p2p4)**2-2*(-(p1p3*p1p4*p2)+p1p2*
     -   p1p4*p2p3+p1p2*p1p3*p2p4-p1*p2p3*p2p4)*p3p4+(p1p2**2-p1*p2)*
     -   p3p4**2)+p5*s12*s23**2*s34+s34*s45*(-(p2*p5*s23)-s12*s15*s23
     -   +p1*s23*s34+p2*s15*s45))
       x(1,1) = (p1-s12)**2*s34**2-4*msq*s34*((p2-s23)*(s15-s23)+s23*s
     -   34)+(p2-s23)**2*(p5-s45)**2+(-s15+s23+s34)**2*(s12-s45)**2+2
     -   *((p1-s12)*(-p2+s23)*s34*(p5-s45)+(p1-s12)*s34*(-s15+s23+s34
     -   )*(s12-s45)+(p2*(s15-s23+s34)+s23*(-s15+s23+s34))*(p5-s45)*(
     -   -s12+s45))
       x(2,1) = p2**2*p5**2+s12sq*s15**2+p5**2*s23**2+s12sq*s23**2+p1*
     -   *2*s34**2-(2*(p2-s15)*(p2*p5-p5*s23+s12*(-s15+s23))-(p2*p5+p
     -   5*s23+s12*(s15+s23))*s34)*s45-p1*s34*(2*(s12*s15-(p5+s12)*s2
     -   3+p2*(p5-s45))+(-2*s15+4*s23+s34)*s45)-2*(p2*p5*s12*s15+p2*p
     -   5**2*s23-p2*p5*s12*s23-p5*s12*s15*s23+s12sq*s15*s23+p5*s12*s
     -   23**2+p5*s12*s23*s34+msq*s34*(s12*(-s15+s23)+p1*s34+s23*(p5-
     -   2*(s15-s23)+s34)+(s15-2*s23-s34)*s45+p2*(-p5+2*(s15-s23)+s45
     -   )))+((p2-s15)**2-(p2+s15)*s34)*s45sq
       x(3,1) = s12sq*s15**2+p5*s12*s15*s23-p5*s12*s23**2+s12sq*s23**2
     -   +p1*p5*s23*s34+p1**2*s34**2+p2**2*(p5-s45)**2-s12*s15**2*s45
     -   +s12*s15*s23*s45+p1*s15*s34*s45+s12*s15*s34*s45+s12*s23*s34*
     -   s45-p1*s34**2*s45-2*(s12sq*s15*s23+p1*s12*s15*s34-p1*s12*s23
     -   *s34+p5*s12*s23*s34+p1*s23*s34*s45-msq*(s15-s23-s34)*(-(p5*s
     -   23)+s12*(-s15+s23)+s34*(p1+s23-s45)+s15*s45))+p2*(-((p5-s45)
     -   *(p5*s23+2*(s12*(s15-s23)+p1*s34)-(s15+s34)*s45))+2*msq*(-2*
     -   (s15-s23)*s34+p5*(s15-s23+s34)-(s15-s23+s34)*s45))
       x(4,1) = s12*s23*(-(p5*s23)+s12*(-s15+s23)+p1*s34)+s23*(-2*p1*s
     -   34+s12*(s15+s34))*s45+p2**2*s45*(-p5+s45)+p2*(p5*s23*(s12+s4
     -   5)+s45*(s12*(s15-2*s23)+p1*s34-(s15+s34)*s45))+2*msq*(p2**2*
     -   (p5-s45)+p2*(-2*p5*s23+s12*(-s15+s23)+(-p1+s23)*s34+(s15+s23
     -   +s34)*s45)+s23*(p5*s23+s12*(s15-s23-2*s34)-s15*s45+s34*(p1-s
     -   23+s45)))
       x(2,2) = p1**2*s34**2-4*msq*s34*(p1*s34+(-p2+s12+s23-s45)*(p5-s
     -   15+s23-s45))+(p2*p5-s12*s15-p5*s23+s12*s23-p2*s45+s15*s45)**
     -   2+2*p1*s34*(-(p2*p5)-s12*s15+p5*s23+s12*s23+(p2+s15-2*s23)*s
     -   45)
       x(3,2) = p1**2*s34**2+p1*s34*(p5*s23-2*(p2*p5+s12*s15-s12*s23)+
     -   (s15+2*(p2-s23))*s45)+(s12*(s15-s23)+p2*(-p5+s45))*(s12*(s15
     -   -s23)+p5*s23-s15*s45+p2*(-p5+s45))-2*msq*(s12*s15**2-p5**2*s
     -   23+s12*s23**2-p1*s15*s34+p1*s23*s34-s15*s23*s34+s23**2*s34-2
     -   *(s12*s15*s23+s12*s15*s34-s12*s23*s34-p1*s34**2)+p2*(p5-s15+
     -   s23-s45)*(p5-2*s34-s45)+((s12-s15)*(s15-s23)+(p1+s15-2*(s12+
     -   s23))*s34)*s45+p5*(-(p1*s34)+s23*(s15-s23+s34)+s12*(-s15+s23
     -   +2*s34)+(s15+s23-s34)*s45)-(s15-s34)*s45sq)
       x(4,2) = (s12*s23-p2*s45)*(-(p5*s23)+s12*(-s15+s23)+p2*(p5-s45)
     -   +s15*s45)+p1*s34*(s12*s23+(p2-2*s23)*s45)+2*msq*(s12sq*(s15-
     -   s23)+s23*(p5*s23+(p1-s23)*s34)+p2**2*(p5-s45)-((p5+s15)*s23-
     -   (p1+2*s23)*s34)*s45+s12*(s23*(p5+s15-s23-s34)-p1*s34+(-2*s15
     -   +s23+s34)*s45)+p2*(-(p1*s34)+s15*s45+p5*(-s12-2*s23+s45)+s12
     -   *(-s15+s23+s45)+(s23-s45)*(s34+s45))+(s15-s34)*s45sq)
       x(3,3) = (-(s12*s15)+s12*s23+p1*s34)**2+p2**2*(p5-s45)**2+2*p2*
     -   (s12*s15-s12*s23+p1*s34)*(-p5+s45)-4*msq*(p2*(p5-s15+s23-s45
     -   )*(p5-s34-s45)+(s15-s23-s34)*(p1*(p5-s34-s45)+s12*(-p5+s15-s
     -   23+s45)))
       x(4,3) = (s12*(-s15+s23)+p2*(p5-s45))*(s12*s23-p2*s45)+p1*s34*(
     -   s12*s23+p2*s45)-2*msq*(-(p1**2*s34)+p2**2*(-p5+s45)+p1*(-(p5
     -   *s23)+s23*s34+s12*(s15-s23+s34)+p2*(p5+s34-s45)+(-s15+2*s23+
     -   s34)*s45)+s12*(p5*s23+s12*(-s15+s23)+(2*s23+s34)*(s23-s45)+s
     -   15*(-2*s23+s45))+p2*(p5*(s12+s23-2*s45)+s12*(s15-s23-s45)+s1
     -   5*s45-(s23-s45)*(s34+2*s45)))
       x(4,4) = (s12*s23-p2*s45)**2-4*msq*(p1*(p2-s23)*(s12-s45)+(p2-s
     -   12-s23+s45)*(-(s12*s23)+p2*s45))
c          
       x(1,0)= -(s12sq*(s15-s23)**2)-p5**2*s23**2-2*p1*p5*s23*s34+p5*s
     -   23**2*s34-p1**2*s34**2+p1*s23*s34**2-p2**2*(p5-s45)**2+(2*p5
     -   *s23*(s15-s34)+s34*(p1*(4*s23-2*(s15-s34))+s23*(-s15+s34)))*
     -   s45+s12*((2*p1-s23)*(s15-s23)*s34+2*(p5*s23*(-s15+s23+2*s34)
     -   +(s15**2-s23*s34-s15*(s23+s34))*s45))-(s15-s34)**2*s45sq+p2*
     -   (p5*(-(s23*s34)+2*(s12*(s15-s23)+p5*s23+p1*s34))-((2*(p1-s15
     -   )+s23)*s34+2*(s12*(s15-s23)+p5*(s15+s23+s34)))*s45+2*(s15+s3
     -   4)*s45sq)
       x(2,0)= -(s12sq*(s15-s23)**2)-p5**2*s23**2-2*p1*p5*s23*s34+p5*s
     -   23**2*s34-p1**2*s34**2+p1*s23*s34**2-p2**2*(p5-s45)**2+(2*p5
     -   *s15*s23-(2*p1*(s15-2*s23)+(p5+s15)*s23)*s34+p1*s34**2)*s45-
     -   s12*(s15*s23*s34-s23**2*s34+s15*s34*s45+s23*s34*s45-2*(p1*s1
     -   5*s34-p1*s23*s34+p5*s23*(-s15+s23+s34)+s15**2*s45-s15*s23*s4
     -   5))+s15*(-s15+s34)*s45sq+p2*(p5*(-(s23*s34)+2*(s12*(s15-s23)
     -   +p5*s23+p1*s34))-(2*(s12*(s15-s23)+p5*(s15+s23))+(p5+2*(p1-s
     -   15)+s23)*s34)*s45+(2*s15+s34)*s45sq)
       x(3,0)= -(s12sq*(s15-s23)**2)-p2**2*(p5-s45)**2-s12*(p5*s23*(s1
     -   5-s23-2*s34)-2*p1*(s15-s23)*s34+s15*s23*s34-s23**2*s34-s15**
     -   2*s45+s15*s23*s45+s15*s34*s45+s23*s34*s45)-p1*s34*(p5*s23+(p
     -   1-s23)*s34+s15*s45-(2*s23+s34)*s45)+p2*(p5*(p5*s23-s23*s34+2
     -   *(s12*(s15-s23)+p1*s34))-(2*s12*(s15-s23)+(2*(p1-s15)+s23)*s
     -   34+p5*(s15+s23+s34))*s45+(s15+s34)*s45sq)
       x(4,0)= p2**2*(p5-s45)*s45+s23*(s12sq*(s15-s23)+2*p1*s34*s45+s1
     -   2*(p5*s23-p1*s34+s23*s34-(s15+s34)*s45))+p2*(-(p5*s23*(s12+s
     -   45))+s45*(-(s12*(s15-2*s23))-(p1+s23)*s34+(s15+s34)*s45))
       endif 
       else 
       If(abs(p5).lt.eps) then 
c  Configuration                 (p1sp2s0p4s0)
       d= 2*(16*msq*(p1p4**2*p2p3**2+p3p4*(2*p1*p2p3*p2p4+p1p2**2*p3p4
     -   -p1*p2*p3p4)-p1*p2p3**2*p4+p1p3**2*(p2p4**2-p2*p4)-2*(p1p4*(
     -   p1p3*p2p3*p2p4-p1p3*p2*p3p4+p1p2*p2p3*p3p4)+p1p2*p1p3*(p2p4*
     -   p3p4-p2p3*p4)))+(p1*p2*p4+s12*s15*s23-p1*s23*s34-p2*s15*s45)
     -   *(p4*s12-s34*s45))
       x(1,1) = (p1-s12)**2*(p4-s34)**2-4*msq*(p2**2*p4+s23*(p4*(s15-s
     -   34)+s34*(-s15+s23+s34))+p2*(p4**2+(s15-s23)*s34-p4*(s15+s23+
     -   s34)))+(-4*p2*p4+(-s15+s23+s34)**2)*(s12-s45)**2+2*((p1-s12)
     -   *(-2*p2*p4+p4*(s15+s23-s34)+s34*(-s15+s23+s34))*(s12-s45)-(p
     -   1-s12)*(p2-s23)*(p4-s34)*s45+(p2*(2*p4-s15+s23-s34)-s23*(-s1
     -   5+s23+s34))*s45*(-s12+s45))+(p2-s23)**2*s45sq
       x(2,1) = -(s12sq*(-(s15-s23)**2+p4*(s15+s23)))+p1**2*(p4-s34)**
     -   2+p1*s12*(-(p4*(4*p2+p4-2*(s15+s23)))+(p4-2*(s15-s23))*s34)+
     -   s12*(s15*(p4-2*(s15-s23))+p2*(p4+2*(s15-s23))+(s15+s23)*s34)
     -   *s45+p1*((2*s15-4*s23-s34)*s34+2*p2*(p4+s34)+p4*(-2*s15+s34)
     -   )*s45-2*msq*(-(p4**2*s12)+p4*s12*s15+p4*s12*s23+p1*(p4-s34)*
     -   *2+p4*s12*s34-s12*s15*s34-p4*s23*s34+s12*s23*s34+s23*s34**2+
     -   2*(p2**2*p4+p4*s15*s23-s15*s23*s34+s23**2*s34)-(p4*s15-(p4+s
     -   15-2*s23)*s34+s34**2)*s45+p2*(p4**2-p4*(2*(s12+s15+s23)+s34-
     -   s45)+s34*(2*(s15-s23)+s45)))+((p2-s15)**2-(p2+s15)*s34)*s45s
     -   q
       x(3,1) = -(s12sq*(-(s15-s23)**2+p4*(s15+s23)))+p1**2*s34*(-p4+s
     -   34)+s12*(-s15**2+p2*(p4+2*(s15-s23))+s23*s34+s15*(s23+s34))*
     -   s45+2*msq*(p4*s12*s15-s12*s15**2+p4*s12*s23-s12*s23**2-2*(p2
     -   **2*p4-s12*s15*s23)-p4*s12*s34+s12*s15*s34-s12*s23*s34+s15*s
     -   23*s34-s23**2*s34-s23*s34**2+p1*(2*p2*p4-p4*(s15+s23)+(p4+s1
     -   5-s23)*s34-s34**2)+(s15-s34)*(s15-s23-s34)*s45+p2*(-2*(s15-s
     -   23)*s34+p4*(s15+s23+s34+2*(s12-s45))-(s15-s23+s34)*s45))+p1*
     -   (2*s12*(-s15+s23)*s34+p4*s12*(s15+s23+s34)+(s15-2*s23-s34)*s
     -   34*s45+p2*(2*s34*s45+p4*(-4*s12+s45)))-p2*(-p2+s15+s34)*s45s
     -   q
       x(4,1) = (s12*s23-p2*s45)*(-(s12*(p4+s15-s23))+(-p2+s15+s34)*s4
     -   5)+p1*(s12*s23*(p4+s34)+p2*(p4+s34)*s45-2*(p2*p4*s12+s23*s34
     -   *s45))+2*msq*(p1*(p2-s23)*(p4-s34)-p2**2*(p4+s45)+s23*(p4*s1
     -   2+s12*(s15-s23-2*s34)-s23*s34-s15*s45+s34*s45)+p2*(s12*(-s15
     -   +s23)+s23*s34+p4*(s12+s23-2*s45)+(s15+s23+s34)*s45))
       x(2,2) = p1**2*(p4-s34)**2+(s12*(s15-s23)+(p2-s15)*s45)**2+2*p1
     -   *(-2*p2*p4*s12+p4*s12*(s15+s23)-p4*s15*s45+p2*(p4+s34)*s45+s
     -   34*(-(s12*s15)+s12*s23+s15*s45-2*s23*s45))-4*msq*(p2**2*p4+p
     -   1*(p4-s34)**2+(s12+s23-s45)*(p4*(s12+s15)-s34*(s15-s23+s45))
     -   +p2*(-(p4*(2*s12+s15+s23-s45))+s34*(s15-s23+s45)))
       x(3,2) = p1**2*s34*(-p4+s34)+(s12*(s15-s23)+p2*s45)*(s12*(s15-s
     -   23)+(p2-s15)*s45)+p1*(p4*s12*(s15+s23)+p2*(2*s34*s45+p4*(-4*
     -   s12+s45))+s34*(s15*s45-2*(s12*s15-s12*s23+s23*s45)))-2*msq*(
     -   p4*s12*s15+s12*s15**2+p4*s12*s23+s12*s23**2-s15*s23*s34+s23*
     -   *2*s34+2*(p2**2*p4+p4*s12sq-s12*s15*s23-s12*s15*s34+s12*s23*
     -   s34)-(p4*s12+(-s12+s15)*(s15-s23)+(-s15+2*(s12+s23))*s34)*s4
     -   5+p2*(-(p4*(4*s12+s15+s23-s45))+(s15-s23+s45)*(2*s34+s45))+p
     -   1*(-2*p2*p4+s34*(-s15+s23+2*s34+s45)+p4*(s15+s23-2*(s12+s34)
     -   +s45))-(s15-s34)*s45sq)
       x(4,2) = (s12*s23-p2*s45)*(s12*(-s15+s23)+(-p2+s15)*s45)+p1*(s1
     -   2*s23*(p4+s34)+p2*(p4+s34)*s45-2*(p2*p4*s12+s23*s34*s45))+2*
     -   msq*(p1*(p4-s34)*(p2+s12-s23-s45)-(p2-s12-s23+s45)*(p2*p4-p4
     -   *s12+s12*s15-s12*s23-s23*s34+(p2-s15+s34)*s45))
       x(3,3) = p1**2*s34**2+(s12*(s15-s23)+p2*s45)**2+2*p1*(s12*(-s15
     -   +s23)*s34+p2*(-2*p4*s12+s34*s45))-4*msq*(p1**2*p4+p2**2*p4+s
     -   12*(p4*s12+(s15-s23-s34)*(s15-s23+s45))+p1*(-2*p4*(p2+s12)-(
     -   s15-s23-s34)*(s34+s45))+p2*(-2*p4*s12+(s15-s23+s45)*(s34+s45
     -   )))
       x(4,3) = (s12*s23-p2*s45)*(s12*(-s15+s23)-p2*s45)+p1*(-2*p2*p4*
     -   s12+s12*s23*s34+p2*s34*s45)-2*msq*(p1**2*(p4-s34)+p2**2*(p4+
     -   s45)+s12*(p4*s12+s12*(-s15+s23)+s23*(-2*(s15-s23)+s34)+(s15-
     -   2*s23-s34)*s45)+p1*(-2*(p2*p4+p4*s12)+s12*s15-s12*s23+p2*s34
     -   +s12*s34+s23*s34-(p2+s15-2*s23-s34)*s45)+p2*(-2*p4*s12+s12*(
     -   s15-s23-s45)+s15*s45-(s23-s45)*(s34+2*s45)))
       x(4,4) = (s12*s23-p2*s45)**2-4*msq*(p1*(p2-s23)*(s12-s45)+(p2-s
     -   12-s23+s45)*(-(s12*s23)+p2*s45))
c          
       x(1,0)= -(p1**2*(p4-s34)**2)+s12*(-(s12*(p4-s15)**2)+p2*p4*(p4-
     -   s15-s23)+2*(s12*s15+p4*(s12+s15))*s23-s12*s23**2-(p4+s15-s23
     -   )*s23*s34)-(-(p2**2*p4)-2*s12*s15*(-p4+s15-s23)+(s15*s23-2*(
     -   p4*s12-s12*(s15+s23)))*s34-s23*s34**2+p2*(2*s12*(s15-s23)+(-
     -   2*s15+s23)*s34+p4*(2*s12+s15+s34)))*s45+p1*(p2*p4*(p4+4*s12-
     -   s34)+2*(p4**2*s12-p2*(p4+s34)*s45)-p4*(s23*s34+2*(s12*(s15+s
     -   23+s34)-s15*s45+s34*s45))+s34*(s23*s34+2*(s12*(s15-s23)+(-s1
     -   5+2*s23+s34)*s45)))-(p2**2+(s15-s34)**2-2*p2*(s15+s34))*s45s
     -   q
       x(2,0)= -(p1**2*(p4-s34)**2)+s12*(-(p2*p4*(s15+s23))+p4*(2*s15*
     -   s23+s12*(s15+s23))-(s15-s23)*(s12*(s15-s23)+s23*s34))-(-(p2*
     -   *2*p4)+s12*s15*(p4-2*(s15-s23))+(s15*s23+s12*(s15+s23))*s34+
     -   p2*(p4*(s12+s15)+s23*s34+2*(s12*s15-s12*s23-s15*s34)))*s45+p
     -   1*(p4**2*s12+p2*p4*(p4+4*s12-s34)-2*p2*(p4+s34)*s45+s34*(2*s
     -   12*(s15-s23)+s23*s34+(-2*s15+4*s23+s34)*s45)-p4*(s12*(2*(s15
     -   +s23)+s34)-2*s15*s45+s34*(s23+s45)))-((p2-s15)**2-(p2+s15)*s
     -   34)*s45sq
       x(3,0)= -(p1**2*s34*(-p4+s34))+p2**2*(p4-s45)*s45-s12*(-(p4*s12
     -   *(s15+s23))+(s15-s23)*(s12*(s15-s23)+s23*s34)+(-s15**2+s23*s
     -   34+s15*(s23+s34))*s45)-p2*(p4*s12*(s15+s23+s45)-s45*(-(s23*s
     -   34)-2*(s12*s15-s12*s23-s15*s34)+(s15+s34)*s45))-p1*(p4*s12*(
     -   s15+s23+s34)-s34*(2*s12*(s15-s23)+s23*s34+(-s15+2*s23+s34)*s
     -   45)+p2*(2*s34*s45+p4*(-4*s12+s34+s45)))
       x(4,0)= p1*s12*(2*p2*p4-s23*(p4+s34))-p1*(-2*s23*s34+p2*(p4+s34
     -   ))*s45+(-(s12*s23)+p2*s45)*(p2*p4-p4*s12-s12*s15+s12*s23-s23
     -   *s34+(-p2+s15+s34)*s45)
       else 
c  Configuration                 (p1sp2s0p4sp5s)
       d= 2*(16*msq*(p1p4**2*p2p3**2+p3p4*(2*p1*p2p3*p2p4+p1p2**2*p3p4
     -   -p1*p2*p3p4)-p1*p2p3**2*p4+p1p3**2*(p2p4**2-p2*p4)-2*(p1p4*(
     -   p1p3*p2p3*p2p4-p1p3*p2*p3p4+p1p2*p2p3*p3p4)+p1p2*p1p3*(p2p4*
     -   p3p4-p2p3*p4)))+p1*(p2*p4-s23*s34)*(p4*s12-s34*s45)+(s12*s23
     -   -p2*s45)*(-(p2*p4*p5)+p4*s12*s15+p5*s23*s34-s15*s34*s45))
       x(1,1) = (p1-s12)**2*(p4-s34)**2-4*msq*(p2**2*p4+s23*(p4*(s15-s
     -   34)+s34*(-s15+s23+s34))+p2*(p4**2+(s15-s23)*s34-p4*(s15+s23+
     -   s34)))+(p2-s23)**2*(p5-s45)**2+(-4*p2*p4+(-s15+s23+s34)**2)*
     -   (s12-s45)**2+2*((p1-s12)*(p2-s23)*(p4-s34)*(p5-s45)+(p1-s12)
     -   *(-2*p2*p4+p4*(s15+s23-s34)+s34*(-s15+s23+s34))*(s12-s45)+(p
     -   2*(-2*p4+s15-s23+s34)+s23*(-s15+s23+s34))*(p5-s45)*(-s12+s45
     -   ))
       x(2,1) = p2**2*p5**2+p2*p4*p5*s12-p4*s12sq*s15+s12sq*s15**2+p4*
     -   p5*s12*s23-p4*s12sq*s23+p5**2*s23**2+s12sq*s23**2+p1**2*(p4-
     -   s34)**2-(2*p2**2*p5+s15*(-(s12*(p4-2*(s15-s23)))+2*p5*s23)-(
     -   p5*s23+s12*(s15+s23))*s34-p2*(p4*(-2*p5+s12)+2*s12*(s15-s23)
     -   +p5*(2*(s15+s23)+s34)))*s45+p1*(-(p4**2*s12)+s34*(-2*(s12*(s
     -   15-s23)-p5*s23)-(-2*s15+4*s23+s34)*s45)+2*p2*(s34*(-p5+s45)+
     -   p4*(p5-2*s12+s45))+p4*(s12*(2*(s15+s23)+s34)+s34*s45-2*(p5*s
     -   23+s15*s45)))-2*(p2*p5*s12*s15+p2*p5**2*s23-p2*p5*s12*s23-p5
     -   *s12*s15*s23+s12sq*s15*s23+p5*s12*s23**2+p5*s12*s23*s34+msq*
     -   (-(p4**2*s12)+p4*s12*s15-p4*p5*s23+p4*s12*s23+p1*(p4-s34)**2
     -   +p4*s12*s34-s12*s15*s34-p4*s23*s34+p5*s23*s34+s12*s23*s34+s2
     -   3*s34**2+2*(p2**2*p4+p4*s15*s23-s15*s23*s34+s23**2*s34)-(p4*
     -   s15-(p4+s15-2*s23)*s34+s34**2)*s45+p2*(p4**2+s34*(-p5+2*(s15
     -   -s23)+s45)+p4*(p5-2*(s12+s15+s23)-s34+s45))))+((p2-s15)**2-(
     -   p2+s15)*s34)*s45sq
       x(3,1) = p2**2*p5**2+p2*p4*p5*s12-p4*s12sq*s15+s12sq*s15**2-p2*
     -   p5**2*s23-p4*s12sq*s23+p5*s12*s15*s23-p5*s12*s23**2+s12sq*s2
     -   3**2+p1**2*s34*(-p4+s34)+(-2*p2**2*p5+s12*(-s15**2+s23*s34+s
     -   15*(s23+s34))+p2*(s12*(p4+2*(s15-s23))+p5*(s15+s23+s34)))*s4
     -   5-2*(p2*p5*s12*s15-p2*p5*s12*s23+s12sq*s15*s23+p5*s12*s23*s3
     -   4+msq*(-(p4*s12*s15)+s12*s15**2-p4*s12*s23+p5*s15*s23-p5*s23
     -   **2+s12*s23**2+2*(p2**2*p4-s12*s15*s23)+p4*s12*s34-s12*s15*s
     -   34-p5*s23*s34+s12*s23*s34-s15*s23*s34+s23**2*s34+s23*s34**2+
     -   p1*(-2*p2*p4+p4*(s15+s23-s34)+s34*(-s15+s23+s34))-p2*(2*(-s1
     -   5+s23)*s34+p5*(s15-s23+s34)+p4*(-2*(p5-s12)+s15+s23+s34))-(s
     -   15-s34)*(s15-s23-s34)*s45+p2*(2*p4+s15-s23+s34)*s45))+p1*(p4
     -   *s12*(s15+s23+s34)+s34*(p5*s23-2*(s12*s15-s12*s23)+(s15-2*s2
     -   3-s34)*s45)+p2*(2*s34*(-p5+s45)+p4*(p5-4*s12+s45)))+p2*(p2-s
     -   15-s34)*s45sq
       x(4,1) = (s12*s23-p2*s45)*(-(s12*(p4+s15))+(-p5+s12)*s23+p2*(p5
     -   -s45)+(s15+s34)*s45)+p1*(s12*s23*(p4+s34)+p2*(p4+s34)*s45-2*
     -   (p2*p4*s12+s23*s34*s45))+2*msq*(p1*(p2-s23)*(p4-s34)-p2**2*(
     -   p4-p5+s45)+s23*(p4*s12+p5*s23+s12*(s15-s23-2*s34)-s23*s34-s1
     -   5*s45+s34*s45)+p2*(-(s12*s15)-2*p5*s23+s12*s23+s23*s34+p4*(s
     -   12+s23-2*s45)+(s15+s23+s34)*s45))
       x(2,2) = p1**2*(p4-s34)**2+(p2*p5-s12*s15-p5*s23+s12*s23-p2*s45
     -   +s15*s45)**2+2*p1*(p4*(-(p5*s23)+s12*(s15+s23)-s15*s45)+s34*
     -   (-(s12*s15)+p5*s23+s12*s23+s15*s45-2*s23*s45)+p2*(s34*(-p5+s
     -   45)+p4*(p5-2*s12+s45)))-4*msq*(p2**2*p4+p1*(p4-s34)**2+(p4*(
     -   -p5+s12+s15)+s34*(p5-s15+s23-s45))*(s12+s23-s45)+p2*(p4*(p5-
     -   2*s12-s15-s23+s45)+s34*(-p5+s15-s23+s45)))
       x(3,2) = p1**2*s34*(-p4+s34)+(s12*(s15-s23)+p2*(-p5+s45))*(s12*
     -   (s15-s23)+p5*s23-s15*s45+p2*(-p5+s45))+p1*(p4*s12*(s15+s23)+
     -   p2*(2*s34*(-p5+s45)+p4*(p5-4*s12+s45))+s34*(p5*s23+s15*s45-2
     -   *(s12*s15-s12*s23+s23*s45)))-2*msq*(-(p4*p5*s12)+p4*s12*s15-
     -   p5*s12*s15+s12*s15**2-p5**2*s23+p4*s12*s23+p5*s12*s23+p5*s15
     -   *s23-p5*s23**2+s12*s23**2+p5*s23*s34-s15*s23*s34+s23**2*s34+
     -   2*(p2**2*p4+p4*s12sq-s12*s15*s23+p5*s12*s34-s12*s15*s34+s12*
     -   s23*s34)-(p4*s12+(-s12+s15)*(s15-s23)-p5*(s15+s23-s34)+(-s15
     -   +2*(s12+s23))*s34)*s45+p2*((p5-s15+s23-s45)*(p5-2*s34-s45)+p
     -   4*(p5-4*s12-s15-s23+s45))+p1*(-2*p2*p4+s34*(-p5-s15+s23+2*s3
     -   4+s45)+p4*(p5+s15+s23-2*(s12+s34)+s45))-(s15-s34)*s45sq)
       x(4,2) = (s12*s23-p2*s45)*(-(s12*s15)-p5*s23+s12*s23+p2*(p5-s45
     -   )+s15*s45)+p1*(s12*s23*(p4+s34)+p2*(p4+s34)*s45-2*(p2*p4*s12
     -   +s23*s34*s45))+2*msq*(p1*(p4-s34)*(p2+s12-s23-s45)-(p2-s12-s
     -   23+s45)*(p2*p4-p2*p5-p4*s12+s12*s15+p5*s23-s12*s23-s23*s34+(
     -   p2-s15+s34)*s45))
       x(3,3) = p1**2*s34**2-2*p1*(s12*(s15-s23)*s34+p2*(2*p4*s12+s34*
     -   (p5-s45)))+(s12*(s15-s23)+p2*(-p5+s45))**2-4*msq*(p1**2*p4+p
     -   2**2*p4+p2*(-2*p4*s12+(p5-s15+s23-s45)*(p5-s34-s45))+s12*(p4
     -   *s12+(s15-s23-s34)*(-p5+s15-s23+s45))+p1*(-2*p4*(p2+s12)-(s1
     -   5-s23-s34)*(-p5+s34+s45)))
       x(4,3) = (s12*(-s15+s23)+p2*(p5-s45))*(s12*s23-p2*s45)+p1*(-2*p
     -   2*p4*s12+s12*s23*s34+p2*s34*s45)-2*msq*(p1**2*(p4-s34)+p2**2
     -   *(p4-p5+s45)+s12*(p4*s12+s12*(-s15+s23)+s23*(p5-2*(s15-s23)+
     -   s34)+(s15-2*s23-s34)*s45)+p1*(s12*s15-p5*s23-s12*s23+s12*s34
     -   +s23*s34+p2*(-2*p4+p5+s34-s45)-s15*s45+s34*s45-2*(p4*s12-s23
     -   *s45))+p2*(s12*s15-s12*s23-s23*s34+p5*(s12+s23-2*s45)-s12*s4
     -   5+s15*s45+s34*s45-2*(p4*s12+s23*s45-s45sq)))
       x(4,4) = (s12*s23-p2*s45)**2-4*msq*(p1*(p2-s23)*(s12-s45)+(p2-s
     -   12-s23+s45)*(-(s12*s23)+p2*s45))
c          
       x(1,0)= -(p4**2*s12sq)-s12sq*s15**2-p5**2*s23**2-s12sq*s23**2+2
     -   *(p4*s12sq*s15-p4*p5*s12*s23+p4*s12sq*s23+p4*s12*s15*s23-p5*
     -   s12*s15*s23+s12sq*s15*s23+p5*s12*s23**2)-p1**2*(p4-s34)**2-p
     -   4*s12*s23*s34+4*p5*s12*s23*s34-s12*s15*s23*s34+p5*s23**2*s34
     -   +s12*s23**2*s34-(s23*(2*p5-s34)*(-s15+s34)-2*(s12*s15*(s15-s
     -   23)-p4*s12*(s15-s34)-s12*(s15+s23)*s34))*s45+p2**2*(-(p5-s45
     -   )**2+p4*(p5+s45))+p2*(p4**2*s12+2*p5**2*s23-p4*(s12*(s15+s23
     -   )+p5*(2*s12+s23-4*s45)+(2*s12+s15+s34)*s45)+s45*(-(s23*s34)-
     -   2*(s12*s15-s12*s23-s15*s34-(s15+s34)*s45))+p5*(-(s23*s34)+2*
     -   (s12*s15-s12*s23-(s15+s23+s34)*s45)))+p1*(2*p4**2*s12-p4*(s2
     -   3*s34-2*(p5*s23-s12*(s15+s23+s34)+s15*s45-s34*s45))+s34*(s23
     -   *s34+2*(s12*(s15-s23)-p5*s23+(-s15+2*s23+s34)*s45))+p2*(p4**
     -   2+2*s34*(p5-s45)-p4*(-4*s12+s34+2*(p5+s45))))-(s15-s34)**2*s
     -   45sq
       x(2,0)= p4*s12sq*s15-s12sq*s15**2-p4*p5*s12*s23+p4*s12sq*s23-p5
     -   **2*s23**2-s12sq*s23**2-p1**2*(p4-s34)**2-s12*s15*s23*s34+p5
     -   *s23**2*s34+s12*s23**2*s34+2*(p4*s12*s15*s23-p5*s12*s15*s23+
     -   s12sq*s15*s23+p5*s12*s23**2+p5*s12*s23*s34)-(s15*(s12*(p4-2*
     -   (s15-s23))-2*p5*s23)+(s12*s15+(p5+s12+s15)*s23)*s34)*s45-p2*
     -   *2*((p5-s45)**2-p4*(p5+s45))-p1*(-(p4**2*s12)-s34*(2*(s12*(s
     -   15-s23)-p5*s23)+s23*s34+(-2*s15+4*s23+s34)*s45)+p4*(s12*(2*(
     -   s15+s23)+s34)+s34*(s23+s45)-2*(p5*s23+s15*s45))+p2*(-p4**2+2
     -   *s34*(-p5+s45)+p4*(-4*s12+s34+2*(p5+s45))))-s15*(s15-s34)*s4
     -   5sq-p2*(p5*(-2*(s12*s15+p5*s23-s12*s23)+s23*s34)+(2*((p5+s12
     -   )*s15+(p5-s12)*s23)+(p5-2*s15+s23)*s34)*s45+p4*(p5*(s12+s23-
     -   2*s45)+s15*s45+s12*(s15+s23+s45))-(2*s15+s34)*s45sq)
       x(3,0)= p1**2*(p4-s34)*s34-p2*(p4*s12*(p5+s15+s23)+p5*(-(p5*s23
     -   )-2*(s12*s15-s12*s23)+s23*s34))-p2*(p4*s12+s23*s34+p5*(s15+s
     -   23+s34)+2*(s12*s15-s12*s23-s15*s34))*s45+s12*(-(s12*(s15-s23
     -   )**2)+p5*s23*(-s15+s23)+p4*s12*(s15+s23)+s23*(2*p5-s15+s23)*
     -   s34-(-s15**2+s23*s34+s15*(s23+s34))*s45)+p2**2*(-(p5-s45)**2
     -   +p4*(p5+s45))-p1*(p4*s12*(s15+s23+s34)+s34*(p5*s23-2*(s12*s1
     -   5-s12*s23)-s23*s34+(s15-2*s23-s34)*s45)+p2*(2*s34*(-p5+s45)+
     -   p4*(p5-4*s12+s34+s45)))+p2*(s15+s34)*s45sq
       x(4,0)= p1*s12*(2*p2*p4-s23*(p4+s34))-p1*(-2*s23*s34+p2*(p4+s34
     -   ))*s45+(-(s12*s23)+p2*s45)*(-(p4*s12)-s12*s15-p5*s23+s12*s23
     -   -s23*s34+p2*(p4+p5-s45)+(s15+s34)*s45)
       endif 
       endif 
       else 
       If(abs(p4).lt.eps) then 
       If(abs(p5).lt.eps) then 
c  Configuration                 (p1sp2sp3s00)
       d= 32*msq*(p1p3**2*p2p4**2-p1*p2p4**2*p3+p1p4**2*(p2p3**2-p2*p3
     -   )+p3p4*(2*p1*p2p3*p2p4+p1p2**2*p3p4-p1*p2*p3p4)-2*(p1p2*p1p3
     -   *p2p4*p3p4+p1p4*(p1p3*p2p3*p2p4-p1p2*p2p4*p3-p1p3*p2*p3p4+p1
     -   p2*p2p3*p3p4)))+2*((s12*s15-p1*s34)*(p3*s15-s23*s34)*s45+p2*
     -   s15*s34*s45sq)
       x(1,1) = s12sq*(s15-s23)**2+p1**2*(p3-s34)**2+4*msq*(p2*(s15-s2
     -   3)*(p3-s34)-(p3+s15-s23-s34)*(p3*s15-s23*s34))+2*(s12*(p2*(s
     -   15-s23)-s15*(2*p3+s15-s23)+(s15+s23)*s34)*s45+p1*(s12*(s15-s
     -   23)*(p3-s34)+(p3*(-p2+s15)+(p2+p3+s15-2*s23)*s34-s34**2)*s45
     -   ))+(p2**2+(s15-s34)**2-2*p2*(s15+s34))*s45sq
       x(2,1) = (s12*(s15-s23)+p1*(p3-s34))**2+(s12*(2*(p2*(s15-s23)-s
     -   15*(p3+s15-s23))+(s15+s23)*s34)+p1*(-2*p3*(p2-s15)+(p3+2*(p2
     -   +s15)-4*s23)*s34-s34**2))*s45-2*msq*(p3**2*s15+p3*s12*s15-p3
     -   *s12*s23+p1*(p3-s34)**2-p3*s15*s34-s12*s15*s34-p3*s23*s34+s1
     -   2*s23*s34+s23*s34**2+2*(p3*s15**2-p3*s15*s23-s15*s23*s34+s23
     -   **2*s34)+((s15-2*s23-s34)*s34+p3*(s15+s34))*s45-p2*(p3-s34)*
     -   (2*(s15-s23)+s45))+((p2-s15)**2-(p2+s15)*s34)*s45sq
       x(3,1) = (s12*(s15-s23)+p1*(p3-s34))**2+(s12*(2*p2*(s15-s23)-s1
     -   5*(2*p3+s15-s23)+(s15+s23)*s34)+p1*(p3*(-2*p2+s15)+(p3+s15+2
     -   *(p2-s23))*s34-s34**2))*s45-2*msq*(-2*p2*(s15-s23)*(p3-s34)+
     -   p1*(p3-s34)*(p3+s15-s23-s34)+p2*(-p3+s15-s23+s34)*s45+(p3+s1
     -   5-s23-s34)*(p3*s15+s12*s15-s12*s23-s23*s34-s15*s45+s34*s45))
     -   -p2*(-p2+s15+s34)*s45sq
       x(4,1) = (p1*p3-s12*s23)*(s12*(s15-s23)+p1*(p3-s34))+(p1*(p3*(-
     -   2*p2+s15)+(p2+p3-2*s23)*s34)+s12*(-2*p3*s15+p2*(s15-2*s23)+s
     -   23*(s15+s34)))*s45+2*msq*(-(p3**2*s15)+p3*s12*s15-p1*p3*(-p2
     -   +p3+2*s15-s23)+p3*s12*s23+p3*s15*s23+s12*s15*s23-s12*s23**2+
     -   p3*s23*s34-2*s12*s23*s34-s23**2*s34+p1*(-p2+p3+s23)*s34-p2**
     -   2*s45+(p3-s23)*(s15-s34)*s45+p2*(p3*s15-s12*s15-2*p3*s23+s12
     -   *s23+s23*s34+(p3+s15+s23+s34)*s45))-p2*(-p2+s15+s34)*s45sq
       x(2,2) = (s12*(s15-s23)+p1*(p3-s34))**2+2*((p2-s15)*(-(p1*p3)+s
     -   12*(s15-s23))+p1*(p2+s15-2*s23)*s34)*s45-4*msq*(p1*(p3-s34)*
     -   *2+(p3*(s12+s15)+p2*(-p3+s34)-s34*(s12+s23-s45))*(s15-s23+s4
     -   5))+(p2-s15)**2*s45sq
       x(3,2) = (s12*(s15-s23)+p1*(p3-s34))**2+((2*p2-s15)*(-(p1*p3)+s
     -   12*(s15-s23))+p1*(s15+2*(p2-s23))*s34)*s45-2*msq*(p1*(p3-s34
     -   )*(s15-s23+2*(p3-s34)-s45)-(s15-s23+s45)*(-(p3*s15)-s12*s15+
     -   s12*s23+s23*s34+2*(p2*p3-p3*s12-p2*s34+s12*s34)-(p2-s15+s34)
     -   *s45))+p2*(p2-s15)*s45sq
       x(4,2) = p1**2*p3*(p3-s34)+(s12*s23-p2*s45)*(s12*(-s15+s23)+(-p
     -   2+s15)*s45)+p1*(p3*(s12*(s15-2*s23)+(-2*p2+s15)*s45)+s34*(s1
     -   2*s23+(p2-2*s23)*s45))+2*msq*(-(p3*s12*s15)+s12sq*s15+2*p3*s
     -   12*s23-s12sq*s23+p3*s15*s23+s12*s15*s23-s12*s23**2-s12*s23*s
     -   34-s23**2*s34-p2**2*s45-(2*s12*s15+p3*(2*s12+s15)-s12*s23+s1
     -   5*s23-(s12+2*s23)*s34)*s45+p1*(-2*p3**2+p2*(p3-s34)+s34*(-s1
     -   2+s23+s45)+p3*(s12+s23-2*(s15-s34)+s45))+p2*(p3*(s15-2*(s23-
     -   s45))+s15*s45+s12*(-s15+s23+s45)+(s23-s45)*(s34+s45))+(s15-s
     -   34)*s45sq)
       x(3,3) = (s12*(s15-s23)+p1*(p3-s34))**2+2*p2*(s12*(s15-s23)+p1*
     -   (-p3+s34))*s45+4*msq*(p1*(p3+s15-s23-s34)*(-p3+s34+s45)-(s15
     -   -s23+s45)*(s12*(p3+s15-s23-s34)+p2*(-p3+s34+s45)))+p2**2*s45
     -   sq
       x(4,3) = (p1*p3-s12*s23)*(s12*(s15-s23)+p1*(p3-s34))-p2*(-(s12*
     -   s15)+2*(p1*p3+s12*s23)-p1*s34)*s45-2*msq*(p1**2*(p3-s34)+p2*
     -   *2*s45+s12*(-(s12*s15)+s12*s23-2*(s15*s23-s23**2)+s23*s34+p3
     -   *(s15-2*(s23-s45))+(s15-2*s23-s34)*s45)-p2*(p3*(s15-2*(s23-s
     -   45))-s15*s45+s12*(-s15+s23+s45)+(s23-s45)*(s34+2*s45))+p1*(s
     -   12*s15-s12*s23+s12*s34+s23*s34-s15*s45+s34*s45-p2*(p3-s34+s4
     -   5)+2*(p3**2+s23*s45)-p3*(s12-s15+2*(s23+s34+s45))))+p2**2*s4
     -   5sq
       x(4,4) = p1**2*p3**2+(s12*s23-p2*s45)**2-2*p1*p3*(s12*s23+p2*s4
     -   5)-4*msq*(p1**2*p3+s12*s23*(-p3+s12+s23)+p2**2*s45+s12*(p3-s
     -   23)*s45-p1*(-p3**2+s23*(s12-s45)+p2*(p3-s12+s45)+p3*(s12+s23
     -   +s45))+p2*((p3-s12)*s23-(p3+s12+s23)*s45+s45sq))
c          
       x(1,0)= -(p1**2*(p3-s34)**2)+p1*(p3-s34)*(p3*s15-2*(s12*s15-s12
     -   *s23)-s23*s34)-s12*(s15-s23)*(s12*(s15-s23)+s23*s34)-2*p1*(p
     -   3*(-p2+s15)+(p2+p3+s15-2*s23)*s34-s34**2)*s45-(2*s12*(p2-s15
     -   )*(s15-s23)+(s15*s23+p2*(-2*s15+s23)+2*s12*(s15+s23))*s34-s2
     -   3*s34**2)*s45+p3*s15*(-((p2-s15+s34)*s45)+s12*(s15-s23+4*s45
     -   ))-(p2**2+(s15-s34)**2-2*p2*(s15+s34))*s45sq
       x(2,0)= -(p1**2*(p3-s34)**2)+p1*(p3-s34)*(p3*s15-2*(s12*s15-s12
     -   *s23)-s23*s34)-s12*(s15-s23)*(s12*(s15-s23)+s23*s34)-(2*s12*
     -   (p2-s15)*(s15-s23)+((-2*p2+s12)*s15+(p2+s12+s15)*s23)*s34)*s
     -   45+p1*(2*p3*(p2-s15)-(p3+2*(p2+s15)-4*s23)*s34+s34**2)*s45+p
     -   3*s15*((-p2+s15)*s45+s12*(s15-s23+2*s45))-((p2-s15)**2-(p2+s
     -   15)*s34)*s45sq
       x(3,0)= -(p1**2*(p3-s34)**2)+s12*(s15-s23)*(p3*s15-s12*s15+s12*
     -   s23-s23*s34)+p1*(p3-s34)*(p3*s15-2*(s12*s15-s12*s23)-s23*s34
     -   )+p1*(p3*(2*p2-s15)-(p3+s15+2*(p2-s23))*s34+s34**2)*s45-(-(s
     -   12*s15*(2*p3+s15-s23))+s12*(s15+s23)*s34+p2*(p3*s15+s23*s34+
     -   2*(s12*s15-s12*s23-s15*s34)))*s45+p2*(-p2+s15+s34)*s45sq
       x(4,0)= -((p1*p3-s12*s23)*(-(p3*s15)+s12*s15-s12*s23+p1*(p3-s34
     -   )+s23*s34))-(p2*p3*s15+p2*s12*s15+s12*s15*s23-2*(p3*s12*s15+
     -   p2*s12*s23)+(p2+s12)*s23*s34+p1*(p3*(-2*p2+s15)+(p2+p3-2*s23
     -   )*s34))*s45+p2*(-p2+s15+s34)*s45sq
       else 
c  Configuration                 (p1sp2sp3s0p5s)
       d= 2*(16*msq*(p1p3**2*p2p4**2-p1*p2p4**2*p3+p1p4**2*(p2p3**2-p2
     -   *p3)+p3p4*(2*p1*p2p3*p2p4+p1p2**2*p3p4-p1*p2*p3p4)-2*(p1p2*p
     -   1p3*p2p4*p3p4+p1p4*(p1p3*p2p3*p2p4-p1p2*p2p4*p3-p1p3*p2*p3p4
     -   +p1p2*p2p3*p3p4)))+p1*(p3*s15-s23*s34)*(p3*p5-s34*s45)+(p5*s
     -   23-s15*s45)*(p2*p3*p5-p3*s12*s15+s12*s23*s34-p2*s34*s45))
       x(1,1) = (p1-s12)**2*(p3-s34)**2-4*msq*(-(p2*(s15-s23)*(p3-s34)
     -   )+(p3+s15-s23-s34)*(p3*s15-s23*s34))+(p2**2+(p3-s23)**2-2*p2
     -   *(p3+s23))*(p5-s45)**2+(p3+s15-s23-s34)**2*(s12-s45)**2+2*((
     -   p1-s12)*(-p3**2+p2*(p3-s34)+s23*s34+p3*(-2*s15+s23+s34))*(p5
     -   -s45)+(p1-s12)*(p3-s34)*(p3+s15-s23-s34)*(s12-s45)-(-((p3-s2
     -   3)*(p3+s15-s23-s34))+p2*(p3-s15+s23-s34))*(p5-s45)*(-s12+s45
     -   ))
       x(2,1) = p2**2*p5**2-p2*p3*p5**2+p3*p5*s12*s15+s12sq*s15**2-p3*
     -   p5**2*s23+p3*p5*s12*s23+p5**2*s23**2+s12sq*s23**2+p1**2*(p3-
     -   s34)**2-(2*p2**2*p5+s15*(-(p3*p5)+2*(p3*s12+s12*s15+p5*s23-s
     -   12*s23))-(p5*s23+s12*(s15+s23))*s34-p2*(p3*p5+2*s12*(s15-s23
     -   )+p5*(2*(s15+s23)+s34)))*s45-2*(p2*p5*s12*s15+p2*p5**2*s23-p
     -   2*p5*s12*s23-p5*s12*s15*s23+s12sq*s15*s23+p5*s12*s23**2+p5*s
     -   12*s23*s34+msq*(-(p3**2*p5)+p3**2*s15+p3*s12*s15+p3*p5*s23-p
     -   3*s12*s23+p1*(p3-s34)**2+p3*p5*s34-p3*s15*s34-s12*s15*s34-p3
     -   *s23*s34+p5*s23*s34+s12*s23*s34+s23*s34**2-2*(p3*p5*s15-p3*s
     -   15**2+p3*s15*s23+s15*s23*s34-s23**2*s34)+p2*(p3-s34)*(p5-2*(
     -   s15-s23)-s45)+((s15-2*s23-s34)*s34+p3*(s15+s34))*s45))+p1*(-
     -   (p3**2*p5)+2*p2*(p3-s34)*(p5-s45)+s34*(-2*(s12*(s15-s23)-p5*
     -   s23)-(-2*s15+4*s23+s34)*s45)+p3*(p5*(-4*s15+2*s23+s34)+s34*s
     -   45+2*(s12*s15-s12*s23+s15*s45)))+((p2-s15)**2-(p2+s15)*s34)*
     -   s45sq
       x(3,1) = p2**2*p5**2-p2*p3*p5**2+p3*p5*s12*s15+s12sq*s15**2-p2*
     -   p5**2*s23+p3*p5*s12*s23+p5*s12*s15*s23-p5*s12*s23**2+s12sq*s
     -   23**2+p1**2*(p3-s34)**2+(-2*p2**2*p5-s12*s15*(2*p3+s15-s23)+
     -   s12*(s15+s23)*s34+p2*(p3*p5+2*s12*(s15-s23)+p5*(s15+s23+s34)
     -   ))*s45+p1*(-(p3**2*p5)+2*p2*(p3-s34)*(p5-s45)+s34*(p5*s23-2*
     -   (s12*s15-s12*s23)+(s15-2*s23-s34)*s45)+p3*(2*s12*(s15-s23)+p
     -   5*(-2*s15+s23+s34)+(s15+s34)*s45))-2*(p2*p5*s12*s15-p2*p5*s1
     -   2*s23+s12sq*s15*s23+p5*s12*s23*s34+msq*(p1*(p3-s34)*(p3+s15-
     -   s23-s34)+(p3+s15-s23-s34)*(p3*(-p5+s15)+s12*(s15-s23)+p5*s23
     -   -s23*s34-s15*s45+s34*s45)+p2*(2*(s15-s23)*s34-p5*(s15-s23+s3
     -   4)+p3*(p5-2*(s15-s23)-s45)+(s15-s23+s34)*s45)))+p2*(p2-s15-s
     -   34)*s45sq
       x(4,1) = s12*s23*(p2*p5+p3*p5-s12*s15-p5*s23+s12*s23)+p1**2*p3*
     -   (p3-s34)+(-(p2**2*p5)+p2*(p3*p5+s12*s15+p5*s23-2*s12*s23)+s1
     -   2*(-2*p3*s15+s23*(s15+s34)))*s45+p1*(-(p3**2*p5)+s23*s34*(s1
     -   2-2*s45)+p2*(p3*(p5-2*s45)+s34*s45)+p3*(s12*s15+p5*s23-2*s12
     -   *s23+(s15+s34)*s45))+2*msq*(p3**2*p5-p3**2*s15+p3*s12*s15-p1
     -   *p3*(-p2+p3+2*s15-s23)+p3*s12*s23+p3*s15*s23+s12*s15*s23+p5*
     -   s23**2-s12*s23**2+p3*s23*s34-s23**2*s34+p1*(-p2+p3+s23)*s34-
     -   2*(p3*p5*s23+s12*s23*s34)+p2**2*(p5-s45)+(p3-s23)*(s15-s34)*
     -   s45+p2*(-(s12*s15)-2*p5*s23+s12*s23+s23*s34+(s15+s23+s34)*s4
     -   5+p3*(s15-2*(p5+s23)+s45)))+p2*(p2-s15-s34)*s45sq
       x(2,2) = p1**2*(p3-s34)**2-4*msq*(p1*(p3-s34)**2+(p3*(p5-s12-s1
     -   5)+p2*(p3-s34)+s34*(s12+s23-s45))*(p5-s15+s23-s45))+(p2*p5-s
     -   12*s15-p5*s23+s12*s23-p2*s45+s15*s45)**2+2*p1*(p2*(p3-s34)*(
     -   p5-s45)+p3*(-2*p5*s15+s12*s15+p5*s23-s12*s23+s15*s45)+s34*(-
     -   (s12*s15)+p5*s23+s12*s23+s15*s45-2*s23*s45))
       x(3,2) = p1**2*(p3-s34)**2+(s12*(s15-s23)+p2*(-p5+s45))*(s12*(s
     -   15-s23)+p5*s23-s15*s45+p2*(-p5+s45))-2*msq*(p1*(p3-s34)*(p5+
     -   s15-s23+2*(p3-s34)-s45)+(p5-s15+s23-s45)*(p3*(p5-2*s12-s15)-
     -   s12*s15-p5*s23+s12*s23+2*s12*s34+s23*s34+p2*(p5+2*(p3-s34)-s
     -   45)+s15*s45-s34*s45))+p1*(2*p2*(p3-s34)*(p5-s45)+p3*(p5*s23-
     -   2*(p5*s15-s12*s15+s12*s23)+s15*s45)+s34*(p5*s23+s15*s45-2*(s
     -   12*s15-s12*s23+s23*s45)))
       x(4,2) = p1**2*p3*(p3-s34)+(s12*s23-p2*s45)*(-(p5*s23)+s12*(-s1
     -   5+s23)+p2*(p5-s45)+s15*s45)+p1*(s23*s34*(s12-2*s45)+p3*(s12*
     -   s15+p5*s23-2*s12*s23+s15*s45)+p2*(p3*(p5-2*s45)+s34*s45))+2*
     -   msq*(p3*p5*s12-p3*s12*s15+s12sq*s15-p3*p5*s23+2*p3*s12*s23+p
     -   5*s12*s23-s12sq*s23+p3*s15*s23+s12*s15*s23+p5*s23**2-s12*s23
     -   **2-s12*s23*s34-s23**2*s34+p2**2*(p5-s45)+(p3*(p5-2*s12-s15)
     -   -s23*(p5+s15-2*s34)+s12*(-2*s15+s23+s34))*s45+p1*(-2*p3**2+p
     -   2*(p3-s34)+s34*(-s12+s23+s45)+p3*(s12+s23-2*(p5+s15-s34)+s45
     -   ))+p2*(-(p5*s12)-s12*s15-2*p5*s23+s12*s23+s23*s34+p3*(-p5+s1
     -   5-2*(s23-s45))+(p5+s12+s15+s23-s34)*s45-s45sq)+(s15-s34)*s45
     -   sq)
       x(3,3) = p1**2*(p3-s34)**2-4*msq*((s12*(-p3-s15+s23+s34)+p2*(p3
     -   +p5-s34-s45))*(p5-s15+s23-s45)+p1*(p3+s15-s23-s34)*(p3+p5-s3
     -   4-s45))+2*p1*(p3-s34)*(p2*p5+s12*s15-s12*s23-p2*s45)+(p2*p5-
     -   s12*s15+s12*s23-p2*s45)**2
       x(4,3) = p1**2*p3*(p3-s34)+(s12*(-s15+s23)+p2*(p5-s45))*(s12*s2
     -   3-p2*s45)+p1*(s12*(p3*(s15-2*s23)+s23*s34)+p2*(p3*(p5-2*s45)
     -   +s34*s45))-2*msq*(p1**2*(p3-s34)+p2**2*(-p5+s45)+s12*(s12*(-
     -   s15+s23)+s23*(p5-2*(s15-s23)+s34)+p3*(-p5+s15-2*(s23-s45))+(
     -   s15-2*s23-s34)*s45)+p1*(s12*s15-p5*s23-s12*s23+s12*s34+s23*s
     -   34+p2*(-p3+p5+s34-s45)-s15*s45+s34*s45+2*(p3**2+s23*s45)+p3*
     -   (p5-s12+s15-2*(s23+s34+s45)))+p2*(s12*s15-s12*s23-s23*s34+p3
     -   *(p5-s15+2*(s23-s45))+p5*(s12+s23-2*s45)-s12*s45+s15*s45+s34
     -   *s45-2*(s23*s45-s45sq)))
       x(4,4) = p1**2*p3**2+(s12*s23-p2*s45)**2-2*p1*p3*(s12*s23+p2*s4
     -   5)-4*msq*(p1**2*p3+s12*s23*(-p3+s12+s23)+p2**2*s45+s12*(p3-s
     -   23)*s45-p1*(-p3**2+s23*(s12-s45)+p2*(p3-s12+s45)+p3*(s12+s23
     -   +s45))+p2*((p3-s12)*s23-(p3+s12+s23)*s45+s45sq))
c          
       x(1,0)= -(p3**2*p5**2)+p3**2*p5*s15+p3*s12*s15**2-s12sq*s15**2-
     -   p3*p5*s15*s23-p3*s12*s15*s23-p5**2*s23**2-s12sq*s23**2-p1**2
     -   *(p3-s34)**2-p3*p5*s23*s34+4*p5*s12*s23*s34-s12*s15*s23*s34+
     -   p5*s23**2*s34+s12*s23**2*s34+p2*p5*(2*(s12*s15+p5*s23-s12*s2
     -   3)+p3*(-s15+2*(p5+s23))-s23*s34)-p2**2*(p5-s45)**2-(s23*(2*p
     -   5-s34)*(-s15+s34)-2*(s12*s15*(s15-s23)-s12*(s15+s23)*s34)+p3
     -   *(2*p5*(s15-s34)+s15*(-4*s12-s15+s34)))*s45-p2*(p3*(2*p5+s15
     -   )+s23*s34+2*(s12*s15-s12*s23-s15*s34+p5*(s15+s23+s34)))*s45-
     -   p1*(-(p3**2*(2*p5+s15))+2*p2*(p3-s34)*(p5-s45)+p3*((s15+s23)
     -   *s34+2*(s12*(s15-s23)+p5*(-2*s15+s23+s34)+(s15+s34)*s45))-s3
     -   4*(s23*s34+2*(s12*(s15-s23)-p5*s23+(-s15+2*s23+s34)*s45)))-(
     -   s15-s34)**2*s45sq-2*(p3*p5*s12*s15-p3*p5**2*s23+p3*p5*s12*s2
     -   3+p5*s12*s15*s23-s12sq*s15*s23-p5*s12*s23**2-p2*(s15+s34)*s4
     -   5sq)
       x(2,0)= -(p3*p5*s12*s15)+p3*s12*s15**2-s12sq*s15**2+p3*p5**2*s2
     -   3-p3*p5*s12*s23-p3*p5*s15*s23-p3*s12*s15*s23-p5**2*s23**2-s1
     -   2sq*s23**2-p1**2*(p3-s34)**2-s12*s15*s23*s34+p5*s23**2*s34+s
     -   12*s23**2*s34+2*(-(p5*s12*s15*s23)+s12sq*s15*s23+p5*s12*s23*
     -   *2+p5*s12*s23*s34)-p2**2*(p5-s45)**2-(p3*(p5-2*s12-s15)*s15-
     -   2*s15*(s12*(s15-s23)+p5*s23)+(s12*s15+(p5+s12+s15)*s23)*s34)
     -   *s45-p1*(-(p3**2*(p5+s15))+2*p2*(p3-s34)*(p5-s45)+p3*(2*s12*
     -   (s15-s23)+(s15+s23)*s34+p5*(-4*s15+2*s23+s34)+(2*s15+s34)*s4
     -   5)-s34*(2*(s12*(s15-s23)-p5*s23)+s23*s34+(-2*s15+4*s23+s34)*
     -   s45))-s15*(s15-s34)*s45sq-p2*(-(p5*(p3*(p5-s15+2*s23)+2*(s12
     -   *s15+p5*s23-s12*s23)-s23*s34))+(p3*(p5+s15)+2*(s12*(s15-s23)
     -   +p5*(s15+s23))+(p5-2*s15+s23)*s34)*s45-(2*s15+s34)*s45sq)
       x(3,0)= -(p1**2*(p3-s34)**2)-p2**2*(p5-s45)**2-s12*(s12*(s15-s2
     -   3)**2+p5*s15*s23-p5*s23**2-2*p5*s23*s34+s15*s23*s34-s23**2*s
     -   34+p3*(p5*(s15+s23)+s15*(-s15+s23-2*s45))-s15**2*s45+s15*s23
     -   *s45+s15*s34*s45+s23*s34*s45)-p1*(-(p3**2*(p5+s15))+2*p2*(p3
     -   -s34)*(p5-s45)+s34*(p5*s23-2*(s12*s15-s12*s23)-s23*s34+(s15-
     -   2*s23-s34)*s45)+p3*(2*s12*(s15-s23)+(s15+s23)*s34+p5*(-2*s15
     -   +s23+s34)+(s15+s34)*s45))+p2*(p5*(2*s12*(s15-s23)+p3*(p5-s15
     -   +2*s23)+s23*(p5-s34))-(p3*(p5+s15)+s23*s34+p5*(s15+s23+s34)+
     -   2*(s12*s15-s12*s23-s15*s34))*s45+(s15+s34)*s45sq)
       x(4,0)= p1**2*p3*(-p3+s34)+s23*(p2*p5*(2*p3-s12)+s12*(-(p3*(p5+
     -   s15))+s12*(s15-s23)+s23*(p5+s34)))-(-(p2**2*p5)+p2*(p3*(p5+s
     -   15)+s12*(s15-2*s23)+s23*(p5+s34))+s12*(-2*p3*s15+s23*(s15+s3
     -   4)))*s45-p1*(-(p3**2*(p5+s15))+s23*s34*(s12-2*s45)+p2*(p3*(p
     -   5-2*s45)+s34*s45)+p3*(s12*(s15-2*s23)+s23*(p5+s34)+(s15+s34)
     -   *s45))-p2*(p2-s15-s34)*s45sq
       endif 
       else 
       If(abs(p5).lt.eps) then 
c  Configuration                 (p1sp2sp3sp4s0)
       d= 2*(16*msq*(p1p3**2*p2p4**2-p1*p2p4**2*p3+p1p4**2*(p2p3**2-p2
     -   *p3)+p1p2**2*p3p4**2-p1*p2*p3p4**2-2*(p1p2*p1p3*p2p4*p3p4-p1
     -   *p2p3*p2p4*p3p4+p1p4*(p1p3*p2p3*p2p4-p1p2*p2p4*p3-p1p3*p2*p3
     -   p4+p1p2*p2p3*p3p4))-(p1p3**2*p2-2*p1p2*p1p3*p2p3+p1*p2p3**2+
     -   p1p2**2*p3-p1*p2*p3)*p4)+p4*s12*(p1*p2*p4-p1*p3*s15+s12*s15*
     -   s23)+p1*p4*(p1*p3-s12*s23)*s34-(p2*p4*(s12*s15+p1*s34)-(s12*
     -   s15-p1*s34)*(p3*s15-s23*s34))*s45+p2*s15*s34*s45sq)
       x(1,1) = s12sq*(p4**2+(s15-s23)**2-2*p4*(s15+s23))+p1**2*(p3**2
     -   +(p4-s34)**2-2*p3*(p4+s34))+4*msq*(-(p2**2*p4)-p3*s15*(p3-p4
     -   +s15)+(p3-p4)*s15*s23+((p4+s15-s23)*s23+p3*(-p4+s15+s23))*s3
     -   4-s23*s34**2+p2*(-p4**2+p3*(p4+s15-s23)+(-s15+s23)*s34+p4*(s
     -   15+s23+s34)))+2*(-(p1*s12*(2*p2*p4+p4*(p4-s15-s23)-p3*(p4+s1
     -   5-s23)-(p4-s15+s23)*s34))+s12*(-(s15*(2*p3-p4+s15-s23))+p2*(
     -   p4+s15-s23)+(-p4+s15+s23)*s34)*s45+p1*(-((p3-p4)*(p2-s15))+(
     -   p2+p3+p4+s15-2*s23)*s34-s34**2)*s45)+(p2**2+(s15-s34)**2-2*p
     -   2*(s15+s34))*s45sq
       x(2,1) = -(s12sq*(-(s15-s23)**2+p4*(s15+s23)))-p1*s12*(4*p2*p4-
     -   p3*(p4+2*(s15-s23))+p4*(p4-2*(s15+s23))-(p4-2*(s15-s23))*s34
     -   )+p1**2*(p3**2+(p4-s34)**2-2*p3*(p4+s34))+s12*(-2*p3*s15+s15
     -   *(p4-2*(s15-s23))+p2*(p4+2*(s15-s23))+(s15+s23)*s34)*s45+p1*
     -   (-2*(p3-p4)*(p2-s15)+(p3+p4+2*(p2+s15)-4*s23)*s34-s34**2)*s4
     -   5-2*msq*(p3*p4*s12-p4**2*s12+p3**2*s15-p3*p4*s15+p3*s12*s15+
     -   p4*s12*s15-p3*s12*s23+p4*s12*s23+p4*s12*s34-p3*s15*s34-s12*s
     -   15*s34-p3*s23*s34-p4*s23*s34+s12*s23*s34+s23*s34**2+2*(p2**2
     -   *p4+p3*s15**2-p3*s15*s23+p4*s15*s23-s15*s23*s34+s23**2*s34)+
     -   p1*(p3**2+(p4-s34)**2-2*p3*(p4+s34))+((p3-p4)*s15+(p3+p4+s15
     -   -2*s23)*s34-s34**2)*s45-p2*(-p4**2+p4*(2*(s12+s15+s23)+s34-s
     -   45)-s34*(2*(s15-s23)+s45)+p3*(p4+2*(s15-s23)+s45)))+((p2-s15
     -   )**2-(p2+s15)*s34)*s45sq
       x(3,1) = -(s12sq*(-(s15-s23)**2+p4*(s15+s23)))-p1*s12*(4*p2*p4-
     -   p3*(p4+2*(s15-s23))-p4*(s15+s23)-(p4-2*(s15-s23))*s34)+p1**2
     -   *(p3**2+s34*(-p4+s34)-p3*(p4+2*s34))+s12*(p2*(p4+2*(s15-s23)
     -   )-s15*(2*p3+s15-s23)+(s15+s23)*s34)*s45+p1*(p2*(p4-2*(p3-s34
     -   ))+(s15-2*s23-s34)*s34+p3*(s15+s34))*s45-2*msq*(p3*p4*s12+p3
     -   **2*s15+p3*s12*s15-p4*s12*s15+p3*s15**2+s12*s15**2-p3*s12*s2
     -   3-p4*s12*s23-p3*s15*s23+s12*s23**2+2*(p2**2*p4-s12*s15*s23)+
     -   p4*s12*s34-p3*s15*s34-s12*s15*s34-p3*s23*s34+s12*s23*s34-s15
     -   *s23*s34+s23**2*s34+s23*s34**2+p1*(p3**2-2*p2*p4+p4*(s15+s23
     -   -s34)+s34*(-s15+s23+s34)-p3*(p4-s15+s23+2*s34))-(s15-s34)*(p
     -   3+s15-s23-s34)*s45-p2*(-2*(s15-s23)*s34+p4*(s15+s23+s34+2*(s
     -   12-s45))-(s15-s23+s34)*s45+p3*(p4+2*(s15-s23)+s45)))-p2*(-p2
     -   +s15+s34)*s45sq
       x(4,1) = -(s12sq*(p4+s15-s23)*s23)+p1**2*p3*(p3-p4-s34)+s12*(-2
     -   *p3*s15+p2*(p4+s15-2*s23)+s23*(s15+s34))*s45+p1*(p3*s12*(p4+
     -   s15-2*s23)+s12*s23*(p4+s34)+p2*(-2*p3+p4+s34)*s45+p3*(s15+s3
     -   4)*s45-2*(p2*p4*s12+s23*s34*s45))+2*msq*(-(p3*p4*s12)-p3**2*
     -   s15+p3*s12*s15+p3*s12*s23+p4*s12*s23+p3*s15*s23+s12*s15*s23-
     -   s12*s23**2+p3*s23*s34-2*s12*s23*s34-s23**2*s34+p1*(-p3**2+p2
     -   *(p3+p4-s34)+s23*(-p4+s34)+p3*(p4-2*s15+s23+s34))+(p3-s23)*(
     -   s15-s34)*s45-p2**2*(p4+s45)+p2*(-(s12*s15)+s12*s23+s23*s34+p
     -   4*(s12+s23-2*s45)+(s15+s23+s34)*s45+p3*(p4+s15-2*s23+s45)))-
     -   p2*(-p2+s15+s34)*s45sq
       x(2,2) = p1**2*(p3**2+(p4-s34)**2-2*p3*(p4+s34))+(s12*(s15-s23)
     -   +(p2-s15)*s45)**2+2*p1*(-2*p2*p4*s12+s12*(p3*(s15-s23)+p4*(s
     -   15+s23)+(-s15+s23)*s34)+p2*(-p3+p4+s34)*s45+(p3*s15-p4*s15+(
     -   s15-2*s23)*s34)*s45)-4*msq*(p2**2*p4+p4*s12sq+p3*s12*s15+p4*
     -   s12*s15+p3*s15**2-p3*s12*s23+p4*s12*s23-p3*s15*s23+p4*s15*s2
     -   3-s12*s15*s34+s12*s23*s34-s15*s23*s34+s23**2*s34+p1*(p3**2+(
     -   p4-s34)**2-2*p3*(p4+s34))+((p3-p4)*(s12+s15)-(s12-s15+2*s23)
     -   *s34)*s45+p2*(-(p4*(2*s12+s15+s23-s45))-(p3-s34)*(s15-s23+s4
     -   5))+s34*s45sq)
       x(3,2) = p1**2*(p3**2+s34*(-p4+s34)-p3*(p4+2*s34))+p1*s12*(-4*p
     -   2*p4+p4*(s15+s23)+2*(p3*(s15-s23)+(-s15+s23)*s34))+p1*(p3*s1
     -   5+p2*(p4-2*(p3-s34))+(s15-2*s23)*s34)*s45+(s12*(s15-s23)+p2*
     -   s45)*(s12*(s15-s23)+(p2-s15)*s45)-2*msq*(p4*s12*s15+p3*s15**
     -   2+s12*s15**2+p4*s12*s23-p3*s15*s23+s12*s23**2-s15*s23*s34+s2
     -   3**2*s34+2*(p2**2*p4+p4*s12sq+p3*s12*s15-p3*s12*s23-s12*s15*
     -   s23-s12*s15*s34+s12*s23*s34)+(-(p4*s12)+p3*(2*s12+s15)+(s12-
     -   s15)*(s15-s23)+(s15-2*(s12+s23))*s34)*s45+p2*(-(p4*(4*s12+s1
     -   5+s23-s45))-(2*(p3-s34)-s45)*(s15-s23+s45))+p1*(2*(p3**2-p2*
     -   p4)+s34*(-s15+s23+2*s34+s45)-p3*(2*p4-s15+s23+4*s34+s45)+p4*
     -   (s15+s23-2*(s12+s34)+s45))-(s15-s34)*s45sq)
       x(4,2) = p1**2*p3*(p3-p4-s34)+(s12*s23-p2*s45)*(s12*(-s15+s23)+
     -   (-p2+s15)*s45)+p1*(s12*s23*(p4+s34)+p2*(-2*p3+p4+s34)*s45+p3
     -   *(s12*s15-2*s12*s23+s15*s45)-2*(p2*p4*s12+s23*s34*s45))+2*ms
     -   q*(-(p4*s12sq)-p3*s12*s15+s12sq*s15+2*p3*s12*s23-p4*s12*s23-
     -   s12sq*s23+p3*s15*s23+s12*s15*s23-s12*s23**2-s12*s23*s34-s23*
     -   *2*s34+(p4*s12-p3*(2*s12+s15)+s12*s23-s15*s23+s12*s34-2*(s12
     -   *s15-s23*s34))*s45-p2**2*(p4+s45)+p2*(p3*(s15-2*(s23-s45))+p
     -   4*(2*s12+s23-s45)+s15*s45+s12*(-s15+s23+s45)+(s23-s45)*(s34+
     -   s45))+p1*(-2*p3**2+p2*(p3+p4-s34)+(p4-s34)*(s12-s23-s45)+p3*
     -   (s12+s23+2*(p4-s15+s34)+s45))+(s15-s34)*s45sq)
       x(3,3) = p1**2*(p3-s34)**2+(s12*(s15-s23)+p2*s45)**2+2*p1*(s12*
     -   (s15-s23)*(p3-s34)+p2*(-2*p4*s12+(-p3+s34)*s45))-4*msq*(p1**
     -   2*p4+p2**2*p4+p1*(-2*p4*(p2+s12)+(p3+s15-s23-s34)*(p3-s34-s4
     -   5))+s12*(p4*s12+(p3+s15-s23-s34)*(s15-s23+s45))+p2*(-2*p4*s1
     -   2+(s15-s23+s45)*(-p3+s34+s45)))
       x(4,3) = p1**2*p3*(p3-s34)+(s12*s23-p2*s45)*(s12*(-s15+s23)-p2*
     -   s45)+p1*(s12*(p3*(s15-2*s23)+s23*s34)+p2*(-2*p4*s12+(-2*p3+s
     -   34)*s45))-2*msq*(p1**2*(p3+p4-s34)+p2**2*(p4+s45)+s12*(p4*s1
     -   2+s12*(-s15+s23)+s23*(-2*(s15-s23)+s34)+p3*(s15-2*(s23-s45))
     -   +(s15-2*s23-s34)*s45)-p2*(2*p4*s12+p3*(s15-2*(s23-s45))-s15*
     -   s45+s12*(-s15+s23+s45)+(s23-s45)*(s34+2*s45))+p1*(s12*s15-s1
     -   2*s23+s12*s34+s23*s34-s15*s45+s34*s45-p2*(p3+2*p4-s34+s45)+2
     -   *(p3**2-p4*s12+s23*s45)+p3*(-s12+s15-2*(s23+s34+s45))))
       x(4,4) = p1**2*p3**2+(s12*s23-p2*s45)**2-2*p1*p3*(s12*s23+p2*s4
     -   5)-4*msq*(p1**2*p3+s12*s23*(-p3+s12+s23)+p2**2*s45+s12*(p3-s
     -   23)*s45-p1*(-p3**2+s23*(s12-s45)+p2*(p3-s12+s45)+p3*(s12+s23
     -   +s45))+p2*((p3-s12)*s23-(p3+s12+s23)*s45+s45sq))
c          
       x(1,0)= s12*((p4-s15)*(p2*p4-p4*s12+(-p3+s12)*s15)+(-(p2*p4)-p3
     -   *s15+2*(s12*s15+p4*(s12+s15)))*s23-s12*s23**2-(p4+s15-s23)*s
     -   23*s34)-p1**2*(p3**2+(p4-s34)**2-2*p3*(p4+s34))-(-(p2**2*p4)
     -   -s15*(p3*(4*s12+s15)-2*s12*(p4-s15+s23))+(p3*s15+s15*s23-2*(
     -   p4*s12-s12*(s15+s23)))*s34-s23*s34**2+p2*(p3*s15+s23*s34+p4*
     -   (2*s12+s15+s34)+2*(s12*s15-s12*s23-s15*s34)))*s45+p1*(p3**2*
     -   s15-p3*p4*s15+p2*p4*(-p3+p4+4*s12-s34)+(-(p3*s15)+2*(p3*p4-p
     -   4*s12+s12*s15)-(p3+p4+2*s12)*s23)*s34+s23*s34**2-2*(p3*p4*s1
     -   2-p4**2*s12+p3*s12*s15+p4*s12*s15-p3*s12*s23+p4*s12*s23-((p3
     -   -p4)*(p2-s15)-(p2+p3+p4+s15-2*s23)*s34+s34**2)*s45))-(p2**2+
     -   (s15-s34)**2-2*p2*(s15+s34))*s45sq
       x(2,0)= -(p1**2*(p3**2+(p4-s34)**2-2*p3*(p4+s34)))+s12*(-(p2*p4
     -   *(s15+s23))+p4*(2*s15*s23+s12*(s15+s23))+(s15-s23)*(p3*s15-s
     -   12*s15+s12*s23-s23*s34))-(-(p2**2*p4)-p3*s15*(2*s12+s15)+s12
     -   *s15*(p4-2*(s15-s23))+(s12*s15+(s12+s15)*s23)*s34+p2*(p3*s15
     -   +p4*(s12+s15)+s23*s34+2*(s12*s15-s12*s23-s15*s34)))*s45+p1*(
     -   -(p3*p4*s12)+p4**2*s12+p3**2*s15-p3*p4*s15-2*(p3*s12*s15+p4*
     -   s12*s15-p3*s12*s23+p4*s12*s23)+p2*p4*(-p3+p4+4*s12-s34)-(p4*
     -   (s12+s23)+2*s12*(-s15+s23)+p3*(s15+s23))*s34+s23*s34**2+(2*(
     -   p3-p4)*(p2-s15)-(p3+p4+2*(p2+s15)-4*s23)*s34+s34**2)*s45)-((
     -   p2-s15)**2-(p2+s15)*s34)*s45sq
       x(3,0)= p1**2*(-p3**2+(p4-s34)*s34+p3*(p4+2*s34))+p2**2*(p4-s45
     -   )*s45+s12*(p4*s12*(s15+s23)+(s15-s23)*(p3*s15-s12*s15+s12*s2
     -   3-s23*s34)+(s15*(2*p3+s15-s23)-(s15+s23)*s34)*s45)+p2*(-(p4*
     -   s12*(s15+s23+s45))+s45*(-(p3*s15)-s23*s34-2*(s12*s15-s12*s23
     -   -s15*s34)+(s15+s34)*s45))-p1*(-(p3**2*s15)+p4*s12*s15+p4*s12
     -   *s23+p4*s12*s34-s23*s34**2-2*(s12*s15*s34-s12*s23*s34)+(s15-
     -   2*s23-s34)*s34*s45+p3*(p4*s12+2*s12*(s15-s23)+(s15+s23)*s34+
     -   (s15+s34)*s45)+p2*(p3*(p4-2*s45)+2*s34*s45+p4*(-4*s12+s34+s4
     -   5)))
       x(4,0)= p1**2*p3*(-p3+p4+s34)+s12*s23*(-(p3*s15)+s12*(p4+s15-s2
     -   3)+s23*s34)-s12*(-2*p3*s15+s23*(s15+s34))*s45+p2**2*(p4-s45)
     -   *s45-p1*(-(p3**2*s15)+s12*s23*(p4+s34)-2*s23*s34*s45+p2*(-2*
     -   p4*s12+p3*(p4-2*s45)+(p4+s34)*s45)+p3*(p4*s12+s12*s15-2*s12*
     -   s23+s23*s34+(s15+s34)*s45))+p2*(-(p4*s12*s23)-(p3*s15+s12*(p
     -   4+s15-2*s23)+s23*s34)*s45+(s15+s34)*s45sq)
       else 
c  Configuration                 (p1sp2sp3sp4sp5s)
       d= 2*(16*msq*(p1p3**2*p2p4**2-p1*p2p4**2*p3+p1p4**2*(p2p3**2-p2
     -   *p3)+p1p2**2*p3p4**2-p1*p2*p3p4**2-2*(p1p2*p1p3*p2p4*p3p4-p1
     -   *p2p3*p2p4*p3p4+p1p4*(p1p3*p2p3*p2p4-p1p2*p2p4*p3-p1p3*p2*p3
     -   p4+p1p2*p2p3*p3p4))-(p1p3**2*p2-2*p1p2*p1p3*p2p3+p1*p2p3**2+
     -   p1p2**2*p3-p1*p2*p3)*p4)+p1**2*p3*p4*s34+s23*((p3*p5-p4*s12)
     -   *(p2*p5-s12*s15)+p5*s12*s23*s34)+((p2*p4-p3*s15)*(p2*p5-s12*
     -   s15)-(p2*p5+s12*s15)*s23*s34)*s45-p1*(-(p3**2*p5*s15)+p3*p4*
     -   s12*s15+p3*p5*s23*s34+p4*s12*s23*s34+s34*(p3*s15-s23*s34)*s4
     -   5+p2*p4*(p3*p5-p4*s12+s34*s45))+p2*s15*s34*s45sq)
       x(1,1) = (p1-s12)**2*(p3**2+(p4-s34)**2-2*p3*(p4+s34))-4*msq*(p
     -   2**2*p4+s15*(p3**2+p4*s23-p3*(p4-s15+s23))-((p4+s15-s23)*s23
     -   +p3*(-p4+s15+s23))*s34+s23*s34**2-p2*(-p4**2+p3*(p4+s15-s23)
     -   +(-s15+s23)*s34+p4*(s15+s23+s34)))+(p2**2+(p3-s23)**2-2*p2*(
     -   p3+s23))*(p5-s45)**2+(-4*p2*p4+(p3+s15-s23-s34)**2)*(s12-s45
     -   )**2+2*((p1-s12)*(-p3**2+p2*(p3+p4-s34)+s23*(-p4+s34)+p3*(p4
     -   -2*s15+s23+s34))*(p5-s45)+(p1-s12)*(p3**2-2*p2*p4+p4*(s15+s2
     -   3-s34)+s34*(-s15+s23+s34)-p3*(p4-s15+s23+2*s34))*(s12-s45)-(
     -   -((p3-s23)*(p3+s15-s23-s34))+p2*(p3+2*p4-s15+s23-s34))*(p5-s
     -   45)*(-s12+s45))
       x(2,1) = p2**2*p5**2-p2*p3*p5**2+p2*p4*p5*s12+p3*p5*s12*s15-p4*
     -   s12sq*s15+s12sq*s15**2-p3*p5**2*s23+p3*p5*s12*s23+p4*p5*s12*
     -   s23-p4*s12sq*s23+p5**2*s23**2+s12sq*s23**2+p1**2*(p3**2+(p4-
     -   s34)**2-2*p3*(p4+s34))-(2*p2**2*p5+s15*(-(p3*(p5-2*s12))-s12
     -   *(p4-2*(s15-s23))+2*p5*s23)-(p5*s23+s12*(s15+s23))*s34-p2*(p
     -   3*p5+p4*s12+2*(-(p4*p5)+p5*s15+s12*s15+p5*s23-s12*s23)+p5*s3
     -   4))*s45+p1*(-(p3**2*p5)-p4**2*s12+p4*s12*s34-(2*p4*s15-(p4+2
     -   *s15-4*s23)*s34+s34**2)*s45+p3*(p4*(p5+s12)+p5*(-4*s15+2*s23
     -   +s34)+s34*s45+2*(s12*s15-s12*s23+s15*s45))+2*(p4*s12*s15-p4*
     -   p5*s23+p4*s12*s23-s12*s15*s34+p5*s23*s34+s12*s23*s34+p2*(p3*
     -   (p5-s45)+s34*(-p5+s45)+p4*(p5-2*s12+s45))))-2*(p2*p5*s12*s15
     -   +p2*p5**2*s23-p2*p5*s12*s23-p5*s12*s15*s23+s12sq*s15*s23+p5*
     -   s12*s23**2+p5*s12*s23*s34+msq*(-(p3**2*p5)+p3*p4*p5+p3*p4*s1
     -   2-p4**2*s12+p3**2*s15-p3*p4*s15+p3*s12*s15+p4*s12*s15+p3*p5*
     -   s23-p4*p5*s23-p3*s12*s23+p4*s12*s23+p3*p5*s34+p4*s12*s34-p3*
     -   s15*s34-s12*s15*s34-p3*s23*s34-p4*s23*s34+p5*s23*s34+s12*s23
     -   *s34+s23*s34**2+2*(p2**2*p4-p3*p5*s15+p3*s15**2-p3*s15*s23+p
     -   4*s15*s23-s15*s23*s34+s23**2*s34)+p1*(p3**2+(p4-s34)**2-2*p3
     -   *(p4+s34))+((p3-p4)*s15+(p3+p4+s15-2*s23)*s34-s34**2)*s45+p2
     -   *(p4**2+s34*(-p5+2*(s15-s23)+s45)-p3*(p4-p5+2*(s15-s23)+s45)
     -   +p4*(p5-2*(s12+s15+s23)-s34+s45))))+((p2-s15)**2-(p2+s15)*s3
     -   4)*s45sq
       x(3,1) = p2**2*p5**2-p2*p3*p5**2+p2*p4*p5*s12+p3*p5*s12*s15-p4*
     -   s12sq*s15+s12sq*s15**2-p2*p5**2*s23+p3*p5*s12*s23-p4*s12sq*s
     -   23+p5*s12*s15*s23-p5*s12*s23**2+s12sq*s23**2+p1**2*(p3**2+s3
     -   4*(-p4+s34)-p3*(p4+2*s34))+(-2*p2**2*p5-s12*s15*(2*p3+s15-s2
     -   3)+s12*(s15+s23)*s34+p2*(p3*p5+s12*(p4+2*(s15-s23))+p5*(s15+
     -   s23+s34)))*s45+p1*(-(p3**2*p5)+p4*s12*s15+p4*s12*s23+p4*s12*
     -   s34+p5*s23*s34-2*(s12*s15*s34-s12*s23*s34)+(s15-2*s23-s34)*s
     -   34*s45+p3*(s12*(p4+2*(s15-s23))+p5*(-2*s15+s23+s34)+(s15+s34
     -   )*s45)+p2*(p4*(p5-4*s12+s45)+2*(p3*(p5-s45)+s34*(-p5+s45))))
     -   -2*(p2*p5*s12*s15-p2*p5*s12*s23+s12sq*s15*s23+p5*s12*s23*s34
     -   +msq*(-(p3**2*p5)+p3*p4*s12+p3**2*s15-p3*p5*s15+p3*s12*s15-p
     -   4*s12*s15+p3*s15**2+s12*s15**2-p3*s12*s23-p4*s12*s23-p3*s15*
     -   s23+p5*s15*s23-p5*s23**2+s12*s23**2+2*(p2**2*p4+p3*p5*s23-s1
     -   2*s15*s23)+p3*p5*s34+p4*s12*s34-p3*s15*s34-s12*s15*s34-p3*s2
     -   3*s34-p5*s23*s34+s12*s23*s34-s15*s23*s34+s23**2*s34+s23*s34*
     -   *2+p1*(p3**2-2*p2*p4+p4*(s15+s23-s34)+s34*(-s15+s23+s34)-p3*
     -   (p4-s15+s23+2*s34))-(s15-s34)*(p3+s15-s23-s34)*s45-p2*(p5*s1
     -   5-p5*s23+p5*s34-2*(s15*s34-s23*s34)-(s15-s23+s34)*s45+p3*(p4
     -   -p5+2*(s15-s23)+s45)+p4*(s15+s23+s34-2*(p5-s12+s45)))))+p2*(
     -   p2-s15-s34)*s45sq
       x(4,1) = s12*s23*(p2*p5+p3*p5-s12*(p4+s15)+(-p5+s12)*s23)+p1**2
     -   *p3*(p3-p4-s34)+(-(p2**2*p5)+p2*(p3*p5+s12*(p4+s15-2*s23)+p5
     -   *s23)+s12*(-2*p3*s15+s23*(s15+s34)))*s45+p1*(-(p3**2*p5)+s12
     -   *s23*(p4+s34)-2*s23*s34*s45+p2*(p3*p5-2*p4*s12+(-2*p3+p4+s34
     -   )*s45)+p3*(s12*(p4+s15-2*s23)+p5*s23+(s15+s34)*s45))+2*msq*(
     -   p3**2*p5-p3*p4*s12-p3**2*s15+p3*s12*s15+p3*s12*s23+p4*s12*s2
     -   3+p3*s15*s23+s12*s15*s23+p5*s23**2-s12*s23**2+p3*s23*s34-s23
     -   **2*s34-2*(p3*p5*s23+s12*s23*s34)+p1*(-p3**2+p2*(p3+p4-s34)+
     -   s23*(-p4+s34)+p3*(p4-2*s15+s23+s34))+(p3-s23)*(s15-s34)*s45-
     -   p2**2*(p4-p5+s45)+p2*(-(s12*s15)-2*p5*s23+s12*s23+s23*s34+p4
     -   *(s12+s23-2*s45)+(s15+s23+s34)*s45+p3*(p4+s15-2*(p5+s23)+s45
     -   )))+p2*(p2-s15-s34)*s45sq
       x(2,2) = p1**2*(p3**2+(p4-s34)**2-2*p3*(p4+s34))+(p2*p5-s12*s15
     -   -p5*s23+s12*s23-p2*s45+s15*s45)**2+2*p1*(p4*s12*s15-p4*p5*s2
     -   3+p4*s12*s23-s12*s15*s34+p5*s23*s34+s12*s23*s34-p4*s15*s45+s
     -   15*s34*s45-2*s23*s34*s45+p3*(-2*p5*s15+s12*s15+p5*s23-s12*s2
     -   3+s15*s45)+p2*(p3*p5+p4*p5-2*p4*s12-p5*s34+(-p3+p4+s34)*s45)
     -   )-4*msq*(p2**2*p4+p3*p5**2-p3*p5*s12-p4*p5*s12+p4*s12sq-2*p3
     -   *p5*s15+p3*s12*s15+p4*s12*s15+p3*s15**2+p3*p5*s23-p4*p5*s23-
     -   p3*s12*s23+p4*s12*s23-p3*s15*s23+p4*s15*s23+p5*s12*s34-s12*s
     -   15*s34+p5*s23*s34+s12*s23*s34-s15*s23*s34+s23**2*s34+p1*(p3*
     -   *2+(p4-s34)**2-2*p3*(p4+s34))-((p3-p4)*(p5-s12-s15)+(p5+s12-
     -   s15+2*s23)*s34)*s45+p2*(p3*(p5-s15+s23-s45)+p4*(p5-2*s12-s15
     -   -s23+s45)+s34*(-p5+s15-s23+s45))+s34*s45sq)
       x(3,2) = p1**2*(p3**2+s34*(-p4+s34)-p3*(p4+2*s34))+(s12*(s15-s2
     -   3)+p2*(-p5+s45))*(s12*(s15-s23)+p5*s23-s15*s45+p2*(-p5+s45))
     -   +p1*(p4*s12*s15+p4*s12*s23+p5*s23*s34-2*(s12*s15*s34-s12*s23
     -   *s34)+(s15-2*s23)*s34*s45+p3*(2*s12*(s15-s23)+p5*(-2*s15+s23
     -   )+s15*s45)+p2*(p4*(p5-4*s12+s45)+2*(p3*(p5-s45)+s34*(-p5+s45
     -   ))))-2*msq*(p3*p5**2-p4*p5*s12+p4*s12*s15-p5*s12*s15+p3*s15*
     -   *2+s12*s15**2+p3*p5*s23-p5**2*s23+p4*s12*s23+p5*s12*s23-p3*s
     -   15*s23+p5*s15*s23-p5*s23**2+s12*s23**2+p5*s23*s34-s15*s23*s3
     -   4+s23**2*s34+2*(p2**2*p4-p3*p5*s12+p4*s12sq-p3*p5*s15+p3*s12
     -   *s15-p3*s12*s23-s12*s15*s23+p5*s12*s34-s12*s15*s34+s12*s23*s
     -   34)-(p4*s12+p3*(p5-2*s12-s15)-(s12-s15)*(s15-s23)-p5*(s15+s2
     -   3-s34)+(-s15+2*(s12+s23))*s34)*s45+p2*(2*p3*(p5-s15+s23-s45)
     -   +(p5-s15+s23-s45)*(p5-2*s34-s45)+p4*(p5-4*s12-s15-s23+s45))+
     -   p1*(2*(p3**2-p2*p4)+s34*(-p5-s15+s23+2*s34+s45)-p3*(2*p4-p5-
     -   s15+s23+4*s34+s45)+p4*(p5+s15+s23-2*(s12+s34)+s45))-(s15-s34
     -   )*s45sq)
       x(4,2) = p1**2*p3*(p3-p4-s34)+(s12*s23-p2*s45)*(-(p5*s23)+s12*(
     -   -s15+s23)+p2*(p5-s45)+s15*s45)+p1*(s12*s23*(p4+s34)-2*s23*s3
     -   4*s45+p3*(s12*s15+p5*s23-2*s12*s23+s15*s45)+p2*(p3*p5-2*p4*s
     -   12+(-2*p3+p4+s34)*s45))+2*msq*(p3*p5*s12-p4*s12sq-p3*s12*s15
     -   +s12sq*s15-p3*p5*s23+2*p3*s12*s23-p4*s12*s23+p5*s12*s23-s12s
     -   q*s23+p3*s15*s23+s12*s15*s23+p5*s23**2-s12*s23**2-s12*s23*s3
     -   4-s23**2*s34+(p4*s12+p3*(p5-2*s12-s15)-p5*s23+s12*s23-s15*s2
     -   3+s12*s34-2*(s12*s15-s23*s34))*s45-p2**2*(p4-p5+s45)+p1*(-2*
     -   p3**2+p2*(p3+p4-s34)+(p4-s34)*(s12-s23-s45)+p3*(s12+s23+2*(p
     -   4-p5-s15+s34)+s45))+p2*(-(p5*s12)-s12*s15-2*p5*s23+s12*s23+s
     -   23*s34+p3*(-p5+s15-2*(s23-s45))+p4*(2*s12+s23-s45)+(p5+s12+s
     -   15+s23-s34)*s45-s45sq)+(s15-s34)*s45sq)
       x(3,3) = p1**2*(p3-s34)**2+2*p1*(s12*(s15-s23)*(p3-s34)+p2*(-2*
     -   p4*s12+(p3-s34)*(p5-s45)))+(s12*(s15-s23)+p2*(-p5+s45))**2-4
     -   *msq*(p1**2*p4+p2**2*p4+p1*(-2*p4*(p2+s12)+(p3+s15-s23-s34)*
     -   (p3+p5-s34-s45))+p2*(-2*p4*s12+(p5-s15+s23-s45)*(p3+p5-s34-s
     -   45))+s12*(p4*s12+(p3+s15-s23-s34)*(-p5+s15-s23+s45)))
       x(4,3) = p1**2*p3*(p3-s34)+(s12*(-s15+s23)+p2*(p5-s45))*(s12*s2
     -   3-p2*s45)+p1*(s12*(p3*(s15-2*s23)+s23*s34)+p2*(-2*p4*s12+p3*
     -   (p5-2*s45)+s34*s45))-2*msq*(p1**2*(p3+p4-s34)+p2**2*(p4-p5+s
     -   45)+s12*(p4*s12+s12*(-s15+s23)+s23*(p5-2*(s15-s23)+s34)+p3*(
     -   -p5+s15-2*(s23-s45))+(s15-2*s23-s34)*s45)+p1*(s12*s15-p5*s23
     -   -s12*s23+s12*s34+s23*s34+p2*(-p3-2*p4+p5+s34-s45)-s15*s45+s3
     -   4*s45+2*(p3**2-p4*s12+s23*s45)+p3*(p5-s12+s15-2*(s23+s34+s45
     -   )))+p2*(s12*s15-s12*s23-s23*s34+p3*(p5-s15+2*(s23-s45))+p5*(
     -   s12+s23-2*s45)-s12*s45+s15*s45+s34*s45-2*(p4*s12+s23*s45-s45
     -   sq)))
       x(4,4) = p1**2*p3**2+(s12*s23-p2*s45)**2-2*p1*p3*(s12*s23+p2*s4
     -   5)-4*msq*(p1**2*p3+s12*s23*(-p3+s12+s23)+p2**2*s45+s12*(p3-s
     -   23)*s45-p1*(-p3**2+s23*(s12-s45)+p2*(p3-s12+s45)+p3*(s12+s23
     -   +s45))+p2*((p3-s12)*s23-(p3+s12+s23)*s45+s45sq))
c          
       x(1,0)= -(p3**2*p5**2)-p4**2*s12sq+p3**2*p5*s15-p3*p4*s12*s15+p
     -   3*s12*s15**2-s12sq*s15**2-p3*p5*s15*s23-p3*s12*s15*s23-p5**2
     -   *s23**2-s12sq*s23**2-p3*p5*s23*s34-p4*s12*s23*s34+4*p5*s12*s
     -   23*s34-s12*s15*s23*s34+p5*s23**2*s34+s12*s23**2*s34-p1**2*(p
     -   3**2+(p4-s34)**2-2*p3*(p4+s34))-p2*(-(p4**2*s12)+p4*s12*(2*p
     -   5+s15)+p4*(p5+s12)*s23+p3*p5*(p4+s15-2*(p5+s23))+p5*(-2*(s12
     -   *s15+p5*s23-s12*s23)+s23*s34))-(-(s15*(p3*(-2*p5+4*s12+s15)+
     -   2*(p5*s23-s12*(p4-s15+s23))))+(p3*(-2*p5+s15)-2*(p4*s12-s12*
     -   s15)+(2*(p5+s12)+s15)*s23)*s34-s23*s34**2)*s45-p2*(p3*(2*p5+
     -   s15)+2*((p5+s12)*s15+(p5-s12)*s23)+(2*(p5-s15)+s23)*s34+p4*(
     -   -4*p5+2*s12+s15+s34))*s45+p2**2*(-(p5-s45)**2+p4*(p5+s45))-p
     -   1*(-(p3**2*(2*p5+s15))+p4*s23*s34-s23*s34**2-2*(p4**2*s12-p4
     -   *s12*s15+p4*p5*s23-p4*s12*s23-p4*s12*s34+s12*s15*s34-p5*s23*
     -   s34-s12*s23*s34+(p4*s15-(p4+s15-2*s23)*s34+s34**2)*s45)+p3*(
     -   p4*(s15+2*(p5+s12-s34))+s15*s34+s23*s34+2*(s12*s15-s12*s23+p
     -   5*(-2*s15+s23+s34)+(s15+s34)*s45))+p2*(-p4**2+p3*(p4+2*(p5-s
     -   45))+2*s34*(-p5+s45)+p4*(-4*s12+s34+2*(p5+s45))))-(s15-s34)*
     -   *2*s45sq+2*(p3*p4*p5*s12-p3*p5*s12*s15+p4*s12sq*s15+p3*p5**2
     -   *s23-p3*p5*s12*s23-p4*p5*s12*s23+p4*s12sq*s23+p4*s12*s15*s23
     -   -p5*s12*s15*s23+s12sq*s15*s23+p5*s12*s23**2+p2*(s15+s34)*s45
     -   sq)
       x(2,0)= -(p3*p5*s12*s15)+p4*s12sq*s15+p3*s12*s15**2-s12sq*s15**
     -   2+p3*p5**2*s23-p3*p5*s12*s23-p4*p5*s12*s23+p4*s12sq*s23-p3*p
     -   5*s15*s23-p3*s12*s15*s23-p5**2*s23**2-s12sq*s23**2-s12*s15*s
     -   23*s34+p5*s23**2*s34+s12*s23**2*s34+2*(p4*s12*s15*s23-p5*s12
     -   *s15*s23+s12sq*s15*s23+p5*s12*s23**2+p5*s12*s23*s34)-p1**2*(
     -   p3**2+(p4-s34)**2-2*p3*(p4+s34))-(s15*(p3*(p5-2*s12-s15)+s12
     -   *(p4-2*(s15-s23))-2*p5*s23)+(s12*s15+(p5+s12+s15)*s23)*s34)*
     -   s45-p2**2*((p5-s45)**2-p4*(p5+s45))-p1*(-(p4**2*s12)-p3**2*(
     -   p5+s15)+p4*s12*s34+p4*s23*s34-s23*s34**2+2*(p4*s12*s15-p4*p5
     -   *s23+p4*s12*s23-s12*s15*s34+p5*s23*s34+s12*s23*s34)-(2*p4*s1
     -   5-(p4+2*s15-4*s23)*s34+s34**2)*s45+p3*(p4*(p5+s12+s15)+s15*s
     -   34+s23*s34+p5*(-4*s15+2*s23+s34)+s34*s45+2*(s12*s15-s12*s23+
     -   s15*s45))+p2*(-p4**2+p3*(p4+2*(p5-s45))+2*s34*(-p5+s45)+p4*(
     -   -4*s12+s34+2*(p5+s45))))-s15*(s15-s34)*s45sq-p2*(-(p3*p5*(p5
     -   -s15+2*s23))+p5*(-2*(s12*s15+p5*s23-s12*s23)+s23*s34)+p3*(p5
     -   +s15)*s45+(2*((p5+s12)*s15+(p5-s12)*s23)+(p5-2*s15+s23)*s34)
     -   *s45+p4*(p5*(s12+s23-2*s45)+s15*s45+s12*(s15+s23+s45))-(2*s1
     -   5+s34)*s45sq)
       x(3,0)= -(p1**2*(p3**2+s34*(-p4+s34)-p3*(p4+2*s34)))-s12*(s12*s
     -   15**2+p5*s15*s23-p5*s23**2+s12*s23**2-p4*s12*(s15+s23)+s15*s
     -   23*s34-s23**2*s34-2*(s12*s15*s23+p5*s23*s34)+p3*(p5*(s15+s23
     -   )+s15*(-s15+s23-2*s45))-s15**2*s45+s15*s23*s45+s15*s34*s45+s
     -   23*s34*s45)-p2**2*((p5-s45)**2-p4*(p5+s45))-p1*(p4*s12*s15-p
     -   3**2*(p5+s15)+p4*s12*s23+p4*s12*s34+p5*s23*s34-s23*s34**2-2*
     -   (s12*s15*s34-s12*s23*s34)+(s15-2*s23-s34)*s34*s45+p3*(p4*s12
     -   +2*s12*(s15-s23)+(s15+s23)*s34+p5*(-2*s15+s23+s34)+(s15+s34)
     -   *s45)+p2*(p3*(p4+2*(p5-s45))+2*s34*(-p5+s45)+p4*(p5-4*s12+s3
     -   4+s45)))-p2*(-(p3*p5*(p5-s15+2*s23))-p5*(2*s12*(s15-s23)+s23
     -   *(p5-s34))+p3*(p5+s15)*s45+(s23*s34+p5*(s15+s23+s34)+2*(s12*
     -   s15-s12*s23-s15*s34))*s45+p4*s12*(p5+s15+s23+s45)-(s15+s34)*
     -   s45sq)
       x(4,0)= -(p1**2*p3*(p3-p4-s34))+s23*(2*p2*p3*p5-p2*(p4+p5)*s12+
     -   s12*(-(p3*(p5+s15))+s12*(p4+s15-s23)+p5*s23+s23*s34))-(-(p2*
     -   *2*(p4+p5))+p2*(p3*(p5+s15)+s12*(p4+s15-2*s23)+p5*s23+s23*s3
     -   4)+s12*(-2*p3*s15+s23*(s15+s34)))*s45-p1*(-(p3**2*(p5+s15))+
     -   s12*s23*(p4+s34)-2*s23*s34*s45+p2*(-2*p4*s12+p3*(p4+p5-2*s45
     -   )+(p4+s34)*s45)+p3*(p4*s12+s12*(s15-2*s23)+s23*(p5+s34)+(s15
     -   +s34)*s45))-p2*(p2-s15-s34)*s45sq
       endif 
       endif 
       endif 
       endif 
       endif 


      d3=d*3.d0
      
      Id=1.d0/d
      I2=Id*0.5d0
      I3=1.d0/d3
      I4=Id*0.25d0

      x(1,2)=x(2,1)
     
      x(1,3)=x(3,1)
      x(2,3)=x(3,2)
     
      x(1,4)=x(4,1)
      x(2,4)=x(4,2)
      x(3,4)=x(4,3)
     
c$$$      
c$$$      Do j=1,4
c$$$      Do i=1,4
c$$$      Print*, "x(",i,",",j,")=",x(i,j)
c$$$      enddo
c$$$      enddo     
C      
      Y(1,1,1)=0
      Y(2,1,1)=(x(1,2)*x(1,0)-x(2,0)*x(1,1))*Id
      Y(3,1,1)=(x(1,3)*x(1,0)-x(3,0)*x(1,1))*Id
      Y(4,1,1)=(x(1,4)*x(1,0)-x(4,0)*x(1,1))*Id
      Y(1,2,1)=-Y(2,1,1)
      Y(2,2,1)=0
      Y(3,2,1)=(x(1,3)*x(2,0)-x(3,0)*x(1,2))*Id
      Y(4,2,1)=(x(1,4)*x(2,0)-x(4,0)*x(1,2))*Id
      Y(1,3,1)=-Y(3,1,1)
      Y(2,3,1)=-Y(3,2,1)
      Y(3,3,1)=0
      Y(4,3,1)=(x(1,4)*x(3,0)-x(4,0)*x(1,3))*Id
      Y(1,4,1)=-Y(4,1,1)
      Y(2,4,1)=-Y(4,2,1)
      Y(3,4,1)=-Y(4,3,1)
      Y(4,4,1)=0
      Y(1,1,2)=0
      Y(2,1,2)=(x(2,2)*x(1,0)-x(2,0)*x(2,1))*Id
      Y(3,1,2)=(x(2,3)*x(1,0)-x(3,0)*x(2,1))*Id
      Y(4,1,2)=(x(2,4)*x(1,0)-x(4,0)*x(2,1))*Id
      Y(1,2,2)=-Y(2,1,2)
      Y(2,2,2)=0
      Y(3,2,2)=(x(2,3)*x(2,0)-x(3,0)*x(2,2))*Id
      Y(4,2,2)=(x(2,4)*x(2,0)-x(4,0)*x(2,2))*Id
      Y(1,3,2)=-Y(3,1,2)
      Y(2,3,2)=-Y(3,2,2)
      Y(3,3,2)=0
      Y(4,3,2)=(x(2,4)*x(3,0)-x(4,0)*x(2,3))*Id
      Y(1,4,2)=-Y(4,1,2)
      Y(2,4,2)=-Y(4,2,2)
      Y(3,4,2)=-Y(4,3,2)
      Y(4,4,2)=0
      Y(1,1,3)=0
      Y(2,1,3)=(x(3,2)*x(1,0)-x(2,0)*x(3,1))*Id
      Y(3,1,3)=(x(3,3)*x(1,0)-x(3,0)*x(3,1))*Id
      Y(4,1,3)=(x(3,4)*x(1,0)-x(4,0)*x(3,1))*Id
      Y(1,2,3)=-Y(2,1,3)
      Y(2,2,3)=0
      Y(3,2,3)=(x(3,3)*x(2,0)-x(3,0)*x(3,2))*Id
      Y(4,2,3)=(x(3,4)*x(2,0)-x(4,0)*x(3,2))*Id
      Y(1,3,3)=-Y(3,1,3)
      Y(2,3,3)=-Y(3,2,3)
      Y(3,3,3)=0
      Y(4,3,3)=(x(3,4)*x(3,0)-x(4,0)*x(3,3))*Id
      Y(1,4,3)=-Y(4,1,3)
      Y(2,4,3)=-Y(4,2,3)
      Y(3,4,3)=-Y(4,3,3)
      Y(4,4,3)=0
      Y(1,1,4)=0
      Y(2,1,4)=(x(4,2)*x(1,0)-x(2,0)*x(4,1))*Id
      Y(3,1,4)=(x(4,3)*x(1,0)-x(3,0)*x(4,1))*Id
      Y(4,1,4)=(x(4,4)*x(1,0)-x(4,0)*x(4,1))*Id
      Y(1,2,4)=-Y(2,1,4)
      Y(2,2,4)=0
      Y(3,2,4)=(x(4,3)*x(2,0)-x(3,0)*x(4,2))*Id
      Y(4,2,4)=(x(4,4)*x(2,0)-x(4,0)*x(4,2))*Id
      Y(1,3,4)=-Y(3,1,4)
      Y(2,3,4)=-Y(3,2,4)
      Y(3,3,4)=0
      Y(4,3,4)=(x(4,4)*x(3,0)-x(4,0)*x(4,3))*Id
      Y(1,4,4)=-Y(4,1,4)     
      Y(2,4,4)=-Y(4,2,4)    
      Y(3,4,4)=-Y(4,3,4)
      Y(4,4,4)=0
c
      Y123=Y(1,3,2)+Y(1,2,3)
      Y423=Y(4,2,3)+Y(4,3,2)
      Y124=Y(1,4,2)+Y(1,2,4)
      Y324=Y(3,2,4)+Y(3,4,2)
      Y412=Y(4,1,2)+Y(4,2,1)
      Y312=Y(3,2,1)+Y(3,1,2)
      Y213=Y(2,3,1)+Y(2,1,3)
      Y413=Y(4,1,3)+Y(4,3,1)
      Y214=Y(2,4,1)+Y(2,1,4)
      Y314=Y(3,1,4)+Y(3,4,1)
      Y134=Y(1,4,3)+Y(1,3,4)
      Y234=Y(2,3,4)+Y(2,4,3)
c  Rank-1
      R(1)=DR1345-DR2345
      R(2)=DR1245-DR1345
      R(3)=DR1235-DR1245
      R(4)=DR1234-DR1235
c 1
      E(1,1)=(R(1)*x(1,1)+R(2)*x(2,1)+R(3)*x(3,1)+R(4)*x(4,1)
     /         -DR2345*x(1,0))*Id
c 2
      E(2,1)=(R(1)*x(1,2)+R(2)*x(2,2)+R(3)*x(3,2)+R(4)*x(4,2)   
     /           -DR2345*x(2,0))*Id
c 3
      E(3,1)=(R(1)*x(1,3)+R(2)*x(2,3)+R(3)*x(3,3)+R(4)*x(4,3)   
     /           -DR2345*x(3,0))*Id
c 4
      E(4,1)=(R(1)*x(1,4)+R(2)*x(2,4)+R(3)*x(3,4)+R(4)*x(4,4)   
     /          -DR2345*x(4,0))*Id
c  Rank-2
      R(1)=DRR1345(7,2)-DRR2345(7,2)
      R(2)=DRR1245(7,2)-DRR1345(7,2)
      R(3)=DRR1235(7,2)-DRR1245(7,2)
      R(4)=DRR1234(7,2)-DRR1235(7,2)

      R1(1)= DRR1345(1,1) + DR2345
      R1(2)= DRR1245(1,1) - DRR1345(1,1)
      R1(3)= DRR1235(1,1) - DRR1245(1,1)
      R1(4)= DRR1234(1,1) - DRR1235(1,1)

      R2(1)= DRR1345(1,1) - DRR2345(1,1)
      R2(2)= DRR1245(2,1) - DRR1345(1,1)
      R2(3)= DRR1235(2,1) - DRR1245(2,1)
      R2(4)= DRR1234(2,1) - DRR1235(2,1) 

      R3(1)= DRR1345(2,1) - DRR2345(2,1)
      R3(2)= DRR1245(2,1) - DRR1345(2,1)
      R3(3)= DRR1235(3,1) - DRR1245(2,1)
      R3(4)= DRR1234(3,1) - DRR1235(3,1) 
     
      R4(1)= DRR1345(3,1) - DRR2345(3,1)
      R4(2)= DRR1245(3,1) - DRR1345(3,1)
      R4(3)= DRR1235(3,1) - DRR1245(3,1)
      R4(4)=               - DRR1235(3,1) 

c 1 1
      E(1,2)=((R(2)*Y(2,1,1)+R(3)*Y(3,1,1)+R(4)*Y(4,1,1))*2.d0+
     \ R1(1)*x(1,1)+R1(2)*x(2,1)+R1(3)*x(3,1)+R1(4)*x(4,1)+DR2345*x(1,0))*Id
c 2 2 
      E(2,2)=((R(1)*Y(1,2,2)+R(3)*Y(3,2,2)+R(4)*Y(4,2,2))*2.d0+
     \R2(1)*x(1,2)+R2(2)*x(2,2)+R2(3)*x(3,2)+R2(4)*x(4,2)-DRR2345(1,1)*x(2,0))*Id
c 3 3
      E(3,2)=((R(1)*Y(1,3,3)+R(2)*Y(2,3,3)+R(4)*Y(4,3,3))*2.d0+
     \R3(1)*x(1,3)+R3(2)*x(2,3)+R3(3)*x(3,3)+R3(4)*x(4,3)-DRR2345(2,1)*x(3,0))*Id
c 4 4
      E(4,2)=((R(1)*Y(1,4,4)+R(2)*Y(2,4,4)+R(3)*Y(3,4,4))*2.d0+
     \R4(1)*x(1,4)+R4(2)*x(2,4)+R4(3)*x(3,4)+R4(4)*x(4,4)-DRR2345(3,1)*x(4,0))*Id
c 1 2
      E(5,2)=((R(1)*Y(1,2,1)+R(2)*Y(2,1,2)+R(3)*Y312+R(4)*Y412)*2.d0+
     \R1(1)*x(1,2)+R1(2)*x(2,2)+R1(3)*x(3,2)+R1(4)*x(4,2)+R2(1)*x(1,1)+
     \R2(2)*x(2,1)+R2(3)*x(3,1)+R2(4)*x(4,1)+DR2345*x(2,0)-DRR2345(1,1)*x(1,0))*I2
c 1 3
      E(6,2)=((R(1)*Y(1,3,1)+R(2)*Y213+R(3)*Y(3,1,3)+R(4)*Y413)*2.d0+
     \R1(1)*x(1,3)+R1(2)*x(2,3)+R1(3)*x(3,3)+R1(4)*x(4,3)+R3(1)*x(1,1)+
     \R3(2)*x(2,1)+R3(3)*x(3,1)+R3(4)*x(4,1)+DR2345*x(3,0)-DRR2345(2,1)*x(1,0))*I2
c 1 4
      E(7,2)=((R(1)*Y(1,4,1)+R(2)*Y214+R(3)*Y314+R(4)*Y(4,1,4))*2.d0+
     \R1(1)*x(1,4)+R1(2)*x(2,4)+R1(3)*x(3,4)+R1(4)*x(4,4)+R4(1)*x(1,1)+
     \R4(2)*x(2,1)+R4(3)*x(3,1)+R4(4)*x(4,1)+DR2345*x(4,0)-DRR2345(3,1)*x(1,0))*I2
c 2 3
      E(8,2)=((R(1)*Y123+R(2)*Y(2,3,2)+R(3)*Y(3,2,3)+R(4)*Y423)*2.d0+
     \R3(1)*x(1,2)+R3(2)*x(2,2)+R3(3)*x(3,2)+R3(4)*x(4,2)+R2(1)*x(1,3)+
     \R2(2)*x(2,3)+R2(3)*x(3,3)+R2(4)*x(4,3)-DRR2345(1,1)*x(3,0)-DRR2345(2,1)*x(2,0))*I2
c 2 4
      E(9,2)=((R(1)*Y124+R(2)*Y(2,4,2)+R(3)*Y324+R(4)*Y(4,2,4))*2.d0+
     \R4(1)*x(1,2)+R4(2)*x(2,2)+R4(3)*x(3,2)+R4(4)*x(4,2)+R2(1)*x(1,4)+
     \R2(2)*x(2,4)+R2(3)*x(3,4)+R2(4)*x(4,4)-DRR2345(1,1)*x(4,0)-DRR2345(3,1)*x(2,0))*I2
c 3 4
      E(10,2)=((R(1)*Y134+R(2)*Y234+R(3)*Y(3,4,3)+R(4)*Y(4,3,4))*2.d0+
     \R4(1)*x(1,3)+R4(2)*x(2,3)+R4(3)*x(3,3)+R4(4)*x(4,3)+R3(1)*x(1,4)+
     \R3(2)*x(2,4)+R3(3)*x(3,4)+R3(4)*x(4,4)-DRR2345(2,1)*x(4,0)-DRR2345(3,1)*x(3,0))*I2
C 00
      E(11,2)=(R(1)*x(1,0)+R(2)*x(2,0)+R(3)*x(3,0)+R(4)*x(4,0))*Id
c  Rank-3
      R1(1)= DRR1345(11,3) + DRR2345(7,2)
      R1(2)= DRR1245(11,3) - DRR1345(11,3)
      R1(3)= DRR1235(11,3) - DRR1245(11,3)
      R1(4)= DRR1234(11,3) - DRR1235(11,3)
c 0 0 1
      E(21,3)=(R(1)*x(1,1)+R(2)*x(2,1)+R(3)*x(3,1)+R(4)*x(4,1)   
     \+(R1(1)*x(1,0)+R1(2)*x(2,0)+R1(3)*x(3,0)+R1(4)*x(4,0))*2.d0 
     \-DRR2345(7,2)*x(1,0))*I3
c
      R11(1)= DRR1345(1,2) - DR2345
      R11(2)= DRR1245(1,2) - DRR1345(1,2)
      R11(3)= DRR1235(1,2) - DRR1245(1,2)
      R11(4)= DRR1234(1,2) - DRR1235(1,2)

c 1 1 1 
      E(1,3)=((R1(2)*Y(2,1,1)+R1(3)*Y(3,1,1)+R1(4)*Y(4,1,1))*4.d0+
     \R11(1)*x(1,1)+R11(2)*x(2,1)+R11(3)*x(3,1)+R11(4)*x(4,1)
     \-DR2345*x(1,0))*Id

      R2(1)= DRR1345(11,3) - DRR2345(11,3)
      R2(2)= DRR1245(12,3) - DRR1345(11,3)
      R2(3)= DRR1235(12,3) - DRR1245(12,3)
      R2(4)= DRR1234(12,3) - DRR1235(12,3)
c 0 0 2
      E(22,3)=(R(1)*x(1,2)+R(2)*x(2,2)+R(3)*x(3,2)+R(4)*x(4,2)   
     \+(R2(1)*x(1,0)+R2(2)*x(2,0)+R2(3)*x(3,0)+R2(4)*x(4,0))*2.d0
     \-DRR2345(7,2)*x(2,0))*I3
c
      R22(1)= DRR1345(1,2) - DRR2345(1,2)
      R22(2)= DRR1245(2,2) - DRR1345(1,2)
      R22(3)= DRR1235(2,2) - DRR1245(2,2)
      R22(4)= DRR1234(2,2) - DRR1235(2,2)

c 2 2 2
      E(2,3)=((R2(1)*Y(1,2,2)+R2(3)*Y(3,2,2)+R2(4)*Y(4,2,2))*4.d0+
     \R22(1)*x(1,2)+R22(2)*x(2,2)+R22(3)*x(3,2)+R22(4)*x(4,2)
     \ -DRR2345(1,2)*x(2,0))*Id
c
      R3(1)= DRR1345(12,3) - DRR2345(12,3)
      R3(2)= DRR1245(12,3) - DRR1345(12,3)
      R3(3)= DRR1235(13,3) - DRR1245(12,3)
      R3(4)= DRR1234(13,3) - DRR1235(13,3)
c 0 0 3
      E(23,3)=(R(1)*x(1,3)+R(2)*x(2,3)+R(3)*x(3,3)+R(4)*x(4,3)   
     \+(R3(1)*x(1,0)+R3(2)*x(2,0)+R3(3)*x(3,0)+R3(4)*x(4,0))*2.d0
     \-DRR2345(7,2)*x(3,0))*I3
c
      R33(1)= DRR1345(2,2) - DRR2345(2,2)
      R33(2)= DRR1245(2,2) - DRR1345(2,2)
      R33(3)= DRR1235(3,2) - DRR1245(2,2)
      R33(4)= DRR1234(3,2) - DRR1235(3,2)
c 3 3 3
      E(3,3)=((R3(1)*Y(1,3,3)+R3(2)*Y(2,3,3)+R3(4)*Y(4,3,3))*4.d0+
     \R33(1)*x(1,3)+R33(2)*x(2,3)+R33(3)*x(3,3)+R33(4)*x(4,3)-DRR2345(2,2)*x(3,0))*Id
c
      R4(1)= DRR1345(13,3) - DRR2345(13,3)
      R4(2)= DRR1245(13,3) - DRR1345(13,3)
      R4(3)= DRR1235(13,3) - DRR1245(13,3)
      R4(4)=                - DRR1235(13,3)
c 0 0 4
      E(24,3)=(R(1)*x(1,4)+R(2)*x(2,4)+R(3)*x(3,4)+R(4)*x(4,4)
     \+(R4(1)*x(1,0)+R4(2)*x(2,0)+R4(3)*x(3,0)+R4(4)*x(4,0))*2.d0
     \-DRR2345(7,2)*x(4,0))*I3

      R44(1)= DRR1345(3,2) - DRR2345(3,2)
      R44(2)= DRR1245(3,2) - DRR1345(3,2)
      R44(3)= DRR1235(3,2) - DRR1245(3,2)
      R44(4)=               - DRR1235(3,2)   
c 4 4 4
      E(4,3)=((R4(1)*Y(1,4,4)+R4(2)*Y(2,4,4)+R4(3)*Y(3,4,4))*4.d0+
     \R44(1)*x(1,4)+R44(2)*x(2,4)+R44(3)*x(3,4)+R44(4)*x(4,4)-DRR2345(3,2)*x(4,0))*Id

      R12(1)= DRR1345(1,2) + DRR2345(1,1)
      R12(2)= DRR1245(4,2) - DRR1345(1,2)
      R12(3)= DRR1235(4,2) - DRR1245(4,2)
      R12(4)= DRR1234(4,2) - DRR1235(4,2)

c 1 1 2 
      E(5,3)=((R1(1)*Y(1,2,1)+R1(2)*Y(2,1,2)+R1(3)*Y312+R1(4)*Y412+
     \R2(2)*Y(2,1,1)+R2(3)*Y(3,1,1)+R2(4)*Y(4,1,1))*4.d0+(R12(1)*x(1,1)+
     \R12(2)*x(2,1)+R12(3)*x(3,1)+R12(4)*x(4,1)+DRR2345(1,1)*x(1,0))*2.d0+
     \R11(1)*x(1,2)+R11(2)*x(2,2)+R11(3)*x(3,2)+ R11(4)*x(4,2)-DR2345*x(2,0))*I3
c  2 2 1
      E(8,3)=((R2(1)*Y(1,2,1)+R2(2)*Y(2,1,2)+R2(3)*Y312+R2(4)*Y412+
     \R1(1)*Y(1,2,2)+R1(3)*Y(3,2,2)+R1(4)*Y(4,2,2))*4.d0+(R12(1)*x(1,2)+
     \+R12(2)*x(2,2)+R12(3)*x(3,2)+R12(4)*x(4,2)+DRR2345(1,1)*x(2,0))*2.d0+
     \R22(1)*x(1,1)+R22(2)*x(2,1)+R22(3)*x(3,1)+R22(4)*x(4,1)-DRR2345(1,2)*x(1,0))*I3
c
      R13(1)= DRR1345(4,2) + DRR2345(2,1)
      R13(2)= DRR1245(4,2) - DRR1345(4,2)
      R13(3)= DRR1235(5,2) - DRR1245(4,2)
      R13(4)= DRR1234(5,2) - DRR1235(5,2)
c 1 1 3 
      E(6,3)=((R1(1)*Y(1,3,1)+R1(2)*Y213+R1(3)*Y(3,1,3)+R1(4)*Y413+
     \R3(2)*Y(2,1,1)+R3(3)*Y(3,1,1)+R3(4)*Y(4,1,1))*4.d0+(R13(1)*x(1,1)+
     \R13(2)*x(2,1)+R13(3)*x(3,1)+R13(4)*x(4,1)+DRR2345(2,1)*x(1,0))*2.d0+
     \R11(1)*x(1,3)+R11(2)*x(2,3)+R11(3)*x(3,3)+R11(4)*x(4,3)-DR2345*x(3,0))*I3
c  3 3 1
      E(11,3)=((R3(1)*Y(1,3,1)+R3(2)*Y213+R3(3)*Y(3,1,3)+R3(4)*Y413+
     \R1(1)*Y(1,3,3)+R1(2)*Y(2,3,3)+R1(4)*Y(4,3,3))*4.d0+(R13(1)*x(1,3)+
     \R13(2)*x(2,3)+R13(3)*x(3,3)+R13(4)*x(4,3)+DRR2345(2,1)*x(3,0))*2.d0+
     \R33(1)*x(1,1)+R33(2)*x(2,1)+R33(3)*x(3,1)+R33(4)*x(4,1)-DRR2345(2,2)*x(1,0))*I3
c
      R14(1)= DRR1345(5,2) + DRR2345(3,1)
      R14(2)= DRR1245(5,2) - DRR1345(5,2)
      R14(3)= DRR1235(5,2) - DRR1245(5,2)
      R14(4)=               - DRR1235(5,2)
c 1 1 4
      E(7,3)=((R1(1)*Y(1,4,1)+R1(2)*Y214+R1(3)*Y314+R1(4)*Y(4,1,4)+
     \R4(2)*Y(2,1,1)+R4(3)*Y(3,1,1)+R4(4)*Y(4,1,1))*4.d0+(R14(1)*x(1,1)+
     \R14(2)*x(2,1)+R14(3)*x(3,1)+R14(4)*x(4,1)+DRR2345(3,1)*x(1,0))*2.d0+
     \R11(1)*x(1,4)+R11(2)*x(2,4)+R11(3)*x(3,4)+R11(4)*x(4,4)-DR2345*x(4,0))*I3
c  4 4 1
      E(14,3)=((R4(1)*Y(1,4,1)+R4(2)*Y214+R4(3)*Y314+R4(4)*Y(4,1,4)+
     \R1(1)*Y(1,4,4)+R1(2)*Y(2,4,4)+R1(3)*Y(3,4,4))*4.d0+(R14(1)*x(1,4)+
     \R14(2)*x(2,4)+R14(3)*x(3,4)+R14(4)*x(4,4)+DRR2345(3,1)*x(4,0))*2.d0+
     \R44(1)*x(1,1)+R44(2)*x(2,1)+R44(3)*x(3,1)+R44(4)*x(4,1)-DRR2345(3,2)*x(1,0))*I3
c
      R23(1)= DRR1345(4,2) - DRR2345(4,2)
      R23(2)= DRR1245(2,2) - DRR1345(4,2)
      R23(3)= DRR1235(6,2) - DRR1245(2,2)
      R23(4)= DRR1234(6,2) - DRR1235(6,2)
c  2 2 3
      E(9,3)=((R2(1)*Y123+R2(2)*Y(2,3,2)+R2(3)*Y(3,2,3)+R2(4)*Y423+
     \R3(1)*Y(1,2,2)+R3(3)*Y(3,2,2)+R3(4)*Y(4,2,2))*4.d0+(R23(1)*x(1,2)+
     \R23(2)*x(2,2)+R23(3)*x(3,2)+R23(4)*x(4,2)-DRR2345(4,2)*x(2,0))*2.d0+
     \R22(1)*x(1,3)+R22(2)*x(2,3)+R22(3)*x(3,3)+R22(4)*x(4,3)-DRR2345(1,2)*x(3,0))*I3
c  3 3 2
      E(12,3)=((R3(1)*Y123+R3(2)*Y(2,3,2)+R3(3)*Y(3,2,3)+R3(4)*Y423+
     \R2(1)*Y(1,3,3)+R2(2)*Y(2,3,3)+R2(4)*Y(4,3,3))*4.d0+(R23(1)*x(1,3)+
     \R23(2)*x(2,3)+R23(3)*x(3,3)+R23(4)*x(4,3)-DRR2345(4,2)*x(3,0))*2.d0+
     \R33(1)*x(1,2)+R33(2)*x(2,2)+R33(3)*x(3,2)+R33(4)*x(4,2)- DRR2345(2,2)*x(2,0))*I3
c 1 2 3
      E(17,3)=((R1(1)*Y123+R1(2)*Y(2,3,2)+R1(3)*Y(3,2,3)+R1(4)*Y423+
     \R2(1)*Y(1,3,1)+R2(2)*Y213+R2(3)*Y(3,1,3)+R2(4)*Y413+R3(1)*Y(1,2,1)
     \+R3(2)*Y(2,1,2)+R3(3)*Y312+R3(4)*Y412)*2.d0+R12(1)*x(1,3)+
     \R12(2)*x(2,3)+R12(3)*x(3,3)+R12(4)*x(4,3)+R13(1)*x(1,2)+
     \R13(2)*x(2,2)+R13(3)*x(3,2)+R13(4)*x(4,2)+R23(1)*x(1,1)+
     \R23(2)*x(2,1)+R23(3)*x(3,1)+R23(4)*x(4,1)+DRR2345(2,1)*x(2,0)+ 
     \DRR2345(1,1)*x(3,0)-DRR2345(4,2)*x(1,0))*I3
c
      R24(1)= DRR1345(5,2) - DRR2345(5,2)
      R24(2)= DRR1245(6,2) - DRR1345(5,2)
      R24(3)= DRR1235(6,2) - DRR1245(6,2)
      R24(4)=               - DRR1235(6,2)
c  2 2 4
      E(10,3)=((R2(1)*Y124+R2(2)*Y(2,4,2)+R2(3)*Y324+R2(4)*Y(4,2,4)+
     \R4(1)*Y(1,2,2)+R4(3)*Y(3,2,2)+R4(4)*Y(4,2,2))*4.d0+(R24(1)*x(1,2)+
     \R24(2)*x(2,2)+R24(3)*x(3,2)+R24(4)*x(4,2)-DRR2345(5,2)*x(2,0))*2.d0+
     \R22(1)*x(1,4)+R22(2)*x(2,4)+R22(3)*x(3,4)+R22(4)*x(4,4)-DRR2345(1,2)*x(4,0))*I3
c  4 4 2
      E(15,3)=((R4(1)*Y124+R4(2)*Y(2,4,2)+R4(3)*Y324+R4(4)*Y(4,2,4)+
     \R2(1)*Y(1,4,4)+R2(2)*Y(2,4,4)+R2(3)*Y(3,4,4))*4.d0+(R24(1)*x(1,4)+
     \R24(2)*x(2,4)+R24(3)*x(3,4)+R24(4)*x(4,4)-DRR2345(5,2)*x(4,0))*2.d0+
     \R44(1)*x(1,2)+R44(2)*x(2,2)+R44(3)*x(3,2)+R44(4)*x(4,2)-DRR2345(3,2)*x(2,0))*I3
c 1 2 4
       E(18,3)=((R1(1)*Y124+R1(2)*Y(2,4,2)+R1(3)*Y324+R1(4)*Y(4,2,4)+
     \R2(1)*Y(1,4,1)+R2(2)*Y214+R2(3)*Y314+R2(4)*Y(4,1,4)+R4(1)*Y(1,2,1)
     \+R4(2)*Y(2,1,2)+R4(3)*Y312+R4(4)*Y412)*2.d0+R12(1)*x(1,4)+
     \R12(2)*x(2,4)+R12(3)*x(3,4)+R12(4)*x(4,4)+R14(1)*x(1,2)+
     \R14(2)*x(2,2)+R14(3)*x(3,2)+R14(4)*x(4,2)+R24(1)*x(1,1)+
     \R24(2)*x(2,1)+R24(3)*x(3,1)+R24(4)*x(4,1)+DRR2345(3,1)*x(2,0)+
     \DRR2345(1,1)*x(4,0)-DRR2345(5,2)*x(1,0))*I3

c
      R34(1)= DRR1345(6,2) - DRR2345(6,2)
      R34(2)= DRR1245(6,2) - DRR1345(6,2)
      R34(3)= DRR1235(3,2) - DRR1245(6,2)
      R34(4)=               - DRR1235(3,2)   
c  3 3 4
      E(13,3)=((R3(1)*Y134+R3(2)*Y234 +R3(3)*Y(3,4,3)+R3(4)*Y(4,3,4)+
     \R4(1)*Y(1,3,3)+R4(2)*Y(2,3,3)+R4(4)*Y(4,3,3))*4.d0+(R34(1)*x(1,3)
     \+R34(2)*x(2,3)+R34(3)*x(3,3)+R34(4)*x(4,3)-DRR2345(6,2)*x(3,0))*2.d0+
     \R33(1)*x(1,4)+R33(2)*x(2,4)+R33(3)*x(3,4)+R33(4)*x(4,4)-DRR2345(2,2)*x(4,0))*I3
c  4 4 3
      E(16,3)=((R4(1)*Y134+R4(2)*Y234+R4(3)*Y(3,4,3)+R4(4)*Y(4,3,4)+
     \R3(1)*Y(1,4,4)+R3(2)*Y(2,4,4)+R3(3)*Y(3,4,4))*4.d0+(R34(1)*x(1,4)+
     \+R34(2)*x(2,4)+R34(3)*x(3,4)+R34(4)*x(4,4)-DRR2345(6,2)*x(4,0))*2.d0+
     \R44(1)*x(1,3)+R44(2)*x(2,3)+R44(3)*x(3,3)+R44(4)*x(4,3)-DRR2345(3,2)*x(3,0))*I3
c 1 3 4
       E(19,3)=((R1(1)*Y134+R1(2)*Y234+R1(3)*Y(3,4,3)+R1(4)*Y(4,3,4)+
     \R3(1)*Y(1,4,1)+R3(2)*Y214+R3(3)*Y314+R3(4)*Y(4,1,4)+R4(1)*Y(1,3,1)
     \+R4(2)*Y213+R4(3)*Y(3,1,3)+R4(4)*Y413)*2.d0+R13(1)*x(1,4)+
     \R13(2)*x(2,4)+R13(3)*x(3,4)+R13(4)*x(4,4)+R14(1)*x(1,3)+
     \R14(2)*x(2,3)+R14(3)*x(3,3)+R14(4)*x(4,3)+R34(1)*x(1,1)+
     \R34(2)*x(2,1)+R34(3)*x(3,1)+R34(4)*x(4,1)+DRR2345(3,1)*x(3,0)+
     \DRR2345(2,1)*x(4,0)-DRR2345(6,2)*x(1,0))*I3
c 2 3 4
       E(20,3)=((R2(1)*Y134+R2(2)*Y234+R2(3)*Y(3,4,3)+R2(4)*Y(4,3,4)+
     \R3(1)*Y124+R3(2)*Y(2,4,2)+R3(3)*Y324+R3(4)*Y(4,2,4)+R4(1)*Y123+
     \R4(2)*Y(2,3,2)+R4(3)*Y(3,2,3)+R4(4)*Y423)*2.d0+R23(1)*x(1,4)+
     \R23(2)*x(2,4)+R23(3)*x(3,4)+R23(4)*x(4,4)+R24(1)*x(1,3)+
     \R24(2)*x(2,3)+R24(3)*x(3,3)+R24(4)*x(4,3)+R34(1)*x(1,2)+
     \R34(2)*x(2,2)+R34(3)*x(3,2)+R34(4)*x(4,2)-DRR2345(4,2)*x(4,0) 
     \-DRR2345(5,2)*x(3,0)-DRR2345(6,2)*x(2,0))*I3

C Rank 4
c$$$CCCCCCCCCCCCC
       R(1)=DRR1345(13,1) - DRR2345(13,1)
       R(2)=DRR1245(13,1) - DRR1345(13,1)
       R(3)=DRR1235(13,1) - DRR1245(13,1)
       R(4)=DRR1234(13,1) - DRR1235(13,1)
c 0 0 0 0
      E(46,4)=(R(1)*x(1,0)+R(2)*x(2,0)+R(3)*x(3,0)+R(4)*x(4,0))*Id
c
      R111(1)=(DRR1345(1,3) +       DR2345)
      R111(2)=(DRR1245(1,3) - DRR1345(1,3)) 
      R111(3)=(DRR1235(1,3) - DRR1245(1,3)) 
      R111(4)=(DRR1234(1,3) - DRR1235(1,3))  
   
      R11(1)=(DRR1345(7,1) -  DRR2345(7,2))
      R11(2)=(DRR1245(7,1) -  DRR1345(7,1)) 
      R11(3)=(DRR1235(7,1) -  DRR1245(7,1)) 
      R11(4)=(DRR1234(7,1) -  DRR1235(7,1))
c 1 1 1 1
      E(1,4)=((R11(2)*Y(2,1,1)+R11(3)*Y(3,1,1)+R11(4)*Y(4,1,1))*6.d0+
     \(R111(1)*x(1,1)+R111(2)*x(2,1)+R111(3)*x(3,1)+R111(4)*x(4,1))+
     \DR2345*x(1,0))*Id
c 0 0 1 1
      E(36,4)=((R11(1)*x(1,0)+R11(2)*x(2,0)+R11(3)*x(3,0)+R11(4)*x(4,0)
     \+R1(1)*x(1,1)+R1(2)*x(2,1)+R1(3)*x(3,1)+R1(4)*x(4,1)+
     \ DRR2345(7,2)*x(1,0))*2.d0+(R(2)*Y(2,1,1)+R(3)*Y(3,1,1)+
     \ R(4)*Y(4,1,1))*4.d0)*I4
c
      R12(1)=(DRR1345(7,1)  +  DRR2345(11,3))
      R12(2)=(DRR1245(10,1) -  DRR1345(7,1)) 
      R12(3)=(DRR1235(10,1) -  DRR1245(10,1)) 
      R12(4)=(DRR1234(10,1) -  DRR1235(10,1))
  
      R112(1)=(DRR1345(1,3) - DRR2345(1,1))
      R112(2)=(DRR1245(4,3) - DRR1345(1,3)) 
      R112(3)=(DRR1235(4,3) - DRR1245(4,3)) 
      R112(4)=(DRR1234(4,3) - DRR1235(4,3))     
c 1 1 1 2 
      E(5,4)=((R12(2)*Y(2,1,1)+R12(3)*Y(3,1,1)+R12(4)*Y(4,1,1))*12.d0+
     \(R11(1)*Y(1,2,1)+R11(2)*Y(2,1,2)+R11(3)*Y312+R11(4)*Y412)*6.d0+
     \(R112(1)*x(1,1)+R112(2)*x(2,1)+R112(3)*x(3,1)+R112(4)*x(4,1))*3.d0
     \+R111(1)*x(1,2)+R111(2)*x(2,2)+R111(3)*x(3,2)+R111(4)*x(4,2)
     \-DRR2345(1,1)*x(1,0)*3.d0 + DR2345*x(2,0))*I4
c 0 0 1 2
      E(40,4)=((R12(1)*x(1,0)+R12(2)*x(2,0)+R12(3)*x(3,0)+R12(4)*x(4,0)+
     \R(1)*Y(1,2,1)+R(2)*Y(2,1,2)+R(3)*Y312+R(4)*Y412)*2.d0+R1(1)*x(1,2)
     \+R1(2)*x(2,2)+R1(3)*x(3,2)+R1(4)*x(4,2)+R2(1)*x(1,1)+R2(2)*x(2,1)+
     \R2(3)*x(3,1)+R2(4)*x(4,1)+DRR2345(7,2)*x(2,0)
     \-DRR2345(11,3)*x(1,0))*I4
c
      R13(1)=(DRR1345(10,1) + DRR2345(12,3))
      R13(2)=(DRR1245(10,1) -  DRR1345(10,1)) 
      R13(3)=(DRR1235(11,1) -  DRR1245(10,1)) 
      R13(4)=(DRR1234(11,1) -  DRR1235(11,1))

      R113(1)=(DRR1345(4,3) - DRR2345(2,1))
      R113(2)=(DRR1245(4,3) - DRR1345(4,3)) 
      R113(3)=(DRR1235(5,3) - DRR1245(4,3)) 
      R113(4)=(DRR1234(5,3) - DRR1235(5,3))
c 1 1 1 3 
      E(6,4)=((R13(2)*Y(2,1,1)+R13(3)*Y(3,1,1)+R13(4)*Y(4,1,1))*12.d0+
     \(R11(1)*Y(1,3,1)+R11(2)*Y213+R11(3)*Y(3,1,3)+R11(4)*Y413)*6.d0+
     \(R113(1)*x(1,1)+R113(2)*x(2,1)+R113(3)*x(3,1)+R113(4)*x(4,1))*3.d0
     \+R111(1)*x(1,3)+R111(2)*x(2,3)+R111(3)*x(3,3)+R111(4)*x(4,3)
     \-DRR2345(2,1)*x(1,0)*3.d0 + DR2345*x(3,0))*I4
c 0 0 1 3
      E(41,4)=((R13(1)*x(1,0)+R13(2)*x(2,0)+R13(3)*x(3,0)+R13(4)*x(4,0)+
     \R(1)*Y(1,3,1)+R(2)*Y213+R(3)*Y(3,1,3)+R(4)*Y413)*2.d0+R1(1)*x(1,3)
     \+R1(2)*x(2,3)+R1(3)*x(3,3)+R1(4)*x(4,3)+R3(1)*x(1,1)+R3(2)*x(2,1)+
     \R3(3)*x(3,1)+R3(4)*x(4,1)+DRR2345(7,2)*x(3,0)
     \ -DRR2345(12,3)*x(1,0))*I4
c     
      R14(1)=(DRR1345(11,1)  +  DRR2345(13,3))
      R14(2)=(DRR1245(11,1)  -  DRR1345(11,1)) 
      R14(3)=(DRR1235(11,1)  -  DRR1245(11,1)) 
      R14(4)=(                -  DRR1235(11,1))
     
      R114(1)=(DRR1345(5,3) - DRR2345(3,1))
      R114(2)=(DRR1245(5,3) - DRR1345(5,3)) 
      R114(3)=(DRR1235(5,3) - DRR1245(5,3)) 
      R114(4)=(              - DRR1235(5,3))
c 1 1 1 4 
      E(7,4)=((R14(2)*Y(2,1,1)+R14(3)*Y(3,1,1)+R14(4)*Y(4,1,1))*12.d0+
     \ (R11(1)*Y(1,4,1)+R11(2)*Y214+R11(3)*Y314+R11(4)*Y(4,1,4))*6.d0+
     \(R114(1)*x(1,1)+R114(2)*x(2,1)+R114(3)*x(3,1)+R114(4)*x(4,1))*3.d0
     \+R111(1)*x(1,4)+R111(2)*x(2,4)+R111(3)*x(3,4)+R111(4)*x(4,4)
     \-DRR2345(3,1)*x(1,0)*3.d0 + DR2345*x(4,0))*I4
c 0 0 1 4
      E(42,4)=((R14(1)*x(1,0)+R14(2)*x(2,0)+R14(3)*x(3,0)+R14(4)*x(4,0)+
     \R(1)*Y(1,4,1)+R(2)*Y214+R(3)*Y314+R(4)*Y(4,1,4))*2.d0+R1(1)*x(1,4)
     \+R1(2)*x(2,4)+R1(3)*x(3,4)+R1(4)*x(4,4)+R4(1)*x(1,1)+R4(2)*x(2,1)+
     \ R4(3)*x(3,1)+R4(4)*x(4,1)+DRR2345(7,2)*x(4,0)
     \ -DRR2345(13,3)*x(1,0))*I4
c
      R22(1)=(DRR1345(7,1) -  DRR2345(7,1))
      R22(2)=(DRR1245(8,1) -  DRR1345(7,1)) 
      R22(3)=(DRR1235(8,1) -  DRR1245(8,1)) 
      R22(4)=(DRR1234(8,1) -  DRR1235(8,1))

      R222(1)=(DRR1345(1,3) - DRR2345(1,3))
      R222(2)=(DRR1245(2,3) - DRR1345(1,3)) 
      R222(3)=(DRR1235(2,3) - DRR1245(2,3)) 
      R222(4)=(DRR1234(2,3) - DRR1235(2,3))
c 2 2 2 2
      E(2,4)=((R22(1)*Y(1,2,2)+R22(3)*Y(3,2,2)+R22(4)*Y(4,2,2))*6.d0+
     \R222(1)*x(1,2)+R222(2)*x(2,2)+R222(3)*x(3,2)+R222(4)*x(4,2)
     \-DRR2345(1,3)*x(2,0))*Id
c 0 0 2 2
      E(37,4)=((R22(1)*x(1,0)+R22(2)*x(2,0)+R22(3)*x(3,0)+
     - R22(4)*x(4,0)+ R2(1)*x(1,2)+R2(2)*x(2,2)+R2(3)*x(3,2)+
     - R2(4)*x(4,2)- DRR2345(11,3)*x(2,0))*2.d0+(R(1)*Y(1,2,2)+ 
     - R(3)*Y(3,2,2)+R(4)*Y(4,2,2))*4.d0)*I4
c
      R221(1)=(DRR1345(1,3) + DRR2345(1,2))
      R221(2)=(DRR1245(6,3) - DRR1345(1,3)) 
      R221(3)=(DRR1235(6,3) - DRR1245(6,3)) 
      R221(4)=(DRR1234(6,3) - DRR1235(6,3))
c 2 2 2 1 
      E(8,4)=((R12(1)*Y(1,2,2)+R12(3)*Y(3,2,2)+R12(4)*Y(4,2,2))*12.d0+
     \ (R22(1)*Y(1,2,1)+R22(2)*Y(2,1,2)+R22(3)*Y312+R22(4)*Y412)*6.d0+
     \R222(1)*x(1,1)+R222(2)*x(2,1)+R222(3)*x(3,1)+R222(4)*x(4,1)+
     \(R221(1)*x(1,2)+R221(2)*x(2,2)+R221(3)*x(3,2)+R221(4)*x(4,2))*3.d0
     \-DRR2345(1,3)*x(1,0)+DRR2345(1,2)*x(2,0)*3.d0)*I4
c 1 1 2 2 
      E(17,4)=((R11(1)*Y(1,2,2)+R11(3)*Y(3,2,2)+R11(4)*Y(4,2,2)+
     \R22(2)*Y(2,1,1)+R22(3)*Y(3,1,1)+R22(4)*Y(4,1,1))*4.d0+
     \ (R12(1)*Y(1,2,1)+R12(2)*Y(2,1,2)+R12(3)*Y312+R12(4)*Y412)*8.d0
     \+(R112(1)*x(1,2)+R112(2)*x(2,2)+R112(3)*x(3,2)+R112(4)*x(4,2)
     \+R221(1)*x(1,1)+R221(2)*x(2,1)+R221(3)*x(3,1)+R221(4)*x(4,1) 
     \ +DRR2345(1,2)*x(1,0)- DRR2345(1,1)*x(2,0))*2.d0)*I4

c
      R23(1)=(DRR1345(10,1) -  DRR2345(10,1))
      R23(2)=(DRR1245(8,1)  -  DRR1345(10,1)) 
      R23(3)=(DRR1235(12,1) -  DRR1245(8,1)) 
      R23(4)=(DRR1234(12,1) -  DRR1235(12,1))

      R223(1)=(DRR1345(4,3) - DRR2345(4,3))
      R223(2)=(DRR1245(2,3) - DRR1345(4,3)) 
      R223(3)=(DRR1235(8,3) - DRR1245(2,3)) 
      R223(4)=(DRR1234(8,3) - DRR1235(8,3))
c 2 2 2 3 
      E(9,4)=((R23(1)*Y(1,2,2)+R23(3)*Y(3,2,2)+R23(4)*Y(4,2,2))*12.d0+
     \(R22(1)*Y123+R22(2)*(Y(2,3,2))+R22(3)*Y(3,2,3)+R22(4)*Y423)*6.d0+
     \R222(1)*x(1,3)+R222(2)*x(2,3)+R222(3)*x(3,3)+R222(4)*x(4,3)+
     \(R223(1)*x(1,2)+R223(2)*x(2,2)+R223(3)*x(3,2)+R223(4)*x(4,2))*3.d0
     \-DRR2345(1,3)*x(3,0)-DRR2345(4,3)*x(2,0)*3.d0)*I4
c 0 0 2 3
      E(43,4)=((R23(1)*x(1,0)+R23(2)*x(2,0)+R23(3)*x(3,0)+R23(4)*x(4,0)+
     \R(1)*Y123+R(2)*Y(2,3,2)+R(3)*Y(3,2,3)+R(4)*Y423)*2.d0+R3(1)*x(1,2)
     \+R3(2)*x(2,2)+R3(3)*x(3,2)+R3(4)*x(4,2)+R2(1)*x(1,3)+R2(2)*x(2,3)+
     \R2(3)*x(3,3)+R2(4)*x(4,3)-DRR2345(12,3)*x(2,0)
     \ -DRR2345(11,3)*x(3,0))*I4
c
      R24(1)=(DRR1345(11,1) -  DRR2345(11,1))
      R24(2)=(DRR1245(12,1) -  DRR1345(11,1)) 
      R24(3)=(DRR1235(12,1) -  DRR1245(12,1)) 
      R24(4)=(               -  DRR1235(12,1))

      R224(1)=(DRR1345(5,3) - DRR2345(5,3))
      R224(2)=(DRR1245(8,3) - DRR1345(5,3)) 
      R224(3)=(DRR1235(8,3) - DRR1245(8,3)) 
      R224(4)=(              - DRR1235(8,3))
c 2 2 2 4 
      E(10,4)=((R24(1)*Y(1,2,2)+R24(3)*Y(3,2,2)+R24(4)*Y(4,2,2))*12.d0+
     \ (R22(1)*Y124+R22(2)*(Y(2,4,2))+R22(3)*Y324+R22(4)*Y(4,2,4))*6.d0+
     \ R222(1)*x(1,4)+R222(2)*x(2,4)+R222(3)*x(3,4)+R222(4)*x(4,4)+
     \(R224(1)*x(1,2)+R224(2)*x(2,2)+R224(3)*x(3,2)+R224(4)*x(4,2))*3.d0
     \ -DRR2345(1,3)*x(4,0)-DRR2345(5,3)*x(2,0)*3.d0)*I4
c 0 0 2 4
      E(44,4)=((R24(1)*x(1,0)+R24(2)*x(2,0)+R24(3)*x(3,0)+R24(4)*x(4,0)+
     \R(1)*Y124+R(2)*Y(2,4,2)+R(3)*Y324+R(4)*Y(4,2,4))*2.d0+R4(1)*x(1,2)
     \+R4(2)*x(2,2)+R4(3)*x(3,2)+R4(4)*x(4,2)+R2(1)*x(1,4)+R2(2)*x(2,4)+
     \R2(3)*x(3,4)+R2(4)*x(4,4)-DRR2345(13,3)*x(2,0)
     \-DRR2345(11,3)*x(4,0))*I4
c
      R33(1)=(DRR1345(8,1) -  DRR2345(8,1))
      R33(2)=(DRR1245(8,1) -  DRR1345(8,1)) 
      R33(3)=(DRR1235(9,1) -  DRR1245(8,1)) 
      R33(4)=(DRR1234(9,1) -  DRR1235(9,1))

      R333(1)=(DRR1345(2,3) - DRR2345(2,3))
      R333(2)=(DRR1245(2,3) - DRR1345(2,3)) 
      R333(3)=(DRR1235(3,3) - DRR1245(2,3)) 
      R333(4)=(DRR1234(3,3) - DRR1235(3,3))
c 3 3 3 3
      E(3,4)=((R33(1)*Y(1,3,3)+R33(2)*Y(2,3,3)+R33(4)*Y(4,3,3))*6.d0+
     \R333(1)*x(1,3)+R333(2)*x(2,3)+R333(3)*x(3,3)+R333(4)*x(4,3)
     \-DRR2345(2,3)*x(3,0))*Id
c 0 0 3 3
      E(38,4)=((R33(1)*x(1,0)+R33(2)*x(2,0)+R33(3)*x(3,0)+
     - R33(4)*x(4,0)+R3(1)*x(1,3)+R3(2)*x(2,3)+R3(3)*x(3,3)+
     - R3(4)*x(4,3)-DRR2345(12,3)*x(3,0))*2.d0+(R(1)*Y(1,3,3)+
     - R(2)*Y(2,3,3)+R(4)*Y(4,3,3))*4.d0)*I4
c
      R331(1)=(DRR1345(6,3) + DRR2345(2,2))
      R331(2)=(DRR1245(6,3) - DRR1345(6,3)) 
      R331(3)=(DRR1235(7,3) - DRR1245(6,3)) 
      R331(4)=(DRR1234(7,3) - DRR1235(7,3))
c 3 3 3 1 
      E(11,4)=((R13(1)*Y(1,3,3)+R13(2)*Y(2,3,3)+R13(4)*Y(4,3,3))*12.d0+
     \(R33(1)*Y(1,3,1)+R33(2)*Y213+R33(3)*Y(3,1,3)+R33(4)*Y413)*6.d0+
     \ R333(1)*x(1,1)+R333(2)*x(2,1)+R333(3)*x(3,1)+R333(4)*x(4,1)+
     \(R331(1)*x(1,3)+R331(2)*x(2,3)+R331(3)*x(3,3)+R331(4)*x(4,3))*3.d0
     \-DRR2345(2,3)*x(1,0) + DRR2345(2,2)*x(3,0)*3.d0)*I4
c 1 1 3 3
      E(18,4)=((R11(1)*Y(1,3,3)+R11(2)*Y(2,3,3)+R11(4)*Y(4,3,3)+
     \R33(2)*Y(2,1,1)+R33(3)*Y(3,1,1)+R33(4)*Y(4,1,1))*4.d0+
     \ (R13(1)*Y(1,3,1)+R13(2)*Y213+R13(3)*Y(3,1,3)+R13(4)*Y413)*8.d0
     \ +(R113(1)*x(1,3)+R113(2)*x(2,3)+R113(3)*x(3,3)+R113(4)*x(4,3)+
     \ R331(1)*x(1,1)+R331(2)*x(2,1)+R331(3)*x(3,1)+R331(4)*x(4,1)
     \ +DRR2345(2,2)*x(1,0)-DRR2345(2,1)*x(3,0))*2.d0)*I4

      R332(1)=(DRR1345(6,3) - DRR2345(6,3))
      R332(2)=(DRR1245(2,3) - DRR1345(6,3)) 
      R332(3)=(DRR1235(9,3) - DRR1245(2,3)) 
      R332(4)=(DRR1234(9,3) - DRR1235(9,3))
c 3 3 3 2 
      E(12,4)=((R23(1)*Y(1,3,3)+R23(2)*Y(2,3,3)+R23(4)*Y(4,3,3))*12.d0+
     \(R33(1)*Y123+R33(2)*Y(2,3,2)+R33(3)*Y(3,2,3)+R33(4)*Y423)*6.d0
     \+R333(1)*x(1,2)+R333(2)*x(2,2)+R333(3)*x(3,2)+R333(4)*x(4,2)+
     \(R332(1)*x(1,3)+R332(2)*x(2,3)+R332(3)*x(3,3)+R332(4)*x(4,3))*3.d0
     \-DRR2345(2,3)*x(2,0) - DRR2345(6,3)*x(3,0)*3.d0)*I4
c 2 2 3 3
      E(20,4)=((R22(1)*Y(1,3,3)+R22(2)*Y(2,3,3)+R22(4)*Y(4,3,3)+
     \ R33(1)*Y(1,2,2)+R33(3)*Y(3,2,2)+R33(4)*Y(4,2,2))*4.d0
     \ +(R23(1)*Y123+R23(2)*Y(2,3,2)+R23(3)*Y(3,2,3)+R23(4)*Y423)*8.d0
     \ +(R223(1)*x(1,3)+R223(2)*x(2,3)+R223(3)*x(3,3)+R223(4)*x(4,3)
     \+R332(1)*x(1,2)+R332(2)*x(2,2)+R332(3)*x(3,2)+R332(4)*x(4,2)
     \ -DRR2345(6,3)*x(2,0)-DRR2345(4,3)*x(3,0))*2.d0)*I4
c
      R34(1)=(DRR1345(12,1) -  DRR2345(12,1))
      R34(2)=(DRR1245(12,1) -  DRR1345(12,1)) 
      R34(3)=(DRR1235(9,1)  -  DRR1245(12,1)) 
      R34(4)=(               -  DRR1235(9,1))

      R334(1)=(DRR1345(8,3) - DRR2345(8,3))
      R334(2)=(DRR1245(8,3) - DRR1345(8,3)) 
      R334(3)=(DRR1235(3,3) - DRR1245(8,3)) 
      R334(4)=(              - DRR1235(3,3))
c 3 3 3 4
      E(13,4)=((R34(1)*Y(1,3,3)+R34(2)*Y(2,3,3)+R34(4)*Y(4,3,3))*12.d0+
     \(R33(1)*Y134+R33(2)*Y234+R33(3)*Y(3,4,3)+R33(4)*Y(4,3,4))*6.d0
     \+R333(1)*x(1,4)+R333(2)*x(2,4)+R333(3)*x(3,4)+R333(4)*x(4,4)+
     \(R334(1)*x(1,3)+R334(2)*x(2,3)+R334(3)*x(3,3)+R334(4)*x(4,3))*3.d0
     \-DRR2345(2,3)*x(4,0) - DRR2345(8,3)*x(3,0)*3.d0)*I4
c 0 0 3 4
      E(45,4)=((R34(1)*x(1,0)+R34(2)*x(2,0)+R34(3)*x(3,0)+R34(4)*x(4,0)+
     \R(1)*Y134+R(2)*Y234+R(3)*Y(3,4,3)+R(4)*Y(4,3,4))*2.d0+R4(1)*x(1,3)
     \+R4(2)*x(2,3)+R4(3)*x(3,3)+R4(4)*x(4,3)+R3(1)*x(1,4)+R3(2)*x(2,4)
     \+R3(3)*x(3,4)+R3(4)*x(4,4)-DRR2345(13,3)*x(3,0)
     \ -DRR2345(12,3)*x(4,0))*I4
c  
      R44(1)=(DRR1345(9,1) -  DRR2345(9,1))
      R44(2)=(DRR1245(9,1) -  DRR1345(9,1)) 
      R44(3)=(DRR1235(9,1) -  DRR1245(9,1)) 
      R44(4)=(              -  DRR1235(9,1))

      R444(1)=(DRR1345(3,3) - DRR2345(3,3))
      R444(2)=(DRR1245(3,3) - DRR1345(3,3)) 
      R444(3)=(DRR1235(3,3) - DRR1245(3,3)) 
      R444(4)=(              - DRR1235(3,3))
c 4 4 4 4
      E(4,4)=((R44(1)*Y(1,4,4)+R44(2)*Y(2,4,4)+R44(3)*Y(3,4,4))*6.d0+
     \R444(1)*x(1,4)+R444(2)*x(2,4)+R444(3)*x(3,4)+R444(4)*x(4,4)
     \-DRR2345(3,3)*x(4,0))*Id
c 0 0 4 4
      E(39,4)=((R44(1)*x(1,0)+R44(2)*x(2,0)+R44(3)*x(3,0)+
     - R44(4)*x(4,0)+R4(1)*x(1,4)+R4(2)*x(2,4)+R4(3)*x(3,4)+
     - R4(4)*x(4,4)-DRR2345(13,3)*x(4,0))*2.d0+(R(1)*Y(1,4,4)+
     - R(2)*Y(2,4,4)+R(3)*Y(3,4,4))*4.d0)*I4    
c
      R441(1)=(DRR1345(7,3) + DRR2345(3,2))
      R441(2)=(DRR1245(7,3) - DRR1345(7,3)) 
      R441(3)=(DRR1235(7,3) - DRR1245(7,3)) 
      R441(4)=(              - DRR1235(7,3))
c 4 4 4 1 
      E(14,4)=((R14(1)*Y(1,4,4)+R14(2)*Y(2,4,4)+R14(3)*Y(3,4,4))*12.d0+
     \(R44(1)*Y(1,4,1)+R44(2)*Y214+R44(3)*Y314+R44(4)*Y(4,1,4))*6.d0+
     \R444(1)*x(1,1)+R444(2)*x(2,1)+R444(3)*x(3,1)+R444(4)*x(4,1)+
     \(R441(1)*x(1,4)+R441(2)*x(2,4)+R441(3)*x(3,4)+R441(4)*x(4,4))*3.d0
     \-DRR2345(3,3)*x(1,0) + DRR2345(3,2)*x(4,0)*3.d0)*I4     
c 1 1 4 4
      E(19,4)=((R11(1)*Y(1,4,4)+R11(2)*Y(2,4,4)+R11(3)*Y(3,4,4)+
     \ R44(2)*Y(2,1,1)+R44(3)*Y(3,1,1)+R44(4)*Y(4,1,1))*4.d0+
     \ (R14(1)*(Y(1,4,1))+R14(2)*Y214+R14(3)*Y314+R14(4)*Y(4,1,4))*8.d0
     \+(R114(1)*x(1,4)+R114(2)*x(2,4)+R114(3)*x(3,4)+R114(4)*x(4,4)+
     \R441(1)*x(1,1)+R441(2)*x(2,1)+R441(3)*x(3,1)+R441(4)*x(4,1)
     \+DRR2345(3,2)*x(1,0)-DRR2345(3,1)*x(4,0))*2.d0)*I4

      R442(1)=(DRR1345(7,3) - DRR2345(7,3))
      R442(2)=(DRR1245(9,3) - DRR1345(7,3)) 
      R442(3)=(DRR1235(9,3) - DRR1245(9,3)) 
      R442(4)=(              - DRR1235(9,3))
c 4 4 4 2 
      E(15,4)=((R24(1)*Y(1,4,4)+R24(2)*Y(2,4,4)+R24(3)*Y(3,4,4))*12.d0+
     \ (R44(1)*Y124+R44(2)*Y(2,4,2)+R44(3)*Y324+R44(4)*Y(4,2,4))*6.d0+
     \ R444(1)*x(1,2)+R444(2)*x(2,2)+R444(3)*x(3,2)+R444(4)*x(4,2)+
     \(R442(1)*x(1,4)+R442(2)*x(2,4)+R442(3)*x(3,4)+R442(4)*x(4,4))*3.d0
     \ -DRR2345(3,3)*x(2,0) - DRR2345(7,3)*x(4,0)*3.d0)*I4
c 2 2 4 4
      E(21,4)=((R22(1)*Y(1,4,4)+R22(2)*Y(2,4,4)+R22(3)*Y(3,4,4)+
     \R44(1)*Y(1,2,2)+R44(3)*Y(3,2,2)+R44(4)*Y(4,2,2))*4.d0+
     \ (R24(1)*Y124+R24(2)*Y(2,4,2)+R24(3)*Y324+R24(4)*Y(4,2,4))*8.d0
     \+(R224(1)*x(1,4)+R224(2)*x(2,4)+R224(3)*x(3,4)+R224(4)*x(4,4)+
     \R442(1)*x(1,2)+R442(2)*x(2,2)+R442(3)*x(3,2)+R442(4)*x(4,2)
     \ -DRR2345(7,3)*x(2,0)-DRR2345(5,3)*x(4,0))*2.d0)*I4

      R443(1)=(DRR1345(9,3) - DRR2345(9,3))
      R443(2)=(DRR1245(9,3) - DRR1345(9,3)) 
      R443(3)=(DRR1235(3,3) - DRR1245(9,3)) 
      R443(4)=(              - DRR1235(3,3))

c 4 4 4 3 
      E(16,4)=((R34(1)*Y(1,4,4)+R34(2)*Y(2,4,4)+R34(3)*Y(3,4,4))*12.d0+
     \ (R44(1)*Y134+R44(2)*Y234+R44(3)*Y(3,4,3)+R44(4)*Y(4,3,4))*6.d0+
     \ R444(1)*x(1,3)+R444(2)*x(2,3)+R444(3)*x(3,3)+R444(4)*x(4,3)+
     \(R443(1)*x(1,4)+R443(2)*x(2,4)+R443(3)*x(3,4)+R443(4)*x(4,4))*3.d0
     \ -DRR2345(3,3)*x(3,0) - DRR2345(9,3)*x(4,0)*3.d0)*I4
c 3 3 4 4
      E(22,4)=((R33(1)*Y(1,4,4)+R33(2)*Y(2,4,4)+R33(3)*Y(3,4,4)+
     \R44(1)*Y(1,3,3)+R44(2)*Y(2,3,3)+R44(4)*Y(4,3,3))*4.d0+
     \ (R34(1)*Y134+R34(2)*Y234+R34(3)*Y(3,4,3)+R34(4)*Y(4,3,4))*8.d0
     \+(R334(1)*x(1,4)+R334(2)*x(2,4)+R334(3)*x(3,4)+R334(4)*x(4,4)+
     \ R443(1)*x(1,3)+R443(2)*x(2,3)+R443(3)*x(3,3)+R443(4)*x(4,3)
     \ -DRR2345(9,3)*x(3,0)-DRR2345(8,3)*x(4,0))*2.d0)*I4

      R123(1)=(DRR1345(4,3) + DRR2345(4,2))
      R123(2)=(DRR1245(6,3) - DRR1345(4,3)) 
      R123(3)=(DRR1235(10,3)- DRR1245(6,3))      
      R123(4)=(DRR1234(10,3)- DRR1235(10,3))
c 1 1 2 3
      E(23,4)=((R23(2)*Y(2,1,1)+R23(3)*Y(3,1,1)+R23(4)*Y(4,1,1)+
     \ R13(1)*Y(1,2,1)+R13(2)*Y(2,1,2)+R13(3)*Y312+R13(4)*Y412+
     \ R12(1)*Y(1,3,1)+R12(2)*Y213+R12(3)*Y(3,1,3)+R12(4)*Y413)*4.d0
     \+(R11(1)*Y123+R11(2)*Y(2,3,2)+R11(3)*Y(3,2,3)+R11(4)*Y423)*2.d0+
     \R112(1)*x(1,3)+R112(2)*x(2,3)+R112(3)*x(3,3)+R112(4)*x(4,3)+
     \R113(1)*x(1,2)+R113(2)*x(2,2)+R113(3)*x(3,2)+R113(4)*x(4,2)+
     \(R123(1)*x(1,1)+R123(2)*x(2,1)+R123(3)*x(3,1)+R123(4)*x(4,1)
     \+DRR2345(4,2)*x(1,0))*2.d0-DRR2345(2,1)*x(2,0)
     \-DRR2345(1,1)*x(3,0))*I4

c 1 2 2 3
      E(26,4)=((R13(1)*Y(1,2,2)+R13(3)*Y(3,2,2)+R13(4)*Y(4,2,2)+
     \ R23(1)*Y(1,2,1)+R23(2)*Y(2,1,2)+R23(3)*Y312+R23(4)*Y412+
     \ R12(1)*Y123+R12(2)*Y(2,3,2)+R12(3)*Y(3,2,3)+R12(4)*Y423)*4.d0
     \+(R22(1)*(Y(1,3,1))+R22(2)*Y213+R22(3)*Y(3,1,3)+R22(4)*Y413)*2.d0+
     \R221(1)*x(1,3)+R221(2)*x(2,3)+R221(3)*x(3,3)+R221(4)*x(4,3)+
     \R223(1)*x(1,1)+R223(2)*x(2,1)+R223(3)*x(3,1)+R223(4)*x(4,1)+
     \(R123(1)*x(1,2)+R123(2)*x(2,2)+R123(3)*x(3,2)+R123(4)*x(4,2)
     \ +DRR2345(4,2)*x(2,0))*2.d0-DRR2345(4,3)*x(1,0)
     \ +DRR2345(1,2)*x(3,0))*I4
c 1 2 3 3
      E(29,4)=((R12(1)*Y(1,3,3)+R12(2)*Y(2,3,3)+R12(4)*Y(4,3,3)+
     \ R23(1)*Y(1,3,1)+R23(2)*Y213+R23(3)*Y(3,1,3)+R23(4)*Y413+
     \ R13(1)*Y123+R13(2)*Y(2,3,2)+R13(3)*Y(3,2,3)+ R13(4)*Y423)*4.d0
     \+(R33(1)*Y(1,2,1)+R33(2)*Y(2,1,2)+R33(3)*Y312+R33(4)*Y412)*2.d0
     \+R331(1)*x(1,2)+R331(2)*x(2,2)+R331(3)*x(3,2)+R331(4)*x(4,2)+
     \ R332(1)*x(1,1)+R332(2)*x(2,1)+R332(3)*x(3,1)+R332(4)*x(4,1)+
     \ (R123(1)*x(1,3)+ R123(2)*x(2,3)+R123(3)*x(3,3)+R123(4)*x(4,3)
     \ +DRR2345(4,2)*x(3,0))*2.d0 - DRR2345(6,3)*x(1,0)
     \ +DRR2345(2,2)*x(2,0))*I4
c
      R124(1)=(DRR1345(5,3)  + DRR2345(5,2))
      R124(2)=(DRR1245(10,3) - DRR1345(5,3)) 
      R124(3)=(DRR1235(10,3) - DRR1245(10,3))      
      R124(4)=(               - DRR1235(10,3))
c 1 1 2 4
      E(24,4)=((R24(2)*Y(2,1,1)+R24(3)*Y(3,1,1)+R24(4)*Y(4,1,1)+
     \ R14(1)*Y(1,2,1)+R14(2)*Y(2,1,2)+R14(3)*Y312+R14(4)*Y412+
     \ R12(1)*Y(1,4,1)+R12(2)*Y214+R12(3)*Y314+R12(4)*Y(4,1,4))*4.d0
     \+(R11(1)*Y124+R11(2)*Y(2,4,2)+R11(3)*Y324+R11(4)*Y(4,2,4))*2.d0+
     \R112(1)*x(1,4)+R112(2)*x(2,4)+R112(3)*x(3,4)+R112(4)*x(4,4)+
     \R114(1)*x(1,2)+R114(2)*x(2,2)+R114(3)*x(3,2)+R114(4)*x(4,2)+
     \(R124(1)*x(1,1)+R124(2)*x(2,1)+R124(3)*x(3,1)+R124(4)*x(4,1)
     \ +DRR2345(5,2)*x(1,0))*2.d0-DRR2345(3,1)*x(2,0)
     \ -DRR2345(1,1)*x(4,0))*I4
c 1 2 2 4
      E(27,4)=((R14(1)*Y(1,2,2)+R14(3)*Y(3,2,2)+R14(4)*Y(4,2,2)+
     \ R24(1)*Y(1,2,1)+R24(2)*Y(2,1,2)+R24(3)*Y312+R24(4)*Y412+
     \ R12(1)*Y124+R12(2)*Y(2,4,2)+R12(3)*Y324+R12(4)*Y(4,2,4))*4.d0
     \+(R22(1)*Y(1,4,1)+R22(2)*Y214+R22(3)*Y314+R22(4)*Y(4,1,4))*2.d0
     \ +R221(1)*x(1,4)+R221(2)*x(2,4)+R221(3)*x(3,4)+R221(4)*x(4,4)+
     \ R224(1)*x(1,1)+R224(2)*x(2,1)+R224(3)*x(3,1)+R224(4)*x(4,1)+
     \ (R124(1)*x(1,2)+R124(2)*x(2,2)+R124(3)*x(3,2)+R124(4)*x(4,2)
     \ +DRR2345(5,2)*x(2,0))*2.d0-DRR2345(5,3)*x(1,0)
     \ +DRR2345(1,2)*x(4,0))*I4
c 1 2 4 4
      E(32,4)=((R12(1)*Y(1,4,4)+R12(2)*Y(2,4,4)+R12(3)*Y(3,4,4)
     \ +R24(1)*Y(1,4,1)+R24(2)*Y214+R24(3)*Y314+R24(4)*Y(4,1,4)+
     \ R14(1)*Y124+R14(2)*Y(2,4,2)+R14(3)*Y324+R14(4)*Y(4,2,4))*4.d0+
     \(R44(1)*Y(1,2,1)+R44(2)*Y(2,1,2)+R44(3)*Y312+R44(4)*Y412)*2.d0
     \ +R441(1)*x(1,2)+R441(2)*x(2,2)+R441(3)*x(3,2)+R441(4)*x(4,2)+
     \ R442(1)*x(1,1)+R442(2)*x(2,1)+R442(3)*x(3,1)+R442(4)*x(4,1)+
     \ (R124(1)*x(1,4)+R124(2)*x(2,4)+R124(3)*x(3,4)+R124(4)*x(4,4)
     \ +DRR2345(5,2)*x(4,0))*2.d0-DRR2345(7,3)*x(1,0)
     \ +DRR2345(3,2)*x(2,0))*I4

      R134(1)=(DRR1345(10,3) + DRR2345(6,2))
      R134(2)=(DRR1245(10,3) - DRR1345(10,3)) 
      R134(3)=(DRR1235(7,3)  - DRR1245(10,3))      
      R134(4)=(               - DRR1235(7,3))
c 1 1 3 4
      E(25,4)=((R34(2)*Y(2,1,1)+R34(3)*Y(3,1,1)+R34(4)*Y(4,1,1)+
     \ R14(1)*Y(1,3,1)+R14(2)*Y213+R14(3)*Y(3,1,3)+R14(4)*Y413+
     \ R13(1)*Y(1,4,1)+R13(2)*Y214+R13(3)*Y314+R13(4)*Y(4,1,4))*4.d0
     \+(R11(1)*Y134+R11(2)*Y234+R11(3)*Y(3,4,3)+R11(4)*Y(4,3,4))*2.d0+
     \R113(1)*x(1,4)+R113(2)*x(2,4)+R113(3)*x(3,4)+R113(4)*x(4,4)+
     \ R114(1)*x(1,3)+R114(2)*x(2,3)+R114(3)*x(3,3)+R114(4)*x(4,3)+
     \ (R134(1)*x(1,1)+R134(2)*x(2,1)+R134(3)*x(3,1)+R134(4)*x(4,1)
     \ +DRR2345(6,2)*x(1,0))*2.d0-DRR2345(3,1)*x(3,0)
     \ -DRR2345(2,1)*x(4,0))*I4
c 1 3 3 4
      E(30,4)=((R14(1)*Y(1,3,3)+R14(2)*Y(2,3,3)+R14(4)*Y(4,3,3)+
     \ R34(1)*Y(1,3,1)+R34(2)*Y213+R34(3)*Y(3,1,3)+R34(4)*Y413+
     \ R13(1)*Y134+R13(2)*Y234+R13(3)*Y(3,4,3)+R13(4)*Y(4,3,4))*4.d0+
     \ (R33(1)*Y(1,4,1)+R33(2)*Y214+R33(3)*Y314+R33(4)*Y(4,1,4))*2.d0
     \ +R331(1)*x(1,4)+R331(2)*x(2,4)+R331(3)*x(3,4)+R331(4)*x(4,4)+
     \ R334(1)*x(1,1)+R334(2)*x(2,1)+R334(3)*x(3,1)+R334(4)*x(4,1)+
     \(R134(1)*x(1,3)+R134(2)*x(2,3)+R134(3)*x(3,3)+R134(4)*x(4,3) 
     \ +DRR2345(6,2)*x(3,0))*2.d0-DRR2345(8,3)*x(1,0)
     \ +DRR2345(2,2)*x(4,0))*I4
c 1 3 4 4
      E(33,4)=((R13(1)*Y(1,4,4)+R13(2)*Y(2,4,4)+R13(3)*Y(3,4,4)
     \+R34(1)*Y(1,4,1)+R34(2)*Y214+R34(3)*Y314+R34(4)*Y(4,1,4)+
     \ R14(1)*Y134+R14(2)*Y234+R14(3)*Y(3,4,3)+R14(4)*Y(4,3,4))*4.d0+
     \(R44(1)*(Y(1,3,1))+R44(2)*Y213+R44(3)*Y(3,1,3)+R44(4)*Y413)*2.d0
     \+R441(1)*x(1,3)+R441(2)*x(2,3)+R441(3)*x(3,3)+R441(4)*x(4,3)+
     \ R443(1)*x(1,1)+R443(2)*x(2,1)+R443(3)*x(3,1)+R443(4)*x(4,1)+
     \ (R134(1)*x(1,4)+R134(2)*x(2,4)+R134(3)*x(3,4)+R134(4)*x(4,4)
     \ +DRR2345(6,2)*x(4,0))*2.d0-DRR2345(9,3)*x(1,0)
     -+DRR2345(3,2)*x(3,0))*I4
c
      R234(1)=(DRR1345(10,3) - DRR2345(10,3))
      R234(2)=(DRR1245(8,3) - DRR1345(10,3)) 
      R234(3)=(DRR1235(9,3)  - DRR1245(8,3))      
      R234(4)=(               - DRR1235(9,3))
c  2 2 3 4
      E(28,4)=((R34(1)*Y(1,2,2)+R34(3)*Y(3,2,2)+R34(4)*Y(4,2,2)+
     \ R24(1)*Y123+R24(2)*Y(2,3,2)+R24(3)*Y(3,2,3)+R24(4)*Y423+
     \ R23(1)*Y124+R23(2)*Y(2,4,2)+R23(3)*Y324+R23(4)*Y(4,2,4))*4.d0
     \+(R22(1)*Y134+R22(2)*Y234+R22(3)*Y(3,4,3)+R22(4)*Y(4,3,4))*2.d0
     \+R223(1)*x(1,4)+R223(2)*x(2,4)+R223(3)*x(3,4)+R223(4)*x(4,4)+
     \R224(1)*x(1,3)+R224(2)*x(2,3)+R224(3)*x(3,3)+R224(4)*x(4,3)+
     \ (R234(1)*x(1,2)+R234(2)*x(2,2)+R234(3)*x(3,2)+R234(4)*x(4,2)
     \ -DRR2345(10,3)*x(2,0))*2.d0-DRR2345(5,3)*x(3,0)
     \ -DRR2345(4,3)*x(4,0))*I4
c 2 3 3 4
      E(31,4)=((R24(1)*Y(1,3,3)+R24(2)*Y(2,3,3)+R24(4)*Y(4,3,3)+
     \ R34(1)*Y123+R34(2)*Y(2,3,2)+R34(3)*Y(3,2,3)+R34(4)*Y423+
     \ R23(1)*Y134+R23(2)*Y234+R23(3)*Y(3,4,3)+R23(4)*Y(4,3,4))*4.d0+
     \(R33(1)*Y124+R33(2)*Y(2,4,2)+R33(3)*Y324+R33(4)*Y(4,2,4))*2.d0
     \ +R332(1)*x(1,4)+R332(2)*x(2,4)+R332(3)*x(3,4)+R332(4)*x(4,4)+
     \ R334(1)*x(1,2)+R334(2)*x(2,2)+R334(3)*x(3,2)+R334(4)*x(4,2)+
     \ (R234(1)*x(1,3)+R234(2)*x(2,3)+R234(3)*x(3,3)+R234(4)*x(4,3)
     \ -DRR2345(10,3)*x(3,0))*2.d0-DRR2345(8,3)*x(2,0)
     \ -DRR2345(6,3)*x(4,0))*I4
c 2 3 4 4
      E(34,4)=((R23(1)*Y(1,4,4)+R23(2)*Y(2,4,4)+R23(3)*Y(3,4,4)
     \+R34(1)*Y124+R34(2)*(Y(2,4,2))+R34(3)*Y324+R34(4)*Y(4,2,4)+
     \ R24(1)*Y134+R24(2)*Y234+R24(3)*Y(3,4,3)+R24(4)*Y(4,3,4))*4.d0+
     \(R44(1)*Y123+R44(2)*Y(2,3,2)+R44(3)*Y(3,2,3)+R44(4)*Y423)*2.d0
     \+R442(1)*x(1,3)+R442(2)*x(2,3)+R442(3)*x(3,3)+R442(4)*x(4,3)+
     \R443(1)*x(1,2)+R443(2)*x(2,2)+R443(3)*x(3,2)+R443(4)*x(4,2)
     \+(R234(1)*x(1,4)+R234(2)*x(2,4)+R234(3)*x(3,4)+R234(4)*x(4,4)
     \-DRR2345(10,3)*x(4,0))*2.d0-DRR2345(9,3)*x(2,0)
     \-DRR2345(7,3)*x(3,0))*I4
c 1 2 3 4
      E(35,4)=((R12(1)*Y134+R12(2)*Y234+R12(3)*Y(3,4,3)+R12(4)*Y(4,3,4)+
     \ R13(1)*Y124+R13(2)*Y(2,4,2)+R13(3)*Y324+R13(4)*Y(4,2,4)+
     \ R14(1)*Y123+R14(2)*Y(2,3,2)+R14(3)*Y(3,2,3)+R14(4)*Y423+
     \R23(1)*Y(1,4,1)+R23(2)*Y214+R23(3)*Y314+R23(4)*Y(4,1,4)
     \+R24(1)*Y(1,3,1)+R24(2)*Y213+R24(3)*Y(3,1,3)+R24(4)*Y413+
     \ R34(1)*Y(1,2,1)+R34(2)*Y(2,1,2)+R34(3)*Y312+R34(4)*Y412)*2.d0+
     \ R123(1)*x(1,4)+R123(2)*x(2,4)+R123(3)*x(3,4)+R123(4)*x(4,4)+
     \ R124(1)*x(1,3)+R124(2)*x(2,3)+R124(3)*x(3,3)+R124(4)*x(4,3)+
     \ R134(1)*x(1,2)+R134(2)*x(2,2)+R134(3)*x(3,2)+R134(4)*x(4,2)+
     \ R234(1)*x(1,1)+R234(2)*x(2,1)+R234(3)*x(3,1)+ R234(4)*x(4,1)
     \+DRR2345(6,2)*x(2,0)+DRR2345(5,2)*x(3,0)
     \+DRR2345(4,2)*x(4,0)-DRR2345(10,3)*x(1,0))*I4

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cccccccccccccccccccccccccccccccccccccccccccccccccc

c  Rank-1
      R(1)=DI1345-DI2345
      R(2)=DI1245-DI1345
      R(3)=DI1235-DI1245
      R(4)=DI1234-DI1235
c 1
      C(1,1)=(R(1)*x(1,1)+R(2)*x(2,1)+R(3)*x(3,1)+R(4)*x(4,1)
     /         -DI2345*x(1,0))*Id
c 2
      C(2,1)=(R(1)*x(1,2)+R(2)*x(2,2)+R(3)*x(3,2)+R(4)*x(4,2)   
     /           -DI2345*x(2,0))*Id
c 3
      C(3,1)=(R(1)*x(1,3)+R(2)*x(2,3)+R(3)*x(3,3)+R(4)*x(4,3)   
     /           -DI2345*x(3,0))*Id
c 4
      C(4,1)=(R(1)*x(1,4)+R(2)*x(2,4)+R(3)*x(3,4)+R(4)*x(4,4)   
     /          -DI2345*x(4,0))*Id
c  Rank-2
      R(1)=DII1345(7,2)-DII2345(7,2)
      R(2)=DII1245(7,2)-DII1345(7,2)
      R(3)=DII1235(7,2)-DII1245(7,2)
      R(4)=DII1234(7,2)-DII1235(7,2)

      R1(1)= DII1345(1,1) + DI2345
      R1(2)= DII1245(1,1) - DII1345(1,1)
      R1(3)= DII1235(1,1) - DII1245(1,1)
      R1(4)= DII1234(1,1) - DII1235(1,1)

      R2(1)= DII1345(1,1) - DII2345(1,1)
      R2(2)= DII1245(2,1) - DII1345(1,1)
      R2(3)= DII1235(2,1) - DII1245(2,1)
      R2(4)= DII1234(2,1) - DII1235(2,1) 

      R3(1)= DII1345(2,1) - DII2345(2,1)
      R3(2)= DII1245(2,1) - DII1345(2,1)
      R3(3)= DII1235(3,1) - DII1245(2,1)
      R3(4)= DII1234(3,1) - DII1235(3,1) 
     
      R4(1)= DII1345(3,1) - DII2345(3,1)
      R4(2)= DII1245(3,1) - DII1345(3,1)
      R4(3)= DII1235(3,1) - DII1245(3,1)
      R4(4)=               - DII1235(3,1) 

c 1 1
      C(1,2)=((R(2)*Y(2,1,1)+R(3)*Y(3,1,1)+R(4)*Y(4,1,1))*2.d0+
     \ R1(1)*x(1,1)+R1(2)*x(2,1)+R1(3)*x(3,1)+R1(4)*x(4,1)+DI2345*x(1,0))*Id
c 2 2 
      C(2,2)=((R(1)*Y(1,2,2)+R(3)*Y(3,2,2)+R(4)*Y(4,2,2))*2.d0+
     \R2(1)*x(1,2)+R2(2)*x(2,2)+R2(3)*x(3,2)+R2(4)*x(4,2)-DII2345(1,1)*x(2,0))*Id
c 3 3
      C(3,2)=((R(1)*Y(1,3,3)+R(2)*Y(2,3,3)+R(4)*Y(4,3,3))*2.d0+
     \R3(1)*x(1,3)+R3(2)*x(2,3)+R3(3)*x(3,3)+R3(4)*x(4,3)-DII2345(2,1)*x(3,0))*Id
c 4 4
      C(4,2)=((R(1)*Y(1,4,4)+R(2)*Y(2,4,4)+R(3)*Y(3,4,4))*2.d0+
     \R4(1)*x(1,4)+R4(2)*x(2,4)+R4(3)*x(3,4)+R4(4)*x(4,4)-DII2345(3,1)*x(4,0))*Id
c 1 2
      C(5,2)=((R(1)*Y(1,2,1)+R(2)*Y(2,1,2)+R(3)*Y312+R(4)*Y412)*2.d0+
     \R1(1)*x(1,2)+R1(2)*x(2,2)+R1(3)*x(3,2)+R1(4)*x(4,2)+R2(1)*x(1,1)+
     \R2(2)*x(2,1)+R2(3)*x(3,1)+R2(4)*x(4,1)+DI2345*x(2,0)-DII2345(1,1)*x(1,0))*I2
c 1 3
      C(6,2)=((R(1)*Y(1,3,1)+R(2)*Y213+R(3)*Y(3,1,3)+R(4)*Y413)*2.d0+
     \R1(1)*x(1,3)+R1(2)*x(2,3)+R1(3)*x(3,3)+R1(4)*x(4,3)+R3(1)*x(1,1)+
     \R3(2)*x(2,1)+R3(3)*x(3,1)+R3(4)*x(4,1)+DI2345*x(3,0)-DII2345(2,1)*x(1,0))*I2
c 1 4
      C(7,2)=((R(1)*Y(1,4,1)+R(2)*Y214+R(3)*Y314+R(4)*Y(4,1,4))*2.d0+
     \R1(1)*x(1,4)+R1(2)*x(2,4)+R1(3)*x(3,4)+R1(4)*x(4,4)+R4(1)*x(1,1)+
     \R4(2)*x(2,1)+R4(3)*x(3,1)+R4(4)*x(4,1)+DI2345*x(4,0)-DII2345(3,1)*x(1,0))*I2
c 2 3
      C(8,2)=((R(1)*Y123+R(2)*Y(2,3,2)+R(3)*Y(3,2,3)+R(4)*Y423)*2.d0+
     \R3(1)*x(1,2)+R3(2)*x(2,2)+R3(3)*x(3,2)+R3(4)*x(4,2)+R2(1)*x(1,3)+
     \R2(2)*x(2,3)+R2(3)*x(3,3)+R2(4)*x(4,3)-DII2345(1,1)*x(3,0)-DII2345(2,1)*x(2,0))*I2
c 2 4
      C(9,2)=((R(1)*Y124+R(2)*Y(2,4,2)+R(3)*Y324+R(4)*Y(4,2,4))*2.d0+
     \R4(1)*x(1,2)+R4(2)*x(2,2)+R4(3)*x(3,2)+R4(4)*x(4,2)+R2(1)*x(1,4)+
     \R2(2)*x(2,4)+R2(3)*x(3,4)+R2(4)*x(4,4)-DII2345(1,1)*x(4,0)-DII2345(3,1)*x(2,0))*I2
c 3 4
      C(10,2)=((R(1)*Y134+R(2)*Y234+R(3)*Y(3,4,3)+R(4)*Y(4,3,4))*2.d0+
     \R4(1)*x(1,3)+R4(2)*x(2,3)+R4(3)*x(3,3)+R4(4)*x(4,3)+R3(1)*x(1,4)+
     \R3(2)*x(2,4)+R3(3)*x(3,4)+R3(4)*x(4,4)-DII2345(2,1)*x(4,0)-DII2345(3,1)*x(3,0))*I2
C 00
      C(11,2)=(R(1)*x(1,0)+R(2)*x(2,0)+R(3)*x(3,0)+R(4)*x(4,0))*Id
c  Rank-3
      R1(1)= DII1345(11,3) + DII2345(7,2)
      R1(2)= DII1245(11,3) - DII1345(11,3)
      R1(3)= DII1235(11,3) - DII1245(11,3)
      R1(4)= DII1234(11,3) - DII1235(11,3)
c 0 0 1
      C(21,3)=(R(1)*x(1,1)+R(2)*x(2,1)+R(3)*x(3,1)+R(4)*x(4,1)   
     \+(R1(1)*x(1,0)+R1(2)*x(2,0)+R1(3)*x(3,0)+R1(4)*x(4,0))*2.d0 
     \-DII2345(7,2)*x(1,0))*I3
c
      R11(1)= DII1345(1,2) - DI2345
      R11(2)= DII1245(1,2) - DII1345(1,2)
      R11(3)= DII1235(1,2) - DII1245(1,2)
      R11(4)= DII1234(1,2) - DII1235(1,2)

c 1 1 1 
      C(1,3)=((R1(2)*Y(2,1,1)+R1(3)*Y(3,1,1)+R1(4)*Y(4,1,1))*4.d0+
     \R11(1)*x(1,1)+R11(2)*x(2,1)+R11(3)*x(3,1)+R11(4)*x(4,1)
     \-DI2345*x(1,0))*Id

      R2(1)= DII1345(11,3) - DII2345(11,3)
      R2(2)= DII1245(12,3) - DII1345(11,3)
      R2(3)= DII1235(12,3) - DII1245(12,3)
      R2(4)= DII1234(12,3) - DII1235(12,3)
c 0 0 2
      C(22,3)=(R(1)*x(1,2)+R(2)*x(2,2)+R(3)*x(3,2)+R(4)*x(4,2)   
     \+(R2(1)*x(1,0)+R2(2)*x(2,0)+R2(3)*x(3,0)+R2(4)*x(4,0))*2.d0
     \-DII2345(7,2)*x(2,0))*I3
c
      R22(1)= DII1345(1,2) - DII2345(1,2)
      R22(2)= DII1245(2,2) - DII1345(1,2)
      R22(3)= DII1235(2,2) - DII1245(2,2)
      R22(4)= DII1234(2,2) - DII1235(2,2)

c 2 2 2
      C(2,3)=((R2(1)*Y(1,2,2)+R2(3)*Y(3,2,2)+R2(4)*Y(4,2,2))*4.d0+
     \R22(1)*x(1,2)+R22(2)*x(2,2)+R22(3)*x(3,2)+R22(4)*x(4,2)
     \ -DII2345(1,2)*x(2,0))*Id
c
      R3(1)= DII1345(12,3) - DII2345(12,3)
      R3(2)= DII1245(12,3) - DII1345(12,3)
      R3(3)= DII1235(13,3) - DII1245(12,3)
      R3(4)= DII1234(13,3) - DII1235(13,3)
c 0 0 3
      C(23,3)=(R(1)*x(1,3)+R(2)*x(2,3)+R(3)*x(3,3)+R(4)*x(4,3)   
     \+(R3(1)*x(1,0)+R3(2)*x(2,0)+R3(3)*x(3,0)+R3(4)*x(4,0))*2.d0
     \-DII2345(7,2)*x(3,0))*I3
c
      R33(1)= DII1345(2,2) - DII2345(2,2)
      R33(2)= DII1245(2,2) - DII1345(2,2)
      R33(3)= DII1235(3,2) - DII1245(2,2)
      R33(4)= DII1234(3,2) - DII1235(3,2)
c 3 3 3
      C(3,3)=((R3(1)*Y(1,3,3)+R3(2)*Y(2,3,3)+R3(4)*Y(4,3,3))*4.d0+
     \R33(1)*x(1,3)+R33(2)*x(2,3)+R33(3)*x(3,3)+R33(4)*x(4,3)-DII2345(2,2)*x(3,0))*Id
c
      R4(1)= DII1345(13,3) - DII2345(13,3)
      R4(2)= DII1245(13,3) - DII1345(13,3)
      R4(3)= DII1235(13,3) - DII1245(13,3)
      R4(4)=                - DII1235(13,3)
c 0 0 4
      C(24,3)=(R(1)*x(1,4)+R(2)*x(2,4)+R(3)*x(3,4)+R(4)*x(4,4)
     \+(R4(1)*x(1,0)+R4(2)*x(2,0)+R4(3)*x(3,0)+R4(4)*x(4,0))*2.d0
     \-DII2345(7,2)*x(4,0))*I3

      R44(1)= DII1345(3,2) - DII2345(3,2)
      R44(2)= DII1245(3,2) - DII1345(3,2)
      R44(3)= DII1235(3,2) - DII1245(3,2)
      R44(4)=               - DII1235(3,2)   
c 4 4 4
      C(4,3)=((R4(1)*Y(1,4,4)+R4(2)*Y(2,4,4)+R4(3)*Y(3,4,4))*4.d0+
     \R44(1)*x(1,4)+R44(2)*x(2,4)+R44(3)*x(3,4)+R44(4)*x(4,4)-DII2345(3,2)*x(4,0))*Id

      R12(1)= DII1345(1,2) + DII2345(1,1)
      R12(2)= DII1245(4,2) - DII1345(1,2)
      R12(3)= DII1235(4,2) - DII1245(4,2)
      R12(4)= DII1234(4,2) - DII1235(4,2)

c 1 1 2 
      C(5,3)=((R1(1)*Y(1,2,1)+R1(2)*Y(2,1,2)+R1(3)*Y312+R1(4)*Y412+
     \R2(2)*Y(2,1,1)+R2(3)*Y(3,1,1)+R2(4)*Y(4,1,1))*4.d0+(R12(1)*x(1,1)+
     \R12(2)*x(2,1)+R12(3)*x(3,1)+R12(4)*x(4,1)+DII2345(1,1)*x(1,0))*2.d0+
     \R11(1)*x(1,2)+R11(2)*x(2,2)+R11(3)*x(3,2)+ R11(4)*x(4,2)-DI2345*x(2,0))*I3
c  2 2 1
      C(8,3)=((R2(1)*Y(1,2,1)+R2(2)*Y(2,1,2)+R2(3)*Y312+R2(4)*Y412+
     \R1(1)*Y(1,2,2)+R1(3)*Y(3,2,2)+R1(4)*Y(4,2,2))*4.d0+(R12(1)*x(1,2)+
     \+R12(2)*x(2,2)+R12(3)*x(3,2)+R12(4)*x(4,2)+DII2345(1,1)*x(2,0))*2.d0+
     \R22(1)*x(1,1)+R22(2)*x(2,1)+R22(3)*x(3,1)+R22(4)*x(4,1)-DII2345(1,2)*x(1,0))*I3
c
      R13(1)= DII1345(4,2) + DII2345(2,1)
      R13(2)= DII1245(4,2) - DII1345(4,2)
      R13(3)= DII1235(5,2) - DII1245(4,2)
      R13(4)= DII1234(5,2) - DII1235(5,2)
c 1 1 3 
      C(6,3)=((R1(1)*Y(1,3,1)+R1(2)*Y213+R1(3)*Y(3,1,3)+R1(4)*Y413+
     \R3(2)*Y(2,1,1)+R3(3)*Y(3,1,1)+R3(4)*Y(4,1,1))*4.d0+(R13(1)*x(1,1)+
     \R13(2)*x(2,1)+R13(3)*x(3,1)+R13(4)*x(4,1)+DII2345(2,1)*x(1,0))*2.d0+
     \R11(1)*x(1,3)+R11(2)*x(2,3)+R11(3)*x(3,3)+R11(4)*x(4,3)-DI2345*x(3,0))*I3
c  3 3 1
      C(11,3)=((R3(1)*Y(1,3,1)+R3(2)*Y213+R3(3)*Y(3,1,3)+R3(4)*Y413+
     \R1(1)*Y(1,3,3)+R1(2)*Y(2,3,3)+R1(4)*Y(4,3,3))*4.d0+(R13(1)*x(1,3)+
     \R13(2)*x(2,3)+R13(3)*x(3,3)+R13(4)*x(4,3)+DII2345(2,1)*x(3,0))*2.d0+
     \R33(1)*x(1,1)+R33(2)*x(2,1)+R33(3)*x(3,1)+R33(4)*x(4,1)-DII2345(2,2)*x(1,0))*I3
c
      R14(1)= DII1345(5,2) + DII2345(3,1)
      R14(2)= DII1245(5,2) - DII1345(5,2)
      R14(3)= DII1235(5,2) - DII1245(5,2)
      R14(4)=               - DII1235(5,2)
c 1 1 4
      C(7,3)=((R1(1)*Y(1,4,1)+R1(2)*Y214+R1(3)*Y314+R1(4)*Y(4,1,4)+
     \R4(2)*Y(2,1,1)+R4(3)*Y(3,1,1)+R4(4)*Y(4,1,1))*4.d0+(R14(1)*x(1,1)+
     \R14(2)*x(2,1)+R14(3)*x(3,1)+R14(4)*x(4,1)+DII2345(3,1)*x(1,0))*2.d0+
     \R11(1)*x(1,4)+R11(2)*x(2,4)+R11(3)*x(3,4)+R11(4)*x(4,4)-DI2345*x(4,0))*I3
c  4 4 1
      C(14,3)=((R4(1)*Y(1,4,1)+R4(2)*Y214+R4(3)*Y314+R4(4)*Y(4,1,4)+
     \R1(1)*Y(1,4,4)+R1(2)*Y(2,4,4)+R1(3)*Y(3,4,4))*4.d0+(R14(1)*x(1,4)+
     \R14(2)*x(2,4)+R14(3)*x(3,4)+R14(4)*x(4,4)+DII2345(3,1)*x(4,0))*2.d0+
     \R44(1)*x(1,1)+R44(2)*x(2,1)+R44(3)*x(3,1)+R44(4)*x(4,1)-DII2345(3,2)*x(1,0))*I3
c
      R23(1)= DII1345(4,2) - DII2345(4,2)
      R23(2)= DII1245(2,2) - DII1345(4,2)
      R23(3)= DII1235(6,2) - DII1245(2,2)
      R23(4)= DII1234(6,2) - DII1235(6,2)
c  2 2 3
      C(9,3)=((R2(1)*Y123+R2(2)*Y(2,3,2)+R2(3)*Y(3,2,3)+R2(4)*Y423+
     \R3(1)*Y(1,2,2)+R3(3)*Y(3,2,2)+R3(4)*Y(4,2,2))*4.d0+(R23(1)*x(1,2)+
     \R23(2)*x(2,2)+R23(3)*x(3,2)+R23(4)*x(4,2)-DII2345(4,2)*x(2,0))*2.d0+
     \R22(1)*x(1,3)+R22(2)*x(2,3)+R22(3)*x(3,3)+R22(4)*x(4,3)-DII2345(1,2)*x(3,0))*I3
c  3 3 2
      C(12,3)=((R3(1)*Y123+R3(2)*Y(2,3,2)+R3(3)*Y(3,2,3)+R3(4)*Y423+
     \R2(1)*Y(1,3,3)+R2(2)*Y(2,3,3)+R2(4)*Y(4,3,3))*4.d0+(R23(1)*x(1,3)+
     \R23(2)*x(2,3)+R23(3)*x(3,3)+R23(4)*x(4,3)-DII2345(4,2)*x(3,0))*2.d0+
     \R33(1)*x(1,2)+R33(2)*x(2,2)+R33(3)*x(3,2)+R33(4)*x(4,2)- DII2345(2,2)*x(2,0))*I3
c 1 2 3
      C(17,3)=((R1(1)*Y123+R1(2)*Y(2,3,2)+R1(3)*Y(3,2,3)+R1(4)*Y423+
     \R2(1)*Y(1,3,1)+R2(2)*Y213+R2(3)*Y(3,1,3)+R2(4)*Y413+R3(1)*Y(1,2,1)
     \+R3(2)*Y(2,1,2)+R3(3)*Y312+R3(4)*Y412)*2.d0+R12(1)*x(1,3)+
     \R12(2)*x(2,3)+R12(3)*x(3,3)+R12(4)*x(4,3)+R13(1)*x(1,2)+
     \R13(2)*x(2,2)+R13(3)*x(3,2)+R13(4)*x(4,2)+R23(1)*x(1,1)+
     \R23(2)*x(2,1)+R23(3)*x(3,1)+R23(4)*x(4,1)+DII2345(2,1)*x(2,0)+ 
     \DII2345(1,1)*x(3,0)-DII2345(4,2)*x(1,0))*I3
c
      R24(1)= DII1345(5,2) - DII2345(5,2)
      R24(2)= DII1245(6,2) - DII1345(5,2)
      R24(3)= DII1235(6,2) - DII1245(6,2)
      R24(4)=               - DII1235(6,2)
c  2 2 4
      C(10,3)=((R2(1)*Y124+R2(2)*Y(2,4,2)+R2(3)*Y324+R2(4)*Y(4,2,4)+
     \R4(1)*Y(1,2,2)+R4(3)*Y(3,2,2)+R4(4)*Y(4,2,2))*4.d0+(R24(1)*x(1,2)+
     \R24(2)*x(2,2)+R24(3)*x(3,2)+R24(4)*x(4,2)-DII2345(5,2)*x(2,0))*2.d0+
     \R22(1)*x(1,4)+R22(2)*x(2,4)+R22(3)*x(3,4)+R22(4)*x(4,4)-DII2345(1,2)*x(4,0))*I3
c  4 4 2
      C(15,3)=((R4(1)*Y124+R4(2)*Y(2,4,2)+R4(3)*Y324+R4(4)*Y(4,2,4)+
     \R2(1)*Y(1,4,4)+R2(2)*Y(2,4,4)+R2(3)*Y(3,4,4))*4.d0+(R24(1)*x(1,4)+
     \R24(2)*x(2,4)+R24(3)*x(3,4)+R24(4)*x(4,4)-DII2345(5,2)*x(4,0))*2.d0+
     \R44(1)*x(1,2)+R44(2)*x(2,2)+R44(3)*x(3,2)+R44(4)*x(4,2)-DII2345(3,2)*x(2,0))*I3
c 1 2 4
       C(18,3)=((R1(1)*Y124+R1(2)*Y(2,4,2)+R1(3)*Y324+R1(4)*Y(4,2,4)+
     \R2(1)*Y(1,4,1)+R2(2)*Y214+R2(3)*Y314+R2(4)*Y(4,1,4)+R4(1)*Y(1,2,1)
     \+R4(2)*Y(2,1,2)+R4(3)*Y312+R4(4)*Y412)*2.d0+R12(1)*x(1,4)+
     \R12(2)*x(2,4)+R12(3)*x(3,4)+R12(4)*x(4,4)+R14(1)*x(1,2)+
     \R14(2)*x(2,2)+R14(3)*x(3,2)+R14(4)*x(4,2)+R24(1)*x(1,1)+
     \R24(2)*x(2,1)+R24(3)*x(3,1)+R24(4)*x(4,1)+DII2345(3,1)*x(2,0)+
     \DII2345(1,1)*x(4,0)-DII2345(5,2)*x(1,0))*I3

c
      R34(1)= DII1345(6,2) - DII2345(6,2)
      R34(2)= DII1245(6,2) - DII1345(6,2)
      R34(3)= DII1235(3,2) - DII1245(6,2)
      R34(4)=               - DII1235(3,2)   
c  3 3 4
      C(13,3)=((R3(1)*Y134+R3(2)*Y234 +R3(3)*Y(3,4,3)+R3(4)*Y(4,3,4)+
     \R4(1)*Y(1,3,3)+R4(2)*Y(2,3,3)+R4(4)*Y(4,3,3))*4.d0+(R34(1)*x(1,3)
     \+R34(2)*x(2,3)+R34(3)*x(3,3)+R34(4)*x(4,3)-DII2345(6,2)*x(3,0))*2.d0+
     \R33(1)*x(1,4)+R33(2)*x(2,4)+R33(3)*x(3,4)+R33(4)*x(4,4)-DII2345(2,2)*x(4,0))*I3
c  4 4 3
      C(16,3)=((R4(1)*Y134+R4(2)*Y234+R4(3)*Y(3,4,3)+R4(4)*Y(4,3,4)+
     \R3(1)*Y(1,4,4)+R3(2)*Y(2,4,4)+R3(3)*Y(3,4,4))*4.d0+(R34(1)*x(1,4)+
     \+R34(2)*x(2,4)+R34(3)*x(3,4)+R34(4)*x(4,4)-DII2345(6,2)*x(4,0))*2.d0+
     \R44(1)*x(1,3)+R44(2)*x(2,3)+R44(3)*x(3,3)+R44(4)*x(4,3)-DII2345(3,2)*x(3,0))*I3
c 1 3 4
       C(19,3)=((R1(1)*Y134+R1(2)*Y234+R1(3)*Y(3,4,3)+R1(4)*Y(4,3,4)+
     \R3(1)*Y(1,4,1)+R3(2)*Y214+R3(3)*Y314+R3(4)*Y(4,1,4)+R4(1)*Y(1,3,1)
     \+R4(2)*Y213+R4(3)*Y(3,1,3)+R4(4)*Y413)*2.d0+R13(1)*x(1,4)+
     \R13(2)*x(2,4)+R13(3)*x(3,4)+R13(4)*x(4,4)+R14(1)*x(1,3)+
     \R14(2)*x(2,3)+R14(3)*x(3,3)+R14(4)*x(4,3)+R34(1)*x(1,1)+
     \R34(2)*x(2,1)+R34(3)*x(3,1)+R34(4)*x(4,1)+DII2345(3,1)*x(3,0)+
     \DII2345(2,1)*x(4,0)-DII2345(6,2)*x(1,0))*I3
c 2 3 4
       C(20,3)=((R2(1)*Y134+R2(2)*Y234+R2(3)*Y(3,4,3)+R2(4)*Y(4,3,4)+
     \R3(1)*Y124+R3(2)*Y(2,4,2)+R3(3)*Y324+R3(4)*Y(4,2,4)+R4(1)*Y123+
     \R4(2)*Y(2,3,2)+R4(3)*Y(3,2,3)+R4(4)*Y423)*2.d0+R23(1)*x(1,4)+
     \R23(2)*x(2,4)+R23(3)*x(3,4)+R23(4)*x(4,4)+R24(1)*x(1,3)+
     \R24(2)*x(2,3)+R24(3)*x(3,3)+R24(4)*x(4,3)+R34(1)*x(1,2)+
     \R34(2)*x(2,2)+R34(3)*x(3,2)+R34(4)*x(4,2)-DII2345(4,2)*x(4,0) 
     \-DII2345(5,2)*x(3,0)-DII2345(6,2)*x(2,0))*I3

C Rank 4
c$$$CCCCCCCCCCCCC
       R(1)=DII1345(13,1) - DII2345(13,1)
       R(2)=DII1245(13,1) - DII1345(13,1)
       R(3)=DII1235(13,1) - DII1245(13,1)
       R(4)=DII1234(13,1) - DII1235(13,1)
c 0 0 0 0
      C(46,4)=(R(1)*x(1,0)+R(2)*x(2,0)+R(3)*x(3,0)+R(4)*x(4,0))*Id
c
      R111(1)=(DII1345(1,3) +       DI2345)
      R111(2)=(DII1245(1,3) - DII1345(1,3)) 
      R111(3)=(DII1235(1,3) - DII1245(1,3)) 
      R111(4)=(DII1234(1,3) - DII1235(1,3))  
   
      R11(1)=(DII1345(7,1) -  DII2345(7,2))
      R11(2)=(DII1245(7,1) -  DII1345(7,1)) 
      R11(3)=(DII1235(7,1) -  DII1245(7,1)) 
      R11(4)=(DII1234(7,1) -  DII1235(7,1))
c 1 1 1 1
      C(1,4)=((R11(2)*Y(2,1,1)+R11(3)*Y(3,1,1)+R11(4)*Y(4,1,1))*6.d0+
     \(R111(1)*x(1,1)+R111(2)*x(2,1)+R111(3)*x(3,1)+R111(4)*x(4,1))+
     \DI2345*x(1,0))*Id
c 0 0 1 1
      C(36,4)=((R11(1)*x(1,0)+R11(2)*x(2,0)+R11(3)*x(3,0)+R11(4)*x(4,0)
     \+R1(1)*x(1,1)+R1(2)*x(2,1)+R1(3)*x(3,1)+R1(4)*x(4,1)+
     \ DII2345(7,2)*x(1,0))*2.d0+(R(2)*Y(2,1,1)+R(3)*Y(3,1,1)+
     \ R(4)*Y(4,1,1))*4.d0)*I4
c
      R12(1)=(DII1345(7,1)  +  DII2345(11,3))
      R12(2)=(DII1245(10,1) -  DII1345(7,1)) 
      R12(3)=(DII1235(10,1) -  DII1245(10,1)) 
      R12(4)=(DII1234(10,1) -  DII1235(10,1))
  
      R112(1)=(DII1345(1,3) - DII2345(1,1))
      R112(2)=(DII1245(4,3) - DII1345(1,3)) 
      R112(3)=(DII1235(4,3) - DII1245(4,3)) 
      R112(4)=(DII1234(4,3) - DII1235(4,3))     
c 1 1 1 2 
      C(5,4)=((R12(2)*Y(2,1,1)+R12(3)*Y(3,1,1)+R12(4)*Y(4,1,1))*12.d0+
     \(R11(1)*Y(1,2,1)+R11(2)*Y(2,1,2)+R11(3)*Y312+R11(4)*Y412)*6.d0+
     \(R112(1)*x(1,1)+R112(2)*x(2,1)+R112(3)*x(3,1)+R112(4)*x(4,1))*3.d0
     \+R111(1)*x(1,2)+R111(2)*x(2,2)+R111(3)*x(3,2)+R111(4)*x(4,2)
     \-DII2345(1,1)*x(1,0)*3.d0 + DI2345*x(2,0))*I4
c 0 0 1 2
      C(40,4)=((R12(1)*x(1,0)+R12(2)*x(2,0)+R12(3)*x(3,0)+R12(4)*x(4,0)+
     \R(1)*Y(1,2,1)+R(2)*Y(2,1,2)+R(3)*Y312+R(4)*Y412)*2.d0+R1(1)*x(1,2)
     \+R1(2)*x(2,2)+R1(3)*x(3,2)+R1(4)*x(4,2)+R2(1)*x(1,1)+R2(2)*x(2,1)+
     \R2(3)*x(3,1)+R2(4)*x(4,1)+DII2345(7,2)*x(2,0)
     \-DII2345(11,3)*x(1,0))*I4
c
      R13(1)=(DII1345(10,1) + DII2345(12,3))
      R13(2)=(DII1245(10,1) -  DII1345(10,1)) 
      R13(3)=(DII1235(11,1) -  DII1245(10,1)) 
      R13(4)=(DII1234(11,1) -  DII1235(11,1))

      R113(1)=(DII1345(4,3) - DII2345(2,1))
      R113(2)=(DII1245(4,3) - DII1345(4,3)) 
      R113(3)=(DII1235(5,3) - DII1245(4,3)) 
      R113(4)=(DII1234(5,3) - DII1235(5,3))
c 1 1 1 3 
      C(6,4)=((R13(2)*Y(2,1,1)+R13(3)*Y(3,1,1)+R13(4)*Y(4,1,1))*12.d0+
     \(R11(1)*Y(1,3,1)+R11(2)*Y213+R11(3)*Y(3,1,3)+R11(4)*Y413)*6.d0+
     \(R113(1)*x(1,1)+R113(2)*x(2,1)+R113(3)*x(3,1)+R113(4)*x(4,1))*3.d0
     \+R111(1)*x(1,3)+R111(2)*x(2,3)+R111(3)*x(3,3)+R111(4)*x(4,3)
     \-DII2345(2,1)*x(1,0)*3.d0 + DI2345*x(3,0))*I4
c 0 0 1 3
      C(41,4)=((R13(1)*x(1,0)+R13(2)*x(2,0)+R13(3)*x(3,0)+R13(4)*x(4,0)+
     \R(1)*Y(1,3,1)+R(2)*Y213+R(3)*Y(3,1,3)+R(4)*Y413)*2.d0+R1(1)*x(1,3)
     \+R1(2)*x(2,3)+R1(3)*x(3,3)+R1(4)*x(4,3)+R3(1)*x(1,1)+R3(2)*x(2,1)+
     \R3(3)*x(3,1)+R3(4)*x(4,1)+DII2345(7,2)*x(3,0)
     \ -DII2345(12,3)*x(1,0))*I4
c     
      R14(1)=(DII1345(11,1)  +  DII2345(13,3))
      R14(2)=(DII1245(11,1)  -  DII1345(11,1)) 
      R14(3)=(DII1235(11,1)  -  DII1245(11,1)) 
      R14(4)=(                -  DII1235(11,1))
     
      R114(1)=(DII1345(5,3) - DII2345(3,1))
      R114(2)=(DII1245(5,3) - DII1345(5,3)) 
      R114(3)=(DII1235(5,3) - DII1245(5,3)) 
      R114(4)=(              - DII1235(5,3))
c 1 1 1 4 
      C(7,4)=((R14(2)*Y(2,1,1)+R14(3)*Y(3,1,1)+R14(4)*Y(4,1,1))*12.d0+
     \ (R11(1)*Y(1,4,1)+R11(2)*Y214+R11(3)*Y314+R11(4)*Y(4,1,4))*6.d0+
     \(R114(1)*x(1,1)+R114(2)*x(2,1)+R114(3)*x(3,1)+R114(4)*x(4,1))*3.d0
     \+R111(1)*x(1,4)+R111(2)*x(2,4)+R111(3)*x(3,4)+R111(4)*x(4,4)
     \-DII2345(3,1)*x(1,0)*3.d0 + DI2345*x(4,0))*I4
c 0 0 1 4
      C(42,4)=((R14(1)*x(1,0)+R14(2)*x(2,0)+R14(3)*x(3,0)+R14(4)*x(4,0)+
     \R(1)*Y(1,4,1)+R(2)*Y214+R(3)*Y314+R(4)*Y(4,1,4))*2.d0+R1(1)*x(1,4)
     \+R1(2)*x(2,4)+R1(3)*x(3,4)+R1(4)*x(4,4)+R4(1)*x(1,1)+R4(2)*x(2,1)+
     \ R4(3)*x(3,1)+R4(4)*x(4,1)+DII2345(7,2)*x(4,0)
     \ -DII2345(13,3)*x(1,0))*I4
c
      R22(1)=(DII1345(7,1) -  DII2345(7,1))
      R22(2)=(DII1245(8,1) -  DII1345(7,1)) 
      R22(3)=(DII1235(8,1) -  DII1245(8,1)) 
      R22(4)=(DII1234(8,1) -  DII1235(8,1))

      R222(1)=(DII1345(1,3) - DII2345(1,3))
      R222(2)=(DII1245(2,3) - DII1345(1,3)) 
      R222(3)=(DII1235(2,3) - DII1245(2,3)) 
      R222(4)=(DII1234(2,3) - DII1235(2,3))
c 2 2 2 2
      C(2,4)=((R22(1)*Y(1,2,2)+R22(3)*Y(3,2,2)+R22(4)*Y(4,2,2))*6.d0+
     \R222(1)*x(1,2)+R222(2)*x(2,2)+R222(3)*x(3,2)+R222(4)*x(4,2)
     \-DII2345(1,3)*x(2,0))*Id
c 0 0 2 2
      C(37,4)=((R22(1)*x(1,0)+R22(2)*x(2,0)+R22(3)*x(3,0)+
     - R22(4)*x(4,0)+ R2(1)*x(1,2)+R2(2)*x(2,2)+R2(3)*x(3,2)+
     - R2(4)*x(4,2)- DII2345(11,3)*x(2,0))*2.d0+(R(1)*Y(1,2,2)+ 
     - R(3)*Y(3,2,2)+R(4)*Y(4,2,2))*4.d0)*I4
c
      R221(1)=(DII1345(1,3) + DII2345(1,2))
      R221(2)=(DII1245(6,3) - DII1345(1,3)) 
      R221(3)=(DII1235(6,3) - DII1245(6,3)) 
      R221(4)=(DII1234(6,3) - DII1235(6,3))
c 2 2 2 1 
      C(8,4)=((R12(1)*Y(1,2,2)+R12(3)*Y(3,2,2)+R12(4)*Y(4,2,2))*12.d0+
     \ (R22(1)*Y(1,2,1)+R22(2)*Y(2,1,2)+R22(3)*Y312+R22(4)*Y412)*6.d0+
     \R222(1)*x(1,1)+R222(2)*x(2,1)+R222(3)*x(3,1)+R222(4)*x(4,1)+
     \(R221(1)*x(1,2)+R221(2)*x(2,2)+R221(3)*x(3,2)+R221(4)*x(4,2))*3.d0
     \-DII2345(1,3)*x(1,0)+DII2345(1,2)*x(2,0)*3.d0)*I4
c 1 1 2 2 
      C(17,4)=((R11(1)*Y(1,2,2)+R11(3)*Y(3,2,2)+R11(4)*Y(4,2,2)+
     \R22(2)*Y(2,1,1)+R22(3)*Y(3,1,1)+R22(4)*Y(4,1,1))*4.d0+
     \ (R12(1)*Y(1,2,1)+R12(2)*Y(2,1,2)+R12(3)*Y312+R12(4)*Y412)*8.d0
     \+(R112(1)*x(1,2)+R112(2)*x(2,2)+R112(3)*x(3,2)+R112(4)*x(4,2)
     \+R221(1)*x(1,1)+R221(2)*x(2,1)+R221(3)*x(3,1)+R221(4)*x(4,1) 
     \ +DII2345(1,2)*x(1,0)- DII2345(1,1)*x(2,0))*2.d0)*I4

c
      R23(1)=(DII1345(10,1) -  DII2345(10,1))
      R23(2)=(DII1245(8,1)  -  DII1345(10,1)) 
      R23(3)=(DII1235(12,1) -  DII1245(8,1)) 
      R23(4)=(DII1234(12,1) -  DII1235(12,1))

      R223(1)=(DII1345(4,3) - DII2345(4,3))
      R223(2)=(DII1245(2,3) - DII1345(4,3)) 
      R223(3)=(DII1235(8,3) - DII1245(2,3)) 
      R223(4)=(DII1234(8,3) - DII1235(8,3))
c 2 2 2 3 
      C(9,4)=((R23(1)*Y(1,2,2)+R23(3)*Y(3,2,2)+R23(4)*Y(4,2,2))*12.d0+
     \(R22(1)*Y123+R22(2)*(Y(2,3,2))+R22(3)*Y(3,2,3)+R22(4)*Y423)*6.d0+
     \R222(1)*x(1,3)+R222(2)*x(2,3)+R222(3)*x(3,3)+R222(4)*x(4,3)+
     \(R223(1)*x(1,2)+R223(2)*x(2,2)+R223(3)*x(3,2)+R223(4)*x(4,2))*3.d0
     \-DII2345(1,3)*x(3,0)-DII2345(4,3)*x(2,0)*3.d0)*I4
c 0 0 2 3
      C(43,4)=((R23(1)*x(1,0)+R23(2)*x(2,0)+R23(3)*x(3,0)+R23(4)*x(4,0)+
     \R(1)*Y123+R(2)*Y(2,3,2)+R(3)*Y(3,2,3)+R(4)*Y423)*2.d0+R3(1)*x(1,2)
     \+R3(2)*x(2,2)+R3(3)*x(3,2)+R3(4)*x(4,2)+R2(1)*x(1,3)+R2(2)*x(2,3)+
     \R2(3)*x(3,3)+R2(4)*x(4,3)-DII2345(12,3)*x(2,0)
     \ -DII2345(11,3)*x(3,0))*I4
c
      R24(1)=(DII1345(11,1) -  DII2345(11,1))
      R24(2)=(DII1245(12,1) -  DII1345(11,1)) 
      R24(3)=(DII1235(12,1) -  DII1245(12,1)) 
      R24(4)=(               -  DII1235(12,1))

      R224(1)=(DII1345(5,3) - DII2345(5,3))
      R224(2)=(DII1245(8,3) - DII1345(5,3)) 
      R224(3)=(DII1235(8,3) - DII1245(8,3)) 
      R224(4)=(              - DII1235(8,3))
c 2 2 2 4 
      C(10,4)=((R24(1)*Y(1,2,2)+R24(3)*Y(3,2,2)+R24(4)*Y(4,2,2))*12.d0+
     \ (R22(1)*Y124+R22(2)*(Y(2,4,2))+R22(3)*Y324+R22(4)*Y(4,2,4))*6.d0+
     \ R222(1)*x(1,4)+R222(2)*x(2,4)+R222(3)*x(3,4)+R222(4)*x(4,4)+
     \(R224(1)*x(1,2)+R224(2)*x(2,2)+R224(3)*x(3,2)+R224(4)*x(4,2))*3.d0
     \ -DII2345(1,3)*x(4,0)-DII2345(5,3)*x(2,0)*3.d0)*I4
c 0 0 2 4
      C(44,4)=((R24(1)*x(1,0)+R24(2)*x(2,0)+R24(3)*x(3,0)+R24(4)*x(4,0)+
     \R(1)*Y124+R(2)*Y(2,4,2)+R(3)*Y324+R(4)*Y(4,2,4))*2.d0+R4(1)*x(1,2)
     \+R4(2)*x(2,2)+R4(3)*x(3,2)+R4(4)*x(4,2)+R2(1)*x(1,4)+R2(2)*x(2,4)+
     \R2(3)*x(3,4)+R2(4)*x(4,4)-DII2345(13,3)*x(2,0)
     \-DII2345(11,3)*x(4,0))*I4
c
      R33(1)=(DII1345(8,1) -  DII2345(8,1))
      R33(2)=(DII1245(8,1) -  DII1345(8,1)) 
      R33(3)=(DII1235(9,1) -  DII1245(8,1)) 
      R33(4)=(DII1234(9,1) -  DII1235(9,1))

      R333(1)=(DII1345(2,3) - DII2345(2,3))
      R333(2)=(DII1245(2,3) - DII1345(2,3)) 
      R333(3)=(DII1235(3,3) - DII1245(2,3)) 
      R333(4)=(DII1234(3,3) - DII1235(3,3))
c 3 3 3 3
      C(3,4)=((R33(1)*Y(1,3,3)+R33(2)*Y(2,3,3)+R33(4)*Y(4,3,3))*6.d0+
     \R333(1)*x(1,3)+R333(2)*x(2,3)+R333(3)*x(3,3)+R333(4)*x(4,3)
     \-DII2345(2,3)*x(3,0))*Id
c 0 0 3 3
      C(38,4)=((R33(1)*x(1,0)+R33(2)*x(2,0)+R33(3)*x(3,0)+
     - R33(4)*x(4,0)+R3(1)*x(1,3)+R3(2)*x(2,3)+R3(3)*x(3,3)+
     - R3(4)*x(4,3)-DII2345(12,3)*x(3,0))*2.d0+(R(1)*Y(1,3,3)+
     - R(2)*Y(2,3,3)+R(4)*Y(4,3,3))*4.d0)*I4
c
      R331(1)=(DII1345(6,3) + DII2345(2,2))
      R331(2)=(DII1245(6,3) - DII1345(6,3)) 
      R331(3)=(DII1235(7,3) - DII1245(6,3)) 
      R331(4)=(DII1234(7,3) - DII1235(7,3))
c 3 3 3 1 
      C(11,4)=((R13(1)*Y(1,3,3)+R13(2)*Y(2,3,3)+R13(4)*Y(4,3,3))*12.d0+
     \(R33(1)*Y(1,3,1)+R33(2)*Y213+R33(3)*Y(3,1,3)+R33(4)*Y413)*6.d0+
     \ R333(1)*x(1,1)+R333(2)*x(2,1)+R333(3)*x(3,1)+R333(4)*x(4,1)+
     \(R331(1)*x(1,3)+R331(2)*x(2,3)+R331(3)*x(3,3)+R331(4)*x(4,3))*3.d0
     \-DII2345(2,3)*x(1,0) + DII2345(2,2)*x(3,0)*3.d0)*I4
c 1 1 3 3
      C(18,4)=((R11(1)*Y(1,3,3)+R11(2)*Y(2,3,3)+R11(4)*Y(4,3,3)+
     \R33(2)*Y(2,1,1)+R33(3)*Y(3,1,1)+R33(4)*Y(4,1,1))*4.d0+
     \ (R13(1)*Y(1,3,1)+R13(2)*Y213+R13(3)*Y(3,1,3)+R13(4)*Y413)*8.d0
     \ +(R113(1)*x(1,3)+R113(2)*x(2,3)+R113(3)*x(3,3)+R113(4)*x(4,3)+
     \ R331(1)*x(1,1)+R331(2)*x(2,1)+R331(3)*x(3,1)+R331(4)*x(4,1)
     \ +DII2345(2,2)*x(1,0)-DII2345(2,1)*x(3,0))*2.d0)*I4

      R332(1)=(DII1345(6,3) - DII2345(6,3))
      R332(2)=(DII1245(2,3) - DII1345(6,3)) 
      R332(3)=(DII1235(9,3) - DII1245(2,3)) 
      R332(4)=(DII1234(9,3) - DII1235(9,3))
c 3 3 3 2 
      C(12,4)=((R23(1)*Y(1,3,3)+R23(2)*Y(2,3,3)+R23(4)*Y(4,3,3))*12.d0+
     \(R33(1)*Y123+R33(2)*Y(2,3,2)+R33(3)*Y(3,2,3)+R33(4)*Y423)*6.d0
     \+R333(1)*x(1,2)+R333(2)*x(2,2)+R333(3)*x(3,2)+R333(4)*x(4,2)+
     \(R332(1)*x(1,3)+R332(2)*x(2,3)+R332(3)*x(3,3)+R332(4)*x(4,3))*3.d0
     \-DII2345(2,3)*x(2,0) - DII2345(6,3)*x(3,0)*3.d0)*I4
c 2 2 3 3
      C(20,4)=((R22(1)*Y(1,3,3)+R22(2)*Y(2,3,3)+R22(4)*Y(4,3,3)+
     \ R33(1)*Y(1,2,2)+R33(3)*Y(3,2,2)+R33(4)*Y(4,2,2))*4.d0
     \ +(R23(1)*Y123+R23(2)*Y(2,3,2)+R23(3)*Y(3,2,3)+R23(4)*Y423)*8.d0
     \ +(R223(1)*x(1,3)+R223(2)*x(2,3)+R223(3)*x(3,3)+R223(4)*x(4,3)
     \+R332(1)*x(1,2)+R332(2)*x(2,2)+R332(3)*x(3,2)+R332(4)*x(4,2)
     \ -DII2345(6,3)*x(2,0)-DII2345(4,3)*x(3,0))*2.d0)*I4
c
      R34(1)=(DII1345(12,1) -  DII2345(12,1))
      R34(2)=(DII1245(12,1) -  DII1345(12,1)) 
      R34(3)=(DII1235(9,1)  -  DII1245(12,1)) 
      R34(4)=(               -  DII1235(9,1))

      R334(1)=(DII1345(8,3) - DII2345(8,3))
      R334(2)=(DII1245(8,3) - DII1345(8,3)) 
      R334(3)=(DII1235(3,3) - DII1245(8,3)) 
      R334(4)=(              - DII1235(3,3))
c 3 3 3 4
      C(13,4)=((R34(1)*Y(1,3,3)+R34(2)*Y(2,3,3)+R34(4)*Y(4,3,3))*12.d0+
     \(R33(1)*Y134+R33(2)*Y234+R33(3)*Y(3,4,3)+R33(4)*Y(4,3,4))*6.d0
     \+R333(1)*x(1,4)+R333(2)*x(2,4)+R333(3)*x(3,4)+R333(4)*x(4,4)+
     \(R334(1)*x(1,3)+R334(2)*x(2,3)+R334(3)*x(3,3)+R334(4)*x(4,3))*3.d0
     \-DII2345(2,3)*x(4,0) - DII2345(8,3)*x(3,0)*3.d0)*I4
c 0 0 3 4
      C(45,4)=((R34(1)*x(1,0)+R34(2)*x(2,0)+R34(3)*x(3,0)+R34(4)*x(4,0)+
     \R(1)*Y134+R(2)*Y234+R(3)*Y(3,4,3)+R(4)*Y(4,3,4))*2.d0+R4(1)*x(1,3)
     \+R4(2)*x(2,3)+R4(3)*x(3,3)+R4(4)*x(4,3)+R3(1)*x(1,4)+R3(2)*x(2,4)
     \+R3(3)*x(3,4)+R3(4)*x(4,4)-DII2345(13,3)*x(3,0)
     \ -DII2345(12,3)*x(4,0))*I4
c  
      R44(1)=(DII1345(9,1) -  DII2345(9,1))
      R44(2)=(DII1245(9,1) -  DII1345(9,1)) 
      R44(3)=(DII1235(9,1) -  DII1245(9,1)) 
      R44(4)=(              -  DII1235(9,1))

      R444(1)=(DII1345(3,3) - DII2345(3,3))
      R444(2)=(DII1245(3,3) - DII1345(3,3)) 
      R444(3)=(DII1235(3,3) - DII1245(3,3)) 
      R444(4)=(              - DII1235(3,3))
c 4 4 4 4
      C(4,4)=((R44(1)*Y(1,4,4)+R44(2)*Y(2,4,4)+R44(3)*Y(3,4,4))*6.d0+
     \R444(1)*x(1,4)+R444(2)*x(2,4)+R444(3)*x(3,4)+R444(4)*x(4,4)
     \-DII2345(3,3)*x(4,0))*Id
c 0 0 4 4
      C(39,4)=((R44(1)*x(1,0)+R44(2)*x(2,0)+R44(3)*x(3,0)+
     - R44(4)*x(4,0)+R4(1)*x(1,4)+R4(2)*x(2,4)+R4(3)*x(3,4)+
     - R4(4)*x(4,4)-DII2345(13,3)*x(4,0))*2.d0+(R(1)*Y(1,4,4)+
     - R(2)*Y(2,4,4)+R(3)*Y(3,4,4))*4.d0)*I4    
c
      R441(1)=(DII1345(7,3) + DII2345(3,2))
      R441(2)=(DII1245(7,3) - DII1345(7,3)) 
      R441(3)=(DII1235(7,3) - DII1245(7,3)) 
      R441(4)=(              - DII1235(7,3))
c 4 4 4 1 
      C(14,4)=((R14(1)*Y(1,4,4)+R14(2)*Y(2,4,4)+R14(3)*Y(3,4,4))*12.d0+
     \(R44(1)*Y(1,4,1)+R44(2)*Y214+R44(3)*Y314+R44(4)*Y(4,1,4))*6.d0+
     \R444(1)*x(1,1)+R444(2)*x(2,1)+R444(3)*x(3,1)+R444(4)*x(4,1)+
     \(R441(1)*x(1,4)+R441(2)*x(2,4)+R441(3)*x(3,4)+R441(4)*x(4,4))*3.d0
     \-DII2345(3,3)*x(1,0) + DII2345(3,2)*x(4,0)*3.d0)*I4     
c 1 1 4 4
      C(19,4)=((R11(1)*Y(1,4,4)+R11(2)*Y(2,4,4)+R11(3)*Y(3,4,4)+
     \ R44(2)*Y(2,1,1)+R44(3)*Y(3,1,1)+R44(4)*Y(4,1,1))*4.d0+
     \ (R14(1)*(Y(1,4,1))+R14(2)*Y214+R14(3)*Y314+R14(4)*Y(4,1,4))*8.d0
     \+(R114(1)*x(1,4)+R114(2)*x(2,4)+R114(3)*x(3,4)+R114(4)*x(4,4)+
     \R441(1)*x(1,1)+R441(2)*x(2,1)+R441(3)*x(3,1)+R441(4)*x(4,1)
     \+DII2345(3,2)*x(1,0)-DII2345(3,1)*x(4,0))*2.d0)*I4

      R442(1)=(DII1345(7,3) - DII2345(7,3))
      R442(2)=(DII1245(9,3) - DII1345(7,3)) 
      R442(3)=(DII1235(9,3) - DII1245(9,3)) 
      R442(4)=(              - DII1235(9,3))
c 4 4 4 2 
      C(15,4)=((R24(1)*Y(1,4,4)+R24(2)*Y(2,4,4)+R24(3)*Y(3,4,4))*12.d0+
     \ (R44(1)*Y124+R44(2)*Y(2,4,2)+R44(3)*Y324+R44(4)*Y(4,2,4))*6.d0+
     \ R444(1)*x(1,2)+R444(2)*x(2,2)+R444(3)*x(3,2)+R444(4)*x(4,2)+
     \(R442(1)*x(1,4)+R442(2)*x(2,4)+R442(3)*x(3,4)+R442(4)*x(4,4))*3.d0
     \ -DII2345(3,3)*x(2,0) - DII2345(7,3)*x(4,0)*3.d0)*I4
c 2 2 4 4
      C(21,4)=((R22(1)*Y(1,4,4)+R22(2)*Y(2,4,4)+R22(3)*Y(3,4,4)+
     \R44(1)*Y(1,2,2)+R44(3)*Y(3,2,2)+R44(4)*Y(4,2,2))*4.d0+
     \ (R24(1)*Y124+R24(2)*Y(2,4,2)+R24(3)*Y324+R24(4)*Y(4,2,4))*8.d0
     \+(R224(1)*x(1,4)+R224(2)*x(2,4)+R224(3)*x(3,4)+R224(4)*x(4,4)+
     \R442(1)*x(1,2)+R442(2)*x(2,2)+R442(3)*x(3,2)+R442(4)*x(4,2)
     \ -DII2345(7,3)*x(2,0)-DII2345(5,3)*x(4,0))*2.d0)*I4

      R443(1)=(DII1345(9,3) - DII2345(9,3))
      R443(2)=(DII1245(9,3) - DII1345(9,3)) 
      R443(3)=(DII1235(3,3) - DII1245(9,3)) 
      R443(4)=(              - DII1235(3,3))

c 4 4 4 3 
      C(16,4)=((R34(1)*Y(1,4,4)+R34(2)*Y(2,4,4)+R34(3)*Y(3,4,4))*12.d0+
     \ (R44(1)*Y134+R44(2)*Y234+R44(3)*Y(3,4,3)+R44(4)*Y(4,3,4))*6.d0+
     \ R444(1)*x(1,3)+R444(2)*x(2,3)+R444(3)*x(3,3)+R444(4)*x(4,3)+
     \(R443(1)*x(1,4)+R443(2)*x(2,4)+R443(3)*x(3,4)+R443(4)*x(4,4))*3.d0
     \ -DII2345(3,3)*x(3,0) - DII2345(9,3)*x(4,0)*3.d0)*I4
c 3 3 4 4
      C(22,4)=((R33(1)*Y(1,4,4)+R33(2)*Y(2,4,4)+R33(3)*Y(3,4,4)+
     \R44(1)*Y(1,3,3)+R44(2)*Y(2,3,3)+R44(4)*Y(4,3,3))*4.d0+
     \ (R34(1)*Y134+R34(2)*Y234+R34(3)*Y(3,4,3)+R34(4)*Y(4,3,4))*8.d0
     \+(R334(1)*x(1,4)+R334(2)*x(2,4)+R334(3)*x(3,4)+R334(4)*x(4,4)+
     \ R443(1)*x(1,3)+R443(2)*x(2,3)+R443(3)*x(3,3)+R443(4)*x(4,3)
     \ -DII2345(9,3)*x(3,0)-DII2345(8,3)*x(4,0))*2.d0)*I4

      R123(1)=(DII1345(4,3) + DII2345(4,2))
      R123(2)=(DII1245(6,3) - DII1345(4,3)) 
      R123(3)=(DII1235(10,3)- DII1245(6,3))      
      R123(4)=(DII1234(10,3)- DII1235(10,3))
c 1 1 2 3
      C(23,4)=((R23(2)*Y(2,1,1)+R23(3)*Y(3,1,1)+R23(4)*Y(4,1,1)+
     \ R13(1)*Y(1,2,1)+R13(2)*Y(2,1,2)+R13(3)*Y312+R13(4)*Y412+
     \ R12(1)*Y(1,3,1)+R12(2)*Y213+R12(3)*Y(3,1,3)+R12(4)*Y413)*4.d0
     \+(R11(1)*Y123+R11(2)*Y(2,3,2)+R11(3)*Y(3,2,3)+R11(4)*Y423)*2.d0+
     \R112(1)*x(1,3)+R112(2)*x(2,3)+R112(3)*x(3,3)+R112(4)*x(4,3)+
     \R113(1)*x(1,2)+R113(2)*x(2,2)+R113(3)*x(3,2)+R113(4)*x(4,2)+
     \(R123(1)*x(1,1)+R123(2)*x(2,1)+R123(3)*x(3,1)+R123(4)*x(4,1)
     \+DII2345(4,2)*x(1,0))*2.d0-DII2345(2,1)*x(2,0)
     \-DII2345(1,1)*x(3,0))*I4

c 1 2 2 3
      C(26,4)=((R13(1)*Y(1,2,2)+R13(3)*Y(3,2,2)+R13(4)*Y(4,2,2)+
     \ R23(1)*Y(1,2,1)+R23(2)*Y(2,1,2)+R23(3)*Y312+R23(4)*Y412+
     \ R12(1)*Y123+R12(2)*Y(2,3,2)+R12(3)*Y(3,2,3)+R12(4)*Y423)*4.d0
     \+(R22(1)*(Y(1,3,1))+R22(2)*Y213+R22(3)*Y(3,1,3)+R22(4)*Y413)*2.d0+
     \R221(1)*x(1,3)+R221(2)*x(2,3)+R221(3)*x(3,3)+R221(4)*x(4,3)+
     \R223(1)*x(1,1)+R223(2)*x(2,1)+R223(3)*x(3,1)+R223(4)*x(4,1)+
     \(R123(1)*x(1,2)+R123(2)*x(2,2)+R123(3)*x(3,2)+R123(4)*x(4,2)
     \ +DII2345(4,2)*x(2,0))*2.d0-DII2345(4,3)*x(1,0)
     \ +DII2345(1,2)*x(3,0))*I4
c 1 2 3 3
      C(29,4)=((R12(1)*Y(1,3,3)+R12(2)*Y(2,3,3)+R12(4)*Y(4,3,3)+
     \ R23(1)*Y(1,3,1)+R23(2)*Y213+R23(3)*Y(3,1,3)+R23(4)*Y413+
     \ R13(1)*Y123+R13(2)*Y(2,3,2)+R13(3)*Y(3,2,3)+ R13(4)*Y423)*4.d0
     \+(R33(1)*Y(1,2,1)+R33(2)*Y(2,1,2)+R33(3)*Y312+R33(4)*Y412)*2.d0
     \+R331(1)*x(1,2)+R331(2)*x(2,2)+R331(3)*x(3,2)+R331(4)*x(4,2)+
     \ R332(1)*x(1,1)+R332(2)*x(2,1)+R332(3)*x(3,1)+R332(4)*x(4,1)+
     \ (R123(1)*x(1,3)+ R123(2)*x(2,3)+R123(3)*x(3,3)+R123(4)*x(4,3)
     \ +DII2345(4,2)*x(3,0))*2.d0 - DII2345(6,3)*x(1,0)
     \ +DII2345(2,2)*x(2,0))*I4
c
      R124(1)=(DII1345(5,3)  + DII2345(5,2))
      R124(2)=(DII1245(10,3) - DII1345(5,3)) 
      R124(3)=(DII1235(10,3) - DII1245(10,3))      
      R124(4)=(               - DII1235(10,3))
c 1 1 2 4
      C(24,4)=((R24(2)*Y(2,1,1)+R24(3)*Y(3,1,1)+R24(4)*Y(4,1,1)+
     \ R14(1)*Y(1,2,1)+R14(2)*Y(2,1,2)+R14(3)*Y312+R14(4)*Y412+
     \ R12(1)*Y(1,4,1)+R12(2)*Y214+R12(3)*Y314+R12(4)*Y(4,1,4))*4.d0
     \+(R11(1)*Y124+R11(2)*Y(2,4,2)+R11(3)*Y324+R11(4)*Y(4,2,4))*2.d0+
     \R112(1)*x(1,4)+R112(2)*x(2,4)+R112(3)*x(3,4)+R112(4)*x(4,4)+
     \R114(1)*x(1,2)+R114(2)*x(2,2)+R114(3)*x(3,2)+R114(4)*x(4,2)+
     \(R124(1)*x(1,1)+R124(2)*x(2,1)+R124(3)*x(3,1)+R124(4)*x(4,1)
     \ +DII2345(5,2)*x(1,0))*2.d0-DII2345(3,1)*x(2,0)
     \ -DII2345(1,1)*x(4,0))*I4
c 1 2 2 4
      C(27,4)=((R14(1)*Y(1,2,2)+R14(3)*Y(3,2,2)+R14(4)*Y(4,2,2)+
     \ R24(1)*Y(1,2,1)+R24(2)*Y(2,1,2)+R24(3)*Y312+R24(4)*Y412+
     \ R12(1)*Y124+R12(2)*Y(2,4,2)+R12(3)*Y324+R12(4)*Y(4,2,4))*4.d0
     \+(R22(1)*Y(1,4,1)+R22(2)*Y214+R22(3)*Y314+R22(4)*Y(4,1,4))*2.d0
     \ +R221(1)*x(1,4)+R221(2)*x(2,4)+R221(3)*x(3,4)+R221(4)*x(4,4)+
     \ R224(1)*x(1,1)+R224(2)*x(2,1)+R224(3)*x(3,1)+R224(4)*x(4,1)+
     \ (R124(1)*x(1,2)+R124(2)*x(2,2)+R124(3)*x(3,2)+R124(4)*x(4,2)
     \ +DII2345(5,2)*x(2,0))*2.d0-DII2345(5,3)*x(1,0)
     \ +DII2345(1,2)*x(4,0))*I4
c 1 2 4 4
      C(32,4)=((R12(1)*Y(1,4,4)+R12(2)*Y(2,4,4)+R12(3)*Y(3,4,4)
     \ +R24(1)*Y(1,4,1)+R24(2)*Y214+R24(3)*Y314+R24(4)*Y(4,1,4)+
     \ R14(1)*Y124+R14(2)*Y(2,4,2)+R14(3)*Y324+R14(4)*Y(4,2,4))*4.d0+
     \(R44(1)*Y(1,2,1)+R44(2)*Y(2,1,2)+R44(3)*Y312+R44(4)*Y412)*2.d0
     \ +R441(1)*x(1,2)+R441(2)*x(2,2)+R441(3)*x(3,2)+R441(4)*x(4,2)+
     \ R442(1)*x(1,1)+R442(2)*x(2,1)+R442(3)*x(3,1)+R442(4)*x(4,1)+
     \ (R124(1)*x(1,4)+R124(2)*x(2,4)+R124(3)*x(3,4)+R124(4)*x(4,4)
     \ +DII2345(5,2)*x(4,0))*2.d0-DII2345(7,3)*x(1,0)
     \ +DII2345(3,2)*x(2,0))*I4

      R134(1)=(DII1345(10,3) + DII2345(6,2))
      R134(2)=(DII1245(10,3) - DII1345(10,3)) 
      R134(3)=(DII1235(7,3)  - DII1245(10,3))      
      R134(4)=(               - DII1235(7,3))
c 1 1 3 4
      C(25,4)=((R34(2)*Y(2,1,1)+R34(3)*Y(3,1,1)+R34(4)*Y(4,1,1)+
     \ R14(1)*Y(1,3,1)+R14(2)*Y213+R14(3)*Y(3,1,3)+R14(4)*Y413+
     \ R13(1)*Y(1,4,1)+R13(2)*Y214+R13(3)*Y314+R13(4)*Y(4,1,4))*4.d0
     \+(R11(1)*Y134+R11(2)*Y234+R11(3)*Y(3,4,3)+R11(4)*Y(4,3,4))*2.d0+
     \R113(1)*x(1,4)+R113(2)*x(2,4)+R113(3)*x(3,4)+R113(4)*x(4,4)+
     \ R114(1)*x(1,3)+R114(2)*x(2,3)+R114(3)*x(3,3)+R114(4)*x(4,3)+
     \ (R134(1)*x(1,1)+R134(2)*x(2,1)+R134(3)*x(3,1)+R134(4)*x(4,1)
     \ +DII2345(6,2)*x(1,0))*2.d0-DII2345(3,1)*x(3,0)
     \ -DII2345(2,1)*x(4,0))*I4
c 1 3 3 4
      C(30,4)=((R14(1)*Y(1,3,3)+R14(2)*Y(2,3,3)+R14(4)*Y(4,3,3)+
     \ R34(1)*Y(1,3,1)+R34(2)*Y213+R34(3)*Y(3,1,3)+R34(4)*Y413+
     \ R13(1)*Y134+R13(2)*Y234+R13(3)*Y(3,4,3)+R13(4)*Y(4,3,4))*4.d0+
     \ (R33(1)*Y(1,4,1)+R33(2)*Y214+R33(3)*Y314+R33(4)*Y(4,1,4))*2.d0
     \ +R331(1)*x(1,4)+R331(2)*x(2,4)+R331(3)*x(3,4)+R331(4)*x(4,4)+
     \ R334(1)*x(1,1)+R334(2)*x(2,1)+R334(3)*x(3,1)+R334(4)*x(4,1)+
     \(R134(1)*x(1,3)+R134(2)*x(2,3)+R134(3)*x(3,3)+R134(4)*x(4,3) 
     \ +DII2345(6,2)*x(3,0))*2.d0-DII2345(8,3)*x(1,0)
     \ +DII2345(2,2)*x(4,0))*I4
c 1 3 4 4
      C(33,4)=((R13(1)*Y(1,4,4)+R13(2)*Y(2,4,4)+R13(3)*Y(3,4,4)
     \+R34(1)*Y(1,4,1)+R34(2)*Y214+R34(3)*Y314+R34(4)*Y(4,1,4)+
     \ R14(1)*Y134+R14(2)*Y234+R14(3)*Y(3,4,3)+R14(4)*Y(4,3,4))*4.d0+
     \(R44(1)*(Y(1,3,1))+R44(2)*Y213+R44(3)*Y(3,1,3)+R44(4)*Y413)*2.d0
     \+R441(1)*x(1,3)+R441(2)*x(2,3)+R441(3)*x(3,3)+R441(4)*x(4,3)+
     \ R443(1)*x(1,1)+R443(2)*x(2,1)+R443(3)*x(3,1)+R443(4)*x(4,1)+
     \ (R134(1)*x(1,4)+R134(2)*x(2,4)+R134(3)*x(3,4)+R134(4)*x(4,4)
     \ +DII2345(6,2)*x(4,0))*2.d0-DII2345(9,3)*x(1,0)
     -+DII2345(3,2)*x(3,0))*I4
c
      R234(1)=(DII1345(10,3) - DII2345(10,3))
      R234(2)=(DII1245(8,3) - DII1345(10,3)) 
      R234(3)=(DII1235(9,3)  - DII1245(8,3))      
      R234(4)=(               - DII1235(9,3))
c  2 2 3 4
      C(28,4)=((R34(1)*Y(1,2,2)+R34(3)*Y(3,2,2)+R34(4)*Y(4,2,2)+
     \ R24(1)*Y123+R24(2)*Y(2,3,2)+R24(3)*Y(3,2,3)+R24(4)*Y423+
     \ R23(1)*Y124+R23(2)*Y(2,4,2)+R23(3)*Y324+R23(4)*Y(4,2,4))*4.d0
     \+(R22(1)*Y134+R22(2)*Y234+R22(3)*Y(3,4,3)+R22(4)*Y(4,3,4))*2.d0
     \+R223(1)*x(1,4)+R223(2)*x(2,4)+R223(3)*x(3,4)+R223(4)*x(4,4)+
     \R224(1)*x(1,3)+R224(2)*x(2,3)+R224(3)*x(3,3)+R224(4)*x(4,3)+
     \ (R234(1)*x(1,2)+R234(2)*x(2,2)+R234(3)*x(3,2)+R234(4)*x(4,2)
     \ -DII2345(10,3)*x(2,0))*2.d0-DII2345(5,3)*x(3,0)
     \ -DII2345(4,3)*x(4,0))*I4
c 2 3 3 4
      C(31,4)=((R24(1)*Y(1,3,3)+R24(2)*Y(2,3,3)+R24(4)*Y(4,3,3)+
     \ R34(1)*Y123+R34(2)*Y(2,3,2)+R34(3)*Y(3,2,3)+R34(4)*Y423+
     \ R23(1)*Y134+R23(2)*Y234+R23(3)*Y(3,4,3)+R23(4)*Y(4,3,4))*4.d0+
     \(R33(1)*Y124+R33(2)*Y(2,4,2)+R33(3)*Y324+R33(4)*Y(4,2,4))*2.d0
     \ +R332(1)*x(1,4)+R332(2)*x(2,4)+R332(3)*x(3,4)+R332(4)*x(4,4)+
     \ R334(1)*x(1,2)+R334(2)*x(2,2)+R334(3)*x(3,2)+R334(4)*x(4,2)+
     \ (R234(1)*x(1,3)+R234(2)*x(2,3)+R234(3)*x(3,3)+R234(4)*x(4,3)
     \ -DII2345(10,3)*x(3,0))*2.d0-DII2345(8,3)*x(2,0)
     \ -DII2345(6,3)*x(4,0))*I4
c 2 3 4 4
      C(34,4)=((R23(1)*Y(1,4,4)+R23(2)*Y(2,4,4)+R23(3)*Y(3,4,4)
     \+R34(1)*Y124+R34(2)*(Y(2,4,2))+R34(3)*Y324+R34(4)*Y(4,2,4)+
     \ R24(1)*Y134+R24(2)*Y234+R24(3)*Y(3,4,3)+R24(4)*Y(4,3,4))*4.d0+
     \(R44(1)*Y123+R44(2)*Y(2,3,2)+R44(3)*Y(3,2,3)+R44(4)*Y423)*2.d0
     \+R442(1)*x(1,3)+R442(2)*x(2,3)+R442(3)*x(3,3)+R442(4)*x(4,3)+
     \R443(1)*x(1,2)+R443(2)*x(2,2)+R443(3)*x(3,2)+R443(4)*x(4,2)
     \+(R234(1)*x(1,4)+R234(2)*x(2,4)+R234(3)*x(3,4)+R234(4)*x(4,4)
     \-DII2345(10,3)*x(4,0))*2.d0-DII2345(9,3)*x(2,0)
     \-DII2345(7,3)*x(3,0))*I4
c 1 2 3 4
      C(35,4)=((R12(1)*Y134+R12(2)*Y234+R12(3)*Y(3,4,3)+R12(4)*Y(4,3,4)+
     \ R13(1)*Y124+R13(2)*Y(2,4,2)+R13(3)*Y324+R13(4)*Y(4,2,4)+
     \ R14(1)*Y123+R14(2)*Y(2,3,2)+R14(3)*Y(3,2,3)+R14(4)*Y423+
     \R23(1)*Y(1,4,1)+R23(2)*Y214+R23(3)*Y314+R23(4)*Y(4,1,4)
     \+R24(1)*Y(1,3,1)+R24(2)*Y213+R24(3)*Y(3,1,3)+R24(4)*Y413+
     \ R34(1)*Y(1,2,1)+R34(2)*Y(2,1,2)+R34(3)*Y312+R34(4)*Y412)*2.d0+
     \ R123(1)*x(1,4)+R123(2)*x(2,4)+R123(3)*x(3,4)+R123(4)*x(4,4)+
     \ R124(1)*x(1,3)+R124(2)*x(2,3)+R124(3)*x(3,3)+R124(4)*x(4,3)+
     \ R134(1)*x(1,2)+R134(2)*x(2,2)+R134(3)*x(3,2)+R134(4)*x(4,2)+
     \ R234(1)*x(1,1)+R234(2)*x(2,1)+R234(3)*x(3,1)+ R234(4)*x(4,1)
     \+DII2345(6,2)*x(2,0)+DII2345(5,2)*x(3,0)
     \+DII2345(4,2)*x(4,0)-DII2345(10,3)*x(1,0))*I4
c         write(*,*)  '************ 1 ***********'
c         write(*,*)  '************ 1 ***********'
       Eij(1,1)=DCMPLX(E(1,1),C(1,1))
       Eij(2,1) = DCMPLX(E(2,1),C(2,1))
       Eij(3,1)=DCMPLX(E(3,1),C(3,1))    
       Eij(4,1)=DCMPLX(E(4,1),C(4,1))
c         write(*,*)  '************ 2 ***********'
c         write(*,*)  '************ 2 ***********'
       Eij(1,2)=DCMPLX(E(1,2),C(1,2))
       Eij(2,2)=DCMPLX(E(2,2),C(2,2))
       Eij(3,2)=DCMPLX(E(3,2),C(3,2))
       Eij(4,2)=DCMPLX(E(4,2),C(4,2))
       Eij(5,2)=DCMPLX(E(5,2),C(5,2))
       Eij(6,2)=DCMPLX(E(6,2),C(6,2))
       Eij(7,2)=DCMPLX(E(7,2),C(7,2) )  
       Eij(8,2)=DCMPLX(E(8,2),C(8,2))
       Eij(9,2)=DCMPLX(E(9,2),C(9,2))
       Eij(10,2)=DCMPLX(E(10,2),C(10,2))
       Eij(11,2)=DCMPLX(E(11,2),C(11,2))
c        write(*,*)  '************ 3 ***********'
c        write(*,*)  '************ 3 ***********'
       Eij(1,3)=DCMPLX(E(1,3), C(1,3))
       Eij(2,3)=DCMPLX(E(2,3), C(2,3))
       Eij(3,3)=DCMPLX(E(3,3), C(3,3))
       Eij(4,3)=DCMPLX(E(4,3), C(4,3))
       Eij(5,3)=DCMPLX(E(5,3), C(5,3))
       Eij(6,3)=DCMPLX(E(6,3), C(6,3))
       Eij(7,3)=DCMPLX(E(7,3), C(7,3) )   
       Eij(8,3)=DCMPLX(E(8,3), C(8,3))
       Eij(9,3)=DCMPLX(E(9,3), C(9,3))
       Eij(10,3)=DCMPLX(E(10,3), C(10,3))
       Eij(11,3)=DCMPLX(E(11,3), C(11,3))
       Eij(12,3)=DCMPLX(E(12,3), C(12,3))
       Eij(13,3)=DCMPLX(E(13,3), C(13,3))
       Eij(14,3)=DCMPLX(E(14,3), C(14,3))
       Eij(15,3)=DCMPLX(E(15,3), C(15,3))
       Eij(16,3)=DCMPLX(E(16,3), C(16,3))
       Eij(17,3)=DCMPLX(E(17,3), C(17,3) )   
       Eij(18,3)=DCMPLX(E(18,3), C(18,3))
       Eij(19,3)=DCMPLX(E(19,3), C(19,3))
       Eij(20,3)=DCMPLX(E(20,3), C(20,3))
       Eij(21,3)=DCMPLX(E(21,3), C(21,3))
       Eij(22,3)=DCMPLX(E(22,3), C(22,3))
       Eij(23,3)=DCMPLX(E(23,3), C(23,3))
       Eij(24,3)=DCMPLX(E(24,3), C(24,3))
c         write(*,*)  '************ 4 ***********'
c         write(*,*)  '************ 4 **0*********'
       Eij(1,4)=DCMPLX(E(1,4), C(1,4))
       Eij(2,4)=DCMPLX(E(2,4), C(2,4))
       Eij(3,4)=DCMPLX(E(3,4), C(3,4))
       Eij(4,4)=DCMPLX(E(4,4), C(4,4))
       Eij(5,4)=DCMPLX(E(5,4), C(5,4))
       Eij(6,4)=DCMPLX(E(6,4), C(6,4))
       Eij(7,4)=DCMPLX(E(7,4), C(7,4))    
       Eij(8,4)=DCMPLX(E(8,4), C(8,4))
       Eij(9,4)=DCMPLX(E(9,4), C(9,4))
       Eij(10,4)=DCMPLX(E(10,4), C(10,4))
       Eij(11,4)=DCMPLX(E(11,4), C(11,4))
       Eij(12,4)=DCMPLX(E(12,4), C(12,4))
       Eij(13,4)=DCMPLX(E(13,4), C(13,4))
       Eij(14,4)=DCMPLX(E(14,4), C(14,4))
       Eij(15,4)=DCMPLX(E(15,4), C(15,4))
       Eij(16,4)=DCMPLX(E(16,4), C(16,4))
       Eij(17,4)=DCMPLX(E(17,4), C(17,4))
       Eij(18,4)=DCMPLX(E(18,4), C(18,4))
       Eij(19,4)=DCMPLX(E(19,4), C(19,4))
       Eij(20,4)=DCMPLX(E(20,4), C(20,4))
       Eij(21,4)=DCMPLX(E(21,4), C(21,4))
       Eij(22,4)=DCMPLX(E(22,4), C(22,4))
       Eij(23,4)=DCMPLX(E(23,4), C(23,4))
       Eij(24,4)=DCMPLX(E(24,4), C(24,4))
       Eij(25,4)=DCMPLX(E(25,4), C(25,4))
       Eij(26,4)=DCMPLX(E(26,4), C(26,4))
       Eij(27,4)=DCMPLX(E(27,4), C(27,4))   
       Eij(28,4)=DCMPLX(E(28,4), C(28,4))
       Eij(29,4)=DCMPLX(E(29,4), C(29,4))
       Eij(30,4)=DCMPLX(E(30,4), C(30,4))
       Eij(31,4)=DCMPLX(E(31,4), C(31,4))
       Eij(32,4)=DCMPLX(E(32,4), C(32,4))
       Eij(33,4)=DCMPLX(E(33,4), C(33,4))
       Eij(34,4)=DCMPLX(E(34,4), C(34,4))
       Eij(35,4)=DCMPLX(E(35,4), C(35,4))
       Eij(36,4)=DCMPLX(E(36,4), C(36,4))
       Eij(37,4)=DCMPLX(E(37,4), C(37,4))    
       Eij(38,4)=DCMPLX(E(38,4), C(38,4))
       Eij(39,4)=DCMPLX(E(39,4), C(39,4))
       Eij(40,4)=DCMPLX(E(40,4), C(40,4))
       Eij(41,4)=DCMPLX(E(41,4), C(41,4))
       Eij(42,4)=DCMPLX(E(42,4), C(42,4))
       Eij(43,4)=DCMPLX(E(43,4), C(43,4))
       Eij(44,4)=DCMPLX(E(44,4), C(44,4))
       Eij(45,4)=DCMPLX(E(45,4), C(45,4))
       Eij(46,4)=DCMPLX(E(46,4), C(46,4))
        end

